/*
 * Frequency-response matrix object
 *
 * Copyright (c) Solexa   2005, 2006
 *               Illumina 2007
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#else
#include <cstdio>
#include <cstdlib>
#include <cstring>
#endif
#include <cassert>
#include <vector>
#include "FrequencyMatrix.h"
#include "DNACodes.h"
#include "LinAlg.h"
#include "Maths.h"


int write_matrix(FILE *fp, float *matrix, int n, char *bases)
{
    int i, j;
    for (i = 0; i < n; i++)
        fprintf(fp, "> %c\n", bases[i]);
    for (i = 0; i < n; i++) {
        char line[256] = "";
        int  pos = 0;
        for (j = 0; j < n; j++)
            pos += snprintf(line+pos, 256-pos, "%.2f ", matrix[i*n+j]);
        snprintf(line+pos, 256-pos, "\n");
        fprintf(fp, line);
    }
    return 0;
}

//---------------------------------------------------------------------------
/*
 * Class that holds the spectral response matrix of the instrument.
 * The main purpose of the object is to keep track of the order
 * of the bases acquired together with the actual frequency
 * cross-talk. It describes the intensity response in a frequency band
 * (determined by the dye's spectrum and the filter) for a given nucleotide.
 * The matrix can be read from a file or set directly. It can be applied
 * to obtain the intensity in all channel for a given combination of
 * different bases, or its inverse can be applied to recover the base
 * proportions from the observed intensities.
 * A change in power level in one channel corresponds to a rescaling
 * of the corresponding matrix row (the matrix columns corrspond to the 
 * nucleotides and the matrix rows to the frequency channels).
 *
 * TODO:
 * - The number of observed channels may be different from
 *   the number of nucleotides incorporated. This cannot be dealt with
 *   by matrix inversion (unless the number of channels is greater than
 *   the number of nucleotides, in which case some channels can simply
 *   be ignored).
 */
FrequencyMatrix::FrequencyMatrix() : m_Matrix(0), m_NumChannels(4), m_Bases(0),
                                     m_ExposureTimes(0), m_Power(0)
{
}

FrequencyMatrix::FrequencyMatrix(float *matrix, 
                                 int    numChannels,
                                 char  *bases,
                                 float *exposureTimes) 
                                  : m_Matrix(0), m_NumChannels(4), m_Bases(0),
                                    m_ExposureTimes(0), m_Power(0)
{
    setMatrix(matrix, numChannels, bases, exposureTimes);
}

void FrequencyMatrix::setMatrix(float *matrix, 
                                int    numChannels,
                                char  *bases,
                                float *exposureTimes)
{
    assert(numChannels > 0);
    assert(numChannels <= 4);
    m_NumChannels = numChannels;
    m_Matrix = new float [ m_NumChannels * m_NumChannels];
    memcpy(m_Matrix, matrix, m_NumChannels * m_NumChannels * sizeof(float));
    m_Bases = new char [ m_NumChannels ];
    if (bases)
        memcpy(m_Bases, bases, m_NumChannels * sizeof(char));
    else {
        int i;
        for (i = 0; i < m_NumChannels; i++)
            m_Bases[i] = char(BaseUpperCase[i]);
    }
    if (exposureTimes) {
        m_ExposureTimes = new float [ m_NumChannels ];
        int i;
        for (i = 0; i < m_NumChannels; i++)
            m_ExposureTimes[i] = exposureTimes[i];
    }
}

void FrequencyMatrix::identity(int n)
{
    float *matrix = new float [n*n];
    int i;
    for (i = 0; i < n*n; i++)
        matrix[i] = 0.f;
    for (i = 0; i < n; i++)
        matrix[i*n+i] = 1.f;
    setMatrix(matrix, n);
    delete [] matrix;
}

/*
 * Update a permutation lookup table <index> so that it holds a mapping
 * from the order of bases given by the <n>-dimenensional array
 * <bases> to the corresponding matrix columns. (This is to work around
 * the fact that the matrix columns may be given in a different order from
 * the actual base order used in the run.)
 */
void FrequencyMatrix::getReorderingIndex(char *bases,
                                         int  *index)
{
    int i, j;
    for (i = 0; i < m_NumChannels; i++)
        index[i] = i;
    if ((bases == 0)  || (m_NumChannels == 0))
        return;
    for (i = 0; i < m_NumChannels; i++)
        for (j = 0; j < m_NumChannels; j++)
            if (m_Bases[j] == bases[i]) {
                index[i] = j;
                break;
            }
}

/*
 * Return a pointer to a copy the <n>-dimensional subsets of rows 
 * and columns of the matrix. The columns and rows are reordered to 
 * fit the base-order given by the <n>-dimensional array <bases>.
 */
float *FrequencyMatrix::getMatrix(char  *bases,
                                  float *exposureTimes)
{
    int i, j;
    int *index = new int [m_NumChannels];
    getReorderingIndex(bases, index);
    float *result = new float [ m_NumChannels * m_NumChannels];
    for (i = 0; i < m_NumChannels; i++)
        for (j = 0; j < m_NumChannels; j++)
            result[i*m_NumChannels+j] = m_Matrix[index[i]*m_NumChannels+index[j]];

    // apply power correction
    if (m_Power)
        for (j = 0; j < m_NumChannels; j++)
            for (i = 0; i < m_NumChannels; i++)
                result[i*m_NumChannels+j] *= m_Power[i];

    // adjust exposure times
    if ((exposureTimes != 0) && (m_ExposureTimes != 0)) {
        int flag = 1;
        for (i = 0; i < m_NumChannels; i++) {
            if ((exposureTimes[i] <= 0.) || (m_ExposureTimes[i] <= 0.))
                flag = 0;         
        }
        if (flag) {
            for (i = 0; i < m_NumChannels; i++)
                for (j = 0; j < m_NumChannels; j++)
                    result[i*m_NumChannels+j] *= exposureTimes[i] / m_ExposureTimes[index[i]];
        }
    }

    delete [] index;
    return result;
}

/*
 * return the inverse matrix of getMatrix().
 */
float *FrequencyMatrix::getInverseMatrix(char  *bases,
                                         float *exposureTimes)
{
    float *matrix = getMatrix(bases, exposureTimes);
    float *inverse = new float [m_NumChannels*m_NumChannels];

    matinvert(matrix, inverse, m_NumChannels);
  
    delete [] matrix;
    return inverse;
}

/* 
 * Load the matrix from a text file.
 * The file has the following format:
 * - Lines starting with '#' contain comments.
 * - Lines starting with "<" describe a fully functional nucleotide
 * - All other lines contain the rows of the matrix
 */
int FrequencyMatrix::read(const char filename[])
{
    char buffer[256];
    FILE *fp = fopen(filename, "r");
    int dim = 0;
    int row = 0;
    std::vector<char>  bases;
    std::vector<float> exposureTimes;
    char  base;
    float exposureTime;
    if (fp) {
        while (fgets(buffer, sizeof(buffer), fp)) {
            // ignore comments
            if (buffer[0] == '#')
                continue;
            // read column descriptors
            if (buffer[0] == '>') {
                int count = sscanf(buffer, "> %c %f", &base, &exposureTime);
                bases.push_back(base);
                if (count == 2)
                    exposureTimes.push_back(exposureTime);
                dim ++;
            }
            else {                    // read matrix rows

                /* allocate matrix */
                if (m_Matrix == 0) {
                    m_NumChannels = dim;
                    m_Matrix = new float [ m_NumChannels * m_NumChannels ];
                    int i;
                    {
                        assert(m_NumChannels == int(bases.size()));
                        m_Bases = new char [bases.size()];
                        for (i = 0; i < m_NumChannels; i++)
                            m_Bases[i] = bases[i];
                    }
                    if (exposureTimes.size()) {
                        assert(m_NumChannels == int(exposureTimes.size()));
                        m_ExposureTimes = new float [exposureTimes.size()];
                        for (i = 0; i < m_NumChannels; i++)
                            m_ExposureTimes[i] = exposureTimes[i];
                    }
                }
                    
                if (row < m_NumChannels) {
                    int col = 0;
                    char *begin, *end;
                    end = buffer;
                    do {
                        begin = end;
                        m_Matrix[row*m_NumChannels+col] = strtod(begin, &end);
                        col++;
                    }
                    while ((col < m_NumChannels) && (end != NULL) && (end != begin));
                    row++;
                }
            }
        }
        fclose(fp);
    }
    return Maths::min(dim, row);
}

int FrequencyMatrix::write(const char filename[])
{
    int code = 0;
    FILE *fp = fopen(filename, "w");
    if (!fp)
        return 1;
    code = write_matrix(fp, m_Matrix, m_NumChannels, m_Bases);
    if (!code)
        code = fclose(fp);
    return code;
}

int FrequencyMatrix::write()
{
    return write_matrix(stdout, m_Matrix, m_NumChannels, m_Bases);
}


/*
 * set a number of scale factors to rescale the powers in different 
 * frequency channels. This can be used to correct for any fluctuations
 * in relative dye brightnesses.
 */
void FrequencyMatrix::setPowerCorrection(float *power)
{
    if (m_Power)
        delete [] m_Power;
    int i;
    m_Power = new float [m_NumChannels];
    for (i = 0; i < m_NumChannels; i++)
        m_Power[i] = power[i];
}

FrequencyMatrix::~FrequencyMatrix()
{
    delete [] m_Matrix;
    delete [] m_Bases;
    delete [] m_ExposureTimes;
    if (m_Power)
        delete [] m_Power;
}
