/*
 * Copyright (c) 2005-2007 Solexa
 * Authors: Klaus Maisinger,
 *          Charles McLachlan (Metropolis Data Consultants)
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#ifndef PROBABILITYINTEGRATION_H_
#define PROBABILITYINTEGRATION_H_

#include <algorithm>
#include <math.h>
#include <iostream>

class ProbabilityIntegration 
{
    int    nsamples, ncol;
    float  max_score;
    float *samples;
public:
    ProbabilityIntegration(int nnsamples, int nncol,
                           float max_score_ = 40.);
    virtual ~ProbabilityIntegration();
    virtual void integrate(float *signal, int ncol,
                           float *error_covariance_decomposition,
                           float *quality_scores);
    virtual void drawSamples(unsigned int seed = 0);
};

// distances to planes in Gaussian Coordinates
// the function you want is calculateDVector() (below)

// preamble - a few utility functions.

// NB: applyMatrix determines the ordering of the matrices that all
// functions in this header expect

// find closest axis to a certain point
template<typename T>
    int quadrantForPoint(size_t ndim, const T *p)
{
    return std::max_element(p, p + ndim) - p;
}

// apply matrix to vector
template<typename T>
    void applyMatrix(size_t ndim, T *vec, const T *matrix)
{
    T *dest= new T [ndim];
    for(size_t i = 0; i < ndim; i++)
    {
        dest[i] = 0.0;
        for(size_t j = 0; j < ndim; j++)
            dest[i] += vec[j] * matrix[j + i*ndim];
    }
    std::copy(dest, dest + ndim, vec);
    delete [] dest;  
}

/// and the transpose
template<typename T>
    void applyTransMatrix(size_t ndim, T *vec, const T *matrix)
{
    T *dest = new T [ndim];
    for(size_t i = 0; i < ndim; i++)
    {
        dest[i] = 0.0;
        for(size_t j = 0; j < ndim; j++)
            dest[i] += vec[j] * matrix[i + j*ndim];
    }
    std::copy(dest, dest + ndim, vec);
    delete [] dest;  
}

// return the Gaussian distance to origin via return();

// For a Gaussian with a covarience matrix K such that
// the exponent is -0.5*(x - mu)^T K^{-1} (x - mu)
// and with K = X X^T, caller must supply X and X^{-1}
// in the arrays X and invX

template<typename T>
    float calculateDistanceVector(size_t ndim, const T *mu, 
                                  const T *X, const T *invX, T *output)
{
    size_t base_call = quadrantForPoint(ndim, mu);
    
    int nd = 0;
    
    for (size_t i = 0; i < ndim; i++)
    {
        if (i == base_call) continue;
        
        T *nhat = new T [ndim]; // spherical point
        std::fill(nhat, nhat + ndim, 0.0);
        nhat[base_call] = 1.0 / sqrt(2.0);
        nhat[i] = -1.0 / sqrt(2.0);

        double dot = 0.0;
        for (size_t i = 0; i < ndim; i++) dot += nhat[i] * mu[i]; 
        
        // if C^{-1} = K = X X^T then apply X^T
        
        // in the simulation X = S R^{T}
        applyTransMatrix(ndim, nhat, X);

        double mag = 0.0;
        for (size_t i = 0; i < ndim; i++) mag += nhat[i]*nhat[i]; 
        
        mag = sqrt(mag);
        
        T d = dot / mag;
#if 0
        std::cerr << "d[" << nd << "] " << d << " check mag " << mag << "\n";
#endif
        output[nd++] = fabs(d);
        delete [] nhat;
    }
    
    // transform origin from general space to spherical space.
    // form -X^{-1} mu
    T *mutrans = new T [ndim];
    std::copy(mu, mu + ndim, mutrans);

    applyMatrix(ndim, mutrans, invX);
    
    double mmag = 0.0;
    for (size_t i = 0; i < ndim; i++) mmag += mutrans[i] * mutrans[i];
    mmag = sqrt(mmag);
    delete [] mutrans;

    return mmag;
}

// given a distance d , what is the probability split?
template<typename T> T PofD(T d) {
    return 0.5 * (1.0 + erf(fabs(d)/sqrt(2.0)));
}

#endif /*PROBABILITYINTEGRATION_H_*/
