/*
 * Copyright (c) Solexa   2004, 2005, 2006
 *               Illumina 2007
 * Author:       Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cassert>
#include "SequenceData.h"
#include "PositionIndex.h"
#include "Message.h"
#include "Maths.h"

SequenceData::SequenceData(const char filename[])
{
    m_Filename = filename;
}

SequenceData::~SequenceData()
{
}

int SequenceData::write(const unsigned char  * sequenceArray,
                        const int              numClusters,
                        const int              numCycles,
                        const int              lane,
                        const int              tile,
                        const Position       * positions)
{
    return write_sequence_file(m_Filename.c_str(),
                               sequenceArray, numClusters, numCycles,
                               lane, tile, positions);
}

int SequenceData::read(unsigned char * sequenceArray,
                       int             numClusters,
                       int             numCycles)
{
    return read_sequence_file(m_Filename.c_str(),
                              sequenceArray, numClusters, numCycles);
}

/**************************************************************************/
int write_sequence_file(const char filename[],
                        const unsigned char *sequence_array,
                        const int nobj,
                        const int ncycle,
                        const int lane,
                        const int tile,
                        const Position * positions)
{
    FILE *seqfile = fopen(filename,"wt");
    if (!seqfile) {
        error("Failed to open %s for writing.", filename);
        return 1;
    }
    char *buffer = new char[21+ncycle+1];
    int i, n;
    int error_code = 0;

    for (n = 0; n < nobj; n++) {
        int idx = 0;
        if (positions != 0) {
            PositionIndex index(lane, tile, positions[n]);
            idx = index.snprint(buffer, 20);
        }

        buffer[idx++] = '\t';

        for (i = 0; i < ncycle; i++) {
            buffer[idx+i] = sequence_array[i*nobj+n];
        }

        sprintf(buffer+idx+ncycle, "\n");
        if (fprintf(seqfile, "%s", buffer) < int(strlen(buffer))) {
            error_code = 2;
            error("Failed to write to %s.", filename);
        };
    }
    delete [] buffer;

    if (fclose(seqfile)) {
        error("Failed to close %s.", filename);
        error_code = 3;
    }
    return error_code;
}

int read_sequence_file(const char      filename[],
                       unsigned char * sequence_array,
                       const int       numObjects,
                       const int       numCycles)
{
    FILE *seqfile = fopen(filename, "r");
    int   i;

    for (i = 0; i < numObjects*numCycles; i++) {
        sequence_array[i] = '.';
    }

    if (!seqfile)
        return 1;

    const int minBufferLength = 400; // hard-coded limit for minimum no of cycles
    int       bufferLength    = Maths::max(numCycles, minBufferLength);
    char    * buffer    = new char [bufferLength+100+1];
    char    * seqbuffer = new char [bufferLength+1];
    int       nread = 0;
    while (fgets(buffer, bufferLength+100, seqfile) && (nread < numObjects)) {
        if (buffer[0] == '#')  // skip comments
            continue;
        sscanf(buffer, "%*d%*d%*d%*d%100s", seqbuffer);
        for (i = 0; i < Maths::min(numCycles, (int)strlen(seqbuffer)); i++)
            sequence_array[i*numObjects+nread] = seqbuffer[i];
        nread++;
    }
    fclose(seqfile);

    delete [] buffer;
    delete [] seqbuffer;

    return 0;
}
