/*
 * Copyright (c) Illumina 2008
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include "IoUtils.hh"

namespace IoUtils {

AutoFile::AutoFile(const std::string &fileName, const std::string &mode) : fh_(fopen(fileName.c_str(), mode.c_str()))
{
    if (0 == fh_)
    {
        const std::string message = std::string("Couldn't open file ") + fileName + ": " + strerror(errno);
        error(message.c_str());
    }
}


AutoGzFile::AutoGzFile(const std::string &fileName, const std::string &mode, const unsigned int bufferSize) throw (std::ios_base::failure)
    : fh_(gzopen(fileName.c_str(), mode.c_str()))
    , line_(NULL)
    , lineSize_(0)
    , bufferSize_(bufferSize)
    , buffer_(new char [bufferSize_ + 1])
    , lineStart_(buffer_)
{
    if (Z_NULL == fh_)
    {
        const std::string message = std::string("Couldn't open compressed file ") + fileName + " in mode " + mode + 
        (errno ? std::string(": ") + std::string(strerror(errno)) : std::string(""));
        throw std::ios_base::failure(message);
    }
    bufferEnd_ = fillBuffer(lineStart_);
}

char *AutoGzFile::fillBuffer(char * const fromPosition) throw (std::ios_base::failure)
{
    const unsigned int remainingSize = bufferSize_ + buffer_ - fromPosition;
    const int numberRead =  gzread(fh_, fromPosition, remainingSize);
    if (numberRead < 0 || (numberRead == 0 && !gzeof(fh_)))
    {
        int errorNumber = 0;
        const std::string errorMessage (gzerror(fh_, &errorNumber));
        throw std::ios_base::failure(std::string("Couldn't read from file: ") + errorMessage);
    }
    *(fromPosition + numberRead) = 0;
    return fromPosition + numberRead;
}

std::string AutoGzFile::getline() throw (std::ios_base::failure)
{
    return std::string(gets());
}

char *AutoGzFile::gets() throw (std::ios_base::failure) 
{
    char *endOfLine = findEndOfLine(lineStart_);
    *endOfLine = 0; // this is alright even if there is no EOL before EOF (buffer_ is oversized by 1 character)
    char * result = lineStart_;
    lineStart_ = endOfLine + 1; // Note, this makes lineStart_ > bufferEnd_ on EOF
    return result;
}

inline char *AutoGzFile::findEndOfLine(char * const fromPosition) throw (std::ios_base::failure)
{
    char * const endOfLine = strchr(fromPosition, '\n');
    if (NULL != endOfLine)
    {
        return endOfLine;
    }
    else  // reached bufferEnd_ without finding any '\n'
    {
        if (gzeof(fh_))
        {
            return bufferEnd_;
        }
        const unsigned int remainingSize = bufferEnd_ - lineStart_;
        // from a performance point of view, and to allow the use of memcpy,
        // the buffer should be at least double the size of a line
        if (lineStart_ <= buffer_ + (bufferSize_ / 2))
        {
            bufferSize_ *= 2;
            char *newBuffer = new char[bufferSize_ + 1];
            memcpy(newBuffer, lineStart_, remainingSize);
            delete [] buffer_;
            buffer_ = newBuffer;
        }
        else
        {
            memcpy(buffer_, lineStart_, remainingSize);
        }
        lineStart_ = buffer_;
        char * const newLineEnd = buffer_ + remainingSize;
        bufferEnd_ = fillBuffer(newLineEnd);
        return findEndOfLine(newLineEnd);
    }
}

} // namespace Iostring 


