//  Time-stamp: <26 Jun 2006 16:57 BST Logger.cpp km>
/*
 * Simple logger class, loosely modelled on the corresponding Java API.
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <ctime>
#include <cstdlib>
#include <cstdio>
#include "Logger.h"

/*---------------------------------------------------------------------*/
/*
 * Handler that writes out messages to a file.
 * The messages are combined with information on time, computer, user, 
 * logging level, software, separated by tabs.
 */
void FileLogHandler::message(int level, const char message[])
{
    if (level < min_level)
       return;
    event_list.push_back(format_message(level, message));
    if ((int)event_list.size() >= line_buffer_size)
        flush();
}

char *FileLogHandler::format_message(int level, const char message[])
{
    time_t tp;
    time(&tp);
    struct tm *tps = gmtime(&tp);
    char timebuffer[10];
    strftime(timebuffer, 10, "%H:%M:%S", tps);
    char datebuffer[12];
    strftime(datebuffer, 12, "%d/%m/%Y", tps);
    char user[20];
    if (getenv("LOGNAME"))
        snprintf(user, 20, getenv("LOGNAME"));
    else if (getenv("USERNAME"))
        snprintf(user, 20, getenv("USERNAME"));
    else
        snprintf(user, 20, "unknown");
             
    char host[25];
    if (getenv("HOSTNAME"))
        snprintf(host, 25, getenv("HOSTNAME"));
    else
        snprintf(host, 25, "unkown");
    char buffer [256];
    snprintf(buffer, 256, "%i\t%s\t%s\t%s\t%s\t%s\t", 
            level, datebuffer, timebuffer, source.c_str(), user, host);
    string record(buffer);
    record +=  string(message) + string("\n");

    int len = strlen(record.c_str())+1;
    char *msg = new char [len];
    snprintf(msg, len, "%s", record.c_str());
    return msg;
}

void FileLogHandler::flush()
{
    int i;
    file_ptr = new std::ofstream(filename.c_str(), ios::app);
    for (i = 0; i < (int)event_list.size(); i++)
        *file_ptr << event_list[i];
    file_ptr->flush();
    delete file_ptr;
    for (i = 0; i < (int)event_list.size(); i++)
        delete [] event_list[i];
    event_list.clear();
}

/*
 * logger object. It manages a list of handlers and passes on messages to all 
 * handlers.
 */
Logger::Logger() : source("Unknown")
{
    LogHandler *handler = new LogHandler();
    handlers.push_back(handler);
}

Logger::Logger(const char fn[]) : source("Unknown")
{
    FileLogHandler *handler = new FileLogHandler(fn);
    handlers.push_back(handler);
}

void Logger::add_handler(LogHandler *handler)
{
    handlers.push_back(handler);
}

Logger::~Logger()
{
    int i;
    for (i = 0; i < (int)handlers.size(); i++)
        delete handlers[i];
}

void Logger::config(const char message[])
{
    send_message(LOGGING_LEVEL_CONFIG, message);
}

void Logger::info(const char message[])
{
    send_message(LOGGING_LEVEL_INFO, message);
}

void Logger::warning(const char message[])
{
    send_message(LOGGING_LEVEL_WARNING, message);
}

void Logger::error(const char message[])
{
    send_message(LOGGING_LEVEL_SEVERE, message);
}


void Logger::set_source(const char src[])
{
    int i;
    for (i = 0; i < (int)handlers.size(); i++)
        handlers[i]->set_source(src);
}

void Logger::send_message(int level, const char message[])
{
    int i;
    for (i = 0; i < (int)handlers.size(); i++)
        handlers[i]->message(level, message);
}


#if 0
int main() {
    Logger *logger = new Logger("test.txt");
    logger->set_source("LogTest");
    int i;
    for (i = 0; i < 30; i++) {
        cout << i << std::endl;
        logger->info("Hello");
        sleep(1);
    }
    delete logger;
}
#endif
