/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <iostream>

#include "Message.h"

#include "File_Buffer.h"

/*****************************************************************************/

enum { ERR_OK=0, ERR_ARGS, ERR_FILE_READ, ERR_FILE_WRITE, ERR_EXTENSION };

/*****************************************************************************/

int usage(const std::string app_name, int ret_val)
{
    error("Usage: %s <in_text_file[.gz|.bz2]> <out_text_file.gz|.bz2]> [[1]",
          app_name.c_str());

    return ret_val;
}

/*****************************************************************************/

void compr_from_name(const std::string& file_path_str,
                     File_Buffer::Compression_Type& compression_type)
{
    compression_type = File_Buffer::compressionNone;

    const unsigned int file_path_str_len(file_path_str.size());

    if (file_path_str_len > 3) {
        if (file_path_str.substr(file_path_str_len - 3) == ".gz") {
            compression_type = File_Buffer::compressionGzip;
            info("Treating file `%s' as gzip.", file_path_str.c_str());
        }
    }

    if (file_path_str_len > 4) {
        if (file_path_str.substr(file_path_str_len - 4) == ".bz2") {
            compression_type = File_Buffer::compressionBzip2;
            info("Treating file `%s' as bzip2.", file_path_str.c_str());
        }
    }
}

/*****************************************************************************/

bool test_extension_addition()
{
    const std::string file_path_str("fred");

    if (add_std_extension(file_path_str, File_Buffer::compressionNone)
        == file_path_str) {
        info("Extension testing : compressionNone OK");
    } else {
        error("Extension testing : Failed on compressionNone");
        return false;
    }

    if (add_std_extension(file_path_str, File_Buffer::compressionGzip)
        == file_path_str + ".gz") {
        info("Extension testing : compressionGzip OK");
    } else {
        error("Extension testing : Failed on compressionGzip");
        return false;
    }

    if (add_std_extension(file_path_str, File_Buffer::compressionBzip2)
        == file_path_str + ".bz2") {
        info("Extension testing : compressionBzip2 OK");
    } else {
        error("Extension testing : Failed on compressionBzip2");
        return false;
    }

    return true;
}

/*****************************************************************************/

int main(int num_args, char* arg_cstr_arr[])
{
    // Check args.

    const std::string app_name(arg_cstr_arr[0]);

    if ((num_args < 3) || (num_args > 4)) {
        return usage(app_name, ERR_ARGS);
    }

    const std::string in_file_path_str(arg_cstr_arr[1]);
    const std::string out_file_path_str(arg_cstr_arr[2]);

    bool one_line_at_a_time(false);

    if (num_args == 4) {
        if (std::string(arg_cstr_arr[3]) == "1") {
            one_line_at_a_time = true;
        } else {
            return usage(app_name, ERR_ARGS);
        }
    }

    File_Buffer::Compression_Type
        in_compression_type(File_Buffer::compressionNone);
    compr_from_name(in_file_path_str, in_compression_type);

    File_Buffer::Compression_Type
        out_compression_type(File_Buffer::compressionNone);
    compr_from_name(out_file_path_str, out_compression_type);

    File_Buffer in_file_buffer(in_file_path_str, in_compression_type);
    File_Buffer out_file_buffer(out_file_path_str, out_compression_type);
    bool at_eof(false);

    File_Buffer::Line_Vec line_vec;

    if (one_line_at_a_time) {
        std::string line_str;

        while (in_file_buffer.read_line(line_str, at_eof)) {
            line_vec.push_back(line_str);

            if (!out_file_buffer.write_some_lines(line_vec)) {
                error("Failed to write_some_lines to `%s'",
                      out_file_path_str.c_str());
                return ERR_FILE_WRITE;
            }

            line_vec.clear();

            if (at_eof) {
                info("Finished reading from `%s'.",
                     in_file_buffer.file_path_str().c_str());
                break;
            }            
        }
    } else {
        while (in_file_buffer.read_some_lines(line_vec, at_eof)) {
            if (!out_file_buffer.write_some_lines(line_vec)) {
                error("Failed to write_some_lines to `%s'.",
                      out_file_path_str.c_str());
                return ERR_FILE_WRITE;
            }

            // for (File_Buffer::Line_Vec_CIter line_citer = line_vec.begin();
            //     line_citer != line_vec.end(); ++line_citer) {
            //    std::cout << *line_citer << std::endl;
            // }

            if (at_eof) {
                info("Finished reading from `%s'.",
                     in_file_buffer.file_path_str().c_str());
                break;
            }
        }
    }

    if (!out_file_buffer.end_file()) {
        error("Failed to end `%s'.", out_file_path_str.c_str());
        return ERR_FILE_WRITE;
    }

    if (!test_extension_addition()) {
        return ERR_EXTENSION;
    }

    return ERR_OK;
}

/*****************************************************************************/
