#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iostream>
#include <vector>
#include <deque>
#include <string>
#include <cmath>
#include <algorithm>
#include <map>
#include <dirent.h>
#include <pthread.h>

#include <sys/resource.h>

#include "GlobalUtilities.h"
#include "ELAND_inner.h"
#include "ELAND_outer.h"

/*
PROJECT: ELAND (Efficient Local Alignment of Nucleotide Data)
MODULE: ELAND.cpp
AUTHOR: A. J. Cox

Copyright (c) 2003-2006 Solexa Limited. All rights reserved.
This source file is covered by the "Solexa Public Source License"
agreement and bound by the terms therein.
*/


// Identifier to embed in source and binary files

const char* eland_cpp_id="@(#) $Id: ELAND.cpp,v 1.6 2007/10/03 15:51:09 tc Exp $";

// ELAND.cpp

// Main program for ELAND.cpp

// Stands for Efficient Local Alignment of Nucleotide Data

// Matches oligos to genome data allowing for substitution errors and
// for ambiguity codes

// The basic plan: oligos are divided into 4 contiguous fragments, e.g.
// a 24-mer into 4 fragments of 6 bases - call them ABCD.  
// These 4 fragments can be partitioned into (assuming a 24 base oligo) 
// a 12 base prefix and a 12 base suffix in 6 different ways:
//
// Prefix  Suffix
// A B     C D
// C D     A B
// A C     B D
// B D     A C
// A D     B C
// B C     A D
//
// The idea is to convert the prefix to a number using 2 bits per base encoding
// and use this number as an index into a look up table.
//
// The look up table gives you the suffixes of all oligos that have the same
// prefix as the oligo you are searching for. These suffixes are then compared
// inexactly with the query suffix by doing an exclusive-OR then using the
// result to index the "suffix score table" - another look up table which
// gives the positions of any differing bases.
//
// The key point is that we can catch all possible 2 substitution error 
// matches by checking the 6 combinations above: the worst case scenario is 
// that the two errors occur in different fragments - for one of the six 
// combinations both these fragments will be in the suffix and so will be 
// caught by the inexact matching step.
//
// 'N's in the data (and some 2 base ambiguity codes) can be handled by
// additionally storing an AND mask for each oligo which is applied after
// the exclusive-OR and prior to accessing the suffix score table. 
//
// It makes sense to cover the six possible partitions in 3 passes
// through the genome, making use of the fact that the partitions can be 
// paired so that the prefix of one is the suffix of another.

// This file contains the "main"
// Most core functionality is in ELAND_inner.h
// Interface functionality is in ELAND_outer.h & ELAND_outer.cpp

int main (int numArgs, char* const args[])
{
  cerr << "ELAND: Efficient Local Alignment of Nucleotide Data" << endl
       << "Copyright (c) 2003-2006 Solexa Limited. All rights reserved."
       << endl << "Author: Anthony J. Cox" << endl << endl;

  cerr 
      << "Publications incorporating data generated by the use of\n"
      << "this software or modified versions thereof should cite:\n"
      << "Anthony J. Cox.\n"
      << "Ultra high throughput alignment of short sequence tags.\n" 
      << "In preparation." << endl << endl;


  cerr << "WARNING: this software is compiled to process oligos of length "
       << OLIGO_LENGTH 
       << ".\nTo change this value, edit OLIGO_LENGTH and recompile."
       << endl << endl;

  cerr << "Will use at most " << MAX_HASH_BITS 
       << " bits in hash table" << endl;
  cerr << "Can process at most " << maxNumOligos 
       << " oligos per batch" << endl;

#ifdef ONE_ERROR_PER_OLIGO
  cerr << "Will find all exact and single error matches" << endl;
#else
  cerr << "Will find all matches having 2 errors or less" << endl;
#endif

#ifdef DONT_SEARCH_REVERSE_STRAND
  cerr << "WARNING: will search for oligos in forward strand only"<< endl
       << "To search both strands, comment out DONT_SEARCH_REVERSE_STRAND "
       << "and recompile." << endl;
#endif


  if ((numArgs<4)||(numArgs>6))
  {
    cerr << "\nsyntax: " << args[0] 
	 << " oligoFile genomeDirectory outputFile[.vmf]\n"
	 << "[--multi[=N0[,N1,N2]] [repeatFile]\n\n" 
	 << "oligoFile - file or directory of files\n"
	 << "  file type deduced from first character of each file:\n"
	 << "  '>' - fasta format\n"
	 << "  '#' - single molecule array format\n"
	 << "  [AGCTNagctn] - raw sequence format\n\n"
	 << "genomeDirectory - directory of genome files\n"
	 << "  preprocessed to 2-bits-per-base format using squashGenome.\n\n"
	 << "outputFile - name of output file\n"
	 << "  if name ends in '.vmf', use verbose match format\n"
	 << "  else use format required by assembly module\n\n"
	 << "--multi - output multiple hits per read\n"
	 << "  at most N0,N1,N2 exact, 1-mismatch, 2-mismatch hits per read\n"
	 << "repeatFile - list of repeats to exclude (optional)\n"
	 << "  must be ASCII and in alphabetical order"
	 << endl;
    exit (1);
  } // ~if

  int oligoLength(OLIGO_LENGTH);
  assert(oligoLength!=0);

  OligoSource* pOligos(OligoSource::getOligoSource(args[1]));

  const char* pTest=pOligos->getNextOligo();

  if (pTest==NULL)
  {
    cerr << "WARNING: there do not appear to be any sequences in file "
	 << args[1] << endl;
  } // ~if
  else if (strlen(pTest)<(uint)OLIGO_LENGTH)
  {
    cerr << "ERROR: first sequence of " << args[1] << " contains "
	 << strlen(pTest) << " bases, but "
	 << oligoLength << " bases are needed for the alignment."
	 << endl
	 << "To change this value, edit OLIGO_LENGTH and recompile."
	 << endl << endl;
    return (-1);
  } // ~else if
  else if (strlen(pTest)>(uint)OLIGO_LENGTH)
  {
    cerr << "WARNING: first sequence of " << args[1] << " contains "
	 << strlen(pTest) << " bases, but only the first "
	 << oligoLength << " bases will be used in the alignment."
	 << endl
	 << "To change this value, edit OLIGO_LENGTH and recompile."
	 << endl << endl;
  } // ~else if

  pOligos->rewind();


  Timer timer;

  cerr << "Starting run! Time now: " << timer.timeNow();
  cerr << "Will read oligos from file " << args[1] << endl;


  MatchTable* pResults(NULL);
  RepeatTable* pRepeats(NULL);


  if (numArgs>=5)
  {
    char *prefix="--multi";
    //    if (strcmp(args[4],"--multi")==0)
    //   {
    if (strncmp(args[4],prefix,strlen(prefix))==0)
    {
      int maxMatchesExact, maxMatchesOneError, maxMatchesTwoErrors;
      if (strlen(args[4])==strlen(prefix))
      {
	pResults=new MatchTableMulti(args[3]);
      }
      else if (sscanf(args[4],"--multi=%d,%d,%d",
		      &maxMatchesExact,
		      &maxMatchesOneError,
		      &maxMatchesTwoErrors)==3)
      {
	pResults=new MatchTableMulti(args[3],
				     maxMatchesExact,
				     maxMatchesOneError,
				     maxMatchesTwoErrors);
      } // ~else if
      else if (sscanf(args[4],"--multi=%d",&maxMatchesExact)==1)
      {
	pResults=new MatchTableMulti(args[3],maxMatchesExact);
      } // ~else if
      else
      {
	cerr << "Could not parse max number of matches from string " 
	     << args[4] << ", proceeding using defaults" << endl;
	pResults=new MatchTableMulti(args[3]);
      } // ~else

      if (numArgs==6)
      {
	pRepeats=new RepeatTable(args[5]);
      } // ~if
    } // ~if
    else
    {
      pRepeats=new RepeatTable(args[4]);
    } // ~else
  } // ~if

  if (pResults==NULL)
  {
    pResults=MatchTable::getMatchTable(args[3]);
  }

  if (pRepeats!=NULL)
  {
    cerr << "Scanning for repeats in list " 
	 << args[numArgs-1] << ": " << timer << endl;
    pRepeats->checkOligos( *pOligos, *pResults );
    cerr << "Scanned repeats: " << timer << endl;
    delete pRepeats;
  } // ~if




  SuffixScoreTable scoreTable;
  HashTableDataStore htds1, htds2;

  const char suffixName[] = ".2bpb";
  cerr << "Trying to open directory " << args[2]  << " ..." << endl;

  DIR* pDir;
  if ( ! ( pDir = opendir( args[2] ) ) )
  {
    cerr << " ... failed!" << endl;
    return -1; 
  } // ~if

  char dirBuffer[80];
  char* pSuffixStart; 
  dirent* dirEntry;

  // chromosome names are indexed starting at 1 
  // (because in OligoInfo a chromosome num of 0 indicates no match found)
  vector<string> chromNames(1); // %%%%% (1);
  vector<MatchPosition> blockStarts(1);

  const string directoryName(((string)args[2])+'/');
  //  string fullChromName;

  // get chromosome names from directory then close it
  while( (dirEntry = readdir(pDir)) )
  {
    strcpy(dirBuffer, dirEntry->d_name);
    if ( (pSuffixStart=strstr(dirBuffer,suffixName)) )
    {
      assert(*(pSuffixStart+5)=='\0'); // make sure it's a suffix
      *pSuffixStart='\0';
      chromNames.push_back((string)dirBuffer);
    } // ~if
  } // ~while
  closedir( pDir );

  // do pass 0
  {
    OligoHashTable<0> hashTable
      (oligoLength, htds1, htds2, scoreTable, *pResults); 
    scanAll<0>( pOligos, directoryName, chromNames, 
	     blockStarts, hashTable, timer );
  } // ~scope of hashTable

#ifndef ONE_ERROR_PER_OLIGO
  // do pass 1
  {
    OligoHashTable<1> hashTable
      (oligoLength, htds1, htds2, scoreTable, *pResults); 
    scanAll<1>( pOligos, directoryName, chromNames, 
	     blockStarts, hashTable, timer );
  } // ~scope of hashTable

  // do pass 2
  {
    OligoHashTable<2> hashTable
      (oligoLength, htds1, htds2, scoreTable, *pResults); 
    scanAll<2>( pOligos, directoryName, chromNames, 
	     blockStarts, hashTable, timer );
  } // ~scope of hashTable
#endif

  // clear some space - pResults->print may need it for MatchTableMulti
  htds1.clear();
  htds2.clear();

  MatchPositionTranslator getMatchPos
    ( chromNames, blockStarts, directoryName );


  cerr << "Outputting results: " << timer << endl;

  pResults->print
    ( *pOligos, getMatchPos, chromNames, blockStarts, scoreTable, oligoLength );

  cerr << "... done " << timer << endl;

  delete pOligos;

  cerr << "Run complete! Time now: " << timer.timeNow();
  return 0;
} // ~main
