#!/usr/bin/env perl

use warnings;
use strict;
use Getopt::Long;

# PROJECT: ELAND (Efficient Large-Scale Alignment of Nucleotide Databases)
# MODULE: ELAND_standalone.pl
# AUTHOR: A. J. Cox
#
# Copyright (c) 2003-2006 Solexa Limited. All rights reserved.
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Driver script to run ELAND analysis as done in the version 0.3 pipeline
# in a standalone fashion

use Data::Dumper;
use File::Basename;

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: ELAND_standalone.pl,v 1.8 2008/02/28 11:51:46 tc Exp $';
    my $cvs_version='$Revision: 1.8 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

my $maxSeedLength=32;

my $usage=<<"END";
Usage: $0 options
--input-file|if 
  must specify at least one file, specify two for paired analysis
--input-type|it 
  type of input file: fastq, fasta for export 
--read-length|rl
  will work it out from input file(s) if not specified
--seed-length|sl
  length of read used for ELAND alignment 
  - default is to min of read-length and 32
--eland-genome|eg
  full path of a squashed genome directory
--output-prefix|op
  will produce a set of output files whose names prefixed by this
  - defaults to 'reanalysis'  
--pipeline-dir|pd
  path of pipeline installation to call 
  - by default, looks in same dir as executable is found)
--base-quality|bq
  in fasta mode assume all bases have this quality (TBD)
--pair-params|pp
  paired read analysis parameters to pass to pickBestPair
  - multiple arguments must go in quotes
  - defaults to '--circular' - treats all chromosomes as circular
  NB: for a paired read analysis, both reads must share same
  input-type, read-length and seed-length
END

die $usage if (@ARGV==0);


my %options=(
'pipeline-dir'=>dirname($0).'/..',
'base-quality'=>30,
'pair-params'=>'--circular',
'output-prefix'=>'./reanalysis');

GetOptions(\%options, 
	   "input-file|if=s@", 
	   "input-type|it=s", 
	   "read-length|rl=i@",
	   "seed-length|sl=i@",
	   "eland-genome|eg=s",
	   "output-prefix|op=s",
	   "pipeline-dir|pd=s",
	   "base-quality|bq=i",
	   "pair-params|pp:s",
	   "test=s");

$options{'seed-length'}[0]=-1 unless (defined($options{'seed-length'}));
$options{'read-length'}[0]=-1 unless (defined($options{'read-length'}));



print Dumper(\%options);

# sub unpackInputFile($$$$$$)
# Takes input file in fasta/fastq/export format and makes
# qraw, seqpre and filt files
# Also works out read length if not already specified
sub unpackInputFile($$$$$$)
{
    my ($inputFileName,$inputFileType,
     $qrawFileName,$seqpreFileName,$filtFileName,$readLength)=@_;

#    my $qrawFileName=$options{'output-prefix'}."_qraw.txt";
#    my $seqpreFileName=$options{'output-prefix'}."_seqpre.txt";
#    my $filtFileName=$options{'output-prefix'}."_filt.txt";
    my ($line);
    unless (-e $inputFileName)
    {
	print STDERR "$0: file $inputFileName does not exist, will not proceed\n";
	exit (-1);
    }
    if (-e $qrawFileName)
    {
	print STDERR "$0: file $qrawFileName exists, will not proceed\n";
	exit (-1);
    }
    elsif (-e $seqpreFileName)
    {
	print STDERR "$0: file $seqpreFileName exists, will not proceed\n";
	exit (-1);
    }
    open(QRAW, ">$qrawFileName") 
	|| die "Failed to open $qrawFileName for writing: $!";
    open(SEQP, ">$seqpreFileName") 
	|| die "Failed to open $qrawFileName for writing: $!";
    open(FILT, ">$filtFileName") 
	|| die "Failed to open $qrawFileName for writing: $!";
    open(READ, $inputFileName) || 
	die "Failed to open $inputFileName: $!";

    if ($inputFileType eq 'fasta')
    {
	if ($readLength==-1)
	{
	    $line=<READ>;
	    $line=<READ>;
	    chomp ($line);
	    $readLength=length($line);
	    print STDERR "$0: deduced read length of $readLength for data in $inputFileName\n";
	    seek READ,0,0;
	}
	die; # TBD make qual string
    } # if
    elsif ($inputFileType eq 'fastq')
    {

	if ($readLength==-1)
	{
	    $line=<READ>;
	    $line=<READ>;
	    chomp ($line);
	    $readLength=length($line);
	    print STDERR "$0: deduced read length of $readLength for data in $inputFileName\n";
	    seek READ,0,0;
	} # if

	while (defined($line=<READ>))
	{
	    chomp($line);
	    unless ($line=~s/^\@//) { die "Could not parse $line"; }
	    print SEQP "$line";
	    $line=<READ>;
	    chomp($line);
	    print SEQP "\t${line}\n";
	    $line=<READ>;
	    unless ($line=~/^\+/) { die "Could not parse $line"; }
	    $line=<READ>;
	    chomp($line);
	    print QRAW "${line}\n";
	    print FILT "Y\n";
	}
    } # elsif
    elsif ($inputFileType eq 'export')
    {
	my @F;
	my $readID;
	if ($readLength==-1)
	{
	    $line=<READ>;
	    chomp ($line);
	    @F=split("\t",$line);
	    $readLength=length($F[8]);
	    print STDERR "$0: deduced read length of $readLength for data in $inputFileName\n";
	    seek READ,0,0;
	} # if

	while (defined($line=<READ>))
	{
	    chomp($line);
	    @F=split("\t",$line);
	    unless (@F==22) { die "Could not parse $line"; }
	    $readID="$F[0]_$F[1]:$F[2]:$F[3]:$F[4]:$F[5]";
	    $readID.='#'.$F[6] unless ($F[6] eq "");
	    $readID.='/'.$F[7] unless ($F[7] eq "");
	    $F[8]=substr($F[8],0,$readLength); # read
	    $F[9]=substr($F[9],0,$readLength); # qv
		
	    print SEQP "${readID}\t$F[8]\n";
	    print QRAW "$F[9]\n";
	    print FILT "$F[-1]\n";
	} # while

    } # elsif
    else
    {
	die "Value ${inputFileType} is not valid for option input-type";
	exit (-1);
    }
    close (QRAW);
    close (SEQP);
    close (FILT);
    close (READ);

    return $readLength;
} # unpackInputFile

# sub makeElandExtended($$$)
# Takes the processing as far as the eland_extended file
# This bit is common to both single read and paired read code
# Assumes read length is already either specified or worked out by
# Works out alignment seed length if necessary

sub makeElandExtended($$$$$)
{
    my ($outputPrefix, $pipelineDir, $elandGenome, 
	$readLength, $seedLength)=@_;
    my $command;

    if ($seedLength==-1) 
    {
	$seedLength =($readLength>$maxSeedLength)?$maxSeedLength:$readLength; 
    } # if
    else
    {
	$seedLength=($seedLength>$maxSeedLength)?$maxSeedLength:$seedLength;
	$seedLength=($seedLength>$readLength)?$readLength:$seedLength;
    } # else

    print STDERR "$0: Setting seed length for ${outputPrefix}_eland_extended.txt to $seedLength\n";



# make eland query file
$command=<<"END";
cat ${outputPrefix}_seqpre.txt | ${pipelineDir}/Gerald/buildSeq.pl --fasta > ${outputPrefix}_eland_query.txt
END

print STDERR "$0: About to run $command\n";
system($command);

#run eland
$command=<<"END";
${pipelineDir}/Eland/eland_${seedLength} ${outputPrefix}_eland_query.txt ${elandGenome} ${outputPrefix}_eland_multi.txt --multi=
END

print STDERR "$0: About to run $command\n";
system($command);

# get frag
$command=<<"END";
cat ${outputPrefix}_eland_multi.txt | ${pipelineDir}/Gerald/unsquash ${elandGenome} ${readLength} ${seedLength} --multi --compact > ${outputPrefix}_frag.txt
END

print STDERR "$0: About to run $command\n";
system($command);


# build the eland extended
$command=<<"END";
paste ${outputPrefix}_eland_multi.txt ${outputPrefix}_frag.txt | ${pipelineDir}/Gerald/buildExtendedAlign.pl ${readLength} ${seedLength} > ${outputPrefix}_eland_extended.txt
END

print STDERR "$0: About to run $command\n";

    system($command);
    return $seedLength;
} # makeElandExtended

sub makeExportSingle($$$$$)
{
    my ($outputPrefix, $pipelineDir, $elandGenome, 
	$readLength, $seedLength)=@_;
    my $command;

# build the saf file
$command=<<"END";
paste ${outputPrefix}_qraw.txt ${outputPrefix}_eland_extended.txt | ${pipelineDir}/Gerald/pickBestAlignment.pl ${seedLength} > ${outputPrefix}_calsaf.txt
END

print STDERR "$0: About to run $command\n";
system($command);

# build the export file
$command=<<"END";
paste ${outputPrefix}_seqpre.txt ${outputPrefix}_qraw.txt ${outputPrefix}_calsaf.txt ${outputPrefix}_filt.txt | ${pipelineDir}/Gerald/buildExport.pl --symbolic > ${outputPrefix}_export.txt
END

print STDERR "$0: About to run $command\n";
system($command);

} # sub makeExportSingle

sub makeExportPaired($$$$$$)
{
    my ($outputPrefix, $pipelineDir, $elandGenome, 
	$seedLength1, $seedLength2, $pairParams)=@_;
    my $command;

# get genome details
$command=<<"END";
${pipelineDir}/Eland/squashGenome ${elandGenome} > ${outputPrefix}_genomesize.xml
END

print STDERR "$0: About to run $command\n";
system($command);

# pair them up
$command=<<"END";
${pipelineDir}/Gerald/pickBestPair.pl ${pairParams} ${seedLength1} ${seedLength2} ${outputPrefix}_1_eland_extended.txt  ${outputPrefix}_2_eland_extended.txt ${outputPrefix}_1_qraw.txt ${outputPrefix}_2_qraw.txt ${outputPrefix}_1_filt.txt ${outputPrefix}_1_calsaf.txt ${outputPrefix}_2_calsaf.txt ${outputPrefix}_anomaly.txt ${outputPrefix}_genomesize.xml > ${outputPrefix}_pair.xml
END

print STDERR "$0: About to run $command\n";
system($command);

# make export file for read 1
$command=<<"END";
paste ${outputPrefix}_1_seqpre.txt ${outputPrefix}_1_qraw.txt ${outputPrefix}_1_calsaf.txt ${outputPrefix}_1_filt.txt | ${pipelineDir}/Gerald/buildExport.pl --paired --symbolic > ${outputPrefix}_1_export.txt
END

print STDERR "$0: About to run $command\n";
system($command);

# make export file for read 1
$command=<<"END";
paste ${outputPrefix}_2_seqpre.txt ${outputPrefix}_2_qraw.txt ${outputPrefix}_2_calsaf.txt ${outputPrefix}_2_filt.txt | ${pipelineDir}/Gerald/buildExport.pl --paired --symbolic > ${outputPrefix}_2_export.txt
END

print STDERR "$0: About to run $command\n";
system($command);

} # sub makeExportPaired


die "No input-file specified" unless (defined($options{'input-file'}));





unless (defined($options{'input-type'}))
{
    if (($options{'input-file'}=~/\.fa$/)
	||($options{'input-file'}=~/\.fasta$/))
    {
	print STDERR 
	    "$0: No input-type specified, guessing fasta from input file name $options{'input-file'}\n";
	$options{'input-type'}='fasta';
    }
    elsif ($options{'input-file'}=~/export.txt$/)
    {
	print STDERR 
	    "$0: No input-type specified, guessing export from input file name $options{'input-file'}\n";
	$options{'input-type'}='fasta';
    }
    else
    {
	die "No input-type specified";
    }

} # unless





if (@{$options{'input-file'}}==1) 
{ 
    print STDERR "$0: single input-file specified, will do single read analysis\n";
    $options{'read-length'}[0]=unpackInputFile
	( $options{'input-file'}->[0],
	  $options{'input-type'},
	  $options{'output-prefix'}."_qraw.txt",
	  $options{'output-prefix'}."_seqpre.txt",
	  $options{'output-prefix'}."_filt.txt",
	  $options{'read-length'}[0]  );

    $options{'seed-length'}[0]=makeElandExtended
	( $options{'output-prefix'},
	  $options{'pipeline-dir'},
	  $options{'eland-genome'},
	  $options{'read-length'}[0],
	  $options{'seed-length'}[0]);

    makeExportSingle
	( $options{'output-prefix'},
	  $options{'pipeline-dir'},
	  $options{'eland-genome'},
	  $options{'read-length'}[0],
	  $options{'seed-length'}[0]);


}
elsif (@{$options{'input-file'}}==2) 
{
    $options{'seed-length'}[1]=$options{'seed-length'}[0]
	if (@{$options{'seed-length'}}==1);
    $options{'read-length'}[1]=$options{'read-length'}[0]
	if (@{$options{'read-length'}}==1);

    print STDERR "$0: two input-files specified, will do paired read analysis\n";
    $options{'read-length'}[0]=unpackInputFile
	( $options{'input-file'}->[0],
	  $options{'input-type'},
	  $options{'output-prefix'}."_1_qraw.txt",
	  $options{'output-prefix'}."_1_seqpre.txt",
	  $options{'output-prefix'}."_1_filt.txt",
	  $options{'read-length'}[0]  );

    $options{'seed-length'}[0]=makeElandExtended
	( $options{'output-prefix'}.'_1',
	  $options{'pipeline-dir'},
	  $options{'eland-genome'},
	  $options{'read-length'}[0],
	  $options{'seed-length'}[0]);

    $options{'read-length'}[1]=unpackInputFile
	( $options{'input-file'}->[1],
	  $options{'input-type'},
	  $options{'output-prefix'}."_2_qraw.txt",
	  $options{'output-prefix'}."_2_seqpre.txt",
	  $options{'output-prefix'}."_2_filt.txt",
	  $options{'read-length'}[0]  );

    $options{'seed-length'}[1]=makeElandExtended
	( $options{'output-prefix'}.'_2',
	  $options{'pipeline-dir'},
	  $options{'eland-genome'},
	  $options{'read-length'}[1],
	  $options{'seed-length'}[1]);

    makeExportPaired
	( $options{'output-prefix'},
	  $options{'pipeline-dir'},
	  $options{'eland-genome'},
	  $options{'seed-length'}[0],
	  $options{'seed-length'}[1],
	  $options{'pair-params'});

}
else
{
    die;
}


#s_4_1_eland_query.txt \
#/node/test/CompBioTestArea/BAC_Ecoli_ELAND s_4_1_eland_multi.txt.tmp --multi= /node/test/CompBioTestArea/polyA.txt \
#&& mv s_4_1_eland_multi.txt.tmp s_4_1_eland_multi.txt

# get genome details
#/node/test/CompBioTestArea/CodeUnderTest/Pipeline/Gerald/../Eland//squashGenome /node/test/CompBioTestArea/BAC_Ecoli_ELAND > s_4_genomesize.xml.tmp && mv s_4_genomesize.xml.tmp s_4_genomesize.xml

# get frag
#cat s_4_1_eland_multi.txt | /node/test/CompBioTestArea/CodeUnderTest/Pipeline/Gerald/unsquash /node/test/CompBioTestArea/BAC_Ecoli_ELAND \
#35 32 --multi --compact > s_4_1_frag.txt.tmp && mv s_4_1_frag.txt.tmp s_4_1_frag.txt

# build the eland extended
#cat s_4_1_eland_multi.txt | /node/test/CompBioTestArea/CodeUnderTest/Pipeline/Gerald/unsquash /node/test/CompBioTestArea/BAC_Ecoli_ELAND \
#35 32 --multi --compact > s_4_1_frag.txt.tmp && mv s_4_1_frag.txt.tmp s_4_1_frag.txt
#paste s_4_1_eland_multi.txt s_4_1_frag.txt \
#| /node/test/CompBioTestArea/CodeUnderTest/Pipeline/Gerald/buildExtendedAlign.pl 35 32 \
#> s_4_1_eland_extended.txt.tmp && mv s_4_1_eland_extended.txt.tmp s_4_1_eland_extended.txt

#paste s_6_qraw.txt s_6_eland_extended.txt \
#| /node/test/CompBioTestArea/CodeUnderTest/Pipeline/Gerald/pickBestAlignment.pl


# make seq file
#paste s_4_1_seqpre.txt s_4_1_qcal.txt \
#s_4_filt.txt | \
#/node/test/CompBioTestArea/CodeUnderTest/Pipeline/Gerald/buildSeq.pl --qualityFilter --fastq --numeric \
#> s_4_1_sequence.txt.tmp && mv s_4_1_sequence.txt.tmp s_4_1_sequence.txt
