#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cmath>
#include <algorithm>
#include <dirent.h>
#include <set>

#include "GlobalUtilities.h"
#include "XMLContentHandler.h"

/*
PROJECT: ELAND (Efficient Local Alignment of Nucleotide Databases) 
MODULE: SquashGenome.cpp
AUTHOR: A. J. Cox

Copyright (c) 2003-2006 Solexa Limited. All rights reserved.
This source file is covered by the "Solexa Public Source License"
agreement and bound by the terms therein.
*/

// Identifier to embed in source and binary files

const char* id="@(#) $Id: SquashGenome.cpp,v 1.9 2008/04/01 16:43:01 tc Exp $";


// SquashGenome.cpp
// Takes a set of DNA sequence files in ASCII format and squashes them
// into a 2 bits per base format. Produces two files for each sequence 
// file.
// originalFileName.2bpb - compressed sequence data
// originalFileName.vld - gives the start and end positions
// of each valid region in the sequence file (i.e. regions containing
// only AGCT - can't handle '-'s, 'N's etc. using 2bpb format)
// Can also unsquash a single compressed file.


enum
{
  seqLineLength = 60
};

// bases in a squashed file are numbered starting at zero 
// (although match positions are output numbered starting at 1)
// start gives the first valid base in a region (0=first base in file)
// finish gives the last valid base of the region (ie not the first invalid
// base after the end of the region)

#ifdef MADE_GLOBAL
struct ValidRegion
{
  uint start;
  uint finish;
}; // struct ValidRegion
#endif

// class IndexMaker
// creates an index to the entries of a multiple-entry fasta file
class IndexBuilder
{
  public:
  IndexBuilder( const string& idxFileName, const char* firstEntry ) :
    idxFile_(idxFileName.c_str())
  {
    // open output files
    if (idxFile_.fail())
    {
      cerr << "Error in squash: could not open file " << idxFileName << endl;
      exit (1);
    } // ~if
    else
    {
      cerr << "Opened file " << idxFileName << endl;
    } // ~else
    addEntry( firstEntry, 0 );
  } // ~ctor

  ~IndexBuilder() 
  { 
    idxFile_.close(); 
  } // ~dtor

  const char* addEntry( const char* pEntry, int entryPos) 
  {
    cerr << "Adding entry" << endl;
    string thisName;
    idxFile_ << entryPos;
    idxFile_.put('\t');
    while ((*pEntry!=' ')&&(*pEntry!='\n'))
    {
      thisName+=(*pEntry);
      idxFile_.put(*pEntry);
      pEntry++;
    } // ~while

    if (strpbrk(thisName.c_str(),":,/")!=NULL)
    {
      cerr << "Error: invalid entry name in fasta file (" 
	   << thisName << ") - cannot contain ',',':' or '/' characters"
	   << endl;
      exit (1);
    } // ~if
    else if (names_.count(thisName)!=0)
    {
      cerr << "Error: duplicate entry name in fasta file (" 
	   << thisName << ") - each entry must have distinct name"
	   << endl;
      exit (1);
    } // ~else if
    names_.insert(thisName);

    idxFile_.put('\n');
    while (*pEntry!='\n') pEntry++;
    pEntry++;
    numEntries_++;
    return pEntry;
  } // ~addEntry

  uint getNumEntries( void ) const
  {
    return numEntries_;
  } // ~getNumEntries( void ) const

private:
  ofstream idxFile_;
  uint numEntries_;
  set<string> names_;

}; // ~class IndexBuilder




Word getNextBase( Word* pWord, uint i)
{
  return ( (pWord[i>>4])>>(2*((i&0xF)^0xF)) ) & 0x3; 
}

void outputSizesToXML( const char* dirName, DIR* pDir )
{
  const char suffixName[] = ".vld";

  //  char dirBuffer[80];
  char* pSuffixStart; 
  char numBuffer[20];

  string vldFileName, vldFileFullPath;

  FILE* pVldFile;
  dirent* dirEntry;

  ValidRegion lastValidRegion;


  XMLFileWriter xml;
  Attributes attributes;
  attributes.addAttribute( (string)"DirectoryName",(string)dirName); 

  xml.setFileHandle(stdout);
  xml.startDocument();
  xml.startElement("SequenceSizes",&attributes);
  xml.characters("\n",0,1);
  while( (dirEntry = readdir(pDir)) )
  {
    vldFileName=dirEntry->d_name;
    if ( (pSuffixStart=strstr(vldFileName.c_str(),suffixName)) )
    {
      assert(strlen(pSuffixStart)
	     ==strlen(suffixName)); // make sure it's a suffix

      vldFileFullPath=(string)dirName+(string)"/"+vldFileName;

      // memory map validity file
      if ((pVldFile = fopen(vldFileFullPath.c_str(), "r"))==NULL)
      {
	cerr << "Error in outputSizesToXML: could not open file " 
	     << vldFileFullPath << endl;
	exit (1);
      } // ~if

      //      cerr << "outputSizesToXML: opened file " << vldFileFullPath << endl;

      if (fseek( pVldFile, -sizeof(ValidRegion), SEEK_END)!=0)
      {
	cerr << "Error in outputSizesToXML: failed to fseek in " 
	     << vldFileFullPath << endl;
	exit (1);
      } // ~fseek
      if (fread( &lastValidRegion, sizeof(ValidRegion), 1, pVldFile)!=1)
      {
	cerr << "Error in outputSizesToXML: failed to fread in " 
	     << vldFileFullPath << endl;
	exit (1);
      } // ~fread
      *pSuffixStart='\0';
      xml.startElement(vldFileName.c_str());
      //      xml.characters("\n",0,1);
      sprintf(numBuffer,"%d",lastValidRegion.finish+1);
      xml.characters(numBuffer,0,strlen(numBuffer));
      xml.endElement(vldFileName.c_str());
      xml.characters("\n",0,1);
      //      cout << vldFileName.c_str() << " " << lastValidRegion.start 
      //	   << " " << lastValidRegion.finish << endl;
    } // ~if
  } // ~while
  closedir( pDir );

  xml.endElement("SequenceSizes");
  xml.characters("\n",0,1);
  xml.endDocument();


} // ~outputSizesToXML


void unsquash( const char* squashName )
{
  string seqFileName(squashName), vldFileName(squashName);
  seqFileName += (string) ".2bpb";
  vldFileName += (string) ".vld";

  void *pMapSeq;
  void *pMapVld;
  int fd;
  int seqFileSize, vldFileSize;

  Word* pWord;
  Word* pLastWord;

  ValidRegion* pValid;
  ValidRegion* pLastValid;
  
  char* pChar;


  // memory map sequence file
  fd = open(seqFileName.c_str(), O_RDONLY, S_IRUSR );
  if (fd==-1)
  {
    cerr << "Error in unsquash: could not open file " << seqFileName << endl;
    exit (1);
  } // ~if
  seqFileSize = lseek(fd, 0, 2);
  cerr << "squash: opened file " << seqFileName << " of size "
       << seqFileSize << " bytes." << endl;

  pMapSeq=mmap(0, seqFileSize, PROT_READ, MAP_SHARED, fd, 0);
  if (pMapSeq==NULL)
  {
    cerr << "Error in unsquash: could not memory map file " 
	 << seqFileName<< endl;
    exit (1);
  } // ~if
  close (fd);

  // memory map validity file
  fd = open(vldFileName.c_str(), O_RDONLY, S_IRUSR );
  if (fd==-1)
  {
    cerr << "Error in unsquash: could not open file " << vldFileName << endl;
    exit (1);
  } // ~if
  vldFileSize = lseek(fd, 0, 2);
  cerr << "squash: opened file " << vldFileName << " of size "
       << vldFileSize << " bytes." << endl;

  pMapVld=mmap(0, vldFileSize, PROT_READ, MAP_SHARED, fd, 0);
  if (pMapVld==NULL)
  {
    cerr << "Error in unsquash: could not memory map file " 
	 << vldFileName<< endl;
    exit (1);
  } // ~if
  close (fd);

  // set up pointers for Word

  pChar= (char*) pMapSeq;
  pWord= (Word*) pMapSeq;
  pLastWord = (Word*)(pChar+seqFileSize);

  // set up last pointer for valid region

  pChar= (char*) pMapVld;
  pLastValid = (ValidRegion*)(pChar+vldFileSize);

  // print out the comment line ...
  while (*pChar != '\n')
  {
    cout.put(*pChar);
    pChar++;
  } // ~while

  // ... and the '\n' afterwards
  //  cout.put(*(pChar++));
  pChar++;
  //  cout << endl; // %%%%%

  // set up 
  pValid = (ValidRegion*) pChar;
  assert( pValid!=pLastValid );

  uint i(0);

  //  char thisChar;

  while (pValid!=pLastValid)
  {
    //    cout << pValid->start << ' ' << pValid->finish << ' ' << i << endl;
    for (; i<pValid->start; i++ )
    {
      if ( (i%seqLineLength)==0 ) cout.put('\n');
      assert( getNextBase(pWord, i) == 0 );
      cout.put('N');
    } // ~for
    for (i=pValid->start; i!=pValid->finish; i++)
    {
      if ( (i%seqLineLength)==0 ) cout.put('\n');
      cout.put(baseNames[getNextBase(pWord, i)]);
    } // ~for
    assert( i==pValid->finish );
    pValid++;
  } // ~while

  cout.put('\n');
} // ~unsquash

void squash( const char* directoryName, const char* fileName )
{
  const char* pChar;
  const char* pLast;
  void *pMap;

  IndexBuilder* pIndex(NULL);

  // make file names
  cerr << "Full file path: " << fileName << endl;
  const char* pFileName( strrchr(fileName,'/') );
  if (pFileName==NULL) pFileName=fileName; 
  else ++pFileName;
  cerr << "Extracted file name:" << pFileName << endl;

  string seqFileName(directoryName), 
    vldFileName(directoryName),
    idxFileName(directoryName);

  seqFileName += '/';
  vldFileName += '/';
  idxFileName += '/';

  seqFileName += (string)pFileName;
  vldFileName += (string)pFileName;
  idxFileName += (string)pFileName;
  
  seqFileName += (string)".2bpb";
  vldFileName += (string)".vld";
  idxFileName += (string)".idx";


  cerr << seqFileName << " " << vldFileName << endl;

  // open output files
  ofstream seqFile(seqFileName.c_str());
  if (seqFile.fail())
  {
    cerr << "Error in squash: could not open file " << seqFileName << endl;
    exit (1);
  } // ~if

  ofstream vldFile(vldFileName.c_str());
  if (vldFile.fail())
  {
    cerr << "Error in squash: could not open file " << seqFileName << endl;
    exit (1);
  } // ~if

  // memory map input file
  int fd(open(fileName, O_RDONLY, S_IRUSR ));
  if (fd==-1)
  {
    cerr << "Error in squash: could not open file " << fileName << endl;
    exit (1);
  } // ~if
  int fileSize(lseek(fd, 0, 2));
  cerr << "squash: opened file " << fileName << " of size "
       << fileSize << " bytes." << endl;
  pMap=mmap(0, fileSize, PROT_READ, MAP_SHARED, fd, 0);
  if (pMap==NULL)
  {
    cerr << "Error in squash: could not perform memory mapping" << endl;
    exit (1);
  } // ~if
  close (fd);

  pChar= (const char*) pMap;
  pLast= pChar+fileSize;

  // scan past first comment line and write to validFile
  //  while( *(pChar++) != '\n' );
  //  pChar--;

  while (*pChar != '\n')
  {
    vldFile.put(*pChar);
    pChar++;
  } // ~while

  if (*(pChar-1)=='\r')
  {
    cerr << "ERROR: can't squash file " << fileName 
	 << " as it's a DOS format text file!!\n";
    exit (-1);
  } // if



  vldFile.put('\n');

  int thisPos(0), totalValid(0);
  vector<Word> words;
  vector<ValidRegion> valids;
  bool inValidRegion(false);
  Word thisBase;

  words.reserve(fileSize/maxBasesPerWord);
  valids.reserve(fileSize/1000);


  while(1)
  {
    ++pChar;
    if (*pChar=='\n') pChar++;
    if (pChar==pLast)
    {
      munmap (pMap, fileSize );
      if (inValidRegion==true) 
      {
	valids.back().finish=thisPos-1;
	totalValid+=valids.back().finish-valids.back().start+1;
      } // ~if
      break;
    } // ~if
    else if (*pChar=='>')
    {
      if (pIndex==NULL)
	pIndex= new IndexBuilder( idxFileName, (const char*) pMap );
      pChar=pIndex->addEntry( pChar, thisPos );
      if (inValidRegion==true)
      {
	valids.back().finish=thisPos;
	totalValid+=valids.back().finish-valids.back().start+1;
	inValidRegion=false;
      } // #if
      //      continue;
    } // ~else

    if ((thisPos%maxBasesPerWord)==0)
    {
      words.push_back(0);
    } // ~if

    //    thisPos++;

    thisBase = whichBase[(uint)*pChar];

    if (thisBase!=nv)
    { 
      if (inValidRegion==false)
      {
	valids.push_back(ValidRegion());
	valids.back().start=thisPos;
	inValidRegion=true;
      } // ~if
    } // ~if
    else
    {
      thisBase=0; // codes for A
      if (inValidRegion==true)
      {
	valids.back().finish=thisPos-1;
	totalValid+=valids.back().finish-valids.back().start+1;
	inValidRegion=false;
      } // #if
    } // ~else
    words.back() <<= numBitsPerBase;
    words.back() |= thisBase;

    thisPos++;
  //  c=*pChar_;

  } // ~while

  words.back() <<= 2*(16-thisPos%maxBasesPerWord);

  // write sequence file
  seqFile.write( (const char*)&words[0], words.size()*sizeof(Word));
  seqFile.close();
  
  // write validFile
  vldFile.write( (const char*)&valids[0], valids.size()*sizeof(ValidRegion));
  vldFile.close();

  cerr << "finished file " << fileName << endl
       << thisPos << " bases" << endl
       << totalValid << " valid bases (" 
       << (100.*totalValid/thisPos) << "\045)" << endl
       << valids.size() << " valid regions" << endl;

  if (pIndex!=NULL)
  {
    cerr << pIndex->getNumEntries() << " entries" << endl;
    delete pIndex;
  } // ~if
  return;



} // ~squash

int main (int numArgs, char* const args[])
{
  //  int fd;
  //  void* file_memory;
  //  char* charMem;
  //  size_t fileSize;

  if (numArgs<2)
  {
    cerr << "syntax:" << endl 
	 << args[0] 
	 << " fileToUnsquash" << endl 
	 << "  - unsquash file to standard output" << endl
	 << args[0] 
	 << " squashDirectory" << endl 
	 << "  - output sizes of squashed files to standard output in XML" 
	 << endl << args[0] 
	 << " targetDirectory fileToSquash1 fileToSquash2 ..." << endl
	 << "  - squash files and place in targetDirectory" << endl;
    exit (1);
  } // ~if

  Timer timer;
  cerr << timer << endl;
  //  resourceUsage();

  if (numArgs==2)
  {
    cerr << "Trying to open directory " << args[1]  << " ..." << endl;
    DIR* pDir;
    if (( pDir = opendir( args[1] ) ))
    { 
      cerr << " ... success, will output file sizes to XML" << endl;
      outputSizesToXML( args[1], pDir );
    } // ~if
    else
    {
      cerr << " ... could not open directory, will try to unsquash file" 
	   << endl;
      unsquash( args[1] );
    } // ~else

  } // ~if
  else
  {
    for ( int fileNum(2); fileNum<numArgs; fileNum++)
    {
      squash( args[1], args[fileNum] );
    } // ~for
  } // ~else
  
  cerr << timer << endl;
  //  resourceUsage();
  return 0;
} // ~main

