/**********************************************************************************
File:        BackEstimator.cpp
Description: Implementation file for the CBackEstimator class. 
Created:     12/12/06  SBanerjee

Copyright (c) 2006-2010, Solexa Inc, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
**********************************************************************************/
#include "BackEstimator.h"
/***********************************************************************************************
Function:    Setup
Description: Sets it up with the initial image.
Returns:     Status / Error code 
Arguments:   Input: The kernel size for box summing.
             const int iAvgBoxSize:(I) The kernel size for box summing
             Input: A Single 16-bit monochrome image.
             const CShortImage & inPicture:(I) Input Picture
Globals:     None
***********************************************************************************************/
int CBackEstimator::Setup(const int iAvgBoxSize,
                          const CShortImage & inPicture)               
{
    int iStatus = OKAY;
    if (iAvgBoxSize < 0)
    {
        iStatus = INVALID_INPUT_PARAMETERS;
        return iStatus;
    }
    m_iAvgBoxSize = (iAvgBoxSize & 1) ? iAvgBoxSize : iAvgBoxSize + 1;

    /*****************************************************************************
    MEMORY ALLOCATION and INITIALIZATION for
    m_SwissCheese, m_BackMeanPicture, m_BackVarPicture, & m_NumPicture
    ******************************************************************************/
    const int iPicWidth = inPicture.GetWidth();
    const int iPicHeight = inPicture.GetHeight();

    m_SwissCheese.Create(iPicWidth, iPicHeight);
    if ((iStatus = m_SwissCheese.GetStatus()) != OKAY) return iStatus;
    m_SwissCheese.SetClear();

    m_BackMeanPicture.Create(iPicWidth + m_iAvgBoxSize - 1,
                             iPicHeight + m_iAvgBoxSize - 1);
    if ((iStatus = m_BackMeanPicture.GetStatus()) != OKAY) return iStatus;
    m_BackMeanPicture.CenterCopy(iPicWidth, iPicHeight, inPicture.GetImage());

    m_BackVarPicture = m_BackMeanPicture;
    m_BackVarPicture *= m_BackMeanPicture;

    m_NumPicture.Create(iPicWidth + m_iAvgBoxSize - 1,
                        iPicHeight + m_iAvgBoxSize - 1);
    if ((iStatus = m_NumPicture.GetStatus()) != OKAY) return iStatus;

    return iStatus;
}

/***********************************************************************************************
Function:    Add
Description: Adds the foreground picture to update the swiss cheese image.
Returns:     None
Arguments:   Input: A Foreground Picture
             const CGrayImage & forePicture:(I) Foreground Picture
Globals:     None
***********************************************************************************************/
void CBackEstimator::Add(const CGrayImage & forePicture)
{
    unsigned int iImageSize = m_SwissCheese.GetWidth() * m_SwissCheese.GetHeight();
    const unsigned char * pSrcBuffer = forePicture.GetImage();
    unsigned char * pDstBuffer = m_SwissCheese.GetImage();

    while (iImageSize--)
    {
        if (*pSrcBuffer++) *pDstBuffer++ |= 1;
        else pDstBuffer++;
    }
}

/***********************************************************************************************
Function:    Run
Description: Runs the CBackEstimator to create the background mean and variance images.
             Note that the variance image is NOT the true variance image. To create the
             true variance image, the square of the mean has to be subtracted from this. This
             subtraction is NOT carried out here. It is done in the last step in method
             IAMDetector::updateAnalysisImages() for reasons of computational efficiency.
Returns:     None
Arguments:   None
Globals:     None
***********************************************************************************************/
void CBackEstimator::Run()
{
    m_NumPicture.CenterCopy(m_SwissCheese.GetWidth(), m_SwissCheese.GetHeight(), m_SwissCheese.GetImage());

    unsigned int iNum = 0;
    float fGlobalMean = 0;
    float fGlobalVariance = 0;

    unsigned int iImageSize = m_NumPicture.GetWidth() * m_NumPicture.GetHeight();
    unsigned int * pNumBuffer = m_NumPicture.GetImage();
    float * pBackMeanBuffer = m_BackMeanPicture.GetImage();
    float * pBackVarBuffer = m_BackVarPicture.GetImage();

    while (iImageSize--)
    {
        if (*pNumBuffer)
        {
            *pNumBuffer++ = 0;
            *pBackMeanBuffer++ = 0;
            *pBackVarBuffer++ = 0;
        }
        else
        {
            *pNumBuffer++ = 1;
            fGlobalMean += *pBackMeanBuffer++;
            fGlobalVariance += *pBackVarBuffer++;
            iNum++;
        }
    }

    if (iNum)
    {
        fGlobalMean /= iNum;
        fGlobalVariance /= iNum;
    }

    m_NumPicture.BoxSum(m_iAvgBoxSize, m_iAvgBoxSize, true);
    m_BackMeanPicture.BoxSum(m_iAvgBoxSize, m_iAvgBoxSize, true);
    m_BackVarPicture.BoxSum(m_iAvgBoxSize, m_iAvgBoxSize, true);

    iImageSize = m_NumPicture.GetWidth() * m_NumPicture.GetHeight();
    pNumBuffer = m_NumPicture.GetImage();
    pBackMeanBuffer = m_BackMeanPicture.GetImage();
    pBackVarBuffer = m_BackVarPicture.GetImage();

    while (iImageSize--)
    {
        iNum = *pNumBuffer++;
        if (iNum)
        {
            *pBackMeanBuffer++ /= iNum;
            *pBackVarBuffer++ /= iNum;
        }
        else
        {
            *pBackMeanBuffer++ = fGlobalMean;
            *pBackVarBuffer++ = fGlobalVariance;
        }
    }
}

