/****************************************************************************
File:        BlobLabeler.cpp
Description: Include file for the image processing function to perform
             connected component analysis and labeling.
Created:     01/27/05  SBanerjee

Copyright (c) 1996-2006, Saibal Banerjee, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
****************************************************************************/
#if defined(__BORLANDC__)
#include <string.h> // For memset() and memcpy() functions
#include <limits.h> // For UINT_MAX
#else
#include <cstring> // For memset() and memcpy() functions
#include <climits> // For UINT_MAX
#endif
#include "ErrorCodes.h"
#include "BlobLabeler.h"

/****************************************************************************
Function:    Deallocate (private function)
Description: Deallocates m_pPicture and sets it to NULL.
             Resets all other member variables
****************************************************************************/
void CBlobLabeler::Deallocate()
{
    m_iWidth = 0;
    m_iHeight = 0;

    m_iAllocatedSize = 0;
    if (m_pPicture)
    {
        delete [] m_pPicture;
        m_pPicture = 0;
    }

    m_nLabels = 0;

    m_iBreadth = 0;
    m_iTall = 0;
    m_iOrigin = 0;
}

/****************************************************************************
Function:    Destructor
Description: Deallocates and resets all member variables.
****************************************************************************/
CBlobLabeler::~CBlobLabeler()
{
    Deallocate();
    m_iStatus = OKAY;
}

/****************************************************************************
Function:    CopyToPicture (private function)
Description: 
****************************************************************************/
void CBlobLabeler::CopyToPicture(const unsigned char * pPixel)
{
    unsigned int * pPic = m_pPicture + m_iOrigin;
    const int iBump = m_iBreadth - m_iWidth;
    int iWidth;
    int iHeight = m_iHeight;
    while (iHeight--)
    {
        iWidth = m_iWidth;
        while (iWidth--)
        {
            *pPic++ = *pPixel++;
        }
        pPic += iBump;
    }
}

/****************************************************************************
Function:    Blobber4 (private function)
Description: 4-connected component analysis and labeling routine.
             Needs a border of pixels marked 0.
             Needs seed pixels marked 2.
             Needs grow pixels marked 1.
             The Blobber8() routine labels all foreground blobs from 12
             onwards.
             This routine uses Depth First Search from a seed pixel to
             grow into other seed and grow pixels. The stack that it
             requires to backtrack to parents are stored during the
             forward traversal in the array itself.
Arguments:   None
Returns:     None
Globals:     None
WARNING:     This routine contains "goto" statements in order to minimize
             execution time. The readability of the code is not compromised.
****************************************************************************/
void CBlobLabeler::Blobber4()
{
    m_nLabels = 7;
    unsigned int * pPic = m_pPicture;
    unsigned int iPicValue;

    unsigned int iSize = m_iTall * m_iBreadth;
    while (iSize--)
    {
        if (*pPic++ == 2)
        {
            if (++m_nLabels == UINT_MAX)
            {
                return;
            }
            *--pPic = 7;

    FLEE:   iPicValue = *(pPic + 1);
            while ((iPicValue == 1) || (iPicValue == 2))
            {
                *++pPic = 3;
                iPicValue = *(pPic + 1);
            }

    FLWW:   iPicValue = *(pPic - 1);
            while ((iPicValue == 1) || (iPicValue == 2))
            {
                *--pPic = 4;
                iPicValue = *(pPic - 1);
            }

    FLSS:   iPicValue = *(pPic + m_iBreadth);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic += m_iBreadth;
                *pPic = 5;
                goto FLEE;
            }

    FLNN:   iPicValue = *(pPic - m_iBreadth);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic -= m_iBreadth;
                *pPic = 6;
                goto FLEE;
            }

    BACK:   switch (*pPic)
            {
            case 3:     *pPic-- = m_nLabels;
                        goto FLWW;

            case 4:     *pPic++ = m_nLabels;
                        goto FLSS;

            case 5:     *pPic = m_nLabels;
                        pPic -= m_iBreadth;
                        goto FLNN;

            case 6:     *pPic = m_nLabels;
                        pPic += m_iBreadth;
                        goto BACK;

            case 7:    *pPic++ = m_nLabels;
                        break;
            }
        }
    }
}

/****************************************************************************
Function:    Blobber8 (private function)
Description: 8-connected component analysis and labeling routine.
             Needs a border of pixels marked 0.
             Needs seed pixels marked 2.
             Needs grow pixels marked 1.
             The Blobber8() routine labels all foreground blobs from 12
             onwards.
             This routine uses Depth First Search from a seed pixel to
             grow into other seed and grow pixels. The stack that it
             requires to backtrack to parents are stored during the
             forward traversal in the array itself.
Arguments:   None
Returns:     None
Globals:     None
WARNING:     This routine contains "goto" statements in order to minimize
             execution time. The readability of the code is not compromised.
****************************************************************************/
void CBlobLabeler::Blobber8()
{
    m_nLabels = 11;
    unsigned int * pPic = m_pPicture;
    unsigned int iPicValue;
    int iSouthEast = m_iBreadth + 1;
    int iSouthWest = m_iBreadth - 1;

    unsigned int iSize = m_iTall * m_iBreadth;
    while (iSize--)
    {
        if (*pPic++ == 2)
        {
            if (++m_nLabels == UINT_MAX)
            {
                return;
            }
            *--pPic = 11;

    FLEE:   iPicValue = *(pPic + 1);
            while ((iPicValue == 1) || (iPicValue == 2))
            {
                *++pPic = 3;
                iPicValue = *(pPic + 1);
            }

    FLWW:   iPicValue = *(pPic - 1);
            while ((iPicValue == 1) || (iPicValue == 2))
            {
                *--pPic = 4;
                iPicValue = *(pPic - 1);
            }

    FLSS:   iPicValue = *(pPic + m_iBreadth);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic += m_iBreadth;
                *pPic = 5;
                goto FLEE;
            }

    FLNN:   iPicValue = *(pPic - m_iBreadth);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic -= m_iBreadth;
                *pPic = 6;
                goto FLEE;
            }

    FLSE:   iPicValue = *(pPic + iSouthEast);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic += iSouthEast;
                *pPic = 7;
                goto FLEE;
            }

    FLNW:   iPicValue = *(pPic - iSouthEast);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic -= iSouthEast;
                *pPic = 8;
                goto FLEE;
            }

    FLSW:   iPicValue = *(pPic + iSouthWest);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic += iSouthWest;
                *pPic = 9;
                goto FLEE;
            }

    FLNE:   iPicValue = *(pPic - iSouthWest);
            if ((iPicValue == 1) || (iPicValue == 2))
            {
                pPic -= iSouthWest;
                *pPic = 10;
                goto FLEE;
            }

    BACK:   switch (*pPic)
            {
            case 3:     *pPic-- = m_nLabels;
                        goto FLWW;

            case 4:     *pPic++ = m_nLabels;
                        goto FLSS;

            case 5:     *pPic = m_nLabels;
                        pPic -= m_iBreadth;
                        goto FLNN;

            case 6:     *pPic = m_nLabels;
                        pPic += m_iBreadth;
                        goto FLSE;

            case 7:     *pPic = m_nLabels;
                        pPic -= iSouthEast;
                        goto FLNW;

            case 8:     *pPic = m_nLabels;
                        pPic += iSouthEast;
                        goto FLSW;

            case 9:     *pPic = m_nLabels;
                        pPic -= iSouthWest;
                        goto FLNE;

            case 10:    *pPic = m_nLabels;
                        pPic += iSouthWest;
                        goto BACK;

            case 11:    *pPic++ = m_nLabels;
                        break;
            }
        }
    }
}

/****************************************************************************
Function:    CleanupImage (private function)
Description: 
****************************************************************************/
void CBlobLabeler::CleanupImage(const unsigned int iPedestal)
{
    unsigned int * pPicDst = m_pPicture;
    const unsigned int * pPicSrc = m_pPicture + m_iOrigin;
    unsigned int iPicValue;

    const int iBump = m_iBreadth - m_iWidth;
    int iWidth;
    int iHeight = m_iHeight;
    while (iHeight--)
    {
        iWidth = m_iWidth;
        while (iWidth--)
        {
            if ((iPicValue = *pPicSrc++) <= iPedestal) *pPicDst++ = 0;
            else *pPicDst++ = iPicValue - iPedestal;
        }
        pPicSrc += iBump;
    }

    m_nLabels -= iPedestal;
}

/****************************************************************************
Function:    Run

Description: Image Processing routine which given a monochrome
             8-bit image with the following attributes:
             All seed pixels in the image are marked 2;
             All grow pixels in the image are marked 1;
             All other pixels (background) are marked 0;
             reports all blobs by creating:
             (1) m_pLabelImage which holds all labels from 1 onwards
                 for all foreground blob pixels; the background pixels
                 are all labeled 0.
             (2) m_nLabels = the number of foreground blobs.
Arguments:
    Input: The 8-bit monochrome image
        const int iWidth               :(I) Width of image.
        const int iHeight              :(I) Height of image.
        const unsigned char * pPixel   :(I) Image array.
        const bool b8ConnForeGnd       :(I) True if foreground is 8 connected.

Output: In m_iStatus:
             0  = Success.
             1  = Invalid input parameters.
             2  = Memory allocation failure.
             These are defined in the enum above.

Globals:     None
****************************************************************************/
int CBlobLabeler::Run(const int             iWidth,         // Width of image.
                      const int             iHeight,        // Height of image.
                      const unsigned char * pPixel,         // Starting address of image buffer.
                      const bool            b8ConnForeGnd)  // True if foreground is 8 connected.
{
    m_iStatus = OKAY;

    // Check irrecoverable input parameters
    if (!pPixel || (iWidth <= 0) || (iHeight <= 0))
    {
        m_iStatus = INVALID_INPUT_PARAMETERS;
        return m_iStatus;
    }

    // Set member variables
    m_iWidth = iWidth;
    m_iHeight = iHeight;
    m_iBreadth = m_iWidth + 2;
    m_iTall = m_iHeight + 2;
    m_iOrigin = m_iBreadth + 1;

    // Allocate and clear
    unsigned int iAllocSize = m_iBreadth * m_iTall;
    if (m_iAllocatedSize < iAllocSize)
    {
        Deallocate();
        m_pPicture = new unsigned int [iAllocSize];
        if (m_pPicture)
        {
            m_iWidth = iWidth;
            m_iHeight = iHeight;
            m_iBreadth = m_iWidth + 2;
            m_iTall = m_iHeight + 2;
            m_iOrigin = m_iBreadth + 1;
            m_iAllocatedSize = iAllocSize;
        }
        else
        {
            m_iStatus = MEMORY_ALLOCATION_FAILURE;
            return m_iStatus;
        }
    }
    memset(m_pPicture, 0, m_iAllocatedSize * sizeof(unsigned int));

    // Process
    CopyToPicture(pPixel);
    if (b8ConnForeGnd)
    {
        Blobber8();
        CleanupImage(11);
    }
    else
    {
        Blobber4();
        CleanupImage(7);
    }

    m_iBreadth = m_iWidth;
    m_iTall = m_iHeight;
    m_iOrigin = 0;

    return m_iStatus;
}

/****************************************************************************
Function:    UpdateLabels
Description: 
****************************************************************************/
void CBlobLabeler::UpdateLabels(unsigned int * pLabelOkay)
{
    if (!pLabelOkay) return;

    pLabelOkay[0] = 0;
    unsigned int iLabel;
    unsigned int nNewLabels = 0;
    for (iLabel = 1; iLabel <= m_nLabels; iLabel++)
    {
        if (pLabelOkay[iLabel])
        {
            nNewLabels++; 
            pLabelOkay[iLabel] = nNewLabels;
        }
    }

    if (nNewLabels == 0)
    {
        m_nLabels = 0;
        memset(m_pPicture, 0, m_iWidth * m_iHeight * sizeof(unsigned int));
    }
    else if (m_nLabels != nNewLabels)
    {
        m_nLabels = nNewLabels;
        unsigned int *pPic = m_pPicture;
        unsigned int iSize = m_iWidth * m_iHeight;
        while (iSize--)
        {
            iLabel = pLabelOkay[*pPic];
            *pPic++ = iLabel;
        }
    }
}
