/****************************************************************************
File:        BlobLabeler.h
Description: Include file for the image processing function to perform
             connected component analysis and labeling.
Created:     01/27/05  SBanerjee

Copyright (c) 1996-2006, Saibal Banerjee, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
****************************************************************************/

#ifndef BLOBLABELER_H
#define BLOBLABELER_H

class CBlobLabeler
{
public:
    /****************************************************************************
    Function:    Default Constructor
    Arguments:   None
    Globals:     None
    ****************************************************************************/
    CBlobLabeler()
        : m_iWidth(0)
        , m_iHeight(0)
        , m_pPicture(0)
        , m_nLabels(0)
        , m_iAllocatedSize(0)
    {};

    /****************************************************************************
    Function:    Run

    Description: Image Processing routine which given a monochrome
                 8-bit image with the following attributes:
                 All seed pixels in the image are marked 2;
                 All grow pixels in the image are marked 1;
                 All other pixels (background) are marked 0;
                 reports all blobs by creating:
                 (1) m_pPicture which holds all labels from 1 onwards
                     for all foreground blob pixels; the background
                     pixels are all labeled 0.
                 (2) m_nLabels = the number of foreground blobs.
    Arguments:
        Input: The 8-bit monochrome image
            const int iWidth               :(I) Width of image.
            const int iHeight              :(I) Height of image.
            const unsigned char * pPixel   :(I) Image array.
            const bool b8ConnForeGnd       :(I) True if foreground is 8 connected.


    Output: In m_iStatus:
                 0  = Success.
                 1  = Invalid input parameters.
                 2  = Memory allocation failure.
                 These are defined in ErrorCodes.h.

    Globals:     None
    ****************************************************************************/
    int Run(const int             iWidth,         // Width of image.
            const int             iHeight,        // Height of image.
            const unsigned char * pPixel,         // Starting address of image buffer.
            const bool            b8ConnForeGnd); // True if foreground is 8 connected.

    /****************************************************************************
    Function:    Destructor
    Description: Deallocates and resets all member variables.
    ****************************************************************************/
    virtual ~CBlobLabeler();

    /****************************************************************************
    Function:    UpdateLabels
    Description: 
    ****************************************************************************/
    void UpdateLabels(unsigned int * pLabelOkay);

    /****************************************************************************
    Accessor Functions
    ****************************************************************************/
    int            GetWidth()   const;
    int            GetHeight()  const;
    unsigned int * GetImage()   const;
    unsigned int   GetNLabels() const;

protected:
    int            m_iWidth;
    int            m_iHeight;
    unsigned int * m_pPicture;
    unsigned int   m_nLabels;

private:
    unsigned int   m_iAllocatedSize;
    int            m_iBreadth;
    int            m_iTall;
    int            m_iOrigin;
    int            m_iStatus;

    /****************************************************************************
    Function:    Deallocate (private function)
    Description: Deallocates m_pPicture and sets it to NULL.
                 Resets all other member variables
    ****************************************************************************/
    void Deallocate();

    /****************************************************************************
    Function:    CopyToPicture (private function)
    Description: 
    ****************************************************************************/
    void CopyToPicture(const unsigned char * pPixel);

    /****************************************************************************
    Function:    Blobber4 (private function)
    Description: 4-connected component analysis and labeling routine.
                 Needs a border of pixels marked 0.
                 Needs seed pixels marked 2.
                 Needs grow pixels marked 1.
                 The Blobber() routine labels all foreground blobs from 1
                 onwards.
                 This routine uses Depth First Search from a seed pixel to
                 grow into other seed and grow pixels. The stack that it
                 requires to backtrack to parents are stored during the
                 forward traversal in the array itself.
    Arguments:   None
    Returns:     None
    Globals:     None
    WARNING:     This routine contains "goto" statements in order to minimize
                 execution time. The readability of the code is not compromised.
    ****************************************************************************/
    void Blobber4();

    /****************************************************************************
    Function:    Blobber8 (private function)
    Description: 8-connected component analysis and labeling routine.
                 Needs a border of pixels marked 0.
                 Needs seed pixels marked 2.
                 Needs grow pixels marked 1.
                 The Blobber() routine labels all foreground blobs from 1
                 onwards.
                 This routine uses Depth First Search from a seed pixel to
                 grow into other seed and grow pixels. The stack that it
                 requires to backtrack to parents are stored during the
                 forward traversal in the array itself.
    Arguments:   None
    Returns:     None
    Globals:     None
    WARNING:     This routine contains "goto" statements in order to minimize
                 execution time. The readability of the code is not compromised.
    ****************************************************************************/
    void Blobber8();

    /****************************************************************************
    Function:    CleanupImage (private function)
    Description: 
    ****************************************************************************/
    void CleanupImage(const unsigned int iPedestal);
};

/* ******* */
/* Inlines */
/* ******* */
inline int            CBlobLabeler::GetWidth()   const {return m_iWidth;}
inline int            CBlobLabeler::GetHeight()  const {return m_iHeight;}
inline unsigned int * CBlobLabeler::GetImage()   const {return m_pPicture;}
inline unsigned int   CBlobLabeler::GetNLabels() const {return m_nLabels;}

#endif // BLOBLABELER_H
