/****************************************************************************
File:        Filterer.cpp
Description: Implementation for filtering using Mexican hats.
Created:     02/09/07  Saibal Banerjee

Copyright (c) 2007-2017, Illumina, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
****************************************************************************/
#include "Convoluter.h"
#include "Filterer.h"

/***************************************************************************
Function:
Description:
Arguments:
Returns:
Globals:     none
***************************************************************************/
int CFilterer::Run(const float            fSigma,
                   const int              iPicWidth,
                   const int              iPicHeight,
                   const unsigned short * pPicBuffer)
{
    m_iStatus = OKAY;
    /*******************************************
          Set up the Sombrero if necessary
    *******************************************/
    if (fSigma <= 0)
    {
        m_iStatus = INVALID_INPUT_PARAMETERS;
        return m_iStatus;
    }
    else if (!m_pSombrero)
    {
        m_pSombrero = new CSombrero (fSigma);
        if (!m_pSombrero || (m_pSombrero->GetArraySize() == 0))
        {
            m_iStatus = MEMORY_ALLOCATION_FAILURE;
            return m_iStatus;
        }
    }
    else if (fabs((double) (fSigma - m_pSombrero->GetFilterSigma())) > 0.01)
    {
        delete m_pSombrero;
        m_pSombrero = 0;

        m_pSombrero = new CSombrero (fSigma);
        if (!m_pSombrero || (m_pSombrero->GetArraySize() == 0))
        {
            m_iStatus = MEMORY_ALLOCATION_FAILURE;
            return m_iStatus;
        }
    }

    /*******************************************
          Allocate for run if necessary
    *******************************************/
    int iFilterSize = m_pSombrero->GetArraySize();

    m_Picture.Create(iPicWidth + iFilterSize - 1,
                     iPicHeight + iFilterSize - 1);
    if ((m_iStatus = m_Picture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }
    m_Picture.CenterCopy(iPicWidth, iPicHeight, pPicBuffer);

    m_SubPicture = m_Picture;
    if ((m_iStatus = m_SubPicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }

    /*******************************************
          Convolve with sombrero filter
    *******************************************/
    CConvoluter::Run(m_Picture.GetWidth(), m_Picture.GetHeight(),
                     m_Picture.GetImage(),
                     m_pSombrero->GetAddArray(), iFilterSize,
                     m_pSombrero->GetAddArray(), iFilterSize);
    m_Picture.Cut(iFilterSize - 1, 0, iFilterSize - 1, 0);

    CConvoluter::Run(m_SubPicture.GetWidth(), m_SubPicture.GetHeight(),
                     m_SubPicture.GetImage(),
                     m_pSombrero->GetSubArray(), iFilterSize,
                     m_pSombrero->GetSubArray(), iFilterSize);
    m_SubPicture.Cut(iFilterSize - 1, 0, iFilterSize - 1, 0);

    /*******************************************
      Compute the result
    *******************************************/
    m_Picture -= m_SubPicture;
    return m_iStatus;
}

/***************************************************************************
Function:
Description:
Arguments:
Returns:
Globals:     none
***************************************************************************/
CFilterer::~CFilterer()
{
    if (m_pSombrero)
    {
        delete m_pSombrero;
        m_pSombrero = 0;
    }
}

//////////////////////////////////////////////////////////////////////////////////
