/****************************************************************************
File:        GaussianMaker.cpp
Description: Implementation file for Gaussian filter maker.
Created:     12/17/03  SBanerjee

Copyright (c) 1996-2006, Saibal Banerjee, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
****************************************************************************/
#if defined(__BORLANDC__)
#include <math.h>
#include <float.h>
#else
#include <cmath>
#include <cfloat>
#endif
#include "GaussianMaker.h"

const double CGaussianMaker::msc_dCeilAdd = 1.0 - FLT_EPSILON;

// Copy Constructor
CGaussianMaker::CGaussianMaker(const CGaussianMaker & source)
    : m_dSigma(0)
    , m_dXCenter(0)
    , m_dYCenter(0)
    , m_dIntensity(0)
{
    *this = source;
}

// Assignment: from CGaussianMaker to CGaussianMaker
CGaussianMaker & CGaussianMaker::operator = (const CGaussianMaker & source)
{
    if (this == &source) return *this;

    m_dSigma = source.m_dSigma;
    m_dXCenter = source.m_dXCenter;
    m_dYCenter = source.m_dYCenter;
    m_dIntensity = source.m_dIntensity;
    return *this;
}

// Set Function: Modifier
void CGaussianMaker::Set(const float fSigma,
                         const float fXCenter,
                         const float fYCenter,
                         const float fIntensity)
{
    m_dSigma = fSigma;
    m_dXCenter = fXCenter;
    m_dYCenter = fYCenter;
    m_dIntensity = fIntensity;
}

// GetValue Function
double CGaussianMaker::GetValue( const int iX,
                                 const int iY ) const
{
    if ((m_dSigma <= 0) || (m_dIntensity <= 0)) return 0;

    double dX = iX - m_dXCenter;
    if (dX < 0) dX = -dX;
    dX /= m_dSigma;

    if (dX > 3.0) return 0;

    double dY = iY - m_dYCenter;
    if (dY < 0) dY = -dY;
    dY /= m_dSigma;

    if (dY > 3.0) return 0;

    return (m_dIntensity * exp(-0.5 * (dX * dX + dY * dY)));
}

// GetROI Function
void CGaussianMaker::GetROI(int & iROIXStart,
                            int & iROIYStart,
                            int & iROIWidth,
                            int & iROIHeight) const
{
    double dValue;

    dValue = m_dXCenter - 3.0 * m_dSigma;
    iROIXStart = (int) (dValue + msc_dCeilAdd);

    dValue = m_dYCenter - 3.0 * m_dSigma;
    iROIYStart = (int) (dValue + msc_dCeilAdd);

    dValue = m_dXCenter + 3.0 * m_dSigma;
    iROIWidth = (int) dValue;
    iROIWidth = iROIWidth - iROIXStart + 1;

    dValue = m_dYCenter + 3.0 * m_dSigma;
    iROIHeight = (int) dValue;
    iROIHeight = iROIHeight - iROIYStart + 1;
}

// GetCenter Function
void CGaussianMaker::GetCenter(float & fXCenter,
                               float & fYCenter) const
{
    fXCenter = (float) m_dXCenter;
    fYCenter = (float) m_dYCenter;
}
