/**********************************************************************************
File:        Matcher.cpp
Description: Implementation file for the CMatcher class. 
Created:     03/01/07  SBanerjee

Copyright (c) 2006-2010, Solexa Inc, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
**********************************************************************************/
#include "Convoluter.h"
#include "Matcher.h"

/***************************************************************************
Function:
Description:
Arguments:
Returns:
Globals:     none
***************************************************************************/
void CMatcher::SetCorrelationThreshold(const float fCorrThresh)
{
    m_fCorrThresh = fCorrThresh;
}

/***************************************************************************
Function:
Description:
Arguments:
Returns:
Globals:     none
***************************************************************************/
int CMatcher::Run(const CGaussianFilter & filter,
                  const CShortImage     & picture)
{
    m_iStatus = OKAY;

    const int iPicWidth = picture.GetWidth();
    const int iPicHeight = picture.GetHeight();
    const unsigned short * pPicBuffer = picture.GetImage();

    CrossCorrelate(filter, iPicWidth, iPicHeight, pPicBuffer);
    if (m_iStatus != OKAY) return m_iStatus;

    NonMaxSuppress(iPicWidth, iPicHeight, pPicBuffer);
    if (m_iStatus != OKAY) return m_iStatus;

    return m_iStatus;
}

/***************************************************************************
Function:
Description:
Arguments:
Returns:
Globals:     none
***************************************************************************/
int CMatcher::CrossCorrelate(const CGaussianFilter & filter,
                             const int               iPicWidth,
                             const int               iPicHeight,
                             const unsigned short *  pPicBuffer)
{
    m_iStatus = OKAY;
    int iFilterSize = filter.GetArraySize();

    /*****************************************************************************
    MEMORY ALLOCATION and INITIALIZATION Part I
    ******************************************************************************/
    m_MatchPicture.Create(iPicWidth + iFilterSize - 1,
                          iPicHeight + iFilterSize - 1);
    if ((m_iStatus = m_MatchPicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }
    m_MatchPicture.CenterCopy(iPicWidth, iPicHeight, pPicBuffer);

    m_SumPicture = m_MatchPicture;
    if ((m_iStatus = m_SumPicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }

    double dMean = m_SumPicture.GetAvg();

    m_SumOfSqPicture = m_MatchPicture;
    if ((m_iStatus = m_SumOfSqPicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }
    m_SumOfSqPicture *= m_MatchPicture;

    float fGlobalVariance = (float) (m_SumOfSqPicture.GetAvg() - dMean * dMean);

    /*****************************************************************************
    PRELIMINARY PROCESSING
    ******************************************************************************/
    CConvoluter::Run(m_MatchPicture.GetWidth(), m_MatchPicture.GetHeight(),
                     m_MatchPicture.GetImage(),
                     filter.GetBlurArray(), iFilterSize,
                     filter.GetBlurArray(), iFilterSize);
    m_MatchPicture.Cut(iFilterSize - 1, 0, iFilterSize - 1, 0);

    m_SumPicture.BoxSum(iFilterSize,
                        iFilterSize,
                        true);

    m_SumOfSqPicture.BoxSum(iFilterSize,
                            iFilterSize,
                            true);

    /*****************************************************************************
    MEMORY ALLOCATION and INITIALIZATION Part II
    ******************************************************************************/
    m_ForePicture.Create(m_MatchPicture.GetWidth(), m_MatchPicture.GetHeight());
    if ((m_iStatus = m_ForePicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }

    m_ScratchPicture = m_SumPicture;
    if ((m_iStatus = m_ScratchPicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }

    /*****************************************************************************
    FINAL PROCESSING
    ******************************************************************************/
    // Proc of Numerator --- kept in *m_MatchPicture
    float fFilterMean = 1.0f / (iFilterSize * iFilterSize);
    m_ScratchPicture *= fFilterMean;
    m_MatchPicture -= m_ScratchPicture;

    // Proc of Denominator --- kept in m_SumOfSqPicture
    m_ScratchPicture = m_SumPicture;
    m_ScratchPicture *= m_SumPicture;
    m_ScratchPicture *= 1.0f / (iFilterSize * iFilterSize);
    m_SumOfSqPicture -= m_ScratchPicture;

    // Scalar divisor
    float fDivisor = filter.Get2DBlurSumOfSquares();
    fDivisor -= fFilterMean;

    // Comparer
    const float fComparer = fDivisor * m_fCorrThresh * m_fCorrThresh;

    // Final division --- no need to divide, Simply Binarize
    unsigned char * pDest = m_ForePicture.GetImage();
    float * pNumerator = m_MatchPicture.GetImage();
    float * pDenominator = m_SumOfSqPicture.GetImage();
    float fDenValue, fNumValue, fSqNumValue;
    int iSize = m_ForePicture.GetWidth() * m_ForePicture.GetHeight();

    while (iSize--)
    {
        fDenValue = *pDenominator++;
        fNumValue = *pNumerator;
        if ((fDenValue > fGlobalVariance) && ((fSqNumValue = fNumValue * fNumValue) > fDenValue * fComparer))
        {
            *pDest++ = 1;
            *pNumerator++ = fSqNumValue / (fDenValue * fDivisor);
        }
        else
        {
            *pDest++ = 0;
            *pNumerator++ = 0;
        }
    }

    return m_iStatus;
}

/***************************************************************************
Function:
Description:
Arguments:
Returns:
Globals:     none
***************************************************************************/
int CMatcher::NonMaxSuppress(const int              iPicWidth,
                             const int              iPicHeight,
                             const unsigned short * pPicBuffer)
{
    m_iStatus = OKAY;
    const int iKernelSize = 3;

    /*****************************************************************************
    MEMORY ALLOCATION and INITIALIZATION Part I
    ******************************************************************************/
    m_DilatePicture.Create(iPicWidth + iKernelSize - 1,
                           iPicHeight + iKernelSize - 1);
    if ((m_iStatus = m_DilatePicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }
    m_DilatePicture.CenterCopy(iPicWidth, iPicHeight, pPicBuffer);

    /*****************************************************************************
    PRELIMINARY PROCESSING
    ******************************************************************************/
    m_DilatePicture.Morph(iKernelSize, iKernelSize, true, true);

    /*****************************************************************************
    MEMORY ALLOCATION and INITIALIZATION Part II
    ******************************************************************************/
    m_PeakPicture.Create(m_DilatePicture.GetWidth(), m_DilatePicture.GetHeight());
    if ((m_iStatus = m_PeakPicture.GetStatus()) != OKAY)
    {
        return m_iStatus;
    }

    /*****************************************************************************
    FINAL PROCESSING
    ******************************************************************************/
    unsigned char * pDest = m_PeakPicture.GetImage();
    const unsigned short * pSrcOrg = pPicBuffer;
    unsigned short * pSrcDil = m_DilatePicture.GetImage();

    unsigned short usOrgValue, usDilValue;
    int iSize = m_PeakPicture.GetWidth() * m_PeakPicture.GetHeight();
    while (iSize--)
    {
        usOrgValue = *pSrcOrg++;
        usDilValue = *pSrcDil++;
        *pDest++ = (unsigned char) ((usOrgValue > 0) && (usDilValue == usOrgValue)) ? 1 : 0;
    }

    return m_iStatus;
}
