/**********************************************************************************
File:        ObjectDetector.cpp
Description: Implementation file for the CObjectDetector class. 
Created:     08/07/06  SBanerjee

Copyright (c) 2006-2010, Solexa Inc, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
**********************************************************************************/
#include "ObjectDetector.h"

/***********************************************************************************************
Function:    Compute
Description: Performs focus metric related calculations on an image.
Returns:     None
Arguments:   Input: Pointer to a Gaussian Filter.
             const CGaussianFilter * pFilter     :(I) Ptr to filter.
             Input: A Single 16-bit monochrome image.
             const int iImageWidth               :(I) Width of image.
             const int iImageHeight              :(I) Height of image.
             const unsigned short * pImageBuffer :(I) Pointer to the image buffer.
Globals:     None
***********************************************************************************************/
int CObjectDetector::Compute(const CGaussianFilter * pFilter,
                             const float             fCorrelationThreshold,
                             const int               iImageWidth,
                             const int               iImageHeight,
                             const unsigned short *  pImageBuffer)
{
    m_iStatus = OKAY;

    /*****************************************************************************
    INPUT CHECK
    ******************************************************************************/
    if (!pFilter ||
        (pFilter->GetArraySize() == 0) ||
        (iImageWidth <= 0) ||
        (iImageHeight <= 0) ||
        !pImageBuffer)
    {
        m_iStatus = INVALID_INPUT_PARAMETERS;
        return m_iStatus;
    }

    /*****************************************************************************
    MEMORY ALLOCATION and INITIALIZATION for m_InPicture
    ******************************************************************************/
    m_InPicture.Create(iImageWidth, iImageHeight);
    if ((m_iStatus = m_InPicture.GetStatus()) != OKAY) return m_iStatus;

    unsigned short * pDstBuffer = m_InPicture.GetImage();
    unsigned int iImageSize = iImageWidth * iImageHeight;
    memcpy(pDstBuffer, pImageBuffer, iImageSize * sizeof( unsigned short ));

    /*****************************************************************************
    PROCESSING
    ******************************************************************************/
    m_ProcPicture = m_InPicture;
    if ((m_iStatus = m_ProcPicture.GetStatus()) != OKAY) return m_iStatus;

    m_Matcher.SetCorrelationThreshold(fCorrelationThreshold);

    m_iStatus = m_Matcher.CrossCorrelate(*pFilter, iImageWidth, iImageHeight, m_ProcPicture.GetImage());
    if (m_iStatus != OKAY) return m_iStatus;

    float fSigma = pFilter->GetFilterSigma();
    int iHalfFiltSize;
    const float * pSource;
    float fValue;
    for (;;)
    {
        iHalfFiltSize = (int) (3.0 * sqrt(2.0) * fSigma + 0.5);
        // iHalfFiltSize = (int) (3.0 * fSigma + 0.5);
        if (iHalfFiltSize == 0) break;

        m_iStatus = m_Filterer.Run(fSigma, iImageWidth, iImageHeight, m_ProcPicture.GetImage());
        if (m_iStatus != OKAY) return m_iStatus;

        pDstBuffer = m_ProcPicture.GetImage();
        pSource = m_Filterer.GetPicture().GetImage();
        iImageSize = iImageWidth * iImageHeight;
        while (iImageSize--)
        {
            fValue = *pSource++;
            if (fValue < 0) *pDstBuffer++ = 0;
            else if (fValue > 65535.0f) *pDstBuffer++ = 65535;
            else *pDstBuffer++ = (unsigned short) (fValue + 0.5f);
        }

        fSigma /= 2.0f;
    }

    m_iStatus = m_Matcher.NonMaxSuppress(iImageWidth, iImageHeight, m_ProcPicture.GetImage());
    if (m_iStatus != OKAY) return m_iStatus;

    m_SymbolPicture = m_Matcher.GetPeakPicture();
    if ((m_iStatus = m_SymbolPicture.GetStatus()) != OKAY) return m_iStatus;

    m_SymbolPicture += m_Matcher.GetForePicture();

    unsigned char * pDest = m_SymbolPicture.GetImage();
    iImageSize = iImageWidth * iImageHeight;
    while (iImageSize--)
    {
        if (*pDest < 2) *pDest = 0;
        pDest++;
    }

    m_iStatus = m_Labeler.Run(iImageWidth, iImageHeight, m_SymbolPicture.GetImage(), false);
    if (m_iStatus != OKAY) return m_iStatus;

    /*****************************************************************************
    FX
    ******************************************************************************/
    const int nFltFeatures = 3;
    const int nIntFeatures = 1;

    m_FltFeatures.Create(m_Labeler.GetNLabels(), nFltFeatures);
    if ((m_iStatus = m_FltFeatures.GetStatus()) != OKAY) return m_iStatus;

    m_IntFeatures.Create(m_Labeler.GetNLabels(), nIntFeatures);
    if ((m_iStatus = m_IntFeatures.GetStatus()) != OKAY) return m_iStatus;

    featureExtract();

    return m_iStatus;
}

static void interpolate( int pt, int imppl, unsigned short *viz, float &fDelta, float &fValue )
{
    double x[3];
    double a, b, tmp, val, del;

    /*    Forward operations    */
    x[0] = *(viz + pt - imppl);
    x[1] = *(viz + pt);
    x[2] = *(viz + pt + imppl);

    x[0] = log(x[0] + 1);
    x[1] = log(x[1] + 1);
    x[2] = log(x[2] + 1);
    /* ======================== */

    del = 0;
    val = x[1];

	a = x[1] - x[0];
	b = x[2] - x[1];
    if (a > b)
    {
        del = (a + b) / (2.0 * (a - b));
        if (del > 1)
        {
            tmp = x[2] + x[1];
            if (tmp > 0)
            {
                del = x[2] / tmp;
                val = del * b + x[1];
            }
        }
        else if (del < -1)
        {
            tmp = x[1] + x[0];
            if (tmp > 0)
            {
                del = -x[0] / tmp;
                val = del * a + x[1];
            }
        }
        else
        {
            val = del * (a + b) / 4.0 + x[1];
        }
    }

    /*    Reverse operations    */
    val = exp(val) - 1;
    /* ======================== */

    fValue = (float) val;
	fDelta = (float) del;
}

void CObjectDetector::featureExtract()
{
    m_FltFeatures.SetClear();
    m_IntFeatures.SetClear();

    const unsigned int nLabels = m_Labeler.GetNLabels();

    float * pXs = m_FltFeatures.GetImage();
    float * pYs = pXs + nLabels;
    float * pGs = pYs + nLabels;

    unsigned int * pAs = m_IntFeatures.GetImage();

    const unsigned int iWidth = m_Labeler.GetWidth();
    const unsigned int iHeight = m_Labeler.GetHeight();
    unsigned int iX, iY, iPt, iLabel, iGray, iNumValue;
    float fValue, fDelta, fSumValue;

    unsigned int * pLabelBuf = m_Labeler.GetImage();
    unsigned short * pInBuf = m_InPicture.GetImage();

    for (iY = 0; iY < iHeight; iY++)
    {
        for (iX = 0; iX < iWidth; iX++)
        {
            iLabel = *pLabelBuf++;
            iGray = *pInBuf++;
            if (iLabel)
            {
                pXs[--iLabel] += iX * iGray;
                pYs[iLabel] += iY * iGray;
                pGs[iLabel] += iGray;
                pAs[iLabel]++;
            }
        }
    }

    for (iLabel = 0; iLabel < nLabels; iLabel++)
    {
        pXs[iLabel] /= pGs[iLabel];
        pYs[iLabel] /= pGs[iLabel];
        pGs[iLabel] /= pAs[iLabel];
    }

    pInBuf = m_InPicture.GetImage();
    for (iLabel = 0; iLabel < nLabels; iLabel++)
    {
        iY = (unsigned int) (pYs[iLabel] + 0.5f);
        iX = (unsigned int) (pXs[iLabel] + 0.5f);
        iPt = iY * iWidth + iX;

        fSumValue = 0;
        iNumValue = 0;

        if ((iY > 0) && (iY < (iHeight - 1)))
        {
            interpolate(iPt, iWidth, pInBuf, fDelta, fValue);
            pYs[iLabel] = iY + fDelta;
            fSumValue += fValue;
            iNumValue++;
        }

        if ((iX > 0) && (iX < (iWidth - 1)))
        {
            interpolate(iPt, 1, pInBuf, fDelta, fValue);
            pXs[iLabel] = iX + fDelta;
            fSumValue += fValue;
            iNumValue++;
        }

        if (iNumValue > 0)
        {
            pGs[iLabel] = fSumValue / iNumValue;
        }
    }
}

void CObjectDetector::ApplyTransform(const int nFeatIndex,
                                     const float fSlope,
                                     const float fAbscissa)
{
    const unsigned int nLabels = m_Labeler.GetNLabels();
    float * pCoords = m_FltFeatures.GetImage() + nLabels * nFeatIndex;
    unsigned int iLabel;
    for (iLabel = 0; iLabel < nLabels; iLabel++)
    {
        pCoords[iLabel] = fSlope * pCoords[iLabel] + fAbscissa;
    }
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////
