/**********************************************************************************
File:        Registrationer.cpp
Description: Implementation file for the CRegistrationer class. 
Created:     03/01/07  SBanerjee

Copyright (c) 2006-2010, Solexa Inc, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
**********************************************************************************/
#if defined(__BORLANDC__)
#include <math.h>
#include <float.h>
#else
#include <cmath>
#include <cfloat>
#endif
#include "Registrationer.h"

const int CRegistrationer::msc_iWidth = 128;
const int CRegistrationer::msc_iHeight = 128;
const int CRegistrationer::m_iXDefaultUncertainty = 3;
const int CRegistrationer::m_iYDefaultUncertainty = 3;

/***************************************************************************
Function:
Description:
Arguments:
Returns:
Globals:     none
***************************************************************************/
int CRegistrationer::Run(const CFloatImage & fixPicture,
                         const CFloatImage & movPicture,
                         const int           iXPriorMove,
                         const int           iYPriorMove,
                         const int           iXMaxUncertainty,
                         const int           iYMaxUncertainty)
{
    double           aXTrans[9], aXOrig[9];
    double           aYTrans[9], aYOrig[9];
    double           aMetric[9];

    m_iStatus = OKAY;

    m_iXPriorMove = iXPriorMove;
    m_iYPriorMove = iYPriorMove;
    m_iXUncertainty = iXMaxUncertainty;
    m_iYUncertainty = iYMaxUncertainty;

    compute(fixPicture, movPicture, fixPicture.GetWidth() / 2, fixPicture.GetHeight() / 2);
    if (m_iStatus != OKAY) return m_iStatus;

    aXTrans[4] = m_iXPriorMove + m_dXDelta;
    aXOrig[4] = fixPicture.GetWidth() / 2;
    aYTrans[4] = m_iYPriorMove + m_dYDelta;
    aYOrig[4] = fixPicture.GetHeight() / 2;
    aMetric[4] = m_dMeasure;

    m_iXPriorMove = (aXTrans[4] < 0) ? (int) (aXTrans[4] - 0.5) : (int) (aXTrans[4] + 0.5);
    m_iYPriorMove = (aYTrans[4] < 0) ? (int) (aYTrans[4] - 0.5) : (int) (aYTrans[4] + 0.5);
    m_iXUncertainty = (m_iXDefaultUncertainty < iXMaxUncertainty) ? m_iXDefaultUncertainty : iXMaxUncertainty;
    m_iYUncertainty = (m_iYDefaultUncertainty < iYMaxUncertainty) ? m_iYDefaultUncertainty : iYMaxUncertainty;

    int kIndex = 0;
    int iXIndex, iYIndex;
    for (iYIndex = 1; iYIndex <= 3; iYIndex++)
    {
        for (iXIndex = 1; iXIndex <= 3; iXIndex++)
        {
            if (kIndex != 4)
            {
                compute(fixPicture, movPicture, fixPicture.GetWidth() * iXIndex / 4, fixPicture.GetHeight() * iYIndex / 4);
                if (m_iStatus != OKAY) return m_iStatus;

                aXTrans[kIndex] = m_iXPriorMove + m_dXDelta;
                aXOrig[kIndex] = fixPicture.GetWidth() * iXIndex / 4;
                aYTrans[kIndex] = m_iYPriorMove + m_dYDelta;
                aYOrig[kIndex] = fixPicture.GetHeight() * iYIndex / 4;
                aMetric[kIndex] = m_dMeasure;
            }
            aXTrans[kIndex] += aXOrig[kIndex];
            aYTrans[kIndex] += aYOrig[kIndex];
            kIndex++;
        }
    }

    getStatLinearCoeffs(9, aXOrig, aXTrans, m_fXSlope, m_fXAbscissa);
    getStatLinearCoeffs(9, aYOrig, aYTrans, m_fYSlope, m_fYAbscissa);

    return m_iStatus;
}

void CRegistrationer::compute(const CFloatImage & fixPicture,
                              const CFloatImage & movPicture,
                              const int           iXROICenter,
                              const int           iYROICenter)
{
    /*****************************************************************************
    MEMORY ALLOCATION and INITIALIZATION
    ******************************************************************************/
    int nImppl = fixPicture.GetWidth();

    int iOffset = iYROICenter - msc_iHeight / 2;
    iOffset *= nImppl;
    iOffset += iXROICenter - msc_iWidth / 2;

    copyToUPicture(fixPicture.GetImage() + iOffset, nImppl);
    if (m_iStatus != OKAY) return;

    iOffset -= m_iYPriorMove * nImppl + m_iXPriorMove;
    copyToVPicture(movPicture.GetImage() + iOffset, nImppl);
    if (m_iStatus != OKAY) return;

    m_Kernel.Create(2 * m_iXUncertainty + 1, 2 * m_iYUncertainty + 1);
    if ((m_iStatus = m_Kernel.GetStatus()) != OKAY) return;

    /*****************************************************************************
    PROCESSING
    ******************************************************************************/
    double * pKerBuf = m_Kernel.GetImage();
    const float * pUBuf = m_UPicture.GetImage();
    const int iUBump = msc_iWidth - m_Kernel.GetWidth();
    int iKerWd;
    int iKerHt = m_Kernel.GetHeight();
    while (iKerHt--)
    {
        iKerWd = m_Kernel.GetWidth();
        while (iKerWd--)
        {
            *pKerBuf++ = getMatch(pUBuf++);
        }
        pUBuf += iUBump;
    }

    getDeltas();
}

void CRegistrationer::copyToUPicture(const float * pSource,
                                     const int     nImppl)
{
    m_UPicture.Create(msc_iWidth, msc_iHeight);
    if ((m_iStatus = m_UPicture.GetStatus()) != OKAY)
    {
        return;
    }
    
    float * pDst = m_UPicture.GetImage();
    int iHt = msc_iHeight;
    while (iHt--)
    {
        memcpy(pDst, pSource, msc_iWidth * sizeof(float));
        pDst += msc_iWidth;
        pSource += nImppl;
    }
}

void CRegistrationer::copyToVPicture(const float * pSource,
                                     const int     nImppl)
{
    const int iPicWidth = msc_iWidth - 2 * m_iXUncertainty;
    const int iPicHeight = msc_iHeight - 2 * m_iYUncertainty;

    m_VPicture.Create(iPicWidth, iPicHeight);
    if ((m_iStatus = m_VPicture.GetStatus()) != OKAY)
    {
        return;
    }
    
    float * pDst = m_VPicture.GetImage();
    pSource += nImppl * m_iYUncertainty + m_iXUncertainty;
    int iHt = iPicHeight;
    while (iHt--)
    {
        memcpy(pDst, pSource, iPicWidth * sizeof(float));
        pDst += iPicWidth;
        pSource += nImppl;
    }
}

double CRegistrationer::getMatch(const float * pUBuf)
{
    double dValue;
    double dSum = 0;
    const float * pVBuf = m_VPicture.GetImage();
    const int iPicWidth = m_VPicture.GetWidth();
    const int iBump = msc_iWidth - iPicWidth;
    int iWd;
    int iHt = m_VPicture.GetHeight();
    while (iHt--)
    {
        iWd = iPicWidth;
        while (iWd--)
        {
            dValue = (*pUBuf++) - (*pVBuf++);
            dSum += dValue * dValue;
        }
        pUBuf += iBump;
    }

    dSum /= iPicWidth * m_VPicture.GetHeight();
    return dSum;
}

void CRegistrationer::getDeltas()
{
    const int iKerWd = m_Kernel.GetWidth();
    const int iKerHt = m_Kernel.GetHeight();
    const int iKerSz = iKerWd * iKerHt;
    double dErrValue;
    double dMinErrValue = DBL_MAX;
    int iMinErrIndex = 0;
    double * pKerBuf = m_Kernel.GetImage();

    for (int iKerIdx = 0; iKerIdx < iKerSz; iKerIdx++)
    {
        if ((dErrValue = *pKerBuf++) < dMinErrValue)
        {
            dMinErrValue = dErrValue;
            iMinErrIndex = iKerIdx;
        }
    }

    // Parabolic interpolation if possible
    pKerBuf = m_Kernel.GetImage();

    double dKerSubY = 0.0;
    double dKerSubX = 0.0;
    int iKerSubSignY = 0;
    int iKerSubSignX = 0;

    int iKerY = iMinErrIndex / iKerWd;
    int iKerX = iMinErrIndex % iKerWd;
    if ((iKerY > 0) && (iKerY < (iKerHt - 1)))
    {
        dKerSubY = parabolicInterpolate(iMinErrIndex, iKerWd, pKerBuf);
        if (dKerSubY < 0.0)
        {
            iKerSubSignY = -1;
            dKerSubY = -dKerSubY;
        }
        else if (dKerSubY > 0.0)
        {
            iKerSubSignY = 1;
        }
    }
    if ((iKerX > 0) && (iKerX < (iKerWd - 1)))
    {
        dKerSubX = parabolicInterpolate(iMinErrIndex, 1, pKerBuf);
        if (dKerSubX < 0.0)
        {
            iKerSubSignX = -1;
            dKerSubX = -dKerSubX;
        }
        else if (dKerSubX > 0.0)
        {
            iKerSubSignX = 1;
        }
    }

    m_dMeasure = (1.0 - dKerSubY) * (1.0 - dKerSubX) * pKerBuf[iMinErrIndex];
    m_dMeasure += (1.0 - dKerSubY) * dKerSubX * pKerBuf[iMinErrIndex + iKerSubSignX];
    m_dMeasure += (1.0 - dKerSubX) * dKerSubY * pKerBuf[iMinErrIndex + iKerSubSignY * iKerWd];
    m_dMeasure += dKerSubX * dKerSubY * pKerBuf[iMinErrIndex + iKerSubSignY * iKerWd + iKerSubSignX];

    m_dYDelta = iKerY - (iKerHt >> 1);
    m_dYDelta += iKerSubSignY * dKerSubY;

    m_dXDelta = iKerX - (iKerWd >> 1);
    m_dXDelta += iKerSubSignX * dKerSubX;
}

const double CRegistrationer::parabolicInterpolate(int      pt,
                                                   int      imppl,
                                                   double * viz) const
{
    double a, b, del;

    del = viz[pt];
	a = del - viz[pt - imppl];
	b = viz[pt + imppl] - del;
	del = a - b;
	if (fabs(del) > DBL_EPSILON) del = (a + b) / (2.0 * del);
	return del;
}

void CRegistrationer::getStatLinearCoeffs(const int n,
                                          const double * x,
                                          const double * y,
                                          float & fSlope,
                                          float & fAbscissa)
{
    int m = 0;
    double mx = 0, my = 0;
    double mxx = 0, mxy = 0, myy = 0;

    for (int index = 0; index < n; index++)
    {
        m++;
        mx += x[index];
        my += y[index];
        mxx += x[index] * x[index];
        mxy += x[index] * y[index];
        myy += y[index] * y[index];
    }

    mx /= m;
    my /= m;
    mxx /= m;
    mxy /= m;
    myy /= m;
    mxx -= mx * mx;
    mxy -= mx * my;
    myy -= my * my;

    // Find unit vector along major principal axis
    double tmp = (mxx - myy) / 2.0;
    double vx = tmp + sqrt(tmp * tmp + mxy * mxy);
    double vy = mxy;
    tmp = vx * vx + vy * vy;
    if (tmp > 0)
    {
        tmp = sqrt(tmp);
        vx /= tmp;
        vy /= tmp;
    }
    else
    {
        vx = 1.0;
        vy = 0.0;
    }

    fSlope = (float) (vy / vx);
    fAbscissa = (float) (my - (vy * mx / vx));
}
