/**********************************************************************************
File:        SmartImage.h
Description: Include file for the smart template images.
             All images are one band.
Created:     04/15/05  SBanerjee

Copyright (c) 1996-2006, Saibal Banerjee, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
**********************************************************************************/

#ifndef SMARTIMAGE_H
#define SMARTIMAGE_H

#if defined(__BORLANDC__)
#include <string.h>
#else
#include <cstring>
#endif
#include "ErrorCodes.h"

template<class T>
class CSmartImage
{
public:

    // Constructor
    CSmartImage()
        : m_iWidth(0)
        , m_iHeight(0)
        , m_pBuffer(0)
        , m_iStatus(OKAY)
        , m_iAllocatedSize(0)

    {};

    // Constructor: construct image given allocation parameters
    CSmartImage(const int iWidth,
                const int iHeight)
        : m_iWidth(0)
        , m_iHeight(0)
        , m_pBuffer(0)
        , m_iStatus(OKAY)
        , m_iAllocatedSize(0)
    {
        if ((iWidth <= 0) || (iHeight <= 0))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return;
        }

        allocate(iWidth, iHeight);
    }


    // Constructor: construct image given allocation parameters
    template<class TPrime>
    CSmartImage(const int iWidth,
                const int iHeight,
                const TPrime * pBuffer,
                const int iPixPerLine = 0)
        : m_iWidth(0)
        , m_iHeight(0)
        , m_pBuffer(0)
        , m_iStatus(OKAY)
        , m_iAllocatedSize(0)
    {
        if ((iWidth <= 0) || (iHeight <= 0) || (pBuffer == 0))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return;
        }

        allocate(iWidth, iHeight);

        if (m_iStatus != OKAY) return;

        if (iPixPerLine <= m_iWidth)
        {
            const TPrime * pSrc = pBuffer;
            T * pDst = m_pBuffer;
            int iSize = m_iWidth * m_iHeight;
            while (iSize--)
            {
                *pDst++ = (T)(*pSrc++);
            }
        }
        else
        {
            int iWd;
            const int iBump = iPixPerLine - m_iWidth;
            const TPrime * pSrc = pBuffer;
            T * pDst = m_pBuffer;
            int iHt = m_iHeight;
            while (iHt--)
            {
                iWd = m_iWidth;
                while (iWd--)
                {
                    *pDst++ = (T)(*pSrc++);
                }
                pSrc += iBump;
            }

        }
    }

    // Copy constructor: from CSmartImage <T> to CSmartImage<T>
    CSmartImage(const CSmartImage<T> & source)
        : m_iWidth(0)
        , m_iHeight(0)
        , m_pBuffer(0)
        , m_iStatus(OKAY)
        , m_iAllocatedSize(0)
    {
        *this = source;
    }

    // Assignment: from CSmartImage <T> to CSmartImage<T>
    CSmartImage & operator = (const CSmartImage<T> & source)
    {
        if (this == &source) return *this;

        if ((m_iWidth != source.m_iWidth) || (m_iHeight != source.m_iHeight))
        {
            allocate(source.GetWidth(), source.GetHeight());
            if (m_iStatus != OKAY) return *this;
        }

        memcpy(m_pBuffer, source.m_pBuffer, m_iWidth * m_iHeight * sizeof(T));

        m_iStatus = source.GetStatus();
        return *this;
    }

    void Create(const int iWidth,
                const int iHeight)
    {
        if ((iWidth <= 0) || (iHeight <= 0))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return;
        }

        allocate(iWidth, iHeight);
    }


    // CenterCopy: Used for copying a smaller to a larger image at the center of larger image
    //             Also Mirror pads the surrounding
    template<class TPrime>
    void CenterCopy(const int      iWidth,
                    const int      iHeight,
                    const TPrime * pBuffer)
    {
        if ((m_iWidth < iWidth) || (m_iHeight < iHeight) || (pBuffer == 0))
        {
            return;
        }

        // Copy
        const int iXOffset = (m_iWidth - iWidth) >> 1;
        const int iYOffset = (m_iHeight - iHeight) >> 1;
        T * pDst = m_pBuffer + iYOffset * m_iWidth + iXOffset;

        int iBump = m_iWidth - iWidth;

        int iWd;
        int iHt = iHeight;
        while (iHt--)
        {
            iWd = iWidth;
            while (iWd--)
            {
                *pDst++ = (T) *pBuffer++;
            }
            pDst += iBump;
        }

        // Mirror
        mirror(iXOffset, iXOffset, iYOffset, iYOffset);
    }

    // += Operator: from CSmartImage to CSmartImage
    CSmartImage & operator += (const T tValue)
    {
        if (m_iStatus != OKAY) return *this;

        int iSize = m_iWidth * m_iHeight;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            *pDst++ += tValue;
        }

        return *this;
    }

    // -= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator -= (const T tValue)
    {
        if (m_iStatus != OKAY) return *this;

        int iSize = m_iWidth * m_iHeight;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            *pDst++ -= tValue;
        }

        return *this;
    }

    // *= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator *= (const T tValue)
    {
        if (m_iStatus != OKAY) return *this;

        int iSize = m_iWidth * m_iHeight;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            *pDst++ *= tValue;
        }

        return *this;
    }

    // |= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator |= (const T tValue)
    {
        if (m_iStatus != OKAY) return *this;

        int iSize = m_iWidth * m_iHeight;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            if (tValue > *pDst) *pDst++ = tValue;
            else pDst++;
        }

        return *this;
    }

    // &= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator &= (const T tValue)
    {
        if (m_iStatus != OKAY) return *this;

        int iSize = m_iWidth * m_iHeight;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            if (tValue < *pDst) *pDst++ = tValue;
            else pDst++;
        }

        return *this;
    }

    // += Operator: from CSmartImage to CSmartImage
    CSmartImage & operator += (const CSmartImage<T> & source)
    {
        if (m_iStatus != OKAY) return *this;

        if ((m_iWidth!= source.m_iWidth) || (m_iHeight!= source.m_iHeight))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return *this;
        }

        int iSize = m_iWidth * m_iHeight;
        const T * pSrc = source.m_pBuffer;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            *pDst++ += *pSrc++;
        }

        m_iStatus = source.m_iStatus;
        return *this;
    }

    // -= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator -= (const CSmartImage<T> & source)
    {
        if (m_iStatus != OKAY) return *this;

        if ((m_iWidth!= source.m_iWidth) || (m_iHeight!= source.m_iHeight))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return *this;
        }

        int iSize = m_iWidth * m_iHeight;
        const T * pSrc = source.m_pBuffer;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            *pDst++ -= *pSrc++;
        }

        m_iStatus = source.m_iStatus;
        return *this;
    }

    // *= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator *= (const CSmartImage<T> & source)
    {
        if (m_iStatus != OKAY) return *this;

        if ((m_iWidth!= source.m_iWidth) || (m_iHeight!= source.m_iHeight))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return *this;
        }

        int iSize = m_iWidth * m_iHeight;
        const T * pSrc = source.m_pBuffer;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            *pDst++ *= *pSrc++;
        }

        m_iStatus = source.m_iStatus;
        return *this;
    }

    // |= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator |= (const CSmartImage<T> & source)
    {
        if (m_iStatus != OKAY) return *this;

        if ((m_iWidth!= source.m_iWidth) || (m_iHeight!= source.m_iHeight))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return *this;
        }

        int iSize = m_iWidth * m_iHeight;
        const T * pSrc = source.m_pBuffer;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            if (*pSrc > *pDst) *pDst++ = *pSrc++;
            else
            {
                pDst++;
                pSrc++;
            }
        }

        m_iStatus = source.m_iStatus;
        return *this;
    }

    // &= Operator: from CSmartImage to CSmartImage
    CSmartImage & operator &= (const CSmartImage<T> & source)
    {
        if (m_iStatus != OKAY) return *this;

        if ((m_iWidth!= source.m_iWidth) || (m_iHeight!= source.m_iHeight))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return *this;
        }

        int iSize = m_iWidth * m_iHeight;
        const T * pSrc = source.m_pBuffer;
        T * pDst = m_pBuffer;
        while (iSize--)
        {
            if (*pSrc < *pDst) *pDst++ = *pSrc++;
            else
            {
                pDst++;
                pSrc++;
            }
        }

        m_iStatus = source.m_iStatus;
        return *this;
    }

    // Destructor
    virtual ~CSmartImage()
    {
        deallocate();
        m_iStatus = OKAY;
    }

    // Accessors for member functions common to all image types
    int GetWidth()  const;
    int GetHeight() const;
    T * GetImage()  const;
    int GetStatus() const;

    // Modifiers common to all image types
    void SetClear() const;
    void SetAbs() const;

    // Functions
    void BoxSum(const int  iFilterWidth,
                const int  iFilterHeight,
                const bool bCut = false);

    void Morph(const int  iFilterWidth,
               const int  iFilterHeight,
               const bool bIsDilate,
               const bool bCut = false);

    T GetMax() const;
    T GetMin() const;
    double GetAvg() const;
    double GetStdev() const;
    double Dot(const CSmartImage<T> & source) const;
    void Clip();
    void Sqrt();

    void Cut(const int iLeftBorder,
             const int iRightBorder,
             const int iTopBorder,
             const int iBottomBorder);

protected:
    int             m_iWidth;     // image width
    int             m_iHeight;    // image height
    T *             m_pBuffer;    // image buffer starting address
    int             m_iStatus;    // image processing error status

    // allocate function: allocate only if necessary, otherwise reuse space.
    void allocate(int iWidth, int iHeight)
    {
        if ((iWidth <= 0) || (iHeight <= 0))
        {
            m_iStatus = INVALID_INPUT_PARAMETERS;
            return;
        }

        unsigned int iAllocSize = iWidth * iHeight * sizeof(T);

        if (m_iAllocatedSize < iAllocSize)
        {
            deallocate();
            m_pBuffer = new T [iWidth * iHeight];
            if (m_pBuffer)
            {
                m_iWidth = iWidth;
                m_iHeight = iHeight;
                m_iStatus = OKAY;
                m_iAllocatedSize = iAllocSize;
            }
            else
            {
                m_iStatus = MEMORY_ALLOCATION_FAILURE;
            }
        }
        else
        {
            m_iWidth = iWidth;
            m_iHeight = iHeight;
            m_iStatus = OKAY;
        }
    }

    // deallocate function
    void deallocate()
    {
        // common area for destruction
        m_iAllocatedSize = 0;
        if (m_pBuffer) delete [] m_pBuffer;
        m_pBuffer = 0;
        m_iHeight = 0;
        m_iWidth = 0;
    }

    // morphological routine helpers
    void maxHorizontally(const int iFilterSize);

    void minHorizontally(const int iFilterSize);

    void maxVertically(const int iFilterSize);

    void minVertically(const int iFilterSize);

    // shifting routines
    void shiftRight(const int iShift);

    void shiftDown(const int iShift);

private:
    unsigned int m_iAllocatedSize; // allocated size of buffer in BYTES >=
                                   // width * height * sizeof(T)
                                   // Init in all Constructors.
                                   // Set and used only in allocate() method.

    void mirror(const int iLeftBorder,
                const int iRightBorder,
                const int iTopBorder,
                const int iBottomBorder);

};

typedef CSmartImage<unsigned char> CGrayImage;

typedef CSmartImage<unsigned short> CShortImage;

typedef CSmartImage<unsigned int> CUIntImage;

typedef CSmartImage<float> CFloatImage;

typedef CSmartImage<double> CDoubleImage;

#include "SmartImage.inl"

#endif // SMARTIMAGE_H
    
    
///////////////////////////////////////////////////////////////////////////

