/****************************************************************************
File:        SmartImage.inl
Description: Implementation file for the smart template images.
             All images are one band.
Created:     04/15/05  SBanerjee

Copyright (c) 1996-2006, Saibal Banerjee, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
****************************************************************************/

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif

/* ******* */
/* Inlines */
/* ******* */
template<class T>
inline int CSmartImage<T>::GetWidth() const {return m_iWidth;}

template<class T>
inline int CSmartImage<T>::GetHeight() const {return m_iHeight;}

template<class T>
inline T * CSmartImage<T>::GetImage() const {return m_pBuffer;}

template<class T>
inline int CSmartImage<T>::GetStatus() const {return m_iStatus;}

template<class T>
inline void CSmartImage<T>::SetClear() const {memset(m_pBuffer, 0, m_iWidth * m_iHeight * sizeof(T));}

/* *************************** */
/* Modifier Function: SetAbs() */
/* *************************** */
template<class T>
void CSmartImage<T>::SetAbs() const
{
    T * pBuffer = m_pBuffer;
    T tData;
    int iSize = m_iWidth * m_iHeight;
    while (iSize--)
    {
        tData = *pBuffer;
        if (tData < 0) *pBuffer++ = -tData;
        else pBuffer++;
    }
}

/* ***************************** */
/* Statistics Functions          */
/* ***************************** */
template<class T>
T CSmartImage<T>::GetMax() const
{
    T * pBuffer = m_pBuffer;
    T tMaxData = *pBuffer++;
    T tData;
    int iSize = m_iWidth * m_iHeight;
    while (--iSize)
    {
        tData = *pBuffer++;
        if (tData > tMaxData) tMaxData = tData;
    }
    return tMaxData;
}

template<class T>
T CSmartImage<T>::GetMin() const
{
    T * pBuffer = m_pBuffer;
    T tMinData = *pBuffer++;
    T tData;
    int iSize = m_iWidth * m_iHeight;
    while (--iSize)
    {
        tData = *pBuffer++;
        if (tData < tMinData) tMinData = tData;
    }
    return tMinData;
}

template<class T>
double CSmartImage<T>::GetAvg() const
{
    T * pBuffer = m_pBuffer;
    double dSumData = 0;
    int iSize = m_iWidth * m_iHeight;
    while (iSize--)
    {
        dSumData += (double) (*pBuffer++);
    }
    return (dSumData / (m_iWidth * m_iHeight));
}

template<class T>
double CSmartImage<T>::GetStdev() const
{
    T * pBuffer = m_pBuffer;
    double dValue;
    double dMeanData = 0;
    double dVarData = 0;
    int iSize = m_iWidth * m_iHeight;
    while (iSize--)
    {
        dValue = (double) (*pBuffer++);
        dMeanData += dValue;
        dVarData += dValue * dValue;

    }
    dMeanData /= m_iWidth * m_iHeight;
    dVarData /= m_iWidth * m_iHeight;
    dVarData -= dMeanData * dMeanData;

    dVarData = (dVarData > 0) ? sqrt(dVarData) : 0;

    return dVarData;
}

template<class T>
double CSmartImage<T>::Dot(const CSmartImage<T> & source) const
{
    T * pBuffer = m_pBuffer;
    T * pSrcBuffer = source.m_pBuffer;
    double dValue;
    double dDotData = 0;
    int iSize = m_iWidth * m_iHeight;
    while (iSize--)
    {
        dValue = (double) (*pBuffer++);
        dValue *= (double) (*pSrcBuffer++);
        dDotData += dValue;

    }
    dDotData /= m_iWidth * m_iHeight;

    return dDotData;
}

template<class T>
void CSmartImage<T>::Clip()
{
    T * pBuffer = m_pBuffer;
    T tValue;
    int iSize = m_iWidth * m_iHeight;
    while (iSize--)
    {
        tValue = *pBuffer;
        if (tValue < 0) *pBuffer++ = 0;
        else pBuffer++;
    }
}

template<class T>
void CSmartImage<T>::Sqrt()
{
    T * pBuffer = m_pBuffer;
    T tValue;
    int iSize = m_iWidth * m_iHeight;
    while (iSize--)
    {
        tValue = *pBuffer;
        if (tValue < 0) *pBuffer++ = 0;
        *pBuffer++ = (T) sqrt((double) tValue);
    }
}

template<class T>
void CSmartImage<T>::Cut(const int iLeftBorder,
                         const int iRightBorder,
                         const int iTopBorder,
                         const int iBottomBorder)
{
    int iNewWidth = m_iWidth - iLeftBorder - iRightBorder;
    int iNewHeight = m_iHeight - iTopBorder - iBottomBorder;

    T * thisDst = m_pBuffer;
    T * thisSrc = m_pBuffer + m_iWidth * iTopBorder + iLeftBorder;
    int iHt = iNewHeight;
    while (iHt--)
    {
        memcpy(thisDst, thisSrc, iNewWidth * sizeof(T));
        thisDst += iNewWidth;
        thisSrc += m_iWidth;
    }

    m_iWidth = iNewWidth;
    m_iHeight = iNewHeight;
}

template<class T>
void CSmartImage<T>::mirror(const int iLeftBorder,
                            const int iRightBorder,
                            const int iTopBorder,
                            const int iBottomBorder)
{
    T * pSrcBuffer;
    T * pDstBuffer;
    int iHt, iWd;
    int iSrcBump, iDstBump;

    if (iLeftBorder > 0)
    {
        pSrcBuffer = m_pBuffer + iLeftBorder;
        pDstBuffer = pSrcBuffer - 1;
        iSrcBump = m_iWidth - iLeftBorder;
        iDstBump = m_iWidth + iLeftBorder;

        iHt = m_iHeight;
        while (iHt--)
        {
            iWd = iLeftBorder;
            while (iWd--) *pDstBuffer-- = *pSrcBuffer++;
            pSrcBuffer += iSrcBump;
            pDstBuffer += iDstBump;
        }
    }

    if (iRightBorder > 0)
    {
        pSrcBuffer = m_pBuffer + m_iWidth - 1 - iRightBorder;
        pDstBuffer = pSrcBuffer + 1;
        iSrcBump = m_iWidth + iRightBorder;
        iDstBump = m_iWidth - iRightBorder;

        iHt = m_iHeight;
        while (iHt--)
        {
            iWd = iRightBorder;
            while (iWd--) *pDstBuffer++ = *pSrcBuffer--;
            pSrcBuffer += iSrcBump;
            pDstBuffer += iDstBump;
        }
    }

    if (iTopBorder > 0)
    {
        pSrcBuffer = m_pBuffer + iTopBorder * m_iWidth;
        pDstBuffer = pSrcBuffer - m_iWidth;

        iHt = iTopBorder;
        while (iHt--)
        {
            memcpy(pDstBuffer, pSrcBuffer, m_iWidth * sizeof(T));
            pSrcBuffer += m_iWidth;
            pDstBuffer -= m_iWidth;
        }
    }

    if (iBottomBorder > 0)
    {
        pSrcBuffer = m_pBuffer + (m_iHeight - 1 - iBottomBorder) * m_iWidth;
        pDstBuffer = pSrcBuffer + m_iWidth;

        iHt = iBottomBorder;
        while (iHt--)
        {
            memcpy(pDstBuffer, pSrcBuffer, m_iWidth * sizeof(T));
            pSrcBuffer -= m_iWidth;
            pDstBuffer += m_iWidth;
        }
    }
}

/* **************** */
/* BoxSum Function  */
/* **************** */
template <class T>
void CSmartImage<T>::BoxSum(const int  iFilterWidth,
                            const int  iFilterHeight,
                            const bool bCut)
{
	int iTempHt, iTempWd, iFiltSize, iImgSize;
	T tSum, tTake;
	T *pStart;
	T *pEnd;

    if ((iFilterWidth <= 1) && (iFilterHeight <= 1))
    {
        return;
    }

    if (iFilterWidth > 1)
    {
        iFiltSize = iFilterWidth - 1;
        iImgSize = m_iWidth - iFiltSize;

	    pStart = m_pBuffer;
	    iTempHt = m_iHeight;
	    while (iTempHt--)
	    {
		    pEnd = pStart;

		    tSum = 0;
		    iTempWd = iFiltSize;
		    while (iTempWd--) tSum += *pEnd++;

		    tTake = 0;
		    iTempWd = iImgSize;
		    while (iTempWd--)
		    {
			    tSum -= tTake - *pEnd++;
			    tTake = *pStart;
			    *pStart++ = tSum;
		    }

		    pStart += iFiltSize;
	    }
    }

    if (iFilterHeight > 1)
    {
        iFiltSize = iFilterHeight - 1;
        iImgSize = m_iHeight - iFiltSize;

        iTempWd = m_iWidth;
        while (iTempWd--)
        {
            pStart = m_pBuffer++;
            pEnd = pStart;

		    tSum = 0;
		    iTempHt = iFiltSize;
		    while (iTempHt--)
		    {
			    tSum += *pEnd;
			    pEnd += m_iWidth;
		    }

		    tTake = 0;
		    iTempHt = iImgSize;
		    while (iTempHt--)
		    {
			    tSum -= tTake - *pEnd;
			    pEnd += m_iWidth;
			    tTake = *pStart;
			    *pStart = tSum;
			    pStart += m_iWidth;
		    }
        }
        m_pBuffer -= m_iWidth;
    }

    if (!bCut)
    {
        int iShift;

        iShift = (iFilterWidth - 1) / 2;
        if (iShift > 0)
        {
            shiftRight(iShift);
        }

        iShift = (iFilterHeight - 1) / 2;
        if (iShift > 0)
        {
            shiftDown(iShift);
        }
    }
    else
    {
        if ((iFilterWidth > 1) && (iFilterHeight > 1))
        {
            Cut(0, iFilterWidth - 1, 0, iFilterHeight - 1);
        }
        else if (iFilterWidth > 1)
        {
            Cut(0, iFilterWidth - 1, 0, 0);
        }
        else if (iFilterHeight > 1)
        {
            Cut(0, 0, 0, iFilterHeight - 1);
        }
    }
}

/* *********************** */
/* Morphological Functions */
/* *********************** */
template <class T>
void CSmartImage<T>::maxHorizontally(const int iFilterSize)
{
    int iGlobalSize = m_iHeight;
    const int iBump = iFilterSize - 1;
    int iLocalSize;

    register T * pStartPosition = m_pBuffer;
    register T * pEndPosition = pStartPosition + iFilterSize - 1;
    register T * pCurrentPosition;
    register T * pMaxPosition;

    while (iGlobalSize--)
    {
        pMaxPosition = pStartPosition - 1;
        iLocalSize = m_iWidth - iFilterSize + 1;
        while (iLocalSize--)
        {
            if (pMaxPosition < pStartPosition)
            {
				pMaxPosition = pStartPosition;
				for (pCurrentPosition = pStartPosition + 1; pCurrentPosition <= pEndPosition; pCurrentPosition++)
					if (*pMaxPosition <= *pCurrentPosition) pMaxPosition = pCurrentPosition;
			}
            else if (*pMaxPosition <= *pEndPosition) pMaxPosition = pEndPosition;

			*pStartPosition++ = *pMaxPosition;
			pEndPosition++;
		}
		pStartPosition += iBump;
		pEndPosition += iBump;
    }
}

template <class T>
void CSmartImage<T>::minHorizontally(const int iFilterSize)
{
    int iGlobalSize = m_iHeight;
    const int iBump = iFilterSize - 1;
    int iLocalSize;

    register T * pStartPosition = m_pBuffer;
    register T * pEndPosition = pStartPosition + iFilterSize - 1;
    register T * pCurrentPosition;
    register T * pMinPosition;

    while (iGlobalSize--)
    {
        pMinPosition = pStartPosition - 1;
        iLocalSize = m_iWidth - iFilterSize + 1;
        while (iLocalSize--)
        {
            if (pMinPosition < pStartPosition)
            {
				pMinPosition = pStartPosition;
				for (pCurrentPosition = pStartPosition + 1; pCurrentPosition <= pEndPosition; pCurrentPosition++)
					if (*pMinPosition >= *pCurrentPosition) pMinPosition = pCurrentPosition;
			}
            else if (*pMinPosition >= *pEndPosition) pMinPosition = pEndPosition;

			*pStartPosition++ = *pMinPosition;
			pEndPosition++;
		}
		pStartPosition += iBump;
		pEndPosition += iBump;
    }
}

template <class T>
void CSmartImage<T>::maxVertically(const int iFilterSize)
{
    int iGlobalSize = m_iWidth;
    const int iBump = m_iWidth * (iFilterSize - 1);
    int iLocalSize;

    register T * pStartPosition;
    register T * pEndPosition;
    register T * pCurrentPosition;
    register T * pMaxPosition;

    while (iGlobalSize--)
    {
        pStartPosition = m_pBuffer++;
        pEndPosition = pStartPosition + iBump;
        pMaxPosition = pStartPosition - m_iWidth;
        iLocalSize = m_iHeight - iFilterSize + 1;
        while (iLocalSize--)
        {
			if (pMaxPosition < pStartPosition)
			{
				pMaxPosition = pStartPosition;
				for (pCurrentPosition = pStartPosition + m_iWidth; pCurrentPosition <= pEndPosition; pCurrentPosition += m_iWidth)
					if (*pMaxPosition <= *pCurrentPosition) pMaxPosition = pCurrentPosition;
			}
			else if (*pMaxPosition <= *pEndPosition) pMaxPosition = pEndPosition;
			*pStartPosition = *pMaxPosition;
			pStartPosition += m_iWidth;
			pEndPosition += m_iWidth;
		}
    }
    m_pBuffer -= m_iWidth;
}

template <class T>
void CSmartImage<T>::minVertically(const int iFilterSize)
{
    int iGlobalSize = m_iWidth;
    const int iBump = m_iWidth * (iFilterSize - 1);
    int iLocalSize;

    register T * pStartPosition;
    register T * pEndPosition;
    register T * pCurrentPosition;
    register T * pMinPosition;

    while (iGlobalSize--)
    {
        pStartPosition = m_pBuffer++;
        pEndPosition = pStartPosition + iBump;
        pMinPosition = pStartPosition - m_iWidth;
        iLocalSize = m_iHeight - iFilterSize + 1;
        while (iLocalSize--)
        {
			if (pMinPosition < pStartPosition)
			{
				pMinPosition = pStartPosition;
				for (pCurrentPosition = pStartPosition + m_iWidth; pCurrentPosition <= pEndPosition; pCurrentPosition += m_iWidth)
					if (*pMinPosition >= *pCurrentPosition) pMinPosition = pCurrentPosition;
			}
			else if (*pMinPosition >= *pEndPosition) pMinPosition = pEndPosition;
			*pStartPosition = *pMinPosition;
			pStartPosition += m_iWidth;
			pEndPosition += m_iWidth;
		}
    }
    m_pBuffer -= m_iWidth;
}

template<class T>
void CSmartImage<T>::Morph(const int  iFilterWidth,
                           const int  iFilterHeight,
                           const bool bIsDilate,
                           const bool bCut)
{
    if (bIsDilate)
    {
        if (iFilterWidth > 1) maxHorizontally(iFilterWidth);
        if (iFilterHeight > 1) maxVertically(iFilterHeight);
    }
    else
    {
        if (iFilterWidth > 1) minHorizontally(iFilterWidth);
        if (iFilterHeight > 1) minVertically(iFilterHeight);
    }

    if (!bCut)
    {
        int iShift;

        if (iFilterWidth > 1)
        {
            iShift = (iFilterWidth - 1) / 2;
            if (iShift > 0)
            {
                shiftRight(iShift);
            }
        }

        if (iFilterHeight > 1)
        {
            iShift = (iFilterHeight - 1) / 2;
            if (iShift > 0)
            {
                shiftDown(iShift);
            }
        }
    }
    else
    {
        if ((iFilterWidth > 1) && (iFilterHeight > 1))
        {
            Cut(0, iFilterWidth - 1, 0, iFilterHeight - 1);
        }
        else if (iFilterWidth > 1)
        {
            Cut(0, iFilterWidth - 1, 0, 0);
        }
        else if (iFilterHeight > 1)
        {
            Cut(0, 0, 0, iFilterHeight - 1);
        }
    }
}

/* ****************** */
/* Shifting Functions */
/* ****************** */
template<class T>
void CSmartImage<T>::shiftRight(const int iShift)
{
    T * pTPicture = m_pBuffer;
    pTPicture += m_iWidth;
    register T * pDst;
    register T * pSrc;
    register int iTempWd;
    register int iTempHt = m_iHeight;
    while (iTempHt--)
    {
        pDst = pTPicture;
        pSrc = pDst - iShift;
        iTempWd = m_iWidth - iShift;
        while (iTempWd--) *--pDst = *--pSrc;
        pTPicture += m_iWidth;
    }
}

template<class T>
void CSmartImage<T>::shiftDown(const int iShift)
{
    T * pTPicture = m_pBuffer;
    register T * pDst = pTPicture + (m_iHeight - 1) * m_iWidth;
    register T * pSrc = pDst - iShift * m_iWidth;
    register int iTempHt = m_iHeight - iShift;
    while (iTempHt--)
    {
        memcpy(pDst, pSrc, m_iWidth * sizeof(T));
        pDst -= m_iWidth;
        pSrc -= m_iWidth;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
