/****************************************************************************
File:        IAMDetector.h
Description: Implementation file for detecting objects with normailzed cross
             correlation based matching and Mexican hat filtering
Created:     10/15/07 
Author:      Saibal Banerjee

Copyright (c) Illumina 2007, All rights reserved.
This information is protected by copyright law and international
treaties.  Unauthorized reproduction or distribution of this file, or
any portion of it, may result in severe civil and criminal penalties,
and will be prosecuted to the maximum extent possible under the law.
****************************************************************************/

#include "IAMDetection/ErrorCodes.h"
#include "IAMDetection/GaussianFilter.h"
#include "IAMDetection/ObjectDetector.h"
#include "IAMDetection/GaussianMaker.h"
#include "ClusterObject.h"
#include "IAMDetector.h"

void IAMDetector::Run(int                  imageWidth,
                      int                  imageHeight,
                      float *              imageBuffers[],
                      AnalysisParameters & params,
                      int &                nObjects,
                      ClusterObject *    & pObjects)
{
    nObjects = 0;
    if (pObjects)
    {
        delete [] pObjects;
        pObjects = 0;
    }

    float fFWHM = params.ClusterFwhm;
    float fPearsonThreshold = params.PearsonThreshold;
    int nPasses = 2;

    int iStatus = OKAY;

    const double cdFWHMMultiplier = sqrt(log(256.0));

    const float fFilterSigma = (float) (fFWHM / cdFWHMMultiplier);
    CGaussianFilter filter(fFilterSigma);

    CObjectDetector * pObjectDetectors = new CObjectDetector [nPasses];
    if (!pObjectDetectors)
    {
        return;
    }

    CShortImage inPicture(imageWidth, imageHeight);
    if ((iStatus = inPicture.GetStatus()) != OKAY)
    {
        delete [] pObjectDetectors;
        return;
    }

    transposeImage(imageBuffers[0], inPicture);

    const int iKernelSize = 127;
    int iPassNumber = 0;

    iStatus = m_BackEstimator.Setup(iKernelSize, inPicture);
    if (iStatus != OKAY)
    {
        goto FILL_OBJ_LABEL;
    }

    iStatus = pObjectDetectors[iPassNumber].Compute(&filter,
                                                    fPearsonThreshold,
                                                    inPicture.GetWidth(),
                                                    inPicture.GetHeight(),
                                                    inPicture.GetImage());
    if (iStatus != OKAY)
    {
        goto FILL_OBJ_LABEL;
    }

    nObjects += pObjectDetectors[iPassNumber].GetFloatFeatures().GetWidth();
    m_BackEstimator.Add(pObjectDetectors[iPassNumber].GetMatcher().GetForePicture());

    while (++iPassNumber < nPasses)
    {
        nextProcess(pObjectDetectors[iPassNumber - 1], fFilterSigma, inPicture);
        iStatus = pObjectDetectors[iPassNumber].Compute(&filter,
                                                        fPearsonThreshold,
                                                        inPicture.GetWidth(),
                                                        inPicture.GetHeight(),
                                                        inPicture.GetImage());
        if (iStatus != OKAY)
        {
            goto FILL_OBJ_LABEL;
        }

        nObjects += pObjectDetectors[iPassNumber].GetFloatFeatures().GetWidth();
        m_BackEstimator.Add(pObjectDetectors[iPassNumber].GetMatcher().GetForePicture());
    }

FILL_OBJ_LABEL:
    if (nObjects == 0)
    {
        delete [] pObjectDetectors;
        return;
    }

    pObjects = new ClusterObject [nObjects];
    if (!pObjects)
    {
        delete [] pObjectDetectors;
        return;
    }

    m_BackEstimator.Run();

    fillObjectStruct(pObjectDetectors,
                     inPicture.GetWidth(),
                     inPicture.GetHeight(),
                     iPassNumber,
                     nObjects,
                     pObjects);

    // updateAnalysisImages(imageBuffers);

    delete [] pObjectDetectors;
}

void IAMDetector::transposeImage(float *       image, 
                                 CShortImage & inPicture)
{
    const int iHeight = inPicture.GetHeight();
    const int iWidth = inPicture.GetWidth();
    const int iBump = iWidth * iHeight - 1;
    unsigned short * pBuffer = inPicture.GetImage();

    int iTmpHt = iHeight;
    int iTmpWd;
    while (iTmpHt--)
    {
        iTmpWd = iWidth;
        while (iTmpWd--)
        {
            *pBuffer++ = (unsigned short) (*image);
            image += iHeight;
        }
        image -= iBump;
    }
}

void IAMDetector::makeSubtractor(const unsigned int nClusters,
                                 CGaussianMaker * clusters,
                                 CDoubleImage & picture)
{
    picture.SetClear();
    const int iWidth = picture.GetWidth();
    const int iHeight = picture.GetHeight();

    CGaussianMaker cluster;
    double * pBuffer;
    int iX, iY;
    unsigned int uiClusters;
    int iROIXStart, iROIYStart, iROIXEnd, iROIYEnd, iROIWidth, iROIHeight, iBump;
    for (uiClusters = 0; uiClusters < nClusters; uiClusters++)
    {
        cluster = clusters[uiClusters];
        cluster.GetROI(iROIXStart, iROIYStart, iROIWidth, iROIHeight);

        iROIXEnd = iROIXStart + iROIWidth - 1;
        if (iROIXEnd >= iWidth) iROIXEnd = iWidth - 1;
        if (iROIXStart < 0) iROIXStart = 0;

        iROIYEnd = iROIYStart + iROIHeight - 1;
        if (iROIYEnd >= iHeight) iROIYEnd = iHeight - 1;
        if (iROIYStart < 0) iROIYStart = 0;

        iROIWidth = iROIXEnd - iROIXStart + 1;
        iROIHeight = iROIYEnd - iROIYStart + 1;

        if ((iROIWidth <= 0) || (iROIHeight <= 0)) continue;

        pBuffer = picture.GetImage() + iROIYStart * iWidth + iROIXStart;
        iBump = iWidth - iROIWidth;

        for (iY = iROIYStart; iY <= iROIYEnd; iY++)
        {
            for (iX = iROIXStart; iX <= iROIXEnd; iX++)
            {
                *pBuffer++ += cluster.GetValue(iX, iY);
            }
            pBuffer += iBump;
        }
    }
}

void IAMDetector::nextProcess(const CObjectDetector & detector,
                              const float fFilterSigma,
                              CShortImage & inPicture)
{
    const CFloatImage features = detector.GetFloatFeatures();
    const unsigned int nClusters = features.GetWidth();
    const float * pXs = features.GetImage();
    const float * pYs = pXs + nClusters;
    const float * pGs = pYs + nClusters;

    CGaussianMaker * clusters = new CGaussianMaker [nClusters];
    unsigned int uiLabel;
    for (uiLabel = 0; uiLabel < nClusters; uiLabel++)
    {
        clusters[uiLabel].Set(fFilterSigma, pXs[uiLabel], pYs[uiLabel], pGs[uiLabel]);
    }

    CDoubleImage picture(inPicture.GetWidth(), inPicture.GetHeight());
    makeSubtractor(nClusters, clusters, picture);

    unsigned short * pInBuffer = inPicture.GetImage();
    double * pBuffer = picture.GetImage();
    double dValue;

    int iSize = inPicture.GetWidth() * inPicture.GetHeight();
    while (iSize--)
    {
        dValue = *pInBuffer;
        dValue -= *pBuffer++;
        if (dValue < 0) 
        {
            *pInBuffer++ = 0;
        }
        else if (dValue > 0xffff)
        {
            *pInBuffer++ = 0xffff;
        }
        else
        {
            *pInBuffer++ = (unsigned short) (dValue + 0.5);
        }
    }

    delete [] clusters;
}

void IAMDetector::fillObjectStruct(const CObjectDetector * pDetectors,
                                   const int               iWidth,
                                   const int               iHeight,
                                   const int               iPassNumber,
                                   const int               nObjects,
                                   ClusterObject *         pObjects)
{
    const unsigned short iMaxXCoord = (unsigned short) (iWidth - 1);
    const unsigned short iMaxYCoord = (unsigned short) (iHeight - 1);
    const float * pBackMeanBuffer = m_BackEstimator.GetBackMeanPicture().GetImage();

    int nLabels, iLabel;
    const float * pXs;
    const float * pYs;
    const float * pGs;
    float fXCoord, fYCoord;
    int iXCoord, iYCoord;

    for (int iPass = 0; iPass < iPassNumber; iPass++)
    {
        nLabels = pDetectors[iPass].GetFloatFeatures().GetWidth();
        pXs = pDetectors[iPass].GetFloatFeatures().GetImage();
        pYs = pXs + nLabels;
        pGs = pYs + nLabels;

        for (iLabel = 0; iLabel < nLabels; iLabel++)
        {
            pObjects->pass_number = iPass;

            pObjects->position.x = fXCoord = pXs[iLabel];
            pObjects->position.y = fYCoord = pYs[iLabel];

            iXCoord = (int) (fXCoord + 0.5f);
            if (iXCoord < 0) iXCoord = 0;
            if (iXCoord >= iWidth) iXCoord = iWidth - 1;

            iYCoord = (int) (fYCoord + 0.5f);
            if (iYCoord < 0) iYCoord = 0;
            if (iYCoord >= iHeight) iYCoord = iHeight - 1;

            pObjects->max_pixel.value = pObjects->intensity = pGs[iLabel] - pBackMeanBuffer[iYCoord * iWidth + iXCoord];

            if (fXCoord < 0) pObjects->max_pixel.x = 0;
            else if (fXCoord > iMaxXCoord) pObjects->max_pixel.x = iMaxXCoord;
            else pObjects->max_pixel.x = (unsigned short) (fXCoord + 0.5f);

            if (fYCoord < 0) pObjects->max_pixel.y = 0;
            else if (fYCoord > iMaxYCoord) pObjects->max_pixel.y = iMaxYCoord;
            else pObjects->max_pixel.y = (unsigned short) (fYCoord + 0.5f);

            pObjects++;
        }
    }
}

void IAMDetector::updateAnalysisImages(float * imageBuffers[])
{
    float * image = imageBuffers[0];
    float * filter_image = imageBuffers[1];
    float * noise_image = imageBuffers[2];

    const int iHeight = m_BackEstimator.GetBackMeanPicture().GetHeight();
    const int iWidth = m_BackEstimator.GetBackMeanPicture().GetWidth();
    const int iBump = iWidth * iHeight - 1;

    const float * pBackMeanBuffer = m_BackEstimator.GetBackMeanPicture().GetImage();
    const float * pBackVarBuffer = m_BackEstimator.GetBackVarPicture().GetImage();

    int iTmpHt = iHeight;
    int iTmpWd;
    float fMean, fStdev;
    while (iTmpHt--)
    {
        iTmpWd = iWidth;
        while (iTmpWd--)
        {
            fMean = *pBackMeanBuffer++;
            fStdev = *pBackVarBuffer++;

            *filter_image = *image - fMean;
            fStdev -= fMean * fMean;
            *noise_image = (fStdev > 0) ? sqrt(fStdev) : 0;

            image += iHeight;
            filter_image += iHeight;
            noise_image += iHeight;
        }
        image -= iBump;
        filter_image -= iBump;
        noise_image -= iBump;
    }
}
