//  Time-stamp: <26 Jun 2006 16:48 BST ImageProcess.cpp km>
/*
 * Main object used for the image image processing tasks in the Solexa
 * software pipeline. The object holds all relevant intermediate data
 * structures and performs a series of analysis steps. These steps are
 * implemented in a different set of modules:
 *
 *   - Filtering                         (ProcessFiler.cpp)
 *   - Noise and Background Estimation   (ProcessNoiseEstimate.cpp)
 *   - Object Extraction and Deblending  (ProcessObjectDetection.cpp)
 *
 * Basic image processing routines are also in ImageUtil.cpp.
 *
 * Copyright (c) Solexa 2003, 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#include <string.h>
#else
#include <cmath>
#include <cstring>
#endif
#include <cstdlib>
#include <cstdio>
#include <cassert>
#include "ImageProcess.h"
#include "FFT.h"
#include "Sorting.h"
#include "Stats.h"
#include "Tools.h"
#include "ImageUtil.h"
#include "ProcessFilter.h"
#include "ProcessNoiseEstimate.h"
#include "ClusterObject.h"
#include "ObjectDetectorIFC.h"
#include "ProcessObjectDetection.h"
#include "QCImage.h"
#include "Message.h"

/*
 * Auxiliary routines to perform some statistics on an image.
 * Much of this information is mostly of historical interest as
 * it corresponds to quantities calculated by the SMA MetaMorph
 * analysis scripts.
 */
static void get_image_analysis_stats(ImageAnalysisRecord *stats,
                                     float *filter_image, float *noise_image,
                                     int nx, int ny, float gauss_fac)
{
    stats->average_noise = Stats::average(noise_image, nx*ny);
    stats->average_threshold = gauss_fac * stats->average_noise;
    get_threshold_average(filter_image, nx*ny,
                          noise_image, gauss_fac,
                          &(stats->average_threshold_signal),
                          &(stats->average_threshold_background));
}

static void get_image_stats(ImageStatsRecord *stats,
                            float *image, int nx, int ny,
                            unsigned short bitsperpixel)
{
    stats->FractionSaturation
        = get_saturation_fraction(image, nx*ny, bitsperpixel);
    get_array_stats(image, nx*ny, &(stats->Average),
                    &(stats->StandardDeviation));
    get_array_min_max(image, nx*ny, &(stats->Minimum), &(stats->Maximum));
    stats->AverageBackground = 0.;
    stats->DispersionBackground = 0.;
#if 0
    determine_noise_hist_stats(image, nx*ny,
                   &(stats->average_background),
                   &(stats->dispersion_background));
#endif
}

//---------------------------------------------------------------------------
// the image analysis object
//---------------------------------------------------------------------------

ImageData::ImageData(float *im, int x, int y, unsigned short bpp)
{
    Init();
    bitsperpixel = bpp;
    nx = x; ny = y;
    image = new float [nx*ny];
    memcpy(image, im, nx*ny*sizeof(float));
    analyseImageStats();
    analyseLocalImageStats();
}

ImageData::ImageData(unsigned short *im, int x, int y, unsigned short bpp)
{
    Init();
    bitsperpixel = bpp;
    nx = x; ny = y;
    image = new float [nx*ny];
    remove_saturated_pixels(im, nx, ny, bitsperpixel);
    convert_image_ushort2float(im, image, nx, ny);
    analyseImageStats();
    analyseLocalImageStats();
}

void ImageData::Init()
{
    analysis_stats = NULL;
    background_image = NULL;
    bitsperpixel = 0;
    estimated_fwhm = 0.f;
    filter_image = 0;
    filtered_object_stats = NULL;
    fnx = 0;
    fny = 0;
    fourier_image = NULL;
    image = NULL;
    image_stats = NULL;
    local_image_stats = NULL;
    nlocal_stats = 0;
    noise_image = NULL;
    nobj = 0;
    nx = 0;
    ny = 0;
    object_stats = NULL;
    objects = NULL;
    positiondata = NULL;
    positions = NULL;
    sequencing = NULL;
    status = 0;
    transform = NULL;
    transform2 = NULL;
    usedTransform = NULL;

    nmax = 0;
}

ImageData::~ImageData()
{
    clear();
}

void ImageData::setSequencingInfo(unsigned char base, int is_deblocked)
{
    if (sequencing)
        delete sequencing;
    sequencing = new SequencingRecord;
    sequencing->base = base;
    sequencing->is_deblocked = is_deblocked;
}

/*
 * Try to estimate the FWHM of the point spread function
 * by inspection of the image power spectrum. This may be useful
 * to detect focus problems.
 */
void ImageData::analyseEstimateFWHM()
{
    if (!(status & IMAGEDATA_STATE_FILTER))
        analyseFilter();
    estimated_fwhm = estimate_fwhm_fourier(fourier_image,
                                           fnx, fny, this->params.ClusterFwhm);
}

/*
 * get some basic image statistics such as average, standard deviation,
 * minimum and maximum
 */
void ImageData::analyseImageStats()
{
    image_stats = new ImageStatsRecord;
    image_stats->Region.XMin = 0;
    image_stats->Region.YMin = 0;
    image_stats->Region.XSize = nx;
    image_stats->Region.YSize = ny;
    get_image_stats(image_stats, image, nx, ny, bitsperpixel);
    status |= IMAGEDATA_STATE_IMAGE_STATS;
}

/*
 * calculate some basic statistics on subregions of an image
 */
void ImageData::analyseLocalImageStats()
{
    int i, j;
    int sizex = get_local_image_size(nx, LOCAL_SIZE);
    int sizey = get_local_image_size(ny, LOCAL_SIZE);
    int numx = nx/sizex;
    int numy = ny/sizey;
    float *subimage = new float [sizex*sizey];
    nlocal_stats = numx*numy;
    local_image_stats = new ImageStatsRecord [nlocal_stats];
    for (i = 0; i < numx; i++)
        for (j = 0; j < numy; j++) {
            int x = i * sizex;
            int y = j * sizey;
            local_image_stats[i*numy+j].Region.XMin = x;
            local_image_stats[i*numy+j].Region.YMin = y;
            local_image_stats[i*numy+j].Region.XSize = sizex;
            local_image_stats[i*numy+j].Region.YSize = sizey;
            copy_image_region(image, nx, ny, subimage, sizex, sizey, x, y);
            get_image_stats(&local_image_stats[i*numy+j],
                            subimage, sizex, sizey, bitsperpixel);
        }
    delete [] subimage;
}

/*
 * Calculate some statistics (e.g. size, intensity) on individual
 * clusters on an image that has had a filter applied.
 * Not much used.
 */
void ImageData::analyseFilteredObjectStats()
{
    if (status & IMAGEDATA_STATE_FILTER_OBJ_STATS)
        return;
    int i, j;
    filtered_object_stats = new ObjectStatsRecord [ nobj ];
    for (i = 0; i < nobj; i++) {
        filtered_object_stats[i].area = objects[i].na;
        filtered_object_stats[i].boundary = objects[i].nb;
        filtered_object_stats[i].integral = 0.;
        for (j = 0; j < objects[i].na; j++)
            filtered_object_stats[i].integral += objects[i].area[j].value;
        filtered_object_stats[i].maximum = objects[i].max_pixel.value;
        filtered_object_stats[i].background = 0.;
        if (objects[i].is_split) {
           filtered_object_stats[i].gaussian_xsigma = 0.;
           filtered_object_stats[i].gaussian_ysigma = 0.;
        }
        else {
            float x, y;
            filtered_object_stats[i].background = 0.;
            fit_gaussian_object(&objects[i], filter_image, nx, ny,
                                &x, &y,
                                &(filtered_object_stats[i].gaussian_maximum),
                                &(filtered_object_stats[i].gaussian_xsigma),
                                &(filtered_object_stats[i].gaussian_ysigma),
                                &(filtered_object_stats[i].background));
        }
    }
    status |= IMAGEDATA_STATE_FILTER_OBJ_STATS;
}

/*
 * Calculate some statistics (e.g. size, intensity) on individual
 * clusters.
 */
void ImageData::analyseObjectStats()
{
    if (status & IMAGEDATA_STATE_OBJ_STATS)
        return;
    object_stats = new ObjectStatsRecord [ nobj ];
    int *lookup = new int [nx*ny];
    create_lookup_array(lookup, nx, ny, objects, nobj);
    expand_lookup_array(lookup, nx, ny, objects, nobj);
    get_background_dispersion(image, lookup, nx, ny, objects, nobj,
                              object_stats, filter_image);
    integrate_objects(lookup, image, nx, ny, objects, nobj, object_stats);

    int i;
    for (i = 0; i < nobj; i++) {
        object_stats[i].xinter = objects[i].position.x;
        object_stats[i].yinter = objects[i].position.y;
        object_stats[i].area = objects[i].na;
        object_stats[i].boundary = objects[i].nb;
        object_stats[i].xdim = objects[i].max.x - objects[i].min.x + 1;
        object_stats[i].ydim = objects[i].max.y - objects[i].min.y + 1;
        object_stats[i].properties = (objects[i].is_split ?
                                      OBJECT_PROPERTY_IS_SPLIT : 0);
        object_stats[i].gaussian_chi2 = 0.;
        object_stats[i].gaussian_xsigma = 0.;
        object_stats[i].gaussian_ysigma = 0.;
        object_stats[i].gaussian_xcentre = 0.;
        object_stats[i].gaussian_ycentre = 0.;
        object_stats[i].gaussian_background = 0.;
        object_stats[i].gaussian_maximum = 0.;
        object_stats[i].gaussian_volume = 0.;
#if 0
        else {
            float x, y;
            object_stats[i].background = 0.;
            fit_gaussian_object(&objects[i], image, nx, ny,
                                &x, &y,
                                &(object_stats[i].maximum),
                                &(object_stats[i].xfwhm),
                                &(object_stats[i].yfwhm),
                                &(object_stats[i].background));
        }
#endif
    }

    float* nonzero_gaussian_xsigmas = new float[nobj];
    float sum_sq_nonzero_gaussian_xsigmas(0.0);
    unsigned int num_nonzero_gaussian_xsigmas(0);
    float curr_gaussian_xsigma(0.0);
    float gaussian_xsigma_mean(0.0);
    const float SIGMA_TO_FWHM_FACTOR(2.f * sqrt(2.f * log(2.f)));

    for (i = 0; i < nobj; i++) {
        int offset = 5;
        if (((objects[i].is_split)
             || ((objects[i].max_pixel.x < offset)
                  || (objects[i].max_pixel.y < offset)
                  || (objects[i].max_pixel.x >= nx-offset)
                  || (objects[i].max_pixel.y >= ny-offset)))
            || (fit_gaussian_area(lookup, image, nx, ny,
                                  &objects[i], i, &object_stats[i], filter_image))) {
            object_stats[i].gaussian_background = 0.;
            object_stats[i].gaussian_maximum = 0.;
        }

        curr_gaussian_xsigma = object_stats[i].gaussian_xsigma;

        if (curr_gaussian_xsigma > 0.0) {
            nonzero_gaussian_xsigmas[num_nonzero_gaussian_xsigmas++]
                = curr_gaussian_xsigma;
            sum_sq_nonzero_gaussian_xsigmas
                += (curr_gaussian_xsigma * curr_gaussian_xsigma);
        }
    }

    cluster_stats.Num = (int) num_nonzero_gaussian_xsigmas;

	if (num_nonzero_gaussian_xsigmas > 0) {
        gaussian_xsigma_mean = Stats::average(nonzero_gaussian_xsigmas,
                                              num_nonzero_gaussian_xsigmas);
        cluster_stats.FwhmMean
            = gaussian_xsigma_mean * SIGMA_TO_FWHM_FACTOR;
        cluster_stats.FwhmMedian
            = (select(nonzero_gaussian_xsigmas,
                      (long) num_nonzero_gaussian_xsigmas, -1)
                      * SIGMA_TO_FWHM_FACTOR);
        cluster_stats.FwhmStandardDeviation
            = (sqrtf((sum_sq_nonzero_gaussian_xsigmas
                    / num_nonzero_gaussian_xsigmas)
                    - (gaussian_xsigma_mean * gaussian_xsigma_mean))
                    * SIGMA_TO_FWHM_FACTOR);
    } else {
        cluster_stats.FwhmMean = 0.0;
        cluster_stats.FwhmMedian = 0.0;
        cluster_stats.FwhmStandardDeviation = 0.0;
    }

    delete[] nonzero_gaussian_xsigmas;

    //fit_gaussian_area();
    delete [] lookup;
    status |= IMAGEDATA_STATE_OBJ_STATS;
}

/*
 * Write the cluster-specific object statistics to a text file.
 */
int ImageData::writeObjectStats(const char filename[],
                                st_fnoisepixel *object_lists,
                                int *indices, int count)
{
    if (!(status & IMAGEDATA_STATE_OBJ_STATS))
        return 0;

    FILE *fp = fopen(filename, "w");
    if (!fp) {
        error("Failed to open file %s\n", filename);
        return 1;
    }
    if (indices == NULL) {
        count = nobj;
    }
    int j, idx;
    for (j = 0; j < count; j++) {
        idx = j;
        if (indices)
            idx = indices[j];
        if (idx >= 0) {
            ObjectStatsRecord *stats = &(object_stats[idx]);
            if (indices && object_lists)
                fprintf(fp, "%6.1f %6.1f %7.1f %6.1f ",
                        object_lists[idx].x, object_lists[idx].y,
                        object_lists[idx].value, object_lists[idx].noise);
            else
                fprintf(fp, "%6.1f %6.1f %7.1f %6.1f ",
                        (float)(positions[idx].x), (float)(positions[idx].y),
                        filter_image[positions[idx].x*ny+positions[idx].y],
                        noise_image [positions[idx].x*ny+positions[idx].y]);
            fprintf(fp, "%1i %2i %2i %3i %3i ",
            stats->properties & OBJECT_PROPERTY_IS_SPLIT,
            stats->xdim, stats->ydim, stats->area, stats->boundary);
            fprintf(fp, "%6.1f %6.1f %3.1f %5.0f %6.0f %4.0f %5.0f ",
                    stats->xcentre, stats->ycentre,
                    stats->width_sigma * (2. * sqrt(2.*log(2.))),
                    stats->maximum, stats->integral, stats->noise,
                    stats->background);
            fprintf(fp, "%6.1f %6.1f %3.1f %3.1f %5.0f %6.0f %6.0f\n",
                    stats->gaussian_xcentre, stats->gaussian_ycentre,
                    stats->gaussian_xsigma * (2. * sqrt(2.*log(2.))),
                    stats->gaussian_ysigma * (2. * sqrt(2.*log(2.))),
                    stats->gaussian_maximum, stats->gaussian_volume,
                    stats->gaussian_chi2);
        }
        else {
            fprintf(fp, "%6.1f %6.1f %7.1f %6.1f ",
                    0., 0., 0., 0.);
            fprintf(fp, "%1i %2i %2i %3i %3i ",
                    0, 0, 0, 0, 0);
            fprintf(fp, "%6.1f %6.1f %3.1f %5.0f %6.0f %4.0f %5.0f ",
                    0., 0., 0., 0., 0., 0., 0.);
            fprintf(fp, "%6.1f %6.1f %3.1f %3.1f %5.0f %6.0f %6.0f\n",
                    0., 0., 0., 0., 0., 0., 0.);
        }
    }
    fclose(fp);
    return 0;
}

int ImageData::writePositions(FILE *filehandle, int indexflag)
{
    int i;
    if (! indexflag) {
        for (i = 0; i < nobj; i++) {
            fprintf(filehandle, "%6.1f %6.1f %7.1f %6.1f\n",
                    positiondata[i].x, positiondata[i].y,
                    positiondata[i].value, positiondata[i].noise);
        }
    }
    else {
        for (i = 0; i < nobj; i++) {
            fprintf(filehandle, "%5i %6.1f %6.1f %7.1f %6.1f\n",
                    i, positiondata[i].x, positiondata[i].y,
                    positiondata[i].value, positiondata[i].noise);
        }
    }
    return 0;
}

void ImageData::clearImages()
{
    if (image) {
        delete [] image;
        image = NULL;
    }
    if (status & IMAGEDATA_STATE_FILTER) {
        delete [] filter_image;
        delete_fourier_array(fourier_image);
        status &= (~IMAGEDATA_STATE_FILTER);
    }
    if (status & IMAGEDATA_STATE_BACKGROUND_NOISE) {
        delete [] background_image;
        delete [] noise_image;
        status &= (~IMAGEDATA_STATE_BACKGROUND_NOISE);
    }
}

/*
 * free all allocated memory
 */
void ImageData::clear()
{
    clearImages();
    delete image_stats;
    delete [] local_image_stats;
    if (! (status & IMAGEDATA_STATE_ANALYSED))
        return;
    if (status & IMAGEDATA_STATE_FILTER_OBJ_STATS)
        delete [] filtered_object_stats;
    if (status & IMAGEDATA_STATE_OBJ_STATS)
        delete [] object_stats;
    delete analysis_stats;
    delete [] positions;
    delete [] positiondata;

    for (int i = 0; i < nobj; i++)
        objects[i].clear();
    delete [] objects;

    if (sequencing)
        delete sequencing;
    status = 0;
    if (transform)
        delete transform;
    if (usedTransform)
        delete usedTransform;
}

/*
 * apply a linear filter to improve the signal-to-noise
 */
void ImageData::analyseFilter()
{
    if (status & IMAGEDATA_STATE_FILTER)
        return;
    get_fourier_dimensions((int)ceil(this->params.ClusterFwhm), nx, ny,
                           &fnx, &fny);
    fourier_image = new_fourier_array(fnx, fny);
    filter_image = new float [nx*ny];
    //matched filter params
    const float gamma = 2.0;	//scaled free power spectrum
    const float alpha = 1.2;	//scaling factor
    const float fwhm = this->params.ClusterFwhm;

    matchedfilter_params_t params={fwhm,gamma,alpha};

    switch (this->params.FilterType) {
    case 0:  // Mexican-Hat filter
        filter_image_fft(image, filter_image, nx, ny,
                         make_mexican_hat_fft_mask,
                         (void *)(&fwhm),
                         fourier_image, fnx, fny); break;
    case 1:  // Gaussian High-Pass-Filter
        filter_image_fft(image, filter_image, nx, ny,
                         make_gaussian_highpass,
                         (void *)(&fwhm),
                         fourier_image, fnx, fny); break;
    case 2:  // Matched filter
         filter_image_fft(image, filter_image, nx, ny,
                         make_matched_fft_mask, (void *)(&params),
                         fourier_image, fnx, fny); break;
    case 3:  // Biparametric filter
         filter_image_fft(image, filter_image, nx, ny,
                         make_biparametric_fft_mask, (void *)(&params),
                         fourier_image, fnx, fny); break;
    default:
        memcpy(image, filter_image, nx*ny*sizeof(float)); break;
    }
    status |= IMAGEDATA_STATE_FILTER;
}

/*
 * estimate the local background and its dispersion
 */
void ImageData::analyseBackgroundNoise()
{
    if (status & IMAGEDATA_STATE_BACKGROUND_NOISE)
        return;
    background_image = new float [nx*ny];
    noise_image = new float [nx*ny];
    estimate_noise_image(filter_image, nx, ny, noise_image, background_image);
    subtract_array(filter_image, filter_image, background_image, nx*ny);
    status |= IMAGEDATA_STATE_BACKGROUND_NOISE;
}

/*
 * Main analysis routine.
 * Performs all image extraction tasks.
 */
void ImageData::analyse()
{
    TRACE("Entering ImageData[%p]::analyse()\n", (void *)this);
     // filtering
    if (status & IMAGEDATA_STATE_ANALYSED) {
        TRACE("Aborting ImageData[%p]::analyse()\n", (void *)this);
        return;
    }

    // object detection
    IObjectDetector * pDetector = IObjectDetector::CreateInstance(this->params.DetectionType);
    if (!pDetector)
    {
        return;
    }

    // filtering
    analyseFilter();
    // noise/background estimation
    analyseBackgroundNoise();

    int i;
    float * imageBuffers[MAX_IMAGE_BUFFERS];
    for (i = 0; i < MAX_IMAGE_BUFFERS; i++) imageBuffers[i] = 0;

    switch (this->params.DetectionType)
    {
    case 1:
        imageBuffers[0] = image;
        imageBuffers[1] = filter_image;
        imageBuffers[2] = noise_image;
        break;
    default:
        imageBuffers[0] = filter_image;
        imageBuffers[1] = image;
        imageBuffers[2] = noise_image;
        break;
    }

    pDetector->Run(nx, ny, imageBuffers, this->params, nobj, objects);
    IObjectDetector::DeleteInstance(pDetector);

    // image statistics
    analysis_stats = new ImageAnalysisRecord;
    get_image_analysis_stats(analysis_stats, filter_image, noise_image,
                             nx, ny, this->params.GaussFactor);

    // positions (for compatibility reasons)
    positions = new st_pixel [nobj];
    for (i = 0; i < nobj; i++)
        memcpy(&positions[i], &(objects[i].max_pixel), sizeof(st_pixel));

    // more compatibility conversions
    positiondata = new st_fnoisepixel[nobj];
    for (i = 0; i < nobj; i++) {
        int x = objects[i].max_pixel.x;
        int y = objects[i].max_pixel.y;
        int idx = x * ny + y;

        if (this->params.DetectionType == 1)
        {
            positiondata[i].x = objects[i].position.x;
            positiondata[i].y = objects[i].position.y;
            positiondata[i].value = objects[i].intensity;
        }
        else
        {
            if ((!objects[i].is_split) || useInterpolatedPositionFlag) {
                positiondata[i].x = objects[i].position.x;
                positiondata[i].y = objects[i].position.y;
                positiondata[i].value = filter_image[idx];
            }
            else {
                // for objects with close neighbours,
                // do not trust interpolated positions
                // TODO: find better position estimate
#if 1
                positiondata[i].x = (float)x;
                positiondata[i].y = (float)y;
#else
                positiondata[i].x = objects[i].position.x;
                positiondata[i].y = objects[i].position.y;
#endif
#if 1
                positiondata[i].value = filter_image[idx];
#else
                positiondata[i].value = objects[i].max_pixel.value; //value of max_pixel is modifief by deblend_amplitudes() for split objects
#endif
            }
        }
        positiondata[i].noise = noise_image[idx];
    }

    status |= IMAGEDATA_STATE_ANALYSED;
    TRACE("Leaving ImageData[%p]::analyse()\n", (void *)this);
}
