/*
 * The Image class holds the data for a 16-bit image.
 * Main operations are loading an image from a tiff or stack file and
 * analysing the image.
 * The ImageList class holds a whole set of images at the same time;
 * Usually the images come from the same stack file. Mulitple stack files
 * can also be held in the ImageList.
 *
 * Copyright (c) Solexa 2003, 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
//---------------------------------------------------------------------------

#include <cstdio>
#include <cmath>
#include <fstream>
#include <algorithm>
#include <exception>
#include <stdexcept>
#include <cassert>
#include "Tools.h"
#include "ImageUtil.h"
#include "Images.h"
#include "ImageProcess.h"
#include "ProcessFilter.h"
#include "Message.h"
//---------------------------------------------------------------------------

ImageProperties::ImageProperties() : MinScale(2000), MaxScale(10000),
                                     StageXPosition(0.), StageYPosition(0.),
                                     BitsPerPixel(16), ExposureTime(0.)
                                     
{
    ImageDescription[0] = '\0';
    StageLabel[0]       = '\0';
}


Image::Image()
{
    snprintf(filename, sizeof(filename), "none");
    image      = NULL;
    state      = 0;
    active     = 1;
    properties = new ImageProperties();
    data       = NULL;
} 

/*
 * load image meta information from a tiff file;
 * use existing tiff file handle <hd> if given.
 */
int Image::loadProperties(TIFF_FILE *hd)
{
    if (state & IMAGE_STATE_PROPERTIES)
        return 0;
    if (!(state & IMAGE_STATE_FILENAME)) {
        error("Image::loadProperties: No filename set!\n");
        return 1;
    }
    int flag = 0;
    int error_code = 0;
    if (!hd) {
        flag = 1;
        hd = fopen_tiff(filename, "rb", 0);
        if (!hd) {
            error("Image::loadProperties: Failed to open %s\n",
                  filename);
            error_code = 1;
            return error_code;
        }
    }
    error_code = loadSize(hd);        
    error_code = error_code || read_image_properties(hd);
    if (flag)
        fclose_tiff(hd);
    state |= IMAGE_STATE_PROPERTIES;
    return error_code;
}

/*
 * load the x, y image dimensions from a tiff file
 */
int Image::loadSize(TIFF_FILE *hd)
{
    if (state & IMAGE_STATE_SIZE)
        return 0;
    if (!(state & IMAGE_STATE_FILENAME))
        return 1;
    int x, y, s;
    int error_code = 0;
    if (hd)  {
        error_code = fget_tiff_size(hd, &x, &y, &s);
    }
    else {
        hd = fopen_tiff(filename, "rb", 0);
        if (hd) {
            error_code = fget_tiff_size(hd, &x, &y, &s);
            fclose_tiff(hd);
        }
        else 
            return 1;
    }
    setSize(x, y, s);
    return error_code;
}

/*
 * set the x,y dimensions of an image
 */
void Image::setSize(int x, int y, int totalpos)
{
    nx = x;
    ny = y;
    if (totalpos > 0)
        nframes = totalpos;
    snprintf(id, sizeof(id), "%s (%i/%i, %ix%i)",
             filename, file_index+1, nframes, nx, ny);
    state |= IMAGE_STATE_SIZE;
}


/*
 * set the filename and the index of the image plane being loaded
 * from a stack file. 
 */
void Image::setFilename(const char fn[], int pos, int totalpos)
{
    params.SetImageDefaults();
    snprintf(filename, sizeof(filename), "%s", fn);
    file_index = pos;
    nframes = pos+1;
    active = 1;
    state = IMAGE_STATE_FILENAME;
    if (totalpos > 0)
        nframes = totalpos;
}

/*
 * do not associate the image with a file;
 * instead pass image contents directly from memory.
 * WARNING:
 * This function has not been tested for a long time;
 * it is likely that not enough information is set to analyse
 * the image.
 */
void Image::setImage(unsigned short *pixels, int pnx, int pny)
{
    clear();
    snprintf(filename, sizeof(filename), "none");
    state = IMAGE_STATE_IMAGE;
    image = pixels;
    params.SetImageDefaults();
    active = 1;
    file_index = 0;
    nframes = 1;
    setSize(pnx, pny);
}

std::string Image::get_id() const { return std::string(id); }
std::string Image::get_filename() const { return std::string(filename); }

int Image::get_frame_index() { return file_index; }
int Image::get_frame_total() { return nframes; }

PositionData *Image::get_objects() {
#if 1
    if (! (state & IMAGE_STATE_OBJECTS))
        if (analyse())
            return 0;
#endif
    return &objects;
}

unsigned short *Image::get_image()
{
    if (load())
        return 0;
    return image;
}

ImageData *Image::getData()
{
    if (analyse())
        return 0;
    return data;
}

void Image::clearData() 
{
    if ((state & IMAGE_STATE_DATA) && (data != NULL)) {
        delete data;
        data = NULL;
        state &= (~IMAGE_STATE_DATA);
    }
}

void Image::clearImages()
{
    if ((state & IMAGE_STATE_IMAGE) && (image != NULL)) {
        delete [] image;
        image = NULL;
        state &= (~IMAGE_STATE_IMAGE);
    }
    if (data)
        data->clearImages();
}

void Image::clear()
{
    clearImages();
    clearData();
}

/*
 * load the 16-bit image data associated with the image.
 * this called has to be preceded by an initialise call setting the filename.
 */
int Image::load(TIFF_FILE *fp)
{
    int error_code = 0;
    if (! (state & IMAGE_STATE_FILENAME)) {
        error("Image::load: No filename set!\n");
        return 1;
    }
    if (! (state & IMAGE_STATE_IMAGE)) {
        int flag = 0;
        if (! fp) {
            flag = 1;
            fp = fopen_tiff(filename, "rb", 1);
        }
        /* NOTE: The whole file is read in one read.
         * This scheme is ok for tiff files with a single image,
         * but not optimal for stack files with multiple layers as
         * the whole file gets read again for each layer.
         */
        if (fp) {
            error_code = loadSize(fp);
            error_code = error_code || loadProperties(fp);

            unsigned short *rawdata = new unsigned short[nx*ny];
            assert(rawdata);
            int code = fget_tiff_data(fp, rawdata, nx*ny, nx*ny*file_index);
            error_code = error_code || code;
            if (code)
                warning("Failed to read image data segment (error code %i)"
                        " from %s. Setting intensities to zero.\n",
                        code, filename);
            image = new unsigned short[nx*ny];
            assert(image);
            // transpose image
            int x, y;
            for (x = 0; x < nx; x++)
                for (y = 0; y < ny; y++)
                    image[x*ny+y] = rawdata[y*nx+x];
            delete [] rawdata;
            if (flag)
                fclose_tiff(fp);
            state |= IMAGE_STATE_IMAGE;
        }
        else
            error_code = 1;
    }
    return error_code;
}

/*
 * read image meta-information from the TIFF or stack file
 * and store it in the properties and instrument structs
 */
int Image::read_image_properties(TIFF_FILE *fp)
{
    if (!fp)
        return 1;

    // read image description
    char *description;
    int desc_len = fget_tiff_field_string(fp, TIFFTAG_IMAGEDESCRIPTION,
                                          &description);

    /* Parse image description for exposure times. This is used by 
     * MetaMorph images and alpha images.
     */
    if (desc_len > 0) {
        // find part of the string that referes to the n-th image
        char *s = description;
        int i = 0;
        while ((i < file_index) && (s < description+desc_len)) {
            i++;
            s += strlen(s)+1;
        }
        if (s >= description+desc_len)
            throw std::range_error("Image description does not have enough fields");
        if (sscanf(s, "Exposure: %f ms", &properties->ExposureTime) == 1)
            properties->ExposureTime /= 1000.;
        else
            sscanf(s, "Exposure: %f s", &properties->ExposureTime);
        strncpy(properties->ImageDescription, s, 
                sizeof(properties->ImageDescription));
        properties->ImageDescription[sizeof(properties->ImageDescription)-1] = '\0';
        free(description);
    }
    else {
        double value;
        // no image description, check if file is produced by Genopsys
        if (fget_tiff_field_double(fp, TIFFTAG_MANTEIA_IMAGE_TEMPERATURE,
                                   &value)
            || fget_tiff_field_double(fp, TIFFTAG_MANTEIA_CAMERA_GAIN,
                                   &value)) {
            snprintf(properties->ImageDescription, 
                     sizeof(properties->ImageDescription), "Manteia instrument");
        }
        else
            snprintf(properties->ImageDescription, 
                     sizeof(properties->ImageDescription), "Unknown");
    }
    fget_tiff_field_uint16(fp, TIFFTAG_BITSPERSAMPLE,
                           &(properties->BitsPerPixel));                       

    // set default values for some parameters
    properties->BitsPerPixel = 16;
    properties->MinScale = (unsigned int) 2000.f;
    properties->MaxScale = (unsigned int)10000.f;
           
    // Parse additional information for Metamorph files
    float fbits = properties->BitsPerPixel;
    fget_tiff_uic1_long_field(fp, UICTAG_MINSCALE, &properties->MinScale);
    fget_tiff_uic1_long_field(fp, UICTAG_MAXSCALE, &properties->MaxScale);
    fget_tiff_image_float_property(fp, "BitsPerPixel", &fbits);
    properties->BitsPerPixel = (unsigned short)fbits;
         
    properties->StageXPosition =  properties->StageYPosition = 0.;
    unsigned int *stagepos = new unsigned int [4*nframes];
    if (fget_tiff_uic4_longs_field(fp, UICTAG_STAGEPOSITION, stagepos, 
                                   4*nframes)) {
        properties->StageXPosition = (float)(int)stagepos[4*file_index]
            / (float)(int)stagepos[4*file_index+1];
        properties->StageYPosition = (float)(int)stagepos[4*file_index+2]
            / (float)(int)stagepos[4*file_index+3];
    }
    else {
        double value;
        if (fget_tiff_field_double(fp, TIFFTAG_MANTEIA_X_STAGE_POSITION,
                                   &value))
            properties->StageXPosition = (float)value;
        if (fget_tiff_field_double(fp, TIFFTAG_MANTEIA_Y_STAGE_POSITION,
                                   &value))
            properties->StageYPosition = (float)value;
        if (fget_tiff_field_double(fp, TIFFTAG_MANTEIA_EXPOSURE_TIME,
                                   &value))
            properties->ExposureTime = (float)value;
    }
    delete [] stagepos;

#if 0
    printf("# MinScale: %5u  MaxScale: %5u  BitsPerPixel: %2i  "
           "StagePosition: %6.1f %6.1f\n",
           properties->MinScale, properties->MaxScale, properties->BitsPerPixel,
           properties->StageXPosition, properties->StageYPosition);
#endif
    return 0;
}

/*
 * run the image analysis routines to identify the molecules in the image
 */
int Image::analyse()
{
    int error_code = 0;
    TRACE("Entering Image[%p]::analyse()\n", (void *)this);
    if (! (state & IMAGE_STATE_IMAGE)) {
        if (load()) {
            error("Image::analyse(): Failed to load.");
            return 1;
        }
    }
    if (! (state & IMAGE_STATE_IMAGE)) {
        return 0;
    }
    if (data != NULL) 
        return 0;
    data = new ImageData(image, nx, ny, properties->BitsPerPixel);
    assert(data);
    data->params = this->params;
    data->useInterpolatedPositionFlag = this->useInterpolatedPositionFlag;
    data->analyse();
    state |= IMAGE_STATE_DATA;

    if (data->nobj > 0) {
        st_fnoisepixel *posdata = new st_fnoisepixel [data->nobj];
        memcpy(posdata, data->positiondata, (data->nobj)*sizeof(st_fnoisepixel));
        objects.setData(posdata, data->nobj);
    }
    state |= IMAGE_STATE_OBJECTS;
    TRACE("Leaving Image[%p]::analyse()\n", (void *)this);
    return error_code;
}

Image::~Image()
{
    clear();
    delete properties;
}

int ImageList::load()
{
    clearImages();
    
    int i;
    for (i = 0; i < (int)filename.size(); i++) {
        TIFF_FILE *fp = fopen_tiff(filename[i], "rb", 1);
        if (!fp) {
            error("ImageList.load() failed to open %s.\n", filename[i]);
            return 1;
        }
         // get image properties
        int n, nx, ny, ni;
        if (fget_tiff_size(fp, &nx, &ny, &ni)) {
            error("ImageList.load() failed on %s.\n", filename[i]);
            return 2;
        }
        for (n = 0; n < ni; n++) {
            Image *image = new Image;
            assert(image);
            image->setFilename(filename[i], n, ni);
            image->setSize(nx, ny);
            int error_code = image->load(fp);
            push_back(image);
            if (error_code) {
                error("Could not open image layer %i from %s.",
                      n+1, filename[i]);
                for (unsigned int j = 0; j < size(); j++)
                    delete at(j);
                return 3 + error_code;
            }
        }
        fclose_tiff(fp);
    }
    return 0; 
}

int ImageList::createEmptyImages(int nx, int ny, int ni)
{
    clearImages();
    int i, n;
    for (n = 0; n < ni; n++) {
        Image *image = new Image;
        unsigned short *pixels = new unsigned short [nx*ny];
        for (i = 0; i < nx*ny; i++)
	    pixels[i] = 0;
        image->setImage(pixels, nx, ny);
        push_back(image); 
    }
    return 0;  
}

/*
 * load a Tiff file or a stack file containing multiple image planes
 */
int ImageList::loadDimensions(const char name[])
{
    addFile(name);
    int code = 0;
    TIFF_FILE *fp = fopen_tiff(name, "rb", 0);
    if (fp) {
        // get image properties
        int n, nx, ny, ni;
        if (fget_tiff_size(fp, &nx, &ny, &ni))
            code = 2;
        fclose_tiff(fp);
        if (code)
            return code;

        // load image information
        for (n = 0; n < ni; n++) {
            Image *image = new Image;
            image->setFilename(name, n, ni);
            image->setSize(nx, ny);
            push_back(image);
        }
    }
    else
        code = 1;
    return code;
}

void ImageList::clearImages()
{
    for (ImageList::iterator i(begin()); i != end(); ++i)
        delete (*i);
    clear();
}

void ImageList::clearList()
{
    clearImages();
    for (std::vector<char *>::iterator i(filename.begin()); i != filename.end(); ++i)
        delete [] (*i);
    filename.clear();
}

ImageList::~ImageList()
{
    clearList();
}

/*
 * replace a file extension with a new filename suffix
 */
static void ReplaceSuffix(char fn[], int len, char newsuffix[])
{
    char *suffix = strrchr(fn, '.');
    if (suffix && (suffix+strlen(newsuffix)+1 < fn+len))
        strncpy(suffix, newsuffix, strlen(newsuffix)+1);
}

/*
 * for each stack file, save another file with additional image info.
 * Used only in the GUI version; mostly obsolete now.
 */
void ImageList::write_info()
{
    AnalysisParameters defaultParams;
    int j;
    char previous_filename[FILENAME_MAX] = "";
    for (j = 0; j < (int)size(); j++) {
        char filename[FILENAME_MAX];
        snprintf(filename, FILENAME_MAX, at(j)->get_filename().c_str());
        if (!strcmp(previous_filename, filename)) {
            snprintf(previous_filename, FILENAME_MAX, filename);
            ReplaceSuffix(filename, FILENAME_MAX, "_info.txt");
            float thresh = at(j)->params.GaussFactor;
            if (thresh != defaultParams.GaussFactor) {
                FILE *fp = fopen(filename, "w");
                if (fp) {
                    fprintf(fp, "GaussFac=%f\n", thresh);
                    fclose(fp);
                }
            }
            else
                remove(filename);
        }
    }
}

/*
 * for each stack file, read a text file with additional image info.
 * Used only in the GUI version; mostly obsolete now.
 */
void ImageList::read_info()
{
    AnalysisParameters defaultParams;
    int j;
    char previous_filename[FILENAME_MAX] = "";
    for (j = 0; j < (int)size(); j++) {
        char filename[FILENAME_MAX];
        float threshold = defaultParams.GaussFactor;
        snprintf(filename, FILENAME_MAX, at(j)->get_filename().c_str());
        if (!strcmp(previous_filename, filename)) {
            snprintf(previous_filename, FILENAME_MAX, filename);
            ReplaceSuffix(filename, FILENAME_MAX, "_info.txt");
            FILE *fp = fopen(filename, "r");
            if (fp) {
                fscanf(fp, "GaussFac=%f", &threshold);
                fclose(fp);
            }
        }
        at(j)->params.GaussFactor = threshold;
    }
}
