/*
 * Routines to handle cluster intensities in different frequency channels
 * and cycles.
 * 
 * Copyright (c) Solexa   2005, 2006
 *               Illumina 2007
 * Author:       Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#else
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>
#endif
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <sys/stat.h>
#include "IntensityData.h"
#include "ImageUtil.h"
#include "Stats.h"
#include "Maths.h"
#include "IOProcess.h"
#include "Message.h"
#include "GzPack.hh"
#include "File_Buffer.h"


Compression::Compression(const std::string &compressionName) throw (UnsupportedCompressionException)
    : compressionType_(getCompressionType(compressionName))
{
}

IntensityData::CompressionType Compression::getCompressionType(const std::string &compressionName) throw (UnsupportedCompressionException)
{
    using namespace std;
    static const std::map<std::string, IntensityData::CompressionType> compressionTypeMap = getCompressionTypeMap();
    const map<string, IntensityData::CompressionType>::const_iterator found = compressionTypeMap.find(compressionName);
    if (compressionTypeMap.end() == found)
    {
        string message = compressionName + string(" is not a supported compression types. Supported types are: ");
        for(map<string, IntensityData::CompressionType>::const_iterator i = compressionTypeMap.begin(); i != compressionTypeMap.end(); ++i)
            message += string(compressionTypeMap.begin() == i ? "" : ", ") + i->first;
        throw UnsupportedCompressionException(message);
    }
    else
    {
        return found->second;
    }
}

std::map<std::string, IntensityData::CompressionType> Compression::getCompressionTypeMap()
{
    std::map<std::string, IntensityData::CompressionType> compressionTypeMap;
    compressionTypeMap[std::string("none")] = IntensityData::compressionNone;
#ifdef HAVE_ZLIB
    compressionTypeMap[std::string("gzip")] = IntensityData::compressionGzip;
#endif
#ifdef HAVE_BZLIB
    compressionTypeMap[std::string("bzip2")] = IntensityData::compressionBzip2;
#endif
    return compressionTypeMap;
}


IntensityData::IntensityData(const char filename[], CompressionType compression, const bool ipar) 
                            : m_SkippedColumns(ipar ? 0 : 4)
                            , m_Ipar(ipar)
{
    m_Filename    = filename;
    m_Compression = compression;
}

IntensityData::~IntensityData() {}

void IntensityData::setCompressionType(CompressionType compression)
{ 
    m_Compression = compression;
}

int  IntensityData::write(const float * intensities,
                          const int     numClusters,
                          const int     numChannels)
{
    return write_signal_array(m_Filename.c_str(), intensities, numClusters, numChannels,
                              m_Compression);
}

int  IntensityData::write(const std::vector<float * > & signal,
                          const int                     numClusters, 
                          const int                     numChannels,
                          const int                     lane, 
                          const int                     tile, 
                          const Position              * positions)
{
    return write_signal_array(m_Filename.c_str(), signal, numClusters, numChannels,
                              m_Compression, lane, tile, positions);
}

// Note: in input, numChannels is not the number of channels, but the maximum number of 
// channels times the maximum number of cycles
int IntensityData::read(float     * signal,
                        int       * numClusters,
                        int       * numChannels,
                        const int   skipColumns)
{
    // this method does not support IPAR format
    assert(!m_Ipar);
    return read_signal_array(m_Filename.c_str(), signal, numClusters, numChannels,
                             m_Compression, skipColumns);
}

int IntensityData::read(std::vector<float * > & signal,
                        int                   * numClusters,
                        int                   * numChannels,
                        PositionIndex        ** indexPositions,
                        const unsigned int      lastCycle,
                        const unsigned int      firstCycle)
{
    if (m_Ipar)
    {
        // Note: no position in IPAR files (they are in the pos file)
        assert(0 == indexPositions);
        // Note: IPAR files are gzipped
        assert(IntensityData::compressionGzip == m_Compression);
        try
        {
            // Note: m_Filename is not the name of the file, but some sort of
            // base name to which an extension should be appended if the file
            // is compressed. GzPack files are gzipped, with extension '.p.gz'.
            const std::string fileName = m_Filename + std::string(".p.gz");
            GzPack::GzPackReader gzPackReader(fileName);
            *numClusters = gzPackReader.getNumberOfClusters();
            *numChannels = gzPackReader.getNumberOfChannels();
            unsigned int currentCycle = 1;
            while (!gzPackReader.eof() && currentCycle < firstCycle)
            {
                gzPackReader.skipCycle();
                ++currentCycle;
            }
            while (!gzPackReader.eof() && (0 == lastCycle || lastCycle >= currentCycle))
            {
                signal.push_back(new float[(*numClusters) * (*numChannels)]);
                gzPackReader.getIntensityData(signal.back());
                ++currentCycle;
            }
            return 0;
        }
        catch (std::logic_error &e)
        {
            error(e.what());
            return 1;
        }
    }
    else
    {
        int    errorCode = 0;
        signal =  read_signal_array(m_Filename.c_str(), numClusters, numChannels,
                                    &errorCode, m_Compression, indexPositions, lastCycle, firstCycle);
        return errorCode;
    }
}

/**
 ** Helper method to convert the IntensityData::CompressionType into the File_Buffer::CompressionType
 **/
File_Buffer::Compression_Type convert(const IntensityData::CompressionType &compression) throw(UnsupportedCompressionException)
{
    switch (compression)
    {
        case IntensityData::compressionNone:
            return File_Buffer::compressionNone;
            break;
        case IntensityData::compressionGzip:
            return File_Buffer::compressionGzip;
            break;
        case IntensityData::compressionBzip2:
            return File_Buffer::compressionBzip2;
            break;
        default:
            throw UnsupportedCompressionException("Unable to convert");
    }
}

/************************************************************************/

/*
 * Write an intensity file.
 * The intensity file is a text file with one line per cluster.
 * Intensities are arranged in groups of <ncol>, each intensity
 * separated by a blank. The number of such groups is the cycle number;
 * each group is separated by a tab.
 * If more than one cycle is present, there may be 4 additional, tab-separated
 * columns at the beginning of each line, identifying the cluster position.
 */
int write_signal_array(const char                             filename[],
                       const std::vector<float * >          & signal,
                       const int                              nobj,
                       const int                              ncol,
                       const IntensityData::CompressionType   compression,
                       const int                              lane,
                       const int                              tile,
                       const Position                       * positions)
{
    int   i, j, k;
    int   idx, /*totalidx,*/ count;
    int   code            = 0;
    int   ncycles         = int(signal.size());
    int   buffersize      = (22+ncycles*ncol*11);
    char *buffer          = new char [buffersize];
    //int   totalbuffersize = nobj * buffersize;
    //char *totalbuffer     = new char [totalbuffersize]; 
    assert(buffer != 0);
    //assert(totalbuffer != 0);

    /*
     * format output and write to a char buffer
     */
    //totalidx = 0;
    const File_Buffer::Compression_Type compr_type(convert(compression));
    File_Buffer file_buffer(add_std_extension(std::string(filename),
                                              compr_type),
                            compr_type);
    for (i = 0; i < nobj; i++) {
        idx = 0;
        if (positions != NULL) {
            PositionIndex index(lane, tile, positions[i]);
            idx = index.snprint(buffer, 20);
            idx += sprintf(buffer + idx, "\t");
        }
        for (j = 0; j < ncycles; j++) {
            for (k = 0; k < ncol; k++) {
                count = snprintf(buffer+idx, 10, "%6.1f", signal[j][i*ncol+k]);
                if (count >= 10) // last character written is '\0'
                    count = 9;
                idx += count;
                if (k != ncol-1)
                    idx += sprintf(buffer+idx, " ");
                else if (j != ncycles-1)
                    idx += sprintf(buffer+idx, "\t");
                else
                    idx += sprintf(buffer+idx,"\n");
            }
        }
        // Instead of writing to totalbuffer, write directly to the file
        //memcpy(totalbuffer+totalidx, buffer, strlen(buffer));
        //totalidx += strlen(buffer);
        file_buffer.write_buffer(buffer, idx);
    }

    /*
     * write formatted buffer to file
     */
#if 0
    if (compression == IntensityData::compressionNone) {
        code = write_file(filename, totalbuffer, totalidx);
    }
    else if (compression == IntensityData::compressionGzip) {
        std::string filename2(filename);
        filename2 += ".gz";
        code = write_gzip_file(filename2.c_str(), totalbuffer, totalidx);
    }
    else if (compression == IntensityData::compressionBzip2) {
        std::string filename2(filename);
        filename2 += ".bz2";
        code = write_bzip2_file(filename2.c_str(), totalbuffer, totalidx);
    }
#endif
    delete [] buffer;
    //delete [] totalbuffer;
    return code;
}

int write_signal_array(const char    filename[],
                       const float * intensity,
                       const int     numClusters,
                       const int     numChannels,
                       const IntensityData::CompressionType compression)
{
    std::vector<float * > intensities;
    intensities.push_back((float *)intensity);
    return write_signal_array(filename, intensities, numClusters, numChannels,
                              compression);
}


std::vector<std::string > read_signal_file(const char * filename, 
                                           const IntensityData::CompressionType  
                                                        compression,
                                           int        * errorCode)
{
    std::vector<std::string > lines;
    std::string               filename2(filename);
    char                     *read_buffer = 0;
    long                      read_buffersize = 0;
    if (compression == IntensityData::compressionNone) {
        *errorCode = read_file(filename2.c_str(), &read_buffer, &read_buffersize);
    }
    else if (compression == IntensityData::compressionGzip) {
        filename2 += ".gz";
        *errorCode = read_gzip_file(filename2.c_str(),
                                    &read_buffer, &read_buffersize);
    }
    else if (compression == IntensityData::compressionBzip2) {
        filename2 += ".bz2";
        *errorCode = read_bzip2_file(filename2.c_str(),
                                     &read_buffer, &read_buffersize);
    }

    if (! (*errorCode)) {
        lines.reserve(MAX_OBJECTS);
        const long  buffersize = 16384;
        char        *s          = new char [buffersize];
        assert(s);
        long        l, l1;
        l1 = 0;
        for (l = 0; l < read_buffersize; l++)
            if (read_buffer[l] == '\n') {
                //printf("%li %li %li\n", l, l1, read_buffersize);
                if (l-l1 >= buffersize) {
                    error("Line too long in %s: %li characters.",
                          filename2.c_str(), l-l1);
                    *errorCode = 1;
                }
                if (read_buffer[l1] != '#') {
                    long lineLength = Maths::min(l-l1+1, buffersize-1);
                    memcpy(s, &(read_buffer[l1]), lineLength);
                    s[lineLength] = '\0';
                    lines.push_back(std::string(s));
                }
                l1 = l+1;
            }
        delete [] s; 
    }
    if (read_buffersize > 0)
        delete [] read_buffer;
    return lines;
}


/*
 * read intensity data from a full text file, as described in the
 * run-folder spec.
 * Note: this function groups the values by cluster (all values for all 
 * clusters, for a given cycle are contiguous)
 */
std::vector<float * > read_signal_array(const char * filename,
                                        int        * nobj,
                                        int        * ncol,
                                        int        * errorno,
                                        const IntensityData::CompressionType   
                                                     compression,
                                        PositionIndex **indexPositions,
                                        const unsigned int lastCycle,
                                        const unsigned int firstCycle)
{
    std::vector<float * >    intensities;
    std::vector<std::string> lines;
    std::string              errormsg = "";
    const int                offset = 4;

    lines = read_signal_file(filename, compression, errorno);
    if (*errorno)
        return intensities;
    std::string buffer;
    *errorno = 0;
    *ncol = 0;
    *nobj = 0;
    buffer.reserve(10000);

    /*
     * count number of objects, cycles and columns
     */
    *nobj = int(lines.size());
    unsigned int ncycles = 0;
    if (*nobj > 0) {
        std::stringstream ss(lines[0]);
        while (getline(ss, buffer, '\t')) {
            ncycles++;
            if (ncycles > offset) {
                float dummy;
                int n = sscanf(buffer.c_str(),
                                  "%f %f %f %f", 
                                  &dummy, &dummy, &dummy, &dummy);
                if (*ncol == 0)
                    *ncol = n;
                if (*ncol != n) {
                    errormsg =  std::string(filename) + ", line 0"+
                                 ": Wrong number of intensities.";   
                    *errorno = 2;
                }
                
             }
        }
    }
    else
    {
        return intensities;
        *indexPositions = NULL;
    }
    ncycles -= 4;
    ncycles = (lastCycle == 0 ? ncycles : std::min(ncycles, lastCycle)) - firstCycle + 1;

    /* 
     * allocate memory
     */
    int i, j, k;
    for (j = 0; j < ncycles; j++) {
        float *array = new float [*nobj * *ncol];
        for (i = 0; i < *nobj * *ncol; i++)
             array[i] = 0.;
        intensities.push_back(array);
    }
    if (indexPositions != 0)
        *indexPositions = new PositionIndex [ *nobj ];

    for (i = 0; i < *nobj; i++) 
    {
#if 0   // slow C++ version
        std::stringstream ss(lines[i]);
        int cycle = 0;
        while (getline(ss, buffer, '\t')) {
            j = cycle - offset;
            if (j >= 0) {
                float dummy[4];
                int n = sscanf(buffer.c_str(), "%f %f %f %f", 
                               &(dummy[0]), &(dummy[1]),
                               &(dummy[2]), &(dummy[3]));
                for (k = 0; k < n; k++)
                    intensities[j][i * *ncol + k] = dummy[k];
                if (*ncol != n) {
                    std::cerr << "Error: " << filename << ", line " << i <<
                                 ": Wrong number of intensities" << std::endl;
                    *errorno = 2;
                }
            }
            cycle++;
            
            if (j >= ncycles) {
                std::cerr << "Error: " << filename << ", line " << i << 
                             ": Wrong number of cycles" << std::endl;
                *errorno = 3;
            }
        }
#else
        j = -offset - *ncol*(firstCycle - 1);
        const int ncolmax = ncycles * *ncol;
        int cycle;
        const char *buffer = lines[i].c_str();
        char *begin, *end;
        float value;
        end = (char *)buffer;
        do {
            begin = end;
            value = (float)strtod(begin, &end);
            if (j >= 0) {
                cycle = j / *ncol;
                k     = j % *ncol;
                intensities[cycle][i * *ncol + k] = value;
            }
            else if (indexPositions != 0) {
                switch (j+offset) {
                case 0: (*indexPositions)[i].lane = int(value); break;
                case 1: (*indexPositions)[i].tile = int(value); break;
                case 2: (*indexPositions)[i].X    = int(value); break;
                case 3: (*indexPositions)[i].Y    = int(value); break;
                }
            }
            j++;
        } while ((j < ncolmax) && (end != NULL) && (end != begin));
        if (j != ncolmax) {
            char msg[1024];
            snprintf(msg, sizeof(msg),
                     "%s, line %i: File corrupted. Expecting %i columns, "
                     "found %i.", filename, i, ncolmax+offset, j);
            errormsg = msg;
            *errorno = 2;
        }
#endif
    }
    
    if (*errorno != 0)
        error(errormsg.c_str());
    return intensities;
}

/*
 * auxiliary function to read intensity data; compare previous function
 * Note: this function groups the values by cycle (all values for all 
 * cycles, for a given cluster are contiguous)
 * Note: see the corresponding read method for the meaning of ncol
 */
int read_signal_array(const char  filename[],
                      float      *signal,
                      int        *nobj,
                      int        *ncol,
                      const IntensityData::CompressionType   compression,
                      const int   skipcols)
{
    int ncolmax          = *ncol;
    int nobjmax          = *nobj;
    *nobj                = 0;
    *ncol                = 0;
  
    int   errorno        = 0;
    std::string errormsg = "";
    std::vector<std::string > lines = read_signal_file(filename, compression,
                                                       &errorno);
    if (errorno)
        return errorno;


    int ncolfirst = 0;
    int lineno = 0;
    int i;
    for (i = 0; i < int(lines.size()); i++) {
        std::string line = lines[i];
        char *buffer = (char *)line.c_str();
        lineno ++;
        // ignore comments
        if (buffer[0] == '#')
            continue;
        // read molecule intensities
        if (*nobj < nobjmax) {
            int j = 0;
            char *begin, *end;
            float value;
            end = buffer;
            do {
                begin = end;
                value = (float)strtod(begin, &end);
                if (j >= skipcols)
                    signal[(*nobj)*ncolmax+j-skipcols] = value;
                j++;
            } while ((j < ncolmax+skipcols) && (end != NULL) && (end != begin));
            *ncol = j - skipcols;
            if (ncolfirst == 0) {
                ncolfirst = *ncol;
            }
            if (*ncol != ncolfirst) {
                char msg[1024];
                snprintf(msg, sizeof(msg),
                         "%s, line %i: File corrupted. Expecting %i columns, "
                         "found %i.", filename, lineno, ncolfirst, *ncol);
                errormsg = msg;
                errorno  = 2;
            }
            for (j = *ncol; j < ncolmax; j++)
                signal[(*nobj)*ncolmax+j] = 0.;
            (*nobj) ++;
        }
    }
    *ncol = ncolfirst;
    
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
}

