/*
 * Journal file parsing routines.
 * 
 * (c) Copyright Solexa   2005, 2006
 *               Illumina 2007
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
/*
 * Journal files provide a list of image filenames to the image processing
 * software, and the nucleotide the image corresponds to.
 */
 
#include <cassert>
#include "JournalParser.h"
#include "XMLSAXParser.h"
#include "XMLContentHandler.h"
#include "Message.h"

class JournalParser : public XMLContentHandler
{
public:
    JournalParser() : XMLContentHandler() {};
    virtual ~JournalParser() { filenames.clear(); }
    virtual void startDocument();
    virtual void startElement(const char name[], const Attributes *attr);
    virtual void endElement(const char name[]);
    virtual void characters(const char ch[], int start, int length);
    virtual void getElements(RunInfo *runInfo, int *ncycle,
                             std::vector<std::string > & filenames,
                             unsigned char**base_list);
private:
    std::string order;
    int numCycles;
    int numChannels;
    std::string compression;
    std::vector<std::string > filenames;
};

void JournalParser::startDocument()
{
    numCycles = 0;
    numChannels = 0;
    filenames.clear();
}

void JournalParser::startElement(const char name[], const Attributes *attr)
{
    std::string tag(name);
    if (tag == "Cycles") {
        numCycles = atoi(attr->getValue("Num").c_str());
    }
    if (tag == "Filename")
        filenames.push_back(attr->getValue("Path"));
    if (tag == "Channels") {
        numChannels = atoi(attr->getValue("Num").c_str());
        order = attr->getValue("Order");
    }
}


void JournalParser::endElement(const char name[])
{
}

void JournalParser::characters(const char ch[], int start, int length)
{
}

void JournalParser::getElements(RunInfo *runInfo, int *numCycle,
                                std::vector<std::string > & filenames2,
                                unsigned char**base_list)
{
    int i;
    runInfo->NumChannels = numChannels;
    *numCycle = numCycles;
    for (i = 0; i < int(filenames.size()); i++)
        filenames2.push_back(filenames[i]);
    assert(numChannels == static_cast<int>(order.length()));
    *base_list = new unsigned char [numChannels];
    for (i = 0; i < numChannels; i++)
        (*base_list)[i] = order[i];
}

int parse_journal(const char journal_filename[],
                  RunInfo *runInfo, int *ncycle,
                  std::vector<std::string > & filenames,
                  unsigned char**base_list)
{
    FILE *fh = fopen(journal_filename, "r");
    if (!fh) {
        error("Failed to open parameter file %s.", journal_filename);
        return 0;
    }

    JournalParser *handler    = new JournalParser();
    XMLSAXParser  *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);
    handler->getElements(runInfo, ncycle, filenames, base_list);
    delete handler;
    delete sax_parser;

    return 0;
}
