/*
 * Copyright (c) 2006-2007 Solexa
 * Author: Lorenzo Milazzo
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */


#include "sampling.h"
#include "savinio_common.h"
#include "Sorting.h"

#include "savinio_debugging.h"


float getYAxisData(const float *bin_data, const int b_start, const int b_end,
                   const MatrixSamplingParameters params)
{
    float *bin_data_t = new float[(b_end - b_start)];

    int i = 0;
    for (int j = b_start; j < b_end; ++j) {
        bin_data_t[i] = bin_data[j];
        i++;
    }

    quicksort(bin_data_t, b_end - b_start);

    int index_lbound = (int)(params.percentileY * (b_end - b_start));

    float yaxis_data_el = bin_data_t[index_lbound];

    delete [] bin_data_t;

    return yaxis_data_el;

}


//
// 'data sampling 1'
//
// sampling the intensities associated with the X and Y axes
//
// (note that there is one X-axis for each of the 12 archetypal directions and
// there are four different X-axis, one for each color)
//
void sampling1(int base_index, const int no_samples, const int no_bases, 
               const float *i_data,
               float *xaxis_data, float **yaxis_data,
               const MatrixSamplingParameters params)
{
    int i = 0, j = 0, k = 0;
    int no_bins = params.numBins;

    // intensities associated with one base (A, C, G, T)
    float *i_baseX = new float[no_samples];

    // intensities stored in the bins
    float **bin_data = new float * [ no_bases-1 ];
    for (j = 0; j < no_bases-1; j++)
        bin_data[j] = new float[no_samples * no_bases];

    // element position list; original position for each sorted element
    long *indices = new long[no_samples];

    // bin occupancy list; occupancy for each bin
    int *bin_occupancy = new int[no_bins];

    float Xel_lbound;
    float Xel_ubound;


    //--
    // sampling the X-axis
    for (j = 0; j < no_samples; j++)
        i_baseX[j] = i_data[base_index + j * no_bases];
    
    // sorting
    for (j = 0; j < no_samples; j++)
        indices[j] = j;

    quicksort_index(i_baseX, indices, long (no_samples));

    // binning
    int index_lbound = (int)(params.percentileXLower * no_samples);
    int index_ubound = (int)(params.percentileXUpper * no_samples);

    Xel_lbound = i_baseX[index_lbound];
    Xel_ubound = i_baseX[index_ubound];
    if (params.percentageXUpper > 0.)
        Xel_ubound = params.percentageXUpper * Xel_ubound;
    if (params.percentageXLower > 0.)
        Xel_lbound = params.percentageXLower * Xel_ubound;

    for (int b = 0; b < no_bins; b++) {
        xaxis_data[b] = Xel_lbound + (b + 0.5) * (Xel_ubound - Xel_lbound) / no_bins;
    }


    //--
    // sampling the Y-axis (three Y-axes)

    // binning
    int r = 0;
    float Xel_max;

    while (i_baseX[r] < Xel_lbound) {
        r++;
    }
    
    // identify dimensions other than the current main axis
    int *other_indices = new int [ no_bases - 1];
    j = 0;
    for (i = 0; i < no_bases; i++) 
        if (i != base_index) {
            other_indices[j] = i;
            j++;
        }
        
    j = 0;
    i = 0;
    for (int b = 0; b < no_bins; b++) {
        Xel_max = Xel_lbound + (b + 1) * (Xel_ubound - Xel_lbound) / no_bins;
        while (i_baseX[r] < Xel_max) {
            
            for (k = 0; k < no_bases-1; k++)
                bin_data[k][i] = i_data[other_indices[k] + indices[r] * no_bases];
            r++;
            i++;
            j++;
        }
        bin_occupancy[b] = j;
        j = 0;
        Xel_max = 0;
    }

    int b_start = 0;
    int b_end = 0;

    for (int b = 0; b < no_bins; b++) {
        b_end = b_start + bin_occupancy[b];
        for (j = 0; j < no_bases - 1; j++)
            yaxis_data[j][b] = getYAxisData(bin_data[j], b_start, b_end, params);
        b_start = b_end;
    }

 
    //--
    delete [] i_baseX;
    for (j = 0; j < no_bases-1; j++)
        delete [] bin_data[j];
    delete [] bin_data;
    delete [] indices;
    delete [] other_indices;
    delete [] bin_occupancy;


    //--
#ifdef DEBUG_SAMPLING1
    std::cout << "\n>> debug - sampling1";

    std::cout << "\nX axis: " << std::endl;
    for (int i = 0; i < numBins; ++i) {
        std::cout << xaxis_data[i] << " ";
    }
    for (j = 0; j < no_bases - 1; j++) {
        std::cout << "\nY axis " << (j+1) << ":" << std::endl;
        for (int i = 0; i < numBins; ++i) {
            std::cout << yaxis_data[j][i] << " ";
        }
    }
    std::cout << "\n";

#endif // DEBUG_SAMPLING1
    //--
}

//
// sequential search
//
int sequentialSearch(const float *data, const int no_el, const float key)
{
    int index;

    for (index = 0; index < no_el; index++) {
        if (key == data[index])
            return index;
    }

    return (-1);
}


//
// 'data sampling 2'
//
// sampling the intensities
//
// . choose the brightest intensities for each clusters
//
void sampling2(const float *i_data, const int no_points, const int no_cols,
               float *i_data_bA, float *i_data_bC, float *i_data_bG,
               float *i_data_bT, int &no_i_values_bA, int &no_i_values_bC,
               int &no_i_values_bG, int &no_i_values_bT)
{
    int index_el_max;
    int ia, ic, ig, it;
    float el_max;

    // intensities associated with a cluster
    float *i_cluster        = new float[no_cols];
    float *i_cluster_sorted = new float[no_cols];

    ia = 0;
    ic = 0;
    ig = 0;
    it = 0;
    for (int r = 0; r < no_points; ++r) {
        for (int c = 0; c < no_cols; ++c) {
            i_cluster[c] = i_data[c + r * no_cols];
        }
        for (int c = 0; c < no_cols; ++c) {
            i_cluster_sorted[c] = i_data[c + r * no_cols];
        }

        quicksort(i_cluster_sorted, no_cols);

        el_max = i_cluster_sorted[(no_cols - 1)];
        index_el_max = sequentialSearch(i_cluster, no_cols, el_max);

        if (index_el_max == 0) {
            i_data_bA[ia] = el_max;
            ia++;
        }
        else if (index_el_max == 1) {
            i_data_bC[ic] = el_max;
            ic++;
        }
        else if (index_el_max == 2) {
            i_data_bG[ig] = el_max;
            ig++;
        }
        else if (index_el_max == 3) {
            i_data_bT[it] = el_max;
            it++;
        }

    }

    no_i_values_bA = ia;
    no_i_values_bC = ic;
    no_i_values_bG = ig;
    no_i_values_bT = it;

    delete [] i_cluster;
    delete [] i_cluster_sorted;
}


//
// 'data sampling 3'
//
// sampling the intensities
//
// . filter out the very low intensities
//
void sampling3(const float * i_data_i,
               const int     no_i_values_i,
               float       * i_data_f,
               int         & no_i_values_f)
{
    int index_ubound;
    float lbound;

    index_ubound = (int)(PERCENTILE_UPPER_X_S3 * (no_i_values_i + 1));
  
    lbound = (PERCENTAGE_S3 * i_data_i[(index_ubound-1)]);

    int j = 0;
    for (int i = 0; i < no_i_values_i; i++) {
        if (i_data_i[i] > lbound) {
            i_data_f[j] = i_data_i[i];
            j++;
        }
    }

    no_i_values_f = j;
}
