/*
 * Copyright (c) 2006-2007 Solexa
 * Author: Lorenzo Milazzo
 *         Klaus Maisinger
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

// A program to estimate the cross-talk matrix for the 'color-correction'
// in four-dye florescence-based DNA sequencing.


#include <iostream>
#include <fstream>
#include <sstream>
#include <cassert>
#include <cmath>
#include <getopt.h>

#include "MatrixEstimator.h"
#include "IntensityData.h"
#include "FrequencyMatrix.h"
#include "Message.h"
#include "CommandLine.hh"


class SavinioOptions : public CommandLine::Options
{
public:
    SavinioOptions()
        : CommandLine::Options()
        , iparFlag(false)
        , cycle()
        , compressionName("none")
        , compression(Compression(compressionName))
        , intensityFileName("")
        , outputFileName("")
    {
    }
    bool iparFlag;
    int  cycle;
private:
    std::string compressionName;
public:
    Compression compression;
    std::string intensityFileName;
    std::string outputFileName;
private:
    virtual int getOptions(int argc, char * const argv[])
        throw (CommandLine::UnknownOptionException, 
               CommandLine::MissingOptionException, 
               CommandLine::InvalidArgumentException);
    virtual void processOptions() throw (CommandLine::InvalidArgumentException);
    virtual void getParameters(int argc, char * const argv[]) throw (CommandLine::InvalidArgumentException);
    virtual int usage(int argc, char * const argv[], const std::string &message = "");
};

int SavinioOptions::getOptions(int argc, char * const argv[])
        throw (CommandLine::UnknownOptionException,
               CommandLine::MissingOptionException,
               CommandLine::InvalidArgumentException)
{
    const char *shortOptionList = "hIc:z:o:";
    const struct option longOptionList[] =
    {   {"help",         no_argument,            0, 'h'}
    ,   {"IPAR",         no_argument,            0, 'I'}
    ,   {"cycle",        required_argument,      0, 'c'}
    ,   {"compression",  required_argument,      0, 'z'}
    ,   {"output",       required_argument,      0, 'o'}
    ,   {0,              0,                      0,  0}
    };
    int longOptionIndex;
    int option;
    // Note: "case 'h'" returns immediately
    while (-1 != (option = getopt_long(argc, argv, shortOptionList, longOptionList, &longOptionIndex)))
    {
        switch (option)
        {
            case 'I':
                iparFlag = true;
                compressionName = "gzip";
                break;
            case 'c':
                cycle = CommandLine::Options::convert<unsigned int>(optarg);
                break;
            case 'z':
                compressionName = optarg;
                break;
            case 'o':
                outputFileName = optarg;
                break;
            case 'h':
                return usage(argc, argv);
                break;
            case '?':
            case ':':
            default:
                throw CommandLine::UnknownOptionException("");
                break;
        };
    }
    return RUN;
}

void SavinioOptions::processOptions() throw (CommandLine::InvalidArgumentException)
{
    try
    {
        compression = Compression(compressionName);
    }
    catch(const UnsupportedCompressionException &e)
    {
        throw CommandLine::InvalidArgumentException(e.what());
    }
    
}

void SavinioOptions::getParameters(int argc,  char * const argv[]) throw (CommandLine::InvalidArgumentException)
{
    if (argc == optind + 1)
    {
        intensityFileName = argv[optind];
        ++optind;
    }
    else
    {
        throw CommandLine::InvalidArgumentException(std::string("The matrix estimate requires exactly one intensity file"));
    }
}

int SavinioOptions::usage(int argc, char * const argv[], const std::string &message)
{
    if (0 != message.size())
    {
        error(message.c_str());
    }
    const std::string program = argv[0];
    const std::string padding = std::string(program.length(), ' ');
    std::ostringstream os;
    os << argv[0] 
       << " [-h] [-I] [-o <output file>] [-z <compression>] [-c <cycle>] <input file>" << std::endl
       << std::endl
       << "-h                help\n"
       << "-I                IPAR data format. [false]\n"
       << "-o <output file>  write output to file\n"
       << "-z <compression>  set the compression type to 'none', 'gzip' or 'bzip2'. [none]\n"
       << "-c <cycle>        set the sequencing cycle for matrix estimation\n"
       << std::endl
       << "Estimate frequency cross-talk matrices."
       << std::endl;
    info(os.str().c_str());
    return NO_RUN;
}


int main(int argc, char *argv[])
{
    SavinioOptions options;
    if (!(CommandLine::Options::RUN == options.init(argc, argv)))
    {
        return 1;
    }

    IntensityData inputData(options.intensityFileName.c_str(),
                            options.compression,
                            options.iparFlag);

    // reading input file
    std::vector<float *> signal;
    int                  no_objects, no_cols;
    const unsigned int   cycleNumber = std::max(1, options.cycle);
    inputData.read(signal, &no_objects, &no_cols, 0, cycleNumber, cycleNumber);
    float               *intensities = signal.front();
    // estimating the cross-talk matrix  
    MatrixEstimator *estimator = new MatrixEstimator();
    FrequencyMatrix matrix;
    float *matrix_data = new float [ no_cols * no_cols ];
    int errorCode =  estimator->estimateMatrix(matrix_data, intensities, no_objects, no_cols);
    if (errorCode == MatrixEstimator::errorNone) {
        matrix.setMatrix(matrix_data, no_cols);
    }
    else {
        // initial estimate of the cross-talk matrix
        std::string message = estimator->getErrorMessage();
        message += std::string(" (input file: ") + options.intensityFileName
                + std::string(").");
        warning(message.c_str());
        matrix.identity(no_cols);
    }

    delete [] matrix_data;
    delete    estimator;
    delete [] intensities;

    // printing out the cross-talk matrix
    matrix.write();

    // writing output file
    if (!options.outputFileName.empty()) {
        matrix.write(options.outputFileName.c_str());
    }

    return (0);
}
