/*
 * Copyright (c) 2006-2007 Solexa
 * Author: Lorenzo Milazzo
 *
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */


#include <cmath>

#include "distributional_adequacy.h"
#include "estimate_xtalk_mx.h"
#include "Optimise.h"
#include "sampling.h"
#include "savinio_common.h"
#include "scale_normalise_xtalk_mx.h"
#include "Sorting.h"

#include "savinio_debugging.h"
#include "savinio_exceptions.h"


//
// scaling two distributions
//
float scaleDistr(const int no_el1, const int no_el2, float *distr1,
                 float *distr2)
{
    int i;
    int funcallcount;
    float ax, bx, cx;
    float fa, fb, fc;
    float xtol;
    float xmin;
    float fmin;
    float xbracket[3];


    // structure which contains the arguments for
    // the function associated to the KS statistic
    struct args_ks s_args_ks;

    s_args_ks.no_el1 = no_el1;
    s_args_ks.no_el2 = no_el2;
    s_args_ks.data1 = new float[no_el1];
    for (i = 0; i < no_el1; ++i) {
        s_args_ks.data1[i] = distr1[i];
    }
    s_args_ks.data2 = new float[no_el2];
    for (i = 0; i < no_el2; ++i) {
        s_args_ks.data2[i] = distr2[i];
    }
    s_args_ks.p_func = scaledCumulDistrFunc;

    struct args_ks *p_args_ks;

    p_args_ks = &s_args_ks;

    void *p_args_ks_v = (void *)p_args_ks;

    ax = (float)1 / 2.0;
    bx = 2.0;

    // 'KSStatisticFunc(x, args)'
    // x = scale factor (independent variable)

    // bracketing procedure
    bracket(KSStatisticFunc, p_args_ks_v, &ax, &bx, &cx, &fa, &fb, &fc);

    xbracket[0] = ax;
    xbracket[1] = bx;
    xbracket[2] = cx;
    xtol = 2.0e-3;

    // minimising the function associated to the KS statistic by using the Brent's method
    brent(KSStatisticFunc, p_args_ks_v, xbracket, xtol, &xmin, &fmin, &funcallcount);

    delete [] s_args_ks.data1;
    delete [] s_args_ks.data2;


    //--
    #ifdef DEBUG_MIN
      std::cout << "\n>> debug - minimisation" << std::endl;

      std::cout.setf(std::ios::fixed, std::ios::floatfield);
      std::cout.precision(7);

      std::cout << "bracketing: " << std::endl;
      std::cout << " ax, bx, cx: " << ax << " " << bx << " " << cx << std::endl;
      std::cout << " fa, fb, fc: " << fa << " " << fb << " " << fc << std::endl;
      std::cout << "d statistic (minimum value): " << fmin << std::endl;
      std::cout << "scale factor: " << xmin << std::endl;

    #endif // DEBUG_MIN
    //--


    // returning the scale factor
    return (xmin);

}


//
// scaling the cross-talk matrix
// . calculate the scale factor by minimising a function associated to
//   the Kolmogorov-Smirnov statistic (KS statistic or D statistic)
//
void scaleXTalkMx(const int no_i_values_bA, const int no_i_values_bC,
                  const int no_i_values_bG, const int no_i_values_bT,
                  float *i_data_bA_s, float *i_data_bC_s,
                  float *i_data_bG_s, float *i_data_bT_s,
                  Matrix & mx_unscaled, Matrix & mx_scaled)
{

    float scale_factors[4];
    Matrix mx_inv;

    for (int i = 0; i < 4; ++i)
        scale_factors[i] = 1.0;

    try {
        // selected data validation
        if ((no_i_values_bA < MIN_NO_SEL_DATA) ||
            (no_i_values_bC < MIN_NO_SEL_DATA) ||
            (no_i_values_bG < MIN_NO_SEL_DATA) ||
            (no_i_values_bT < MIN_NO_SEL_DATA)) {
            Exception02 e2;
            throw e2;
        }

        scale_factors[1] =
            scaleDistr(no_i_values_bA, no_i_values_bC, i_data_bA_s,
                       i_data_bC_s);
        scale_factors[2] =
            scaleDistr(no_i_values_bA, no_i_values_bG, i_data_bA_s,
                       i_data_bG_s);
        scale_factors[3] =
            scaleDistr(no_i_values_bA, no_i_values_bT, i_data_bA_s,
                       i_data_bT_s);

        mx_inv = mx_unscaled.inverse();

    }
    catch(Exception02 e2) {
        e2.display();

        mx_inv.identity_matrix();

    }


    for (int r = 0; r < mx_inv.rows; ++r) {
        for (int c = 0; c < mx_inv.cols; ++c) {
            mx_inv.data[c + r * mx_inv.cols] /= scale_factors[r];
        }
    }

    mx_scaled = mx_inv.inverse();


    //--
    #ifdef DEBUG_SCALING
      std::cout << "\n>> debug - scaling" << std::endl;

      std::cout.setf(std::ios::fixed, std::ios::floatfield);
      std::cout.precision(7);

      std::cout << "scale factors: "
                << scale_factors[0] << " " << scale_factors[1] << " "
                << scale_factors[2] << " " << scale_factors[3] << std::endl;

    #endif // DEBUG_SCALING
    //--
}


//
// normalising the cross-talk matrix
// . normalising the matrix to get its determinant equal to 1
//
void normaliseXTalkMx(Matrix & mx_scaled, Matrix & mx_normalised)
{
    float det_mx;
    float norm_factor;

    det_mx = mx_scaled.det();

    norm_factor = 1.0 / pow(fabs(det_mx), 1.0 / 4.0);

    mx_normalised = mx_scaled * norm_factor;
}


void scale_normaliseXTalkMx(const float * i_data,
                            const int     no_objects,
                            const int     no_cols,
                            Matrix      & mx_unscaled,
                            Matrix      & mx_normalised)
{
    // intensities (temporary)
    float *i_data_1 = new float[no_objects * no_cols];

    // selected data (after preliminary base-calling)
    // associated with the bases A, C, G, T
    float *i_data_bA = new float[no_objects];
    float *i_data_bC = new float[no_objects];
    float *i_data_bG = new float[no_objects];
    float *i_data_bT = new float[no_objects];

    // number of intensites values for each group of selected data
    int no_i_values_bA;
    int no_i_values_bC;
    int no_i_values_bG;
    int no_i_values_bT;


    applyXTalkMx(i_data, no_objects, no_cols, mx_unscaled, i_data_1);

    // preliminary base-calling
    sampling2(i_data_1, no_objects, no_cols,
              i_data_bA, i_data_bC, i_data_bG, i_data_bT,
              no_i_values_bA, no_i_values_bC, no_i_values_bG, no_i_values_bT);


    delete [] i_data_1;


    // selected data (after preliminary base-calling and sorting)
    // associated with the bases A, C, G, T
    float *i_data_bA_s1 = new float[no_i_values_bA];
    float *i_data_bC_s1 = new float[no_i_values_bC];
    float *i_data_bG_s1 = new float[no_i_values_bG];
    float *i_data_bT_s1 = new float[no_i_values_bT];


    int i;
    for (i = 0; i < no_i_values_bA; ++i) {
        i_data_bA_s1[i] = i_data_bA[i];
    }
    for (i = 0; i < no_i_values_bC; ++i) {
        i_data_bC_s1[i] = i_data_bC[i];
    }
    for (i = 0; i < no_i_values_bG; ++i) {
        i_data_bG_s1[i] = i_data_bG[i];
    }
    for (i = 0; i < no_i_values_bT; ++i) {
        i_data_bT_s1[i] = i_data_bT[i];
    }


    delete [] i_data_bA;
    delete [] i_data_bC;
    delete [] i_data_bG;
    delete [] i_data_bT;


    // sorting selected data
    quicksort(i_data_bA_s1, no_i_values_bA);
    quicksort(i_data_bC_s1, no_i_values_bC);
    quicksort(i_data_bG_s1, no_i_values_bG);
    quicksort(i_data_bT_s1, no_i_values_bT);

    // selected data (after preliminary base-calling, sorting, and sampling)
    // associated with the bases A, C, G, T
    float *i_data_bA_s2 = new float[no_i_values_bA];
    float *i_data_bC_s2 = new float[no_i_values_bC];
    float *i_data_bG_s2 = new float[no_i_values_bG];
    float *i_data_bT_s2 = new float[no_i_values_bT];

    // number of intensites values for each group of selected data
    int no_i_values_bA_s2;
    int no_i_values_bC_s2;
    int no_i_values_bG_s2;
    int no_i_values_bT_s2;


    // sampling
    sampling3(i_data_bA_s1, no_i_values_bA, i_data_bA_s2, no_i_values_bA_s2);
    sampling3(i_data_bC_s1, no_i_values_bC, i_data_bC_s2, no_i_values_bC_s2);
    sampling3(i_data_bG_s1, no_i_values_bG, i_data_bG_s2, no_i_values_bG_s2);
    sampling3(i_data_bT_s1, no_i_values_bT, i_data_bT_s2, no_i_values_bT_s2);


    delete [] i_data_bA_s1;
    delete [] i_data_bC_s1;
    delete [] i_data_bG_s1;
    delete [] i_data_bT_s1;


    // cross-talk matrix (scaled)
    Matrix mx_scaled;

    // scaling
    scaleXTalkMx(no_i_values_bA_s2, no_i_values_bC_s2,
                 no_i_values_bG_s2, no_i_values_bT_s2,
                 i_data_bA_s2, i_data_bC_s2,
                 i_data_bG_s2, i_data_bT_s2,
                 mx_unscaled, mx_scaled);

    // normalisation
    normaliseXTalkMx(mx_scaled, mx_normalised);


    delete [] i_data_bA_s2;
    delete [] i_data_bC_s2;
    delete [] i_data_bG_s2;
    delete [] i_data_bT_s2;


    //--
    #ifdef DEBUG_XTALK_MATRIX_2
      std::cout << "\n>> debug - unscaled cross-talk matrix, last estimate" << std::endl;

      mx_unscaled.display();

    #endif // DEBUG_XTALK_MATRIX_2
    //--

    //--
    #ifdef DEBUG_XTALK_MATRIX_3
      std::cout << "\n>> debug - scaled cross-talk matrix" << std::endl;

      mx_scaled.display();

    #endif // DEBUG_XTALK_MATRIX_3
    //--

    //--
    #ifdef DEBUG_XTALK_MATRIX_4
      std::cout << "\n>> debug - normalised cross-talk matrix" << std::endl;

      mx_normalised.display();

    #endif // DEBUG_XTALK_MATRIX_4
    //--
}
