// [in], [out] data types for analysis plugins
#ifndef _PARAMETER_TYPES_H_
#define _PARAMETER_TYPES_H_

extern "C"
{
	// Marshalled from AnalysisInterface\Direct\TileInfo.cs
	struct TileInfo
	{
		int Lane;
		int Column;
		int Row;
		int Location;// Linear, unique, 0-based identifier of the tile. Increments in the direction of the data acquisition.
		int TileIndex;
		int Cycle;
		int RunId;// Token allowing distinguishing between overlapping callbacks from adjacent runs.
        int IsDeblocked; // 1 for deblock images, 0 otherwise
	};

	// Marshalled from AnalysisInterface\Direct\ImageInfo.cs
	struct ImageInfo
	{
		int	 Channel;
		char Base;
		int	 Width;
		int	 Height;

		int NumBytes() const
		{
			return Width * Height * sizeof( unsigned short );
		}
	};

	struct TileImageInfo
	{
		struct TileInfo	 TileInfo;
		struct ImageInfo ImageInfo;
	};

    const int   image_default_detection_type    = 0;
    const int   image_default_filter_type       = 0;  // mexican hat
    const float image_default_fwhm              = 2.7f;
    const float image_default_distance          = 1.5f;
    const float image_default_gauss_factor      = 4.f;
    const float image_default_pearson_threshold = 0.3f;
    const float image_default_max_stage_offset  = 500.f;
    const float image_default_max_color_offset  = 5.f;
	struct AnalysisParameters
	{
        int   DetectionType;
        int   FilterType;
        float ClusterFwhm;
        float ClusterDistance;
        float GaussFactor;
        float PearsonThreshold;
		float MaxStageOffset; // default = 500
		float MaxColourOffset;// default = 5
        
        AnalysisParameters()
            : DetectionType(image_default_detection_type)
            , FilterType(image_default_filter_type)
            , ClusterFwhm(image_default_fwhm)
            , ClusterDistance(image_default_distance)
            , GaussFactor(image_default_gauss_factor)
            , PearsonThreshold(image_default_pearson_threshold)
            , MaxStageOffset(image_default_max_stage_offset)
            , MaxColourOffset(image_default_max_color_offset)
        {}

        // Copy Constructor
        AnalysisParameters(const AnalysisParameters & source)
        {
            *this = source;
        }

        AnalysisParameters(const float value)
            : DetectionType(image_default_detection_type)
            , FilterType(image_default_filter_type)
            , ClusterFwhm(value)
            , ClusterDistance(value)
            , GaussFactor(value)
            , PearsonThreshold(value)
            , MaxStageOffset(image_default_max_stage_offset)
            , MaxColourOffset(image_default_max_color_offset)
        {}
 
        void CopyIfPositive(const AnalysisParameters & source) // Only a subset of the parameters
        {
            if (source.ClusterFwhm > 0.) ClusterFwhm = source.ClusterFwhm;
            if (source.ClusterDistance > 0.) ClusterDistance = source.ClusterDistance;
            if (source.GaussFactor > 0.) GaussFactor = source.GaussFactor;
            if (source.PearsonThreshold > 0.) PearsonThreshold = source.PearsonThreshold;
						if (source.FilterType > 0) FilterType = source.FilterType;
						DetectionType = source.DetectionType;
        }
        
        void SetImageDefaults()
        {
            ClusterFwhm = image_default_fwhm;
            ClusterDistance = image_default_distance;
            GaussFactor = image_default_gauss_factor;
            PearsonThreshold = image_default_pearson_threshold;
            MaxStageOffset = 0.;
            MaxColourOffset = 0.;
        }
	};

	struct RunInfo
	{
		int			 ImageWidth; // Pixels
		int			 ImageHeight;// Pixels
		int			 NumChannels;
		int			 NumTiles;// number of tile positions in a cycle.
		int			 NumCycles;
		const char * RunFolder;
		const char * TiffPathTemplate;
	};

	// Marshalled from Stargazer\Config\Alignment.cs
	struct Alignment
	{
		float XOffset;
		float YOffset;
		float XScale;
		float YScale;
	};

	struct Position
	{
		float X;
		float Y;
	};

    struct ImageRegion {
        int XMin;
        int YMin;
        int XSize;
        int YSize;
    };
    
    struct ImageStats {
        float Minimum;
        float Maximum;
        float Average;
        float StandardDeviation;
        float AverageBackground;
        float FractionSaturation; // >= 0., << 0.0001
        float DispersionBackground;
    };

    struct ClusterStats {
        int Num;
        float FwhmMean;
        float FwhmMedian;
        float FwhmStandardDeviation;

	    void operator=(const ClusterStats& cluster_stats) {
		    Num = cluster_stats.Num;
		    FwhmMean = cluster_stats.FwhmMean;
		    FwhmMedian = cluster_stats.FwhmMedian;
			FwhmStandardDeviation = cluster_stats.FwhmStandardDeviation;
		}
    };

    struct AlignmentStats : public Alignment {
        struct ImageRegion Region;
        float CorrelationMinimum;
        float CorrelationMaximum;
        float CorrelationWidth;   // > 0., < 2.
    };

    const int NumAlignmentRegions = 4;

    struct TileQCValues
    {
        float  EstimatedFWHM; // should be between 2. and 8. pixels
        int    OffsetsValidated; // should be 1
        struct ImageStats     IntensityStats;
        struct Alignment      UsedAlignment;
        struct AlignmentStats RawAlignment;
        struct AlignmentStats RegionAlignments[NumAlignmentRegions];
        struct ClusterStats   Fwhm;
		int					  NumClusters;
    };
}

#endif//_PARAMETER_TYPES_H_
