/*
 * Object positions and binary operations on them.
 *
 * Time-stamp: <26 Jun 2006 16:58 BST Positions.cpp km>
 * Copyright (c) Solexa 2003-2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
/*
 * "Binary" operations check for the presence or absence of an object 
 * within a given distance on two different images. This is useful to 
 * answer questions such as "Is this object present on both images?",
 * regardless of the object's intensity.
 */

#include <fstream>
#include <cmath>
#include "Maths.h"
#include "Stats.h"
#include "Positions.h"
#include "ImageProcess.h"
#include "ProcessAlign.h"

using namespace std;

//---------------------------------------------------------------------------

int write_positions(FILE *filehandle, st_fnoisepixel *positions, int n)
{
    int i;
    for (i = 0; i < n; i++)
        fprintf(filehandle, "%6.1f %6.1f %7.1f %6.1f\n",
                positions[i].x, positions[i].y,
                positions[i].value, positions[i].noise);
    return 0;
}

int write_positions(const char filename[], st_fnoisepixel *positions, int n)
{
    FILE *posfile = fopen(filename, "wt");
    if (posfile) {
        if (write_positions(posfile, positions, n))
            return 1;
        fclose(posfile);
        return 0;
    }
    return 1;
}

int read_positions(const char filename[], st_fnoisepixel *positions, int nmax)
{
    int nread = 0;
    FILE *posfile = fopen(filename, "r");
    if (posfile) {
        char buffer[200+1];
        while (fgets(buffer, 200, posfile)) {
            float x, y, s, n;
            if (buffer[0] == '#') // discard lines marked as comments
                continue;
            if (sscanf(buffer, "%f %f %f %f", &x, &y, &s, &n) == 4) {
                if (nmax > 0) { // read positions from file
                    if (nread < nmax) {
                        positions[nread].x = x;
                        positions[nread].y = y;
                        positions[nread].value = s;
                        positions[nread].noise = n;
                        nread++;
                    }
                }
                else {         // count number of lines in file
                    nread++;
                }
            }
        }
        fclose(posfile);
        return nread;
    }
    return -1;
}

/*
 * LEGACY version - code commonly used is now in ImageUtil.cpp!
 * auxiliary function to read an array of positions from disk.
 * if <nmax> is negative, count the number of lines in the file
 */
int read_positions(const char filename[], st_fposition *positions, int nmax)
{
    int nread = 0;
    FILE *posfile = fopen(filename, "r");
    if (posfile) {
        char buffer[200+1];
        while (fgets(buffer, 200, posfile)) {
            float x, y;
            if (buffer[0] == '#') // discard lines marked as comments
                continue;
            if (sscanf(buffer, "%f %f", &x, &y) == 2) {
                if (nmax > 0) { // read positions from file
                    if (nread < nmax) {
                        positions[nread].x = x;
                        positions[nread].y = y;
                        nread++;
                    }
                }
                else {         // count number of lines in file
                    nread++;
                }
            }
        }
        fclose(posfile);
        return nread;
    }
    return -1;
}

//---------------------------------------------------------------------------
// binary operations
//---------------------------------------------------------------------------

/*
 * auxiliary function to fill all pixels within a maximum distance
 * of min_dist from the point (x0,y0) in an (nx,ny)-image array with 
 * a given value
 */
static void helper_fill_pixels(int *array, int nx, int ny,
                               int x0, int y0, int value, float min_dist)
{
    int size = (int)ceil(min_dist);
    int xmin = Maths::max(0, x0-size); int xmax = Maths::min(nx-1, x0+size);
    int ymin = Maths::max(0, y0-size); int ymax = Maths::min(ny-1, y0+size);
    int x, y, x2;
    float min_dist2 = min_dist * min_dist;
    for (x = xmin; x <= xmax; x++) {
        x2 = (x-x0)*(x-x0);
        for (y = ymin; y <= ymax; y++)
            if ((float)(x2+(y-y0)*(y-y0)) <= min_dist2)
                array[x*ny+y] = value;
    }
}

/*
 * take two lists of positions and return a list that contains 
 * positions that appear in either the first or the second list.
 * the caller has to make sure enough memory has been reserved
 * for the returned list
 */
void or_positions(st_fposition *positions, int *count,
                  st_fposition *positions1, int count1,
                  st_fposition *positions2, int count2,
                  int nx, int ny, float min_dist)
{
    int i;
    // fill temporary image array
    int *temp = new int [nx*ny];
    for (i = 0; i < nx*ny; i++)
        temp[i] = 0;
    for (i = 0; i < count1; i++)
        helper_fill_pixels(temp, nx, ny, 
                           (int)positions1[i].x, (int)positions1[i].y,
                           1, min_dist);

    // copy first list
    if (positions1 != positions)
        memcpy(positions, positions1, count1*sizeof(st_fposition));
    *count = count1;

    // loop over second list
    int xpos, ypos;
    for (i = 0; i < count2; i++) {
        xpos = (int)positions2[i].x; ypos = (int)positions2[i].y;
        if ((xpos >= 0) && (xpos < nx) && (ypos >= 0) && (ypos < ny)
            && (temp[xpos*ny+ypos] == 0)) {
            helper_fill_pixels(temp, nx, ny, xpos, ypos, 1, min_dist);
            positions[(*count)].x = (float)xpos;
            positions[(*count)].y = (float)ypos;
            (*count)++;
        }
    }

    delete [] temp;
}


/*
 * take two lists of positions and return a list of positions that
 * appear in the first, but not in the second list
 */
void subtract_positions(st_fposition *positions, int *count,
                        st_fposition *positions1, int count1,
                        st_fposition *positions2, int count2,
                        int nx, int ny, float min_dist)
{
    int i;
    int *temp = new int [nx*ny];
    for (i = 0; i < nx*ny; i++)
        temp[i] = 0;
    for (i = 0; i < count2; i++)
        helper_fill_pixels(temp, nx, ny,
                           (int)positions2[i].x, (int)positions2[i].y,
                            1, min_dist);
    *count = 0;
    int xpos, ypos;
    for (i = 0; i < count1; i++) {
        xpos = (int)positions1[i].x; ypos = (int)positions1[i].y;
        if ((xpos >= 0) && (xpos < nx) && (ypos >= 0) && (ypos < ny)
            && (temp[xpos*ny+ypos] == 0)) {
            positions[(*count)].x = (float)xpos;
            positions[(*count)].y = (float)ypos;
            (*count)++;
        }
    }
    delete [] temp;
}

/*
 * take two lists of positions and return the a list of positions
 * that appear in both
 */
void and_positions(st_fposition *positions, int *count,
                   st_fposition *positions1, int count1,
                   st_fposition *positions2, int count2,
                   int nx, int ny, float min_dist)
{
    int i;
    int *temp = new int [nx*ny];
    for (i = 0; i < nx*ny; i++)
        temp[i] = 0;
    for (i = 0; i < count2; i++)
        helper_fill_pixels(temp, nx, ny,
                           (int)positions2[i].x, (int)positions2[i].y,
                           1, min_dist);
    *count = 0;
    int xpos, ypos;
    for (i = 0; i < count1; i++) {
        xpos = (int)positions1[i].x; ypos = (int)positions1[i].y;
        if ((xpos >= 0) && (xpos < nx) && (ypos >= 0) && (ypos < ny)
            && (temp[xpos*ny+ypos] > 0)) {
            positions[(*count)].x = (float)xpos;
            positions[(*count)].y = (float)ypos;
            (*count)++;
        }
    }
    delete [] temp;
}

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void Positions::initialise()
{
    positions = NULL;
    count = 0;
    shiftX = shiftY = 0.;
    nx = ny = DEFAULT_PIXEL_DIM;
}

Positions::Positions()
{
    initialise();
}

/*
 * copy constructor
 */
Positions::Positions(const Positions& temp)
{
    initialise();
    resize(temp.count);
    nx = temp.nx; ny = temp.ny;
    shiftX = temp.shiftX; shiftY = temp.shiftY;
    memcpy(positions, temp.positions, count*sizeof(st_fposition));
}

Positions::Positions(int n, int x, int y)
{
    initialise();
    resize(n);
    nx = x;
    ny = y;
}

Positions::~Positions()
{
    if (count > 0)
        delete [] positions;
}

/*
 * set the maximum number of Positions than can be held and allocate
 * memory accordingly. copy positions to new memory locations if necessary.
 */
void Positions::resize(int newsize)
{
    if ((newsize == count) || (newsize <= 0))
        return;

    st_fposition *newpositions = new st_fposition [ newsize ];

    // copy existing content across
    int ncopy = Maths::min(newsize, count);
    if (ncopy > 0)
        memcpy(newpositions, positions, ncopy*sizeof(st_fposition));

    if (count > 0)
        delete [] positions;
    positions = newpositions;
    count = newsize;
}

/*
 * save molecule list to text file
 */
int Positions::write(const char filename[])
{
    return write_positions(filename, positions, count);
}

/*
 * load molecule list from text file
 */
int Positions::read(const char filename[])
{
    int ct = read_positions(filename, positions, -1);
    if (ct > 0) {
        resize(ct);
        read_positions(filename, positions, ct);
    }
    return ct;
}

/*
 * shift position of each molecule in the list by the same amount
 */
void Positions::shift(float x, float y)
{
    int i;
    for (i = 0; i < count; i++) {
        positions[i].x += x;
        positions[i].y += y;
    }
    shiftX = x;
    shiftY = y;
}

void Positions::align(const Positions& reference, int maxlag,
                      float fwhm)
{
    float xlag = 0., ylag = 0.;
    st_pixel *pos1 = pixel_from_fposition(reference.positions,
                                          reference.getSize());
    st_pixel *pos2 = pixel_from_fposition(positions, getSize());
    AlignmentRecord alignment;
    align_positions(pos1, reference.getSize(),
                    pos2, getSize(), nx, ny, fwhm,
                    maxlag, maxlag, &alignment);
    xlag = alignment.XOffset; ylag = alignment.YOffset;
    shift(-xlag, -ylag);
    delete [] pos1;
    delete [] pos2;
}

Positions *Positions::bin_or(const Positions &s2, float distance)
{
    Positions *result = new Positions(getSize()+s2.getSize(), nx, ny);
    or_positions(result->positions, &result->count,
                 positions, getSize(),
                 s2.positions, s2.getSize(), nx, ny, distance);
    return result;
}

Positions *Positions::bin_subtract(const Positions &s2, float distance)
{
    Positions *result = new Positions(getSize(), nx, ny);
    subtract_positions(result->positions, &result->count,
                       positions, getSize(),
                       s2.positions, s2.getSize(), nx, ny, distance);
    return result;
}

Positions *Positions::bin_and(const Positions &s2, float distance)
{
    Positions *result = new Positions(getSize(), nx, ny);
    and_positions(result->positions, &result->count,
                  positions, getSize(),
                  s2.positions, s2.getSize(), nx, ny, distance);
    return result;
}

//---------------------------------------------------------------------------

int PositionData::write(const char filename[])
{
    return write_positions(filename, data, count);
}

int PositionData::read(const char filename[])
{
    int ct = read_positions(filename, data, -1);
    if (ct > 0) {
         st_fnoisepixel *newdata = new st_fnoisepixel [ct];
         read_positions(filename, newdata, ct);
         setData(newdata, ct);
    }
    return ct;
}

void PositionData::shift(float x, float y)
{
    int i;
    for (i = 0; i < count; i++) {
        data[i].x += x;
        data[i].y += y;
    }
    shiftX += x;
    shiftY += y;
}

st_pixel *PositionData::getPixelsCopy() const
{
    st_pixel *pixel = new st_pixel [count];
    int i;
    for (i = 0; i < count; i++) {
        pixel[i].x = (unsigned short)Stats::round(data[i].x);
        pixel[i].y = (unsigned short)Stats::round(data[i].y);
        pixel[i].value = data[i].value;
    }
    return pixel;
}

void PositionData::align(const PositionData& ref, int maxlag,
                         float fwhm, int nx, int ny)
{
    float xlag = 0., ylag = 0.;
    shiftX = 0.; shiftY = 0.;
    st_pixel *pos1 = ref.getPixelsCopy();
    st_pixel *pos2 = getPixelsCopy();
    AlignmentRecord alignment;
    align_positions(pos1, ref.getSize(), pos2, getSize(), nx, ny, fwhm,
                    maxlag, maxlag, &alignment);
    xlag = alignment.XOffset; ylag = alignment.YOffset;
    shift(-xlag, -ylag);
    delete [] pos1;
    delete [] pos2;
}
