/*
 * Class for image alignment and intensity extraction.
 *
 * Time-stamp: <26 Jan 2007 17:18 GMT Process.cpp km>
 *
 * Performs automated alignment of iteratively passed images.
 * The individual images are processed elsewhere. This class simply
 * combines images taken in different colours, 
 * remaps coordinates suitably (image registration),
 * identifies common objects and extracts intensity informtion for multiple 
 * images. The output is written to disk.
 * 
 * Copyright (c) Solexa 2003, 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <ctype.h>
#include <math.h>
#include <string.h>
#else
#include <cctype>
#include <cmath>
#include <cstring>
#endif
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include "FFT.h"
#include "DNACodes.h"
#include "Process.h"
#include "IOProcess.h"
#include "Maths.h"
#include "Tools.h"
#include "ImageUtil.h"
#include "QCImage.h"
#include "Positions.h"
#include "IntensityData.h"
#include "ProcessRemapping.h"

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
/*
 * Class that accumulates data from the camera and performs 
 * image registration and intensity extraction.
 */

DataProcessor::DataProcessor(const RunInfo & runInfo)
{
    memcpy(&run_info, &runInfo, sizeof(RunInfo));

    default_transforms   = new AlignmentTransform [run_info.NumChannels];
    bases                = new unsigned char [run_info.NumChannels];
    offset_tracker       = new OffsetTracker(run_info.NumTiles, 
                                             run_info.NumChannels);
    images               = new Image * [run_info.NumChannels];
    memset( images, 0, run_info.NumChannels * sizeof( Image * ) );
    max_object_number = run_info.NumChannels 
                       * run_info.ImageWidth * run_info.ImageHeight
                       / MIN_PIXELS_PER_OBJECT;
    aligner = 0;
    useRemappedIntensityFlag    = true;
}

DataProcessor::~DataProcessor()
{
    delete [] images;
    delete [] default_transforms;
    delete [] bases;
    delete    offset_tracker;
}

void DataProcessor::SetTile(const TileInfo & tileInfo)
{
    memcpy(&current_tile, &tileInfo, sizeof(TileInfo));
}

void DataProcessor::SetParameters(const AnalysisParameters & params)
{
    memcpy(&analysis_params, &params, sizeof(AnalysisParameters));
}

void DataProcessor::SetAlignment(const Alignment * defaultAlignment)
{
    int i;
    for (i = 0; i < run_info.NumChannels; i++)
        default_transforms[i].copy(defaultAlignment[i],
                                   run_info.ImageWidth, run_info.ImageHeight);
}

void DataProcessor::setUseRemappedIntensityFlag(bool value)
{
    useRemappedIntensityFlag = value;
}


/*
 * This routine calculates the coordinate transforms needed
 * to align images back to the first cycle reference.
 *
 * WARNING: The caller MUST NOT deallocate any image data until
 * all 4 colours in a tile have been completed.
 */
void DataProcessor::AnalyseImage(Image          *image,
                                 const int       colour,
                                 const char      base,
                                 const Position *referencepositions,
                                 const int       referencecount)
{
    assert(image);
    assert(image->nx == run_info.ImageWidth);
    assert(image->ny == run_info.ImageHeight);
    assert(image->getData());
    assert(image->getData()->status & IMAGEDATA_STATE_ANALYSED);
    assert(image->getData()->filter_image != 0);
    assert(image->getData()->noise_image  != 0);
    assert(image->get_objects()->getSize() < 
           run_info.ImageWidth * run_info.ImageHeight / MIN_PIXELS_PER_OBJECT);

    image->data->setSequencingInfo(base, current_tile.IsDeblocked);
    images[colour] = image;
    bases[colour]  = base;
    image->getData()->analyseEstimateFWHM();

    /*
     * image realignment
     */
    assert(default_transforms[colour].Region.XSize == run_info.ImageWidth);
    assert(default_transforms[colour].Region.YSize == run_info.ImageHeight);
    image->data->transform     = new AlignmentTransform;
    image->data->usedTransform = new AlignmentTransform;
    image->data->transform->copy(default_transforms[colour]);

    /* 
     * alignment on subregions of the image
     * to get an idea of image distortions 
     */
    if (current_tile.Cycle > 0) {
        aligner->alignRegionPixelPositions(image->get_objects()->getData(),
                                           image->get_objects()->getSize(),
                                           image->data->transform);
    }
    image->data->usedTransform->copy(*(image->data->transform));
}

/*
 * Initialise the image alignment lookup table and cache Fourier transforms
 * for faster later access.
 */
void DataProcessor::SetReference(const Position *referencepositions,
                                 const int       referencecount)
{
    aligner = new ProcessAlign((st_fposition *)referencepositions, referencecount, 
                               run_info.ImageWidth, run_info.ImageHeight,
                               analysis_params.ClusterFwhm);
    aligner->setMaximumOffsets((int)analysis_params.MaxStageOffset);
    aligner->setRegionParameters(2, 2);
}

void DataProcessor::ClearReference()
{
    if (aligner)
        delete aligner;
    aligner = 0;
}

int DataProcessor::validateOffsets()
{
    int i;
    int is_deblocked = images[0]->data->sequencing->is_deblocked;
    AlignmentTransform *position_transforms 
             = new AlignmentTransform [run_info.NumChannels];
    int *object_counts = new int [ run_info.NumChannels ]; 
    for (i = 0; i < run_info.NumChannels; i++) {
        position_transforms[i].copy(*(images[i]->data->usedTransform));
        object_counts[i] = images[i]->get_objects()->getSize();
    }

    /* 
     * QC of image offsets 
     */

    /* QC for image realignment values */
    int valid_flag = 0;
    int valid_offsets = 0;

    valid_flag = validate_object_counts(object_counts, run_info.NumChannels, 200);
    for (i = 0; i < run_info.NumChannels; i++)
        if ((fabs(position_transforms[i].XOffset) 
                > analysis_params.MaxStageOffset) ||
            (fabs(position_transforms[i].YOffset) 
                > analysis_params.MaxStageOffset))
            valid_flag &= 0;

    if (valid_flag)
        valid_offsets = validate_offsets(position_transforms,
                                         run_info.NumChannels,
                                         analysis_params.MaxColourOffset,
                                         default_transforms);
    
    if ((valid_offsets == run_info.NumChannels) && (current_tile.Cycle > 0) 
                                                && (!(is_deblocked))
        && (valid_flag))
        offset_tracker->add_offsets(position_transforms, run_info.NumChannels,
                                    0);

    delete [] position_transforms;
    delete [] object_counts;

    return (valid_flag && valid_offsets);
}


void DataProcessor::BuildReference(Position       **newReference,
                                   int             *numNewReference,
                                   const Position  *oldReference,
                                   const int        numOldReference)
{
    int i;
    int *object_counts = new int [run_info.NumChannels];
    st_fnoisepixel **object_lists = new st_fnoisepixel * [run_info.NumChannels];
    for (i = 0; i < run_info.NumChannels; i++) {
        object_counts[i] = images[i]->get_objects()->getSize();
        object_lists[i] = new st_fnoisepixel [object_counts[i]];
        memcpy(object_lists[i], images[i]->get_objects()->getData(),
               object_counts[i] * sizeof(st_fnoisepixel));
    }

    for (i = 0; i < run_info.NumChannels; i++)
        images[i]->getData()->usedTransform->transformReverse(object_lists[i], 
                                                              object_counts[i]);

    *numNewReference = 0;
    RemappingLookup *remapper = new RemappingLookup(run_info.ImageWidth,
                                                    run_info.ImageHeight);
    remapper->setReference((st_fposition *)oldReference, numOldReference);
    remapper->setMaxNewPositions(max_object_number - numOldReference);
    for (i = 0; i < run_info.NumChannels; i++) {
        remapper->addNewPositions(object_lists[i], object_counts[i],
                                  analysis_params.ClusterDistance);
    }
    remapper->findNewPositions(analysis_params.ClusterDistance);
    remapper->addNewPositionsToReference();
    *numNewReference = remapper->getNumberNewPositions();
    *newReference = (Position *)(new st_fposition [*numNewReference]);
    memcpy(*newReference, remapper->getNewPositions(), 
           *numNewReference * sizeof(st_fposition));
    for (i = 0; i < run_info.NumChannels; i++)
        delete [] object_lists[i];
    delete [] object_counts;
    delete [] object_lists;
    delete remapper;
}

void DataProcessor::CompleteTile(float *intensities, float *noise,
                                Position *referencePositions,
                                int &numReference, int numElements,
                                const char **error_message,
                                int *flag, 
                                int *refIndices)
{
    int i;
    int validFlag = 0;
    validFlag = validateOffsets();
    if (validFlag) 
         analyseTile(referencePositions, numReference,
                     intensities, noise, refIndices);
    else {
        for (i = 0; i < run_info.NumChannels * numReference; i++) {
            intensities[i] = 0.;
            noise[i] = 0.;
            refIndices[i] = -1;
        }
    }
    if (flag)
        *flag = validFlag;
}

/**
 * Performs the remapping of all colours and extracts the remapped intensities.
 */
void DataProcessor::analyseTile(Position *positions,
                                int referencecount,
                                float *intensities, float *noise,
                                int *referenceindices)
{
    st_fposition *referencepositions = (st_fposition *)positions;
    int i;
    int *object_counts = new int [run_info.NumChannels];
    st_fnoisepixel **object_lists = new st_fnoisepixel * [run_info.NumChannels];

    for (i = 0; i < run_info.NumChannels; i++) {
        object_counts[i] = images[i]->get_objects()->getSize();
        object_lists[i] = new st_fnoisepixel [object_counts[i]];
        memcpy(object_lists[i], images[i]->get_objects()->getData(),
               object_counts[i] * sizeof(st_fnoisepixel));
    }

    /*
     * remap molecules for the current position and produce array
     * with molecule intensities in all colours
     */
    int *indices = new int [referencecount * run_info.NumChannels];
    for (i = 0; i < referencecount*run_info.NumChannels; i++)
        indices[i] = -1;

    for (i = 0; i < run_info.NumChannels; i++)
        images[i]->getData()->usedTransform->transformReverse(object_lists[i], 
                                                              object_counts[i]);

    if (useRemappedIntensityFlag) {
        RemappingLookup *remapper = new RemappingLookup(run_info.ImageWidth,
                                                        run_info.ImageHeight);
        remapper->setReference(referencepositions, referencecount);
        for (i = 0; i < run_info.NumChannels; i++) {
            remapper->remap(object_lists[i], object_counts[i],
                            analysis_params.ClusterDistance);
            memcpy(indices + i * referencecount,
                   remapper->getIndex(), referencecount*sizeof(int));
        }
        delete remapper;

        for (i = 0; i < run_info.NumChannels; i++) {
            int col = get_base_index(bases[i], bases, run_info.NumChannels);
            remap_colour(referencepositions, indices + i * referencecount,
                         referencecount,
                         object_lists[i], object_counts[i],
                         intensities, noise, run_info.NumChannels, col);
        }
    }

    /*
     * go back to images to read out missing intensity values
     * at the reference positions
     */
    for (i = 0; i < run_info.NumChannels; i++) {
        int col = get_base_index(bases[i], bases, run_info.NumChannels);
        complete_intensities(referencepositions,
                             indices + i * referencecount, referencecount,
                             run_info.NumChannels, col,
                             images[i]->getData()->filter_image,
                             images[i]->getData()->noise_image,
                             run_info.ImageWidth, run_info.ImageHeight,
                             intensities, noise,
                             analysis_params.ClusterFwhm, 0.51,
                             images[i]->getData()->usedTransform);
    }

    if (referenceindices != NULL)
        memcpy(referenceindices, indices,
               referencecount * run_info.NumChannels * sizeof(int));
    delete [] indices;

    for (i = 0; i < run_info.NumChannels; i++)
        delete [] object_lists[i];
    delete [] object_counts;
    delete [] object_lists;
}


void DataProcessor::displayOffsets()
{
    if (offset_tracker)
        offset_tracker->display(default_transforms);
}
