//  Time-stamp: <26 Jun 2006 17:00 BST ProcessFilter.cpp km>
/*
 * Image filter functions in pixel space and Fourier space.
 *
 * Copyright (c) Solexa 2003-2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include "Maths.h"
#include "Tools.h"
#include "ImageUtil.h"
#include "FFT.h"
#include "ProcessFilter.h"

//---------------------------------------------------------------------------
// filtering (using Fourier space filters)
//---------------------------------------------------------------------------

/**
 * Pad the image with a linear ramp on the outer pixels in order to avoid
 * edge effects from a convolution
 * 
 * ARGUMENTS:
 *    image            sizex * sizeY dimensional array
 *    sizex, sizey     dimensions of image array
 *    offsetx, offsety beginning of the real image region
 *    nx, ny           dimensions of the image region (nx <= sizex)
 */
static void pad_linear(float     * image, 
                       const int   sizex,
                       const int   sizey,
                       const int   offsetx,
                       const int   offsety,
                       const int   nx,
                       const int   ny)
{
    TRACE("pad_linear(%p, %i, %i, %i, %i, %i, %i)\n",
          (void *)image, sizex, sizey, offsetx, offsety, nx, ny);
    int x, y;
    int offsetx2 = sizex - offsetx - nx; // offset from right edge
    int offsety2 = sizey - offsety - ny; // offset from top edge
    int ntx = offsetx + offsetx2 + 1;
    int nty = offsety + offsety2 + 1;

    for (x = offsetx; x < sizex-offsetx2; x++) {
        for (y = 0; y < offsety; y++)
            image[(x)*sizey+(y)]
                = ((y+offsety2+1) * image[(x)*sizey+(offsety)]
                + (offsety-y)     * image[(x)*sizey+(sizey-offsety2-1)])/nty;
        for (y = 0; y < offsety2; y++)
            image[(x)*sizey+(sizey-y-1)]
                = ((offsety2-y)   * image[(x)*sizey+(offsety)]
                + (y+offsety+1)   * image[(x)*sizey+(sizey-offsety2-1)])/nty;
    }

    for (y = 0; y < sizey; y++) {
        for (x = 0 ; x < offsetx; x++)
            image[(x)*sizey+(y)]
                = ((x+offsetx2+1) * image[(offsetx)*sizey+(y)]
                + (offsetx-x)     * image[(sizex-offsetx2-1)*sizey+(y)])/ntx;
        for (x = 0 ; x < offsetx2; x++)
            image[(sizex-x-1)*sizey+(y)]
                = ((offsetx2-x)   * image[(offsetx)*sizey+(y)]
                + (x+offsetx+1)   * image[(sizex-offsetx2-1)*sizey+(y)])/ntx;
    }
}

/**
 * Zero-pad an image on the outer pixels.
 * ARGUMENTS: Compare pad_linear()
 */
static void pad_zero(float *image,
                     const int sizex,
                     const int sizey,
                     const int offsetx,
                     const int offsety,
                     const int nx,
                     const int ny)
{
    int offsetx2 = sizex - offsetx - nx; // offset from right edge
    int offsety2 = sizey - offsety - ny; // offset from top edge

    int x, y;
    for (x = 0 ; x < sizex; x++) {
        for (y = 0; y < offsety; y++)
            image[(x)*sizey+(y)] = 0.;
        for (y = 0; y < offsety2; y++)
            image[(x)*sizey+(sizey-y-1)] = 0.;
    }
    for (x = 0 ; x < offsetx; x++)
        for (y = offsety; y < sizey-offsety2; y++)
            image[(x)*sizey+(y)] = 0.;
    for (x = 0 ; x < offsetx2; x++)
        for (y = offsety; y < sizey-offsety2; y++)
            image[(sizex-x-1)*sizey+(y)] = 0.;
}

void get_fourier_dimensions(int minwidth, int nx, int ny, int *fnx, int *fny)
{
    *fnx = get_optimal_fft_length(nx+2*minwidth);
    *fny = get_optimal_fft_length(ny+2*minwidth);
}

/**
 * apply linear filtering to enhance the signal-to-noise in the maps.
 * the image is of dimension n*n; the filtered image is returned in fimage.
 * width is the FWHM of the PSF of the instrument.
 */
void filter_image_fft(float *image, float *mask,
                      float *filter_image, int nx, int ny,
                      float *fourier_image, int fnx, int fny)
{
    int   offsetx = (fnx - nx) / 2, offsety = (fny - ny) / 2;
    float *dummy1 = new float[fnx*fny], *dummy2 = new float[fnx*fny];

    insert_image_region(dummy1, fnx, fny, image, nx, ny, offsetx, offsety);

    // pad the image edges with a linear ramp to avoid edge effects
    pad_linear(dummy1, fnx, fny, offsetx, offsety, nx, ny);

    fft_2d_r2c(dummy1, fourier_image, fnx, fny);
    // apply the mask
    float *fourier_tmp = new_fourier_array(fnx, fny);
    copy_fourier_array(fourier_image, fourier_tmp, fnx, fny);
    apply_fft_mask(fourier_tmp, dummy2, mask, fnx, fny);
    delete_fourier_array(fourier_tmp);

    // now copy image region to output matrix
    copy_image_region(dummy2, fnx, fny, filter_image, nx, ny, offsetx, offsety);

    delete [] dummy2; 
    delete [] dummy1;
}

void filter_image_fft(float *image, float *filter_image, int nx, int ny, 
                      void (*func)(float *, int, int, void *), void *args,
                      float *fourier_image, int fnx, int fny)
{
    float *mask = new float [fnx*fny];
    (*func)(mask, fnx, fny, args);
    filter_image_fft(image, mask, filter_image, nx, ny,
                     fourier_image, fnx, fny);
    delete [] mask;
}

void filter_image_fft(float *image, float *filter_image, int nx, int ny, 
                      void (*func)(float *, int, int, void *), void *args,
                      float width)
{
    int fnx, fny;
    get_fourier_dimensions((int)ceil(width), nx, ny, &fnx, &fny);
    float *fourier_image = new_fourier_array(fnx, fny);
    filter_image_fft(image, filter_image, nx, ny,
                     func, args, fourier_image, fnx, fny);
    delete_fourier_array(fourier_image);
}

//---------------------------------------------------------------------------
// filtering (using pixel space filters).
// These functions are not currently in use.
//---------------------------------------------------------------------------

/**
 * rescale an array of size n*n such that its sum equals val
 */
void rescale_sum_2d_mask(float *mask, int n, float val)
{
    int i;
    float mask_sum = 0.;
    for (i = 0; i < n*n; i++)
        mask_sum += mask[i];
    for (i = 0; i < n*n; i++)
        mask[i] /= (mask_sum/val);
}

/**
 * apply a convolution mask of size (nm * nm)  to a 2-d image of
 * size (n * n). assumes periodic boundary conditions.
 */
void apply_2d_mask(float *src, float *target, int n, float *mask, int nm)
{
    float filter;
    int    offset = nm/2;
    int    i, j, l, i2, j2, i3, j3, l3, istep, jstep;


    for (i = 0; i < n*n; i++)
        target[i] = 0.;

    for (i2 = 0; i2 < nm; i2++)
        for (j2 = 0; j2 < nm; j2++) {
            filter = mask[i2*nm+j2];
            istep = i2-offset;
            jstep = j2-offset;
            if (istep < 0)
                istep += n;
            if (jstep < 0)
                jstep += n;
            for (i = 0; i < n; i++) {
                i3 = (i + istep) % n;
                l = i*n;
                l3 = i3*n;
                for (j = 0; j < n; j++) {
                    j3 = (j + jstep) % n;
                    target[l+j] += filter * src[l3+j3];
                }
            }
        }
}

#if 0
/* various filters:
 * high-pass filter:
 *  float matrix1[5][5] =  {{0,-1,-1,-1,0},
 *			    {-1,-2,-2,-2,-1},
 *			    {-1,-2,31,-2,-1},
 *			    {-1,-2,-2,-2,-1},
 *			    {0,-1,-1,-1,0}};
 *
 * very primitive low-pass filter - this one has _huge_ sidelobes
 *   float matrix2[3][3] = {{1,1,1},
 *			    {1,1,1},
 *			   {1,1,1}};
 */
  /* other possible low-pass filters:
   *    Jason:               {{ 1, 2, 1}, {2, 4, 2}, {1, 2, 1}} / 16
   *    Gaussian 3*3:        {{ 1, 3, 1}, {3, 7, 3}, {1, 3, 1}} / 23
   *    Gaussian 5*5         {{ 0, 1, 3, 1, 0}, { 1, 6, 10, 6, 1}, {3, 10, 15, 10, 3},
   *                          { 1, 6, 10, 6, 1}, { 0, 1, 3, 1, 0}};
   */
void filter_image_real(unsigned short *image, float *fimage, int nx, int ny,
                       float width)
{
    // determine size of padding depending on the width of the PSF
    int   offset = (int)ceil(width);
    // make sure the image size is a multiple of 8 to speed up FFTs
    int   ntx = get_optimal_fft_length(nx+2*offset);
    int   nty = get_optimal_fft_length(ny+2*offset);
    int   offsetx = (ntx - nx) / 2;
    int   offsety = (nty - ny) / 2;
    float *dummy1 = new float[ntx*nty];
    float *dummy2 = new float[ntx*nty];

    insert_image_region(dummy1, ntx, nty, image, nx, ny, offsetx, offsety);

    // pad the image edges with a linear ramp
    pad_linear(dummy1, ntx, nty, offsetx, offsety);

    // Mexican Hat filter for a FWHM of 3 pixels
    float matrix3[11][11] =
        {{-0.000001,-0.000015,-0.000105,-0.000410,-0.000913,-0.001188,
          -0.000913,-0.000410,-0.000105,-0.000015,-0.000001},
         {-0.000015,-0.000182,-0.001188,-0.004271,-0.008834,-0.011148,
          -0.008834,-0.004271,-0.001188,-0.000182,-0.000015},
         {-0.000105,-0.001188,-0.006964,-0.021483,-0.037484,-0.043455,
          -0.037484,-0.021483,-0.006964,-0.001188,-0.000105},
         {-0.000410,-0.004271,-0.021483,-0.048856,-0.045421,-0.026568,
          -0.045421,-0.048856,-0.021483,-0.004271,-0.000410},
         {-0.000913,-0.008834,-0.037484,-0.045421,0.081312,0.199447,
           0.081312,-0.045421,-0.037484,-0.008834,-0.000913},
         {-0.001188,-0.011148,-0.043455,-0.026568,0.199447,0.392241,
           0.199447,-0.026568,-0.043455,-0.011148,-0.001188},
         {-0.000913,-0.008834,-0.037484,-0.045421,0.081312,0.199447,
           0.081312,-0.045421,-0.037484,-0.008834,-0.000913},
         {-0.000410,-0.004271,-0.021483,-0.048856,-0.045421,-0.026568,
          -0.045421,-0.048856,-0.021483,-0.004271,-0.000410},
         {-0.000105,-0.001188,-0.006964,-0.021483,-0.037484,-0.043455,
          -0.037484,-0.021483,-0.006964,-0.001188,-0.000105},
         {-0.000015,-0.000182,-0.001188,-0.004271,-0.008834,-0.011148,
          -0.008834,-0.004271,-0.001188,-0.000182,-0.000015},
         {-0.000001,-0.000015,-0.000105,-0.000410,-0.000913,-0.001188,
          -0.000913,-0.000410,-0.000105,-0.000015,-0.000001}};

    // Now to apply the filters
    apply_2d_mask(dummy1, dummy2, ntx, &matrix3[0][0], 11);

    // Now fill into output matrix
    copy_image_region(dummy2, ntx, nty, fimage, nx, ny, offsetx, offsety);

    delete [] dummy2; delete [] dummy1;
}
#endif

//---------------------------------------------------------------------------
// various filter masks
//---------------------------------------------------------------------------

static void copy_filter_quarter(float *mask, int nx, int ny)
{
    int i, j;
    for (i = nx/2+1; i < nx; i++)
        for (j = ny/2+1; j < ny; j++)
            mask[i*ny+j] = mask[(nx-i)*ny+(ny-j)];
    for (i = 0; i <= nx/2; i++)
        for (j = ny/2+1; j < ny; j++)
            mask[i*ny+j] = mask[(i)*ny+(ny-j)];
    for (i = nx/2+1; i < nx; i++)
        for (j = 0; j <= ny/2; j++)
            mask[i*ny+j] = mask[(nx-i)*ny+(j)];
}	

void make_identity_filter(float *mask, int nx, int ny, void *args)
{
    int i;
    for (i = 0; i < nx*ny; i++)
        mask[i] = 1.;
}

void make_gaussian_highpass(float *mask, int nx, int ny, void *arg)
{
    float *arg2 = (float *)arg;
  	float fwhm = *arg2;
  	// optimised version; reduces number of function calls
    double sigma = fwhm / (2. * sqrt(2.*log(2.)));
    double fac = 4.*PI*PI*sigma*sigma;
    double x2, xy2;
    double cellx2, celly2;
    int i, j;

    cellx2 = 1./(double)(nx*nx);
    celly2 = 1./(double)(ny*ny);

    // start first quarter
    for (i = 0; i <= nx/2; i++) {
        x2 = i*i*cellx2;
        for (j = 0; j <= ny/2; j++) {
            xy2 = x2+j*j*celly2;
            mask[i*ny+j] = 1.f - (float)exp(-fac*xy2/2.f);
        }
    }
    // copy the other 3 image quarters
    copy_filter_quarter(mask, nx, ny);
}

/**
 * provide function for the Mexican hat mask in Fourier space.
 */
/* TODO: find factor of 2.*PI */
float mex_hat_2d_fft(float sigma2, float x2)
{
    double r, sigma, x, fac;
    sigma = (double)sigma2;
    x = (double)x2;
    fac = 4.*PI*PI*x*x*sigma*sigma;
    r =  2.*exp(-fac/2.)*(fac);
    return (float)r;
}

void make_mexican_hat_fft_mask(float *mask, int nx, int ny, void *arg)
{
#if 0
    float sigma = fwhm / (2. * sqrt(2.*log(2.)));
    float x;
    int i, j, i2, j2;
    float cellx, celly;

    cellx = 1./(float)nx;
    celly = 1./(float)ny;

    for (i = 0; i < nx; i++) {
        i2 = (i > nx/2 ? i-nx : i);
        for (j = 0; j < ny; j++) {
            j2 = (j > ny/2 ? j-ny : j);
            x = sqrt((float)(i2*i2*cellx*cellx+j2*j2*celly*celly));
            mask[i*ny+j] = mex_hat_2d_fft(sigma, x);
        }
    }
#else
    float *arg2 = (float *)arg;
    float fwhm = *arg2;
    // optimised version; reduces number of function calls
    double sigma = fwhm / (2. * sqrt(2.*log(2.)));
    double fac = 4.*PI*PI*sigma*sigma;
    double x2, xy2;
    double cellx2, celly2;
    int i, j;

    cellx2 = 1./(double)(nx*nx);
    celly2 = 1./(double)(ny*ny);

    // start first quarter
    for (i = 0; i <= nx/2; i++) {
       x2 = i*i*cellx2;
       for (j = 0; j <= ny/2; j++) {
          xy2 = x2+j*j*celly2;
          mask[i*ny+j] = (float)(2.*exp(-fac*xy2/2.)*(fac*xy2));
       }
    }
    // copy the other 3 image quarters
    copy_filter_quarter(mask, nx, ny);
#endif
}

void make_matched_fft_mask(float *mask, int nx, int ny, void *arg)
{
	const matchedfilter_params_t *params = (matchedfilter_params_t *) arg;

	const double alphasq=(params->alpha)*(params->alpha);
	const double delta=2.0*alphasq/(1.0+alphasq);
	const double m=(params->gamma+2.0)/2.0;
	
	const double sigma = params->fwhm/sqrt(8.0*M_LN2);
	const double fac = 4.0*M_PI*M_PI*sigma*sigma*alphasq;
	const double denom=tgamma(m)*pow(delta,m)/alphasq;

	double cellx2 = 1.0/(double(nx)*double(nx));
	double celly2 = 1.0/(double(ny)*double(ny));

	double x2, xy2;
	// start first quarter
	for (int i = 0; i <= nx/2; i++) {
		x2 = cellx2*i*i;
		for (int j = 0; j <= ny/2; j++) {
			xy2 = x2+celly2*j*j;
			mask[i*ny+j] = (float)(2.0*exp(-fac*xy2/2.0)*pow(fac*xy2,params->gamma/2.0)/denom);
		}
	}
	// copy the other 3 image quarters
	copy_filter_quarter(mask, nx, ny);

}

void make_biparametric_fft_mask(float *mask, int nx, int ny, void *arg)
{
	const biparametric_params_t *params = (biparametric_params_t *) arg;

	const double alphasq=(params->alpha)*(params->alpha);
	const double delta=2.0*alphasq/(1.0+alphasq);
	const double m=(2.0+params->gamma)/2.0;
	const double t=(2.0-params->gamma)/2.0;
	const double c=2.0*t/(m*params->gamma);
	
	const double sigma = params->fwhm/sqrt(8.0*M_LN2);
	const double fac = 4.0*M_PI*M_PI*sigma*sigma*alphasq;
	const double denom=tgamma(m)*pow(delta,m)*(1.0+c*m*delta)/alphasq;

	double cellx2 = 1.0/(double(nx)*double(nx));
	double celly2 = 1.0/(double(ny)*double(ny));

	double x2, xy2;
	// start first quarter
	for (int i = 0; i <= nx/2; i++) {
		x2 = cellx2*i*i;
		for (int j = 0; j <= ny/2; j++) {
			xy2 = x2+celly2*j*j;
			mask[i*ny+j] = (float)(2.0*exp(-fac*xy2/2.0)*pow(fac*xy2,params->gamma/2.0)*(1.0+c*fac*xy2)/denom);
		}
	}
	// copy the other 3 image quarters
	copy_filter_quarter(mask, nx, ny);
}
