/*
 * Determine the background noise dispersion of an image.
 *
 * Copyright (c) Solexa   2003-2005, 2006
 *               Illumina 2007
 * Author:       Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include "Maths.h"
#include "Tools.h"
#include "ImageUtil.h"
#include "Optimise.h"
#include "Message.h"
#include "ImageRegionList.h"
#include "ProcessNoiseEstimate.h"

//---------------------------------------------------------------------------
// image noise estimation
//---------------------------------------------------------------------------

struct CountHistogram {
    int   *values;
    int    size;
    float  argmin;
    float  argwidth;
};

void bin(float *image, int n, CountHistogram *hist)
{
    bin(image, n, hist->values, hist->size, hist->argmin, hist->argwidth);
}


/*
 * Determine the maximum value of a histogram and discard values less than
 * lofactor*maximum on the left side and less than hifactor*maximum on the 
 * right side of the maximum (allowance is given for 1-sigma errors on the
 * number counts). Return the last valid indices.
 */
static void clip_histogram(CountHistogram *histogram,
                           float lofactor, float hifactor,
                           int *maxindex, int *loindex, int *hiindex)
{
    float threshold;
    int histmax = 0;
    int i;

    // find maximum
    *maxindex = 0;
    for (i = 0; i < histogram->size; i++)
        if (histogram->values[i] > histmax) {
            histmax = histogram->values[i];
            *maxindex = i;
        }

    // find first and last index
    *loindex = *maxindex;
    threshold = (histmax-sqrt((float)histmax))*lofactor;
    while ((*loindex-1 >= 0) 
            && (histogram->values[*loindex-1]
                +sqrt((float)histogram->values[*loindex-1]) >= threshold))
        (*loindex)--;
    *hiindex = *maxindex;
    threshold = (histmax-sqrt((float)histmax))*hifactor;
    while ((*hiindex+1 < histogram->size) 
            && (histogram->values[*hiindex+1]
                +sqrt((float)histogram->values[*hiindex+1]) >= threshold))
        (*hiindex)++;
}

/*
 * Auxiliary function to get a percentile from a distribution
 * given by its histogram.
 * <percentile> is quoted as a number between 0. (smallest element) 
 * and 1. (largest element).
 * The histogram <hist> contains <n> equally spaced bins.
 * The bin width is <width>, and the left border of the leftmost bin
 * is given by <minarg>.
 */
static float get_histogram_percentile(float percentile,
                                      CountHistogram *histogram,
                                      float minval = 0., float maxval = 0.)
{
    int    i;
    int    n      = histogram->size;
    int   *hist   = histogram->values;
    float  minarg = histogram->argmin;
    float  width  = histogram->argwidth;
    if (minval == maxval) {
        minval = minarg;
        maxval = minarg + n * width;
    }

    // add up all values in histogram
    long allsum = 0, target;
    for (i = 0; i < n; i++)
        if ((minarg+i*width >= minval) && (minarg+i*width < maxval))
            allsum += hist[i];
    target = (long)(percentile * allsum);

    if (allsum <= 1) {
        printf("Not enough counts in target interval!\n");
        return (maxval+minval)/2.f;
    }
    long sum = 0;
    i = 0;
    while ((i < n) && (sum + hist[i] < target)) {
        if ((minarg+i*width >= minval) && (minarg+i*width < maxval)) 
            sum += hist[i];
        i++;
    }
    float x = minarg + width * i;
    if (i < n) {
        assert(hist[i] > 0);
        x += (width*(target-sum))/hist[i];
    }
#if 0
    printf("Perc %f %6li %6li %6li %5i %5i %f %f %f %f %f\n",
           percentile, target, sum, allsum, i, n, x, minarg, width, minval, maxval);
#endif
    return x;
}


/*
 * Yet another attempt to get a reliable background estimate:
 * Smooth the histogram by mean filtering (moving average)
 * and find the maximum, plus an estimate of the standard deviation.
 */
void get_smoothed_average(CountHistogram *hist, float width,
                          float *average, float *sigma)
{
    assert(width > 0.);
    int nwindow = int(width / hist->argwidth);
    if (nwindow <= 0)
        nwindow = 1;
    if (nwindow >= hist->size/4-1)
        nwindow = hist->size/4-1;

    // calculate the moving average
    long *moving_average = new long [hist->size];
    int    i;
    for (i = 0; i < hist->size; i++)
        moving_average[i] = 0;
    for (i = 0; i <= nwindow; i++)
        moving_average[0] += hist->values[i];
    for (i = 1; i < hist->size; i++) {
        moving_average[i] = moving_average[i-1];
        if (i-1-nwindow > 0)
            moving_average[i] -= hist->values[i-1-nwindow];
        if (i + nwindow < hist->size)
            moving_average[i] += hist->values[i+nwindow];
    }
    
    // identify the maximum
    long maximum = 0;
    int  imax    = 0;
    int  i1      = 0;
    int  i2      = 0;
    for (i = 0; i < hist->size; i++)
        if (moving_average[i] > maximum) {
            imax = i;
            maximum = moving_average[i];
        }

    // identify the point of the distribution where it has fallen off
    // to the level expected for a Gaussian at 1 sigma.
    long ysigma = (long)(exp(-0.5) * maximum);
    for (i = imax; i >= 0; i--)
        if (moving_average[i] >= ysigma)
            i1 = i;
    for (i = imax; i < hist->size; i++)
        if (moving_average[i] >= ysigma)
            i2 = i;
    *average = hist->argmin + imax * hist->argwidth;
#if 1
    *sigma   = Maths::min(imax-i1, i2-imax) * hist->argwidth;
#else
    *sigma   = (i2-i1) * hist->argwidth / 2.;
#endif

#if 0
    for (i = 0; i < hist->size; i++)
        printf("M %i %7.2f %6i %6i\n", i, hist->argmin + i * hist->argwidth,
               moving_average[i], hist->values[i]);
#endif
    delete [] moving_average;
}

/*
 * Estimate average and standard deviation based on rank statistics
 * Assume background is given by median and noise by the 68th percentile 
 * of the pixel distribution.
 * Iterate and discard extreme values beyond 3 sigma to converge on an estimate.
 */
void get_rank_average(CountHistogram *hist,
                      float *average, float *sigma)
{
    int iter2 = 0;
    float old_av, old_sdev;
    do {
        float minval = *average - 3.f * *sigma;
        float maxval = *average + 3.f * *sigma;
        old_av   = *average;
        old_sdev = *sigma;
        *average  =  get_histogram_percentile(0.5f,  hist, minval, maxval);
        *sigma    = (get_histogram_percentile(0.84f, hist, minval, maxval)
                   - get_histogram_percentile(0.16f, hist, minval, maxval))/2.f;
        iter2++;
    } while ((iter2 < 10) && (fabs((*sigma-old_sdev)/old_sdev) > 0.01f));                                  
}

/*
 * For high-density images the pixel distribution of the filtered image
 * has not only a positive wing caused by the clusters,
 * ut also a negative shoulder caused by the 'ringing'
 * introduced by the filter.
 * This routine contains lots of different (fairly ad-hoc) optimisations
 * to clip the histogram for corner cases to avoid being contaminated
 * by the negative shoulder of the histogram.
 */
void clip_peak_histogram(CountHistogram *hist,
                         int *maxidx, int *lowidx, int *higidx)
{
    assert(hist->argwidth > 0.);
    const float y1sigma  = exp(-0.5);
    float       twosigma = (*higidx - *lowidx) * hist->argwidth;
    int         isig     = int(ceil(twosigma / 2. / hist->argwidth));
    int         i, j;

    {
        /*
         * first check for local maxima to the right of the 
         * current maximum. make sure that there are at least
         * two points to the right of and above a local minimum,
         * or two points below and to the left of a local maximum
         */  
        int ioff = 0;
        for (i = 1; i <= *higidx - *maxidx; i++)
            if (hist->values[*maxidx + i] * 0.9 > hist->values[*maxidx + i - 1]) {
                int count = 0;
                for (j = *maxidx + 1; j < *maxidx + i; j++)
                    if (hist->values[j] < 0.9*hist->values[*maxidx + i])
                        count++;
                if (count >= 2)
                ioff = i;
                count = 0;
                for (j = *maxidx + i; j <= *higidx; j++)
                    if (hist->values[j]*0.9 > hist->values[*maxidx+i-1]) 
                        count ++;
                if (count >= 2)
                    ioff = i;
            }
        if (ioff > 0) {
#if defined(DEBUG)
            printf("  Update: local max, low hi max: %i %i %i\n",
                   *lowidx, *higidx, *maxidx);
#endif
            *maxidx += ioff;
        }
        /*
         * now check that the peak isn't too broad; otherwise
         * use outermost point of the peak that is still within 0.8
         * of the global maximum
         */
        ioff = 0;
        for (i = isig; i <= *higidx - *maxidx; i++)
            if (hist->values[*maxidx + i] > 0.9*hist->values[*maxidx])
                ioff = i;

        if (ioff > 0) {
            *maxidx = *maxidx + ioff;
            *lowidx = *maxidx-1;
            if (hist->values[*lowidx] > hist->values[*maxidx])
                *maxidx = *lowidx;

#if defined(DEBUG)
            printf("  Update: broad peak, low hi max: %i %i %i\n",
                   *lowidx, *higidx, *maxidx);
#endif
        }
        ioff = 0;

        /*
         * now cut down on the left shoulder. In fact, the further
         * the shoulder extends, the more cut off.
         */
        assert(*lowidx >= 0);
        for (i = isig; i <= *maxidx-*lowidx; i++)
            if (hist->values[*maxidx - i] > 0.8*hist->values[*maxidx])
                ioff = i;
        if (ioff > 0) {
            *lowidx = *maxidx - isig;
            if (ioff > 2*isig)
                ioff = 2*isig;
            if (ioff > isig)
                *lowidx += ioff-isig;
#if defined(DEBUG)
            printf("  Update: shoulder reduced, low hi max: %i %i %i\n",
                   *lowidx, *higidx, *maxidx);
#endif
        }
        assert(*lowidx >= 0);
        /*
         * make sure we capture the slope on the right properly
         */
        while ((hist->values[*higidx] > y1sigma * hist->values[*maxidx])
               && (*higidx + 1 < hist->size))
            (*higidx) ++;
        /*
         * make sure that the left shoulder is no larger than the right
         */
        if (*higidx - *maxidx < *maxidx - *lowidx)
            *lowidx = *maxidx - (*higidx - *maxidx);
    }

    while ((*higidx - *lowidx+1 < 5) && (*lowidx - 1 > 0))
        *lowidx -= 1;

    assert(*lowidx >= 0);
    assert(*higidx < hist->size);
    assert((*maxidx >= *lowidx) && (*maxidx <= *higidx));
}

/*
 * Try to fit a 1d-Gaussian to the peak of the histogram
 * using Levenberg-Marquardt optimisation.
 */
void fit_gaussian_histogram(CountHistogram *histogram,
                            int maxidx, int lowidx, int higidx,
                            float *fitted_max, float *fitted_av, float *fitted_sdev)
{
    int    i;
    int    ndata = higidx-lowidx+1;
    float *x     = new float [ndata];
    float *y     = new float [ndata];
    float *sig   = new float [ndata];
#if defined(DEBUG)
    printf("  low hi max: %i %i %i ndata %i bins %i\n",
           lowidx, higidx, maxidx, ndata, histogram->size);
#endif
    for (i = 0; i < ndata; i++) {
        int idx = i + lowidx;
        x[i] = histogram->argmin + (idx+0.5f) * histogram->argwidth;
        y[i] = (float)histogram->values[idx]/histogram->argwidth; // normalise by bin width
        sig[i] = sqrt(y[i]); // Poisson standard deviation
        sig[i] = Maths::max(sig[i], 1.f);
        sig[i] = 1;
    }
    // initial estimates
    int    na = 3;
    float *a  = new float [na];
    int   *ia = new int [na];
    for (i = 0; i < na; i++)
        ia[i]  = 1;
    if (maxidx == lowidx)
        ia[1] = 0;
    a[0] = y[maxidx-lowidx];
    a[1] = x[maxidx-lowidx];
    a[2] = ndata * histogram->argwidth / 2.3f;

#if defined(DEBUG)
    printf("  hist: \n");
    for (i = 0; i < ndata; i++)
        printf(" %7.2f %7.2f %7.2f \n", x[i], y[i], sig[i]);
    printf("  Input:  f(x)=a*exp(-(x-b)*(x-b)/(2.*s*s)), a=%f, b=%f, s=%f\n",
           a[0], a[1], a[2]);
#endif
    float chisq = 0.;
    LM_1d(x, y, sig, ndata, a, ia, na, &chisq, LM_gauss_1d_der);
    *fitted_max  = a[0];
    *fitted_av   = a[1];
    *fitted_sdev = a[2];

    delete [] x; delete [] y; delete [] sig; delete [] a; delete [] ia;

#if defined(DEBUG)
        printf("  chisquared: %f\n", chisq);
    printf("  Fitted: f(x)=a*exp(-(x-b)*(x-b)/(2.*s*s)), a=%f, b=%f, s=%f\n",
           *fitted_max, *fitted_av, *fitted_sdev);
#endif
}


/*
 * Determine the noise dispersion from the pixel histogram.
 *
 * The whole procedure is a bit of a mess, but the basic idea is to determine
 * the mean and standard deviations of a (assumed to be Gaussian)
 * pixel histogram in the presence of a foreground (signal from the objects)
 * and any 'ringing' introduced by an image filter.
 * 
 * Because a good method is difficult to find, the function tries various methods
 * of generally increasing accuracy but decreasing robustness. An improvement
 * is usually assumed to be given when the dispersion decreases, based
 * on the assumption that most methods systematically overestimate the 
 * dispersion.
 * 
 * An initial estimate is obtained from a mean filtered histogram.
 * A refinement is obtained by fitting a Gaussian to the peak
 * of a histogram of pixel values.
 */
void determine_noise_hist_stats(float *image, int n,
                                float *average, float *dispersion)
{
    /* Specify number of bins per sigma. This is a heuristic value
     * and will depend on the sample size (i.e. number of pixels). */
    const int nbins = 7;
    /* Avoid small dispersions (less than one camera count). 
     * They are most likely due to blank/saturated/corrupted images. */
    const float minimum_dispersion = 0.9f;
    float filter_av, filter_sdev;
    float back_av,   back_sdev, back_max;
    float smooth_av, smooth_sdev;

    /* 
     * Determine standard deviation and average of original image as initial
     * estimates
     */
    get_array_stats(image, n, &filter_av, &filter_sdev);

    if (filter_sdev < minimum_dispersion) {
        *average = filter_av;
        *dispersion = minimum_dispersion;
        return;
    }
    back_av   = filter_av;
    back_sdev = filter_sdev;
    back_max  = 0.;

    /*
     * Create a histogram with a very fine initial binning.
     * The binning at this stage is mostly a performance
     * hack for the quick calculation of statistics.
     * 
     * The following binning is chosen:
     * Bin by a factor of 10 more finely than any interval; an additional
     * factor of 5. to take into account that the initial standard deviation
     * is going to overestimate the true noise dispersion.
     * Round down to powers of 10. A bin width of less than 0.1 is
     * not useful as camera data only comes in steps of 1 unit.
     */
    CountHistogram fine_histogram;
    fine_histogram.argwidth = pow(10.f, floor(log10(filter_sdev/nbins/50.f)));
    if (fine_histogram.argwidth < 0.1f)
        fine_histogram.argwidth = 0.1f;
    fine_histogram.argmin = (floor((filter_av-2.5f*filter_sdev)/fine_histogram.argwidth)-0.5f) 
                    * fine_histogram.argwidth;
    fine_histogram.size   = (int)ceil(5.f*filter_sdev/fine_histogram.argwidth);

#if defined(DEBUG)
    printf("Global Av %g Sdev %g Minval %g Cell %g n %i\n",
           filter_av, filter_sdev,
           fine_histogram.argmin, fine_histogram.argwidth, fine_histogram.size);
#endif

    /* perform binning to obtain inital histogram */
    fine_histogram.values = new int [fine_histogram.size];
    bin(image, n, &fine_histogram);

#if defined(DEBUG) && 0
    long sum = 0.;
    for (i = 0; i < fine_histogram.size; i++)
        printf("%7.2f %5i %6i\n",
               fine_histogram.argmin+(i+0.5)*fine_histogram.argwidth,
               fine_histogram.values[i], sum+= fine_histogram.values[i]); 
#endif

#if 0
    /*  
     * Now determine background and noise from rank statistics.
     */
    {
        float rank_av, rank_sdev;
        rank_av   = filter_av;
        rank_sdev = filter_sdev;
        get_rank_average(&fine_histogram, &rank_av, &rank_sdev);
#if defined(DEBUG)
        printf("Rank Av %f Sdev %f\n", rank_av, rank_sdev); 
#endif                                     

        if (rank_sdev >= minimum_dispersion) {
            back_av   = rank_av;
            back_sdev = rank_sdev;
        }
    }
#endif

    /* 
     * Now try to estimate dispersion from a mean filtered histogram.
     * Iterate the estimation with adaptive window sizes
     * to avoid being confused by outliers.
     */
    smooth_av   = filter_av;
    smooth_sdev = filter_sdev/2.5f;
    int iter = 0;
    float window_width = smooth_sdev;
    do {
        iter ++;
        window_width = smooth_sdev/2.f;
        if (window_width < 0.) {
            error("ProcessNoiseEstimate.cpp:%i: Width %f < 0., iter %i",
                  __LINE__, window_width, iter);
            exit(1);
        }
        get_smoothed_average(&fine_histogram, window_width, 
                             &smooth_av, &smooth_sdev);
        if (smooth_sdev < 2.f) { // Prevent pathologically small widths
            smooth_sdev = 2.f * window_width;
        }
#if defined(DEBUG)
        printf("Smooth Av %f Sdev %f window %f\n",
               smooth_av, smooth_sdev, window_width);
#endif
    } while ((iter < 3) && (window_width > smooth_sdev));

    if (smooth_av < back_av)
        back_av = smooth_av;
    if ((smooth_sdev < back_sdev) && (smooth_sdev > minimum_dispersion))
        back_sdev = smooth_sdev;

    /*
     * Create a refined histogram based on the new more reliable
     * bin size estimated from the previous steps.
     */
    CountHistogram histogram;
    histogram.size     = 6*nbins+1;
    histogram.values   = new int [histogram.size];
    histogram.argwidth = back_sdev / nbins;
    histogram.argmin   = back_av - 2. * back_sdev - 0.5 * histogram.argwidth;

    {
        int lowidx, higidx, maxidx;
        const float y1sigma  = exp(-0.5);

        /*
         * Determine the pixel histogram for a subsequent fit.
         * Discard wings of the histogram since they are potentially not 
         * noise-dominated.
         */
        bin(image, n, &histogram);

        // discard low values in histogram since they are potentially not
        // noise-dominated
        clip_histogram(&histogram, y1sigma, y1sigma,
                       &maxidx, &lowidx, &higidx);

        float centre = histogram.argmin + (maxidx + 0.5) * histogram.argwidth;
        float sigma  = ((higidx - lowidx)/2.+0.5) * histogram.argwidth;
        float maxval = histogram.values[maxidx] / histogram.argwidth;
        back_max = maxval;

#if defined(DEBUG)
        printf("  Smooth:  f(x)=a*exp(-(x-b)*(x-b)/(2.*s*s)), a=%f, b=%f, s=%f\n",
               maxval, smooth_av, smooth_sdev);

        printf("  bin-width %g minarg %g centre %g sigma %g\n",
               histogram.argwidth, histogram.argmin, centre, sigma);
        printf("  hist: \n");
        for (int i = nbins; i < histogram.size-nbins; i++)
            printf("%7.2f %7.2f\n", histogram.argmin+histogram.argwidth/2+i*histogram.argwidth, histogram.values[i]/histogram.argwidth);
        printf("  low hi max: %i %i %i %7.2f %7.2f %7.2f\n",
               lowidx, higidx, maxidx,
               histogram.argmin+histogram.argwidth/2+lowidx*histogram.argwidth, histogram.argmin+histogram.argwidth/2+higidx*histogram.argwidth,
               histogram.argmin+histogram.argwidth/2+maxidx*histogram.argwidth);
#endif // defined(DEBUG)
        if (higidx - lowidx < 5) { // fit not constrained by data
#if defined(DEBUG)
            fprintf(stdout,
                    "Not enough histogram bins for fitting (%i %i).\n",
                    lowidx, higidx);
#endif // defined(DEBUG)
            if (higidx+1 < histogram.size) higidx += 1;
            if (lowidx-1 >= 0) lowidx -= 1;
        }
        if (higidx - lowidx < 5) { // fit not constrained by data
#if defined(DEBUG)
            fprintf(stdout,
                    "Not enough histogram bins for fitting (%i %i).\n",
                    lowidx, higidx);
#endif // defined(DEBUG)
            goto noise_hist_terminate;
        }

#if defined(DEBUG)
        printf("  low hi max: %i %i %i\n", lowidx, higidx, maxidx);
#endif // defined(DEBUG)
        clip_peak_histogram(&histogram, &maxidx, &lowidx, &higidx);
#if defined(DEBUG)
        printf("  low hi max: %i %i %i\n", lowidx, higidx, maxidx);
#endif // defined(DEBUG)

        /* 
         * Now try to fit a Gaussian to the histogram peak.
         */
        float fitted_max, fitted_av, fitted_sdev;
        fit_gaussian_histogram(&histogram, maxidx, lowidx, higidx,
                               &fitted_max, &fitted_av, &fitted_sdev);
        // try to catch failure of fitting procedure by comparing
        // to values obtained from previous methods and reset to input values
        if (((fitted_sdev < 1.5*sigma)  &&  (fitted_sdev < 1.1 * smooth_sdev) 
             && (fitted_sdev > minimum_dispersion) && (fitted_sdev > smooth_sdev/3.))
            && (fitted_av < smooth_av + smooth_sdev)
            && (fabs(fitted_av-centre) < back_sdev)
            && (fabs((fitted_max-maxval)/maxval) < .25)) {
            back_av   = fitted_av;
            back_sdev = fitted_sdev;
            back_max  = fitted_max;
        }
        else {
#if defined(DEBUG)
            printf("Fit (%f %f %f) too far from input (%f %f %f); resetting (%f %f %f)\n",
                   fitted_max, fitted_av, fitted_sdev,
                   maxval, centre, sigma,
                   back_max, back_av, back_sdev);
#endif
        }
    }

noise_hist_terminate:
    delete [] histogram.values;
    delete [] fine_histogram.values;

    /* a very small std dev is most likely due to image corruption: */
    if (fabs(back_sdev) < minimum_dispersion) {
        back_av   = smooth_av;
        back_sdev = smooth_sdev;
    }
#if defined(DEBUG)
    printf("  Final: f(x)=a*exp(-(x-b)*(x-b)/(2.*s*s)), a=%f, b=%f, s=%f\n",
           back_max, back_av, back_sdev);
#endif
    *average    = back_av;
    *dispersion = back_sdev;
}

/*
 * determine the image threshold to use for the binarisation of the image
 */
void determine_region_noise(float *image, int nx, int ny,
                            int x, int y, int sizex, int sizey,
                            float *background, float *noise)
{
#if defined(DEBUG)
    printf("x y: %i %i\n", x, y);
#endif

    // copy relevant image region to single array
    float *subimage = new float [sizex * sizey ];
    copy_image_region(image, nx, ny, subimage, sizex, sizey, x, y);

    determine_noise_hist_stats(subimage, sizex*sizey, background, noise);

#if defined(DEBUG) 
    printf("final background/noise (%4i %4i) %7.2f %7.2f\n",
           x, y, *background, *noise);
#endif

    delete [] subimage;
}

/*
 * estimate the local noise dispersion of the image.
 * the image is split into a number of tiles that need to be large enough
 * to allow statistically significant estimates of image properties,
 * but small enough to catch local variations in background and noise.
 */
void estimate_noise_image(float *image, int nx, int ny,
                          float *noise, float *background)
{
    // split the image into a set of subregions
    ImageRegionList list(nx, ny);
    list.setRegionSizes(get_local_image_size(nx, LOCAL_SIZE),
                        get_local_image_size(ny, LOCAL_SIZE));
    std::vector<ImageRegion> regions;
    list.getRegions(regions);
    int i;
    for (i = 0; i < int(regions.size()); i++) {
        ImageRegion region = regions[i];
        float noise_val, background_val;

        /* estimate noise dispersion in local image tile */
        determine_region_noise(image, nx, ny, region.XMin, region.YMin,
                               region.XSize, region.YSize,
                               &background_val, &noise_val);
        /* copy local noise dispersion to noise image */
        set_image_region_value(noise, nx, ny, region.XMin, region.YMin,
                               region.XSize, region.YSize, noise_val);

        /* copy local noise dispersion to background image */
        if (background)
            set_image_region_value(background, nx, ny, region.XMin, region.YMin,
                                   region.XSize, region.YSize, background_val);
    }
}
