//  Time-stamp: <26 Jun 2006 16:48 BST ProcessObjectDetection.cpp km>
/*
 * Routines to detect objects on a thresholded image
 *
 * Copyright (c) Solexa 2003-2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#include <string.h>
#else
#include <cmath>
#include <cstring>
#endif
#include <cstdio>
#include <cassert>
#include "Maths.h"
#include "Tools.h"
#include "Message.h"
#include "ProcessObjectDetection.h"

#define BINARY_DEFAULT_VALUE -1000.

void ClusterDetector::Run(int                  imageWidth,
                          int                  imageHeight,
                          float *              imageBuffers[],
                          AnalysisParameters & params,
                          int &                nObjects,
                          ClusterObject *  &   pObjects)
{
    nObjects = 0;
    if (pObjects)
    {
        delete [] pObjects;
        pObjects = 0;
    }

    int nmax = imageWidth * imageHeight / MIN_PIXELS_PER_OBJECT;
    pObjects = new ClusterObject [nmax];

    float * filter_image = imageBuffers[0];
    // float * image = imageBuffers[1];
    float * noise = imageBuffers[2];

    float * spatial = new float [imageWidth * imageHeight];

    // threshold image
    threshold_array(spatial, filter_image, noise, params.GaussFactor, imageWidth * imageHeight, BINARY_DEFAULT_VALUE);

    // object identification
    identify_objects(spatial, imageWidth, imageHeight, pObjects, nmax, &nObjects,
                     MIN_OBJ_PIXELS, BINARY_DEFAULT_VALUE);

    // deblending
    deblend_objects(pObjects, nmax, &nObjects, params.ClusterFwhm);

    // try to obtain better estimate of position of maximum by interpolating
    // peak
    interpolate_objects(filter_image, imageWidth, imageHeight, pObjects, nObjects);

    delete [] spatial;
}
//---------------------------------------------------------------------------
// pattern recognition
//---------------------------------------------------------------------------

/*
 * Given an initial pixel (x0, y0) on the boundary of a connected area,
 * return an array of pixels delineating the boundary of the area (or exclusion set)
 * such that each consecutive  pixel is just one coordinate change away from
 * the previous (i.e. no diagonal steps; the area is 4-connected) and the order
 * of the pixels corresponds to a clockwise walk.
 * Pixel values must be above threshold for pixels to belong to the area.
 */
void ClusterDetector::find_pixel_boundary(float *spatial, int nsx, int nsy,
                                          int x0, int y0, float threshold,
                                          st_pixel *boundary, int nmax, int *size)
{
    /* Directions are labelled as follows:
     *
     *  y            0
     *  ^            |
     *  |        3 --.-- 1
     *  .--> x       |
     *               2
     */
    const int up = 0, right = 1, down = 2, left = 3, none = -1;
    int direction0, new_direction, return_direction;
    int x, y, direction;
    int num;

    *size = 0;
    if (spatial[x0*nsy+y0] <= threshold)
        return;

    /* Identify an initial direction. Since the pixels on the boundary are
     * supposed to describe a closed loop and directions matter, this has
     * to be the direction from the previous pixel on the periodic boundary,
     * i.e. the last pixel in the whole list. So we need to pick a suitable pixel.
     * Since the parent routine is most likely scanning along the x and y axis,
     * a direction opposite to these is most likely to contain undetected pixels.
     * Hence, as an initial guess go left (i.e. last pixel is to the right)
     * instead of up or right.
     * Furthermore, make sure that
     *   1. no pixel is to the left of direction from the previous pixel
     *      (otherwise we might walk the boundary counter-clockwise)
     *   2. there is a previous pixel in the first place
     */
#if 0

    direction0 = left;
    if ((x0+1 < nsx) && (spatial[(x0+1)*nsy+(y0)] <= threshold)) {
        direction0 = down;
        if ((y0+1 < nsy) && (spatial[(x0)*nsy+(y0+1)] <= threshold)) {
            direction0 = right;
            if ((x0-1 >= 0) && (spatial[(x0-1)*nsy+(y0)] <= threshold)) {
                direction0 = up;
            }
        }
    }

#else
    direction0 = left;
    /* check if there is actually a previous pixel from the initial direction */
    if ((x0+1 < nsx) && (spatial[(x0+1)*nsy+(y0)] > threshold)) {
        /* there is a pixel, now make sure that there is no pixel to the left
         * of the previous pixel */
        if ((y0-1 >= 0) && (spatial[(x0)*nsy+(y0-1)] > threshold)) {
            direction0 = up;
            if ((x0-1 >= 0) && (spatial[(x0-1)*nsy+(y0)] > threshold)) {
                direction0 = right;
                if ((y0+1 < nsy) && (spatial[(x0)*nsy+(y0+1)] > threshold)) {
                    warning("find_pixel_boundary(): Point not on boundary!");
                    return;
                }
            }
        }
    }
    else {
        /* there is no previous pixel, so look for another direction */
        if ((y0+1 < nsy) && (spatial[(x0)*nsy+(y0+1)] > threshold))
            direction0 = down;
        else if ((x0-1 >= 0) && (spatial[(x0-1)*nsy+(y0)] > threshold))
            direction0 = right;
        else if ((y0-1 >= 0) && (spatial[(x0)*nsy+(y0-1)] > threshold))
            direction0 = up;
    }
#endif


    /* now walk along the boundary (clockwise) */
    x = x0; y = y0;
    direction = direction0;
    num = 0;
    do {
        /* add pixel to list */
        boundary[num].x = (unsigned short)x; boundary[num].y = (unsigned short)y;
        boundary[num].value = 1.;

        return_direction = (direction+2) % 4; // where we came from
        /* Determine a new direction to walk in.
         * Always try to turn left if possible. */
        direction = return_direction;
        new_direction = none;
        do {
            direction = (direction+1) % 4; /* turn right */
            switch (direction) {
            case up:
                if (((y+1) < nsy) && (spatial[(x)*nsy+(y+1)] > threshold)) {
                    y++; new_direction = direction;
                } break;
            case right:
                if (((x+1) < nsx) && (spatial[(x+1)*nsy+y] > threshold)) {
                    x++; new_direction = direction;
                } break;
            case down:
                if (((y-1) >= 0) && (spatial[(x)*nsy+(y-1)] > threshold)) {
                    y--; new_direction = direction;
                } break;
            case left:
                if (((x-1) >= 0) && (spatial[(x-1)*nsy+y] > threshold)) {
                    x--; new_direction = direction;
                } break;
            }
        } while ((direction != return_direction) && (new_direction == none));
        num++;
    } while ((new_direction != none) && (num < nmax)
             && ((x != x0) || (y != y0) || (direction0 != direction)));
    if (num == nmax)
        warning("Object boundary too long (%i)!", num);
    *size = num;
    return;
}

// find the maximum in an array of pixels
void ClusterDetector::find_pixel_maximum(st_pixel *pixels, int n_pixels, st_pixel *max_pixel)
{
    int i;
    max_pixel->x = 0; max_pixel->y = 0; max_pixel->value = 0;
    for (i = 0; i < n_pixels; i++)
        if (pixels[i].value > max_pixel->value)
            memcpy(max_pixel, &pixels[i], sizeof(st_pixel));
}

// find the smallest square containing all pixels
void ClusterDetector::find_pixel_extent(st_pixel *pixels, int n_pixels,
                                        struct st_position *min,
                                        struct st_position *max)
{
    assert(n_pixels > 0);
    min->x = pixels[0].x; max->x = pixels[0].x;
    min->y = pixels[0].y; max->y = pixels[0].y;
    int i;
    for (i = 1; i < n_pixels; i++) {
        if (pixels[i].x < min->x) min->x = pixels[i].x;
        if (pixels[i].x > max->x) max->x = pixels[i].x;
        if (pixels[i].y < min->y) min->y = pixels[i].y;
        if (pixels[i].y > max->y) max->y = pixels[i].y;
    }
}

void ClusterDetector::list_pixels(st_pixel *pixels, int n_pixels)
{
    int i;
    for (i = 0; i < n_pixels; i++)
        printf("%2d: %2d %2d %7.1f\n", i, pixels[i].x, pixels[i].y, pixels[i].value);
}

/*
 * Given a <boundary> of <nb> pixel delineating a compact region
 * on the binarised <nsx>*<nsy>-image <spatial>, return an array containing
 * all <na> pixels lying within this <area>. <min.x>, <xmax>, <ymin>, <ymax>
 * indicate the corner points of the smallest rectangle the area fits in.
 */
void ClusterDetector::find_pixel_area(st_pixel           *boundary,
                                      int                 nb,
                                      struct st_position *min,
                                      struct st_position *max,
                                      float              *spatial,
                                      int                 nsx,
                                      int                 nsy,
                                      st_pixel           *area,
                                      int                *na)
{
    const unsigned short BOUNDARY_UPPER = (1<<1);
    const unsigned short BOUNDARY_LOWER = (1<<2);
    int i, x, y;
    int nx = max->x - min->x + 1;
    int ny = max->y - min->y + 1;
    unsigned short *bd_array = new unsigned short [nx*ny];

    for (i = 0; i < nx*ny; i++)
        bd_array[i] = 0;

    /* create 2-d array with points on upper and lower boundary
     * marking the area */
    int prev_x, next_x, prev_y, next_y;
    x = boundary[nb-1].x; y = boundary[nb-1].y;
    for (i = 0; i < nb; i++) {
        prev_x = x; prev_y = y;
        x = boundary[i].x; y = boundary[i].y;
        next_x = boundary[(i+1)%nb].x; next_y = boundary[(i+1)%nb].y;

        // mark points on upper boundary (walk is clockwise, so left to right)
        if ((y >= prev_y) && (y >= next_y) && (next_x >= prev_x))
            bd_array[(x-min->x)*ny+(y-min->y)] |= BOUNDARY_UPPER;
        // mark points on lower boundary (walk is clockwise, so right to left)
        if ((y <= prev_y) && (y <= next_y) && (next_x <= prev_x))
            bd_array[(x-min->x)*ny+(y-min->y)] |= BOUNDARY_LOWER;
    }

    /* for each x-value, now scan pixels along y-direction
     * between lower and upper boundary */
    int inarea;
    int idx = 0;
    for (x = min->x; x <= max->x; x++) {
        unsigned short *bd_line = bd_array+(x-min->x)*ny;

        inarea = 0;
        for (y = min->y; y <= max->y; y++) {
            if (bd_line[y-min->y] & BOUNDARY_LOWER) // entering area
                inarea = 1;
            if (inarea) {  // found a pixel
                area[idx].x = x; area[idx].y = y; area[idx].value = spatial[x*nsy+y];
                idx++;
            }
            if (bd_line[y-min->y] & BOUNDARY_UPPER) // leaving area
                inarea = 0;
        }
    }
    delete [] bd_array;

    *na = idx;
    return;
}

void ClusterDetector::fill_pixel_area(st_pixel *area, int na, float *spatial,
                                      int nsx, int nsy, float value)
{
    int i, x, y;
    for (i = 0; i < na; i++) {
        x = area[i].x; y = area[i].y;
        spatial[x*nsy+y] = value;
    }
}

/*
 * apply the object recognition on a thresholded image and identify
 * new objects above a threshold.
 * the image <spatial> has dimensions <ns>*<ns> and will be set
 * to zero after exit.
 */
void ClusterDetector::identify_objects(float *spatial, int nsx, int nsy,
                                       ClusterObject *objs, int nmax, int *nobjs,
                                       unsigned short min_obj_size, float threshold)
{
    int nb, na;
    int nbmax = 4*(nsx+nsy);
    int namax = nsx * nsy;
    st_pixel *boundary = new st_pixel [ nbmax ];
    st_pixel *area = new st_pixel [ namax ];

    int x, y;
    for (x = 0; x < nsx; x++)
        for (y = 0; y < nsy; y++) {
            if (spatial[x*nsy+y] > threshold) {
                if (*nobjs >= nmax) {
                    warning("Too many objects (%i >= %i)!", *nobjs, nmax);
                    return;
                }
                ClusterObject *obj = &objs[*nobjs]; // SB: Replaces: st_object *obj = &objs[*nobjs];
                find_pixel_boundary(spatial, nsx, nsy, x, y, threshold,
                                    boundary, nbmax, &nb);
                obj->SetObjectBoundary(boundary, nb); // SB: Replaces: set_object_boundary(obj, boundary, nb);
                find_pixel_extent(obj->boundary, obj->nb,
                                  &obj->min, &obj->max);
                find_pixel_area(obj->boundary, obj->nb,
                                &obj->min, &obj->max, spatial, nsx, nsy,
                                area, &na);
                if (nb == nbmax) {
                    info("Large Object: %i %i %i %i %i %i\n", x, y, nb, nbmax, na, namax);
#if 0
                    for (i = 0; i < nb; i++)
                        printf("%i %i %i %i %i\n", x, y, i, boundary[i].x, boundary[i].y);
#endif
                }

                obj->SetObjectArea(area, na); // SB: Replaces: set_object_area(obj, area, na);
                find_pixel_maximum(obj->area, obj->na, &obj->max_pixel);
                // set all image pixels within the area to zero to avoid counting
                // them again
                fill_pixel_area(obj->area, obj->na, spatial, nsx, nsy, BINARY_DEFAULT_VALUE);
                if ((nb == nbmax) || (na == namax))
                    spatial[x*nsy+y] = BINARY_DEFAULT_VALUE;
#if 0
                    printf("Area %i: boundary %i area %i\n", *nobjs, area.nb, area.na);
                    printf("  max: %i %i %i\n", area.max_pixel.x, area.max_pixel.y,
                           area.max_pixel.value);
#endif

                // discard objects with too few pixels
                if (obj->na >= min_obj_size)
                    (*nobjs)++;
                else
                    obj->clear(); // SB: Replaces: free_object(obj);
            }
        }
    delete [] area;
    delete [] boundary;
}


//---------------------------------------------------------------------------
// deblending
//---------------------------------------------------------------------------

/*
 * Find the global <maximum> of an <image>.
 */
void ClusterDetector::find_global_maximum(float *image, int nx, int ny, st_pixel *maximum)
{
    maximum->x     = 0;
    maximum->y     = 0;
    maximum->value = 0.;

    int i, j;
    for (i = 0; i < nx; i++)
        for (j = 0; j < ny; j++) {
            if (image[i*ny+j] > maximum->value) {
                maximum->x     = i;
                maximum->y     = j;
                maximum->value = image[i*ny+j];
            }
        }
}

/*
 * Identify local maxima above the <threshold> on an <nx>x<ny>-dimensional
 * <image>.
 * The factor <fudge> is a factor between the maximum pixel and the next
 * best pixel.
 * Positions of local maxima must be separated by at least <min_dist>.
 */
void ClusterDetector::find_local_maxima(st_pixel *maxima, int nm_max, int *nmax,
                                        float *image, int nx, int ny,
                                        float threshold, float min_dist, float fudge)
{
    int      size = (int)floor(min_dist);
    int      xmin, xmax, ymin, ymax, x2, y2;
    int      flag;
    float    value, fudge_value;

    *nmax = 0;
    int x, y;
    for (x = 0;  x < nx; x++)
        for (y = 0;  y < ny; y++) {
            value = image[x*ny+y];
            fudge_value = fudge * value;
            if (value > threshold) {
                flag = 1;
                xmin = Maths::max(0, x-size); xmax = Maths::min(nx-1, x+size);
                ymin = Maths::max(0, y-size); ymax = Maths::min(ny-1, y+size);
                x2 = xmin;
                while ((flag) && (x2 <= xmax)) {
                    for (y2 = ymin; y2 <= ymax; y2++) {
                        if (((float)((x2-x)*(x2-x)+(y2-y)*(y2-y)) < min_dist*min_dist)
                            && ((x != x2) || (y != y2))
                            && (image[x2*ny+y2] > fudge_value))
                            flag = 0;
                    }
                    x2++;
                }

#if defined(SIMULATION)
                // if only four points contributed to this area, check for saddle
                // points. possible configurations:
                //   xoo  oox
                //   oxo  oxo
                //   oox  xoo
                if ((flag) && (min_dist < sqrt(2.))) {
                    if ((x-1 >= 0) && (y-1 >= 0) && (x+1 < nx) && (y+1 < ny)) {
                        if ((image[(x-1)*ny+(y-1)] > fudge_value)
                            && (image[(x+1)*ny+(y+1)] > fudge_value))
                            flag = 0;
                        else if ((image[(x-1)*ny+(y+1)] > fudge_value)
                                 && (image[(x+1)*ny+(y-1)] > fudge_value))
                            flag = 0;
                    }
                }
#endif

                // found local maximum
                if ((flag == 1) && (*nmax < nm_max)) {
                    maxima[(*nmax)].x = x; maxima[(*nmax)].y = y;
                    maxima[(*nmax)].value = value;
                    (*nmax)++;
                }
            }
        }

    /* Now make sure that the global maximum is included in the list
     * (depending on the fudge factor, this may not be the case).
     */
    if (*nmax < nm_max) {
        st_pixel global_max;
        find_global_maximum(image, nx, ny, &global_max);
        int flag = 0;
        int i;
        for (i = 0; i < *nmax; i++)
            if ((maxima[i].x == global_max.x) && (maxima[i].y == global_max.y)) {
                flag = 1;
                break;
            }
        if ((!flag)) {
            maxima[(*nmax)].x = global_max.x;
            maxima[(*nmax)].y = global_max.y;
            maxima[(*nmax)].value = global_max.value;
            (*nmax)++;
        }
    }
}


/*
 * Given a list of local <maxima> and an image region <image> of
 * dimension <nx>*<ny>, create a set of areas around each maximum
 * by assigning each pixel to the maximum that is closest to it.
 * The caller must provide arrays <objs> and <maxima> of minimum
 * length <n_maxima>.
 */
void ClusterDetector::split_local_maxima(ClusterObject *objs, st_pixel *maxima, int n_maxima,
                                         float *image, int nx, int ny, float threshold)
{
    // for each pixel, find nearest maximum;
    // the index of the nearest maximum is stored in the
    // array <basins>. also keep track of how many pixels were
    // assigned to each maximum using the array <counts>
    int i, j, k;
    int *basins = new int [nx*ny];
    int *counts = new int [n_maxima];
    for (k = 0; k < n_maxima; k++)
        counts[k] = 0;
    for (i = 0; i < nx*ny; i++)
        basins[i] = 0;

    for (i = 0; i < nx; i++)
        for (j = 0; j < ny; j++) {
            if (image[i*ny+j] > threshold) {
                int mindist2 = nx*nx + ny*ny;
                for (k = 0; k < n_maxima; k++) {
                    int xdist = i - maxima[k].x;
                    int ydist = j - maxima[k].y;
                    int dist2 = xdist*xdist + ydist*ydist;
                    if (dist2 < mindist2) {
                        mindist2 = dist2;
                        basins[i*ny+j] = k+1;
                    }
                }
                counts[basins[i*ny+j]-1] ++;
            }
        }

    /* Allocate arrays to hold all pixels for the maxima.
     * make sure that the maximal number of objects is not exceeded.
     */
    for (k = 0; k < n_maxima; k++) {
        st_pixel *area = new st_pixel[counts[k]];
        objs[k].SetObjectArea(area, counts[k]); // SB: Replaces: set_object_area(&objs[k], area, counts[k]);
        delete [] area;
        objs[k].nb = 0;
    }

    /* Fill the arrays with all pixels assinged to a given maximum.
     * Also assign maximum and a minimal rectangle around the pixels.
     */
    for (k = 0; k < n_maxima; k++)
        counts[k] = 0;
    for (i = 0; i < nx; i++)
        for (j = 0; j < ny; j++) {
            k = basins[i*ny+j]-1; assert(k < n_maxima);
            if (k < 0)
                continue;
            objs[k].area[counts[k]].x = i;
            objs[k].area[counts[k]].y = j;
            objs[k].area[counts[k]].value = image[i*ny+j];
            counts[k] ++;
        }
    for (k = 0; k < n_maxima; k++) {
        find_pixel_extent(objs[k].area, objs[k].na,
                          &(objs[k].min), &(objs[k].max));
        find_pixel_maximum(objs[k].area, objs[k].na, &(objs[k].max_pixel));
        objs[k].is_split = 1;
    }

    delete [] counts;
    delete [] basins;
}

/*
 * Given an image area <image> of dimension (<nx>, <ny>), find all local maxima
 * in this area and add a list of corresponding objects to the list of <objects>.
 * the image <image> is assumed to contain at most <npixels> pixels with values
 * above <threshold>
 */
void ClusterDetector::find_local_objects(ClusterObject *objects, int nmax,
                                         float *image, int nx, int ny, int npixels,
                                         float threshold, float min_dist, int *n_maxima)
{
    TRACE("find_local_objects(%p, %i, %p, %i, %i, %i, %f, %f, %i)\n",
          (void *)objects, nmax, (void *)image, nx, ny, npixels,
          threshold, min_dist, *n_maxima);
    st_pixel *maxima = new st_pixel[npixels];

    // find local maxima
    find_local_maxima(maxima, npixels, n_maxima,
                      image, nx, ny, threshold, min_dist);
    // now identify new areas
    if (*n_maxima > nmax)
       *n_maxima = nmax;
    if (*n_maxima > 1) {
        split_local_maxima(objects, maxima, *n_maxima,
                           image, nx, ny, threshold);
    }
    delete [] maxima;

    TRACE("~find_local_objects(%p, %i, %p, %i, %i, %i, %f, %f, %i)\n",
          (void* )objects, nmax, (void *)image, nx, ny, npixels,
          threshold, min_dist, *n_maxima);
}

/*
 * Given the object <original>, try to split into new objects around
 * local minima. If there is more than one new object,
 * the new objects are added to the list <objs> up to a maximum of <nmax>
 * entries.
 */
void ClusterDetector::split_object(ClusterObject *original,
                                   ClusterObject *objs, int nmax, int *nobjs,
                                   float threshold, float min_dist)
{
    /* Make sure that distance between objects is at least one pixel. */
    const float tiny = 1.e-12;
    if (min_dist <= 1.f + tiny)
        min_dist = 1.f + tiny;

    /* Find local maxima. */
    int    nsx, nsy;
    int    n_maxima = 1;
    float *sp = original->MakeNewImage(&nsx, &nsy, threshold); // SB: Replaces: float *sp = new_image_from_object(&nsx, &nsy, original, threshold);
    find_local_objects(objs + (*nobjs), nmax - (*nobjs),
                       sp, nsx, nsy, original->na,
                       threshold, min_dist, &n_maxima);
    delete [] sp;

    if (n_maxima > 1) {
        /* If the original object was split into at least two new ones
         * overwrite original and add new objects
         */
#if defined(DEBUG) && 0
        printf("Object split into %i.\n", n_maxima);
#endif
        /* Adjust object coordinates. */
        unsigned short x0 = original->min.x;
        unsigned short y0 = original->min.y;
        int i;
        for (i = 0; i < n_maxima; i++)
            objs[*nobjs+i].Shift(x0, y0); // SB: Replaces: shift_object(&objs[*nobjs+i], x0, y0);

#if defined(DEBUG) && 0
        int j;
        for (i = *nobjs; i < *nobjs + n_maxima; i++)
            for (j = 0; j < objs[i].na; j++)
                printf("Split %3d: %3d: %d %d %.1f\n", i+1, j,
                       objs[i].area[j].x, objs[i].area[j].y, objs[i].area[j].value);
#endif

        original->clear(); // SB: Replaces: free_object(original);
        *original = objs[*nobjs+n_maxima-1]; // SB: Replaces: copy_object(&objs[*nobjs+n_maxima-1], original);

        *nobjs += (n_maxima-1);
    }
    else {
#if defined(DEBUG) && 0
        printf("Did not split object.\n");
#endif
    }

    return;
}

/*
 * Check for local maxima and split up objects if required.
 */
void ClusterDetector::deblend_objects(ClusterObject *objs,
                                      int nmax,
                                      int *nobjs,
                                      float fwhm)
{
    // split up objects of a size above <max_obj_area>
    int max_obj_area = (int)ceil(2.*fwhm*fwhm);
    // discard objects that are too large and most likely artifacts
    // based on some heuristic values
    int max_feature_size = (int)ceil(8.*fwhm);
    int max_feature_area = (int)ceil(16.*fwhm*fwhm);
    int nobjs2 = *nobjs;
    int i;
    for (i = 0; i < nobjs2; i++) {
        // if object exceeds a maximum size
        objs[i].is_split = 0;
        if (objs[i].na > max_obj_area) {
            // if object is too large, discard
            // TODO: actually discard the object rather than
            // just not splitting it up
            if ((objs[i].na < max_feature_area)
                && (objs[i].max.x - objs[i].min.x + 1 < max_feature_size)
                && (objs[i].max.y - objs[i].min.y + 1 < max_feature_size)) {
                float min_dist = Maths::max(1.1f, float(fwhm/2.));
                split_object(&objs[i], objs, nmax, nobjs,
                             BINARY_DEFAULT_VALUE, min_dist);
            }
        }
    }
#if defined(DEBUG) && 0
    printf("number of objects: %4i, before deblending: %4i, new: %4i\n",
           *nobjs, nobjs2, *nobjs-nobjs2);
    for (i = 0; i < *nobjs; i++)
        printf("%4i: %d %d area size %d %d\n", i, objs[i].max_pixel.x,
               objs[i].max_pixel.y, objs[i].na, objs[i].max_pixel.value);
#endif
}


#if defined(DEBUG)
static void print_vector(float *v, int n, const char *comment = NULL)
{
    int i;
    if (comment != NULL)
        printf(comment);
    printf(" (");
    for (i = 0; i < n-1; i++)
        printf("%f,", v[i]);
    printf("%f)\n", v[n-1]);
}
#endif


/*
 * given a a set of 3 points (xi[i], yi[i]), find a parabola that
 * interpolates the points and determine the maximum of the parabola.
 * this function is used to interpolate the maximum of the correlation
 * function between two images to non-integer lags.
 */
float ClusterDetector::parabola_extremum(float xi[], float yi[])
{
    float numerator = 0.;
    float denominator = 0.;
    int i;
    for (i = 0; i < 3; i++) {
        int i2 = (i+1) % 3;
        int i3 = (i+2) % 3;
        numerator += yi[i] * (xi[i2]*xi[i2]-xi[i3]*xi[i3]);
        denominator += yi[i] * (xi[i2]-xi[i3]);
    }
    if (fabs(denominator) > 1.e-14)
        return numerator/(2.f*denominator);
    return xi[1];
}

/*
 * Estimate the position of the peak of the object intensity
 * by performing a 3-point parabolic interpolation through
 * maximum pixel value along the x- and y-axis.
 */
void ClusterDetector::interpolate_object(float * buffer, int nx, int ny, ClusterObject * object)
{
    int x = object->max_pixel.x;
    int y = object->max_pixel.y;
    object->position.x = (float)x;
    object->position.y = (float)y;
    float xv[3], yv[3];
    int i;
    if ((x > 0) && (x < nx-1)) {
        for (i = 0; i < 3; i++) {
            xv[i] = x+i-1;
            yv[i] = buffer[(x+i-1)*ny+y];
        }
        object->position.x = parabola_extremum(xv, yv);
    }
    if ((y > 0) && (y < ny-1)) {
        for (i = 0; i < 3; i++) {
            xv[i] = y+i-1;
            yv[i] = buffer[x*ny+(y+i-1)];
        }
        object->position.y = parabola_extremum(xv, yv);
    }
}

void ClusterDetector::interpolate_objects(float * buffer,
                                          int nx,
                                          int ny,
                                          ClusterObject * objects,
                                          int nobj)
{
    int i;
    for (i = 0; i < nobj; i++)
        interpolate_object(buffer, nx, ny, &(objects[i]));
}
