/*
 * Detect objects on a thresholded image
 *
 * Copyright (c) Solexa 2005, 2006
 * Author:
 * Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#ifndef _PROCESSOBJECTDETECTION_H_
#define _PROCESSOBJECTDETECTION_H_

#include "ClusterObject.h"
#include "ObjectDetectorIFC.h"

#define MIN_OBJ_PIXELS 3

class ClusterDetector : public IObjectDetector
{
public:
    ClusterDetector()
        : IObjectDetector()
    {}

    void Run(int                  imageWidth,
             int                  imageHeight,
             float *              imageBuffers[],
             AnalysisParameters & params,
             int &                nObjects,
             ClusterObject *    & pObjects);

protected:
    void interpolate_objects(float         * buffer, 
                             int             nx,
                             int             ny,
                             ClusterObject * objects,
                             int             nobj);

    void identify_objects(float         * spatial,
                          int             nx,
                          int             ny,
                          ClusterObject * objs,
                          int             nmax,
                          int           * nobjs,
                          unsigned short  min_obj_size = MIN_OBJ_PIXELS,
                          float           threshold = 0.);

    void deblend_objects(ClusterObject * objs,
                         int             nmax,
                         int           * nobjs,
                         float           fwhm);

private:
    /*********************************
    *     HELPER FUNCTIONS
    *********************************/

    /*
     * Given an initial pixel (x0, y0) on the boundary of a connected area,
     * return an array of pixels delineating the boundary of the area (or exclusion set)
     * such that each consecutive  pixel is just one coordinate change away from
     * the previous (i.e. no diagonal steps; the area is 4-connected) and the order
     * of the pixels corresponds to a clockwise walk.
     * Pixel values must be above threshold for pixels to belong to the area.
     */
    void find_pixel_boundary(float *spatial, int nsx, int nsy,
                             int x0, int y0, float threshold,
                             st_pixel *boundary, int nmax, int *size);

    // find the maximum in an array of pixels
    void find_pixel_maximum(st_pixel *pixels, int n_pixels, st_pixel *max_pixel);

    // find the smallest square containing all pixels
    void find_pixel_extent(st_pixel *pixels, int n_pixels,
                           struct st_position *min,
                           struct st_position *max);

    void list_pixels(st_pixel *pixels, int n_pixels);

    /*
     * Given a <boundary> of <nb> pixel delineating a compact region
     * on the binarised <nsx>*<nsy>-image <spatial>, return an array containing
     * all <na> pixels lying within this <area>. <min.x>, <xmax>, <ymin>, <ymax>
     * indicate the corner points of the smallest rectangle the area fits in.
     */
    void find_pixel_area(st_pixel           *boundary,
                         int                 nb,
                         struct st_position *min,
                         struct st_position *max,
                         float              *spatial, 
                         int                 nsx,
                         int                 nsy,
                         st_pixel           *area,
                         int                *na);

    void fill_pixel_area(st_pixel *area, int na, float *spatial,
                         int nsx, int nsy, float value);
    /*
     * Find the global <maximum> of an <image>.
     */
    void find_global_maximum(float *image, int nx, int ny, st_pixel *maximum);

    /*
     * Identify local maxima above the <threshold> on an <nx>x<ny>-dimensional
     * <image>.
     * The factor <fudge> is a factor between the maximum pixel and the next
     * best pixel.
     * Positions of local maxima must be separated by at least <min_dist>.
     */
    void find_local_maxima(st_pixel *maxima, int nm_max, int *nmax,
                           float *image, int nx, int ny,
                           float threshold, float min_dist, float fudge = 1.0f);

    /*
     * Given a list of local <maxima> and an image region <image> of
     * dimension <nx>*<ny>, create a set of areas around each maximum
     * by assigning each pixel to the maximum that is closest to it.
     * The caller must provide arrays <objs> and <maxima> of minimum
     * length <n_maxima>.
     */
    void split_local_maxima(ClusterObject *objs, st_pixel *maxima, int n_maxima,
                            float *image, int nx, int ny, float threshold);

    /*
     * Given an image area <image> of dimension (<nx>, <ny>), find all local maxima
     * in this area and add a list of corresponding objects to the list of <objects>.
     * the image <image> is assumed to contain at most <npixels> pixels with values 
     * above <threshold>
     */
    void find_local_objects(ClusterObject *objects, int nmax,
                            float *image, int nx, int ny, int npixels,
                            float threshold, float min_dist, int *n_maxima);

    /*
     * Given the object <original>, try to split into new objects around
     * local minima. If there is more than one new object,
     * the new objects are added to the list <objs> up to a maximum of <nmax>
     * entries.
     */ 
    void split_object(ClusterObject *original,
                      ClusterObject *objs, int nmax, int *nobjs,
                      float threshold, float min_dist);

    /*
     * given a a set of 3 points (xi[i], yi[i]), find a parabola that
     * interpolates the points and determine the maximum of the parabola.
     * this function is used to interpolate the maximum of the correlation
     * function between two images to non-integer lags.
     */
    float parabola_extremum(float xi[], float yi[]);

    /*
     * Estimate the position of the peak of the object intensity
     * by performing a 3-point parabolic interpolation through 
     * maximum pixel value along the x- and y-axis.
     */
    void interpolate_object(float * buffer, int nx, int ny, ClusterObject * object);
};

#endif //_PROCESSOBJECTDETECTION_H_
