/* 
 * Copyright (c) 2007 Illumina 
 * Authors: Mirian Karbelashvili, Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
#ifndef TILE_ANALYSER_H_
#define TILE_ANALYSER_H_

struct TileInfo;
struct RunInfo;
struct AnalysisParameters;
struct Alignment;
struct Position;
struct TileQCValues;

class DataProcessor;
class Image;
class RunFolderFileName;

class TileAnalyser
{
private:
	const TileInfo	          * m_TileInfo;
	const Position			  * m_Positions;
	int							m_NumPositions;
	const RunInfo	          * m_RunInfo;
	struct AnalysisParameters * m_Parameters;
			/* num Alignment elements = NUM_CHANNELS */
	struct Alignment          * m_DefaultAlignment;
    int                       * m_RefIndices;
    Position                  * m_Reference;
    class DataProcessor       * m_Processor;
    class Image              ** m_Images;

public:

	// Online version
	TileAnalyser(// Data from previous cycles. firstCycle + numCycles - 1 < tileInfo.Cycle.
		const TileInfo			 & tileInfo,
		const Position			 * positions,
		int						   numPositions,
		const RunInfo			 & runInfo,
		const AnalysisParameters & parameters,
		const Alignment			 * defaultAlignment// num elements = num channels
	);

	// Offline version
	TileAnalyser( const RunInfo & runInfo );

	virtual ~TileAnalyser();

	int GetNumPositions() const
	{
		return m_NumPositions;
	}

	// Used in the offline version.
	void SetParameters( const AnalysisParameters & parameters );

    // Used in the offline version.
    void SetTile( const TileInfo & tileInfo );
    
	void SetAlignment( const Alignment * defaultAlignment );// num elements = NumChannels

	///**** Called 'NumChannels' times in random order.
	// Num bytes in the 'image' = ImageWidth * ImageHeight.
	virtual void AnalyseImage( unsigned short * image, int colour, char base );

	///**** Called during first cycle only, after all AnalyseImage(), before CompleteTile().
	// TileAnalyser object allocates memory for 'positions', deallocates it in the destructor.
	virtual void BuildReference( Position ** positions, int * numClusters,
                                 Position *oldPositions = 0, int numOldClusters = 0 );

	///**** Called after calls to AnalyseImage() and BuildReference().
	// Number of elements in each array equals 'numElements'.
	// Number of used elements in each array equals 'numClusters' <= 'numElements'.
	// 'alignment': pointer to an array containing NumChannels elements - filled in during this call.
	// 'errorMessage': if nonzero, returns a pointer to the internal buffer containing error message.
	virtual void CompleteTile(
		float	     * intensity,	// [out]. 'NumChannels' values per cluster.
		float	     * noise,		// [out]. 'NumChannels' values per cluster.
		Position     * positions,	// [out]. One value per cluster.
		int		     & numClusters,	// [in].  Number of clusters (used elements) in the above arrays.
		int		       numElements,	// [in].  Total number of elements in the above arrays, numElements >= numClusters.
		Alignment    * alignment,	// [out]. Array of 'NumChannels' elements, allocated by the caller.
		TileQCValues * qcValues,	// [out]. Array of 'NumChannels' elements, allocated by the caller.
		const char  ** errorMessage	// [out]. If nonzero pointer is supplied, returns an internally allocated error message.
	);

	// Return some cluster size QC-s from here?
	void SaveClusterDetails( const TileInfo & tileInfo, const char * folder,
                             int numClusters );
    void SaveQCFile( const TileInfo & tileInfo, const char * folder );

	static void SaveIntensity( const TileInfo & tileInfo, const char * folder,
                               int numClusters, int numChannels, float * intensity );
	static void SaveNoise( const TileInfo & tileInfo, const char * folder,
                           int numClusters, int numChannels, float * noise );
	static void SavePositions( const TileInfo & tileInfo, const char * folder,
                               int numClusters, Position * positions );
    static void SaveIndexPositions( const TileInfo & tileInfo, const char * folder,
                                    int numClusters, Position * positions );


    // Offline version interfaces
    DataProcessor * GetAnalyser();
	void SetReference( const Position * positions, int numPositions );
    virtual void AnalyseImage( Image * image, int colour, char base );
private:
    static RunFolderFileName *GetName( const TileInfo & tileInfo, const char * folder );
};

#endif
