/*
 * run single analysis on image
 *
 * Copyright (c) Solexa 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 * 
 * The utility can be compiled with a graphical display if the PGPLOT plotting
 * library is installed on the system.
 */

#include <iostream>
#include <cstdio>
#include "plot.h"
#include "FFT.h"
#include "ParameterTypes.h"
#include "ImageProcess.h"
#include "Images.h"
#include "BinaryImage.h"
#include "Tools.h"
#include "ImageUtil.h"
#include "ProcessFilter.h"
#include "ProcessNoiseEstimate.h"
#include "Message.h"
// #include "ProcessObjectDetection.h"

#if defined(PLOT_IMAGE) && defined(HAVE_PGPLOT)
#include <cpgplot.h>
#include "plot_image.h"
#endif

#define N_POS_FILES 10

#if defined(PLOT_IMAGE) && defined(HAVE_PGPLOT)
static void scale_shift_fpositions(st_fposition *positions, int npos,
                                   float xoff, float yoff,
                                   float xscale, float yscale,
                                   float xcentre, float ycentre)
{
    int i;
    for (i = 0; i < npos; i++) {
        float x, y;
        x = positions[i].x - xoff;
        y = positions[i].y - yoff;
        positions[i].x = x + (x-xcentre)*xscale;
        positions[i].y = y + (y-ycentre)*yscale;
    }
}
#endif

int main(int argc, char *argv[])
{
    int nx = DEFAULT_PIXEL_DIM, ny = DEFAULT_PIXEL_DIM;
    int sizex = 0, sizey = 0, offx = 0, offy = 0;
    BinaryImage mImage;
    ImageData *data;
    unsigned short *image = NULL;
    float *imager;
    AnalysisParameters params;
    char *filename = NULL;
    float shiftx[N_POS_FILES], shifty[N_POS_FILES];
    float scalex[N_POS_FILES], scaley[N_POS_FILES];
    char *posfilenames [N_POS_FILES];
    int nposfiles = 0;
    int nshiftx = 0, nshifty = 0;
    int nscalex = 0, nscaley = 0;
    int nox, noy;
    unsigned short bitsperpixel = 12;
    int dumprawflag = 0, dumpfilterflag = 0;
    int summary_flag = 0;
    int image_index = 0;
    int bytes = 2;
    int indexflag = 0;
    int saveflag = 0;
    int object_stats_flag = 0;
    int i, j;
    params.ClusterFwhm        = 0.f;
    params.ClusterDistance    = 0.f;

    for (i = 0; i < N_POS_FILES; i++) {
        shiftx[i] = 0.; shifty[i] = 0.; 
        scalex[i] = 0.; scaley[i] = 0.; 
        posfilenames[i] = NULL;
    }

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-')
            switch (argv[i][1]) {
            case 'i': sscanf(argv[i], "-i%i", &image_index); break;
            case 'd':
                if (argv[i][2])
                    switch (argv[i][2]) {
                    case 'r': dumprawflag = 1; break;
                    case 'f': dumpfilterflag = 1; break;
                    default: sscanf(argv[i], "-d%f", &(params.ClusterDistance)); break;
                    }
                break;
            case 'f':
                if (argv[i][2])
                    switch (argv[i][2]) {
                    case 'm': params.FilterType = 0; break;
                    case 'h': params.FilterType = 1; break;
                    case 'a': params.FilterType = 2; break; // Matched 
                    case 'b': params.FilterType = 3; break; // Biaparametric
											
                    default: sscanf(argv[i], "-f%f", &(params.ClusterFwhm)); break;
                    }
                break;
            case 'n':
                if (argv[i][2]) {
                    switch (argv[i][2]) {
                    case 'x': sscanf(argv[i], "-nx%i", &nx); break;
                    case 'y': sscanf(argv[i], "-ny%i", &ny); break;
                    case 'o': 
                        if (strcmp(argv[i], "-nofilter") == 0)
                            params.FilterType = -1;
                    }
                }
                else
                    indexflag = 1;
                break;
            case 'o': 
                if (argv[i][2])
                    switch (argv[i][2]) {
                    case 'x': sscanf(argv[i], "-ox%i", &offx); break;
                    case 'y': sscanf(argv[i], "-oy%i", &offy); break;
                    }
                break;
            case 's': 
                if (argv[i][2]) {
                    switch (argv[i][2]) {
                    case 'x': sscanf(argv[i], "-sx%i", &sizex); break;
                    case 'y': sscanf(argv[i], "-sy%i", &sizey); break;
                    case 't': object_stats_flag = 1; break;
                    }
                }
                else 
                    saveflag = 1;
                break;
            case 'p':
                if (argv[i][2]) {
                    switch (argv[i][2]) {
                    case 'o':
                        if (argv[i][3]) {
                            switch (argv[i][3]) {
                            case 'x': 
                                if (nshiftx < N_POS_FILES) {
                                    sscanf(argv[i], "-pox%f", &shiftx[nshiftx]); nshiftx++; 
                                } break;
                            case 'y': 
                                if (nshifty < N_POS_FILES) {
                                    sscanf(argv[i], "-poy%f", &shifty[nshifty]); nshifty++;
                                } break;
                            }
                        }
                        break;
                    case 's':
                        if (argv[i][3]) {
                            switch (argv[i][3]) {
                            case 'x': 
                                if (nscalex < N_POS_FILES) {
                                    sscanf(argv[i], "-psx%f", &scalex[nscalex]); nscalex++; 
                                } break;
                            case 'y': 
                                if (nscaley < N_POS_FILES) {
                                    sscanf(argv[i], "-psy%f", &scaley[nscaley]); nscaley++;
                                } break;
                            }
                        }
                        break;
                    }
                } break;
            case 'b': sscanf(argv[i], "-b%i", &bytes); break;
            case 't': sscanf(argv[i], "-t%f", &(params.GaussFactor)); break;
            case 'v': summary_flag += 1; break;
        }
        else {
            if (filename == NULL)
                filename = argv[i];
            else if (nposfiles < N_POS_FILES) {
                posfilenames[nposfiles] = argv[i];
                nposfiles ++;
            }
        }
    }

    if ((argc < 2) || (filename == NULL)) {
        printf("Usage: %s [-i<x>] [-r] [-f<x>] [-n<x>] [-b<x>] <filename>\n"
               "Options:\n"
               "  -i<x>    load x-th frame from a multi-dimensional stack image file\n"  
               "  -dr      dump unfiltered image\n"
               "  -df      dump filtered image\n"
               "  -s       save object list to file - filename is autogenerated\n"
               "  -sx<x>, -sy<x>    crop image dimensions to <x> pixels\n"
               "  -ox<x>, -oy<x>    crop first <x> pixels of the image\n"
               "  -f<x>    Set instrumental FWHM\n"
               "  -fm      Use Mexican Hat filter\n"
               "  -fh      Use Gaussian highpass\n"
               "  -fa      Use Matched filter\n"
               "  -fb      Use Biparametric filter\n"
               "  -nx<x>, -ny<x>    set image dimensions to x,y\n"
               "  -n       number clusters in output - like \"cat -n\" \n"
               "  -nofilter   do not apply filter\n"
               "  -b<x>    set number of bytes per pixel for raw binary files\n"
               "  -t<x>    set 'Gauss factor' - cluster detection threshold\n"
               "  -v       print summary rather than complete object list\n",
              argv[0]);
        exit(0);
    }

    mImage.setFilename(filename, image_index);
    if (strstr(filename, "TIF") || strstr(filename, "tif") || 
        strstr(filename, "stk") || strstr(filename, "STK")) {
        // TIFF images

        if (mImage.load()) {
            error("Unable to open image %s.\n", filename);
            exit(0);
        }

        if (params.ClusterFwhm == 0.)
            params.ClusterFwhm = mImage.params.ClusterFwhm;
        if (params.ClusterDistance == 0.)
            params.ClusterDistance = mImage.params.ClusterDistance;
        nx = mImage.nx; ny = mImage.ny;
    }
    else {                        // raw binary images
        mImage.LoadBinaryImage(bytes, nx, ny, params.ClusterFwhm);
    }
    bitsperpixel = mImage.properties->BitsPerPixel;

    // validate image dimensions
    if ((offx >= nx) || (offx < 0))
        offx = 0;
    if ((offy >= ny) || (offy < 0))
        offy = 0;
    if ((sizex == 0) || (offx + sizex > nx))
        sizex = nx-offx;
    if ((sizey == 0) || (offy + sizey > ny))
        sizey = ny-offy;

    if ((offx != 0) || (offy != 0) || (sizex != nx) || (sizey != ny))
        printf("Using (%i,%i) subimage (%i,%i) to (%i,%i)\n", sizex, sizey,
               offx, offy, offx+sizex, offy+sizey);
    image = new unsigned short [ sizex *sizey ];
    unsigned short *rimage = mImage.get_image();
    for (i = 0; i < sizex; i++)
        for (j = 0; j < sizey; j++)
            image[i*sizey+j] = rimage[(i+offx)*ny+(j+offy)];
    nox = nx; noy = ny;
    nx = sizex; ny = sizey;

#if defined(DEBUG)
  // printf("%i %i %i %i %f\n", rawflag, nx, ny, image_index, fwhm);
  // printf("%i %i %f\n", dark_count, bitsperpixel, fwhm);
#endif

    // dump raw image
    if (dumprawflag) {
        for (i = 0; i < nx*ny; i++)
        printf("%hu\n", image[i]);
        goto terminate;
    }

    for (i = 0; i < nposfiles; i++) {
        shiftx[i] += offx;
        shifty[i] += offy;
    }

#if defined(PLOT_IMAGE) && defined(HAVE_PGPLOT)
    {
        unsigned int min, max;
        cpgopen("?");
        get_genopsys_autoscale(image, nx, ny, &min, &max); 
        plot_image(image, nx, ny, (float)min, (float)max);
        if (nposfiles > 0) {
            int nread = 0;
            st_fposition *positions;
            for (i = 0; i < nposfiles; i++) {
                read_positions(posfilenames[i], &positions, &nread); 
                scale_shift_fpositions(positions, nread,
                                    shiftx[i], shifty[i], scalex[i], scaley[i],
                                    (float)nox/2, (float)noy/2);
                cpgsci(i+2);
                plot_fpositions(positions, 0, nread, nx, ny);
                printf("No clusters %i in %s.\n", nread, posfilenames[i]);
            }
            delete [] positions;
        }
        cpgclos();
    }
#endif

#if defined(PLOT_IMAGE) && defined(HAVE_PGPLOT) && 0
    {
        float *mask = new float [nx* ny];
        make_gaussian_highpass(mask, nx, ny, (void *)(&(params.ClusterFwhm)));
        printf("%f\n", fwhm);
        cpgopen("?");
        plot_image(mask, nx, ny, 0., 1.);
        cpgclos();
        delete [] mask;
    }
#endif

    imager = new float [nx*ny];
    remove_saturated_pixels(image, nx, ny, bitsperpixel);
    convert_image_ushort2float(image, imager, nx, ny);

    data = new ImageData(imager, nx, ny, bitsperpixel);
    
    switch (summary_flag) {
    case 1:
        printf("Average          %10.1f\n", data->image_stats->Average);
        printf("Standard dev.    %10.1f\n", data->image_stats->StandardDeviation);
        printf("Minimum          %10.1f\n", data->image_stats->Minimum);
        printf("Maximum          %10.1f\n", data->image_stats->Maximum);
        printf("Frac. saturation        %8.6f\n",
               data->image_stats->FractionSaturation);
        break;
    case 2:
        printf("%30s %8.0f %5.0f %6.0f %6.0f %8.6f",
               filename, 
               data->image_stats->Average,
               data->image_stats->StandardDeviation,
               data->image_stats->Minimum, data->image_stats->Maximum,
               data->image_stats->FractionSaturation);
        break;
    }
    
    data->params = params;
    data->analyse();


#if defined(PLOT_IMAGE) && defined(HAVE_PGPLOT)
    {
        float min = -20., max = 100.;
        cpgopen("?");
        get_genopsys_autoscale(data->filter_image, nx, ny, &min, &max);
        plot_image(data->filter_image, nx, ny, min, max);
        if (nposfiles > 0) {
            int nread = 0;
            st_fposition *positions;
            for (i = 0; i < nposfiles; i++) {
                read_positions(posfilenames[i], &positions, &nread); 
                scale_shift_fpositions(positions, nread,
                              shiftx[i], shifty[i], scalex[i], scaley[i],
                              (float)nox/2, (float)noy/2);
                cpgsci(i+2);
                plot_fpositions(positions, 0, nread, nx, ny);
            }
            delete [] positions;
        }
        cpgclos();
    }
#endif

    if (dumpfilterflag) {
        for (i = 0; i < nx*ny; i++)
            printf("%f\n", data->filter_image[i]);
        goto terminate;
    }

#if defined(PLOT_IMAGE) && defined(HAVE_PGPLOT)
    { 
        unsigned int min, max;
        printf("Plotting areas...\n");
        cpgopen("?");
        cpgsci(3);
        get_genopsys_autoscale(image, nx, ny, &min, &max);
        plot_image(image, nx, ny, (float)min, (float)max);
        for (i = 0; i < data->nobj; i++) {
            for (j = 0; j < data->objects[i].na; j++) {
                cpgpt1(data->objects[i].area[j].x-nx/2, 
                       ((ny - data->objects[i].area[j].y-1)-ny/2), -1);
            }
        }
        cpgsci(2);
        for (i = 0; i < data->nobj; i++) {
            for (j = 0; j < data->objects[i].nb; j++) {
                cpgpt1(data->objects[i].boundary[j].x-nx/2, 
                       ((ny - data->objects[i].boundary[j].y-1)-ny/2), -1);
            }
        }
        cpgclos();
    }
#endif

  
#if defined(PLOT_IMAGE) && defined(HAVE_PGPLOT)
    {
        unsigned int min = 450, max = 650;
        cpgopen("?");
        get_genopsys_autoscale(image, nx, ny, &min, &max);
        plot_image(data->image, nx, ny, (float)min, (float)max);
        cpgsci(2);
        plot_pixels(data->positions, 0, data->nobj, nx, ny);
        cpgclos();
    }
    {
        float min = -40., max = 100.;
        cpgopen("?");
        get_genopsys_autoscale(data->filter_image, nx, ny, &min, &max);
        plot_image(data->filter_image, nx, ny, min, max);
        cpgsci(2);
        plot_pixels(data->positions, 0, data->nobj, nx, ny);
        cpgclos();
    }
#endif

    switch (summary_flag) {
    case 1:
        printf("Av threshold       %8.1f\n", data->analysis_stats->average_threshold);
        printf("Av thr signal      %8.1f\n", data->analysis_stats->average_threshold_signal);
        printf("Av thr back        %8.1f\n", data->analysis_stats->average_threshold_background);
        printf("Number obj         %6i\n", data->nobj);
        break;
    case 2:
        printf(" %8.1f %8.1f %8.1f %6i", 
               data->analysis_stats->average_threshold,
               data->analysis_stats->average_threshold_signal,
               data->analysis_stats->average_threshold_background,
               data->nobj);
        break;
    }

        if (object_stats_flag) {
            data->analyseObjectStats();
        if (summary_flag == 1) {
            printf("Cluster size median    %4.1f\n", data->cluster_stats.FwhmMedian);
            printf("Cluster size mean      %4.1f\n", data->cluster_stats.FwhmMean);
            printf("Cluster size sdev      %4.1f\n", data->cluster_stats.FwhmStandardDeviation);
        }
        else if (summary_flag == 2) {
            printf(" %4.1f %4.1f %4.1f",
                   data->cluster_stats.FwhmMedian,
                   data->cluster_stats.FwhmMean,
                   data->cluster_stats.FwhmStandardDeviation);
        }
        else
            print_object_stats(stdout, data->filter_image,
                               data->noise_image, nx, ny,
                               data->positions, data->object_stats,
                               data->nobj);
        }
        else {
        if ((!saveflag) && (!summary_flag)) {
                data->writePositions(stdout, indexflag);
            }
        }

    if (summary_flag == 2)
        printf("\n");

    if (saveflag) {
        char fn[256];
        int i;
        for (i = strlen(filename)-1; i > 0; i--)
            if (filename[i] == '.') {
                filename[i] = '\0';
                break;
            }
        if (image_index > 0)
            snprintf(fn, sizeof(fn), "%s_%i_pos.txt", filename, image_index+1);
        else
            snprintf(fn, sizeof(fn), "%s_pos.txt", filename);
        FILE *fp = fopen(fn, "w");
        if (fp) {
            data->writePositions(fp);
            fclose(fp);
        }
        else
            error("Could not open output file %s.\n", fn);
    }

    delete data;
    delete [] imager;
    delete [] image;

  terminate:
    return 0;
}
