/*
 * Main program for image analysis module in the Solexa data analysis
 * pipeline.
 * Copyright (c) 2005, 2006 Solexa
 *               2007 Illumina
 * Author: Klaus Maisinger
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */

#include <cstdio>
#include <cassert>
#include <string>
#include "Images.h"
#include "TileAnalyser.h"
#include "Process.h"
#include "Tiling.h"
#include "Logger.h"
#include "QCWriter.h"
#include "JournalParser.h"
#include "ImageProcess.h"
#include "IntensityData.h"
#include "ParameterTypes.h"
#include "Message.h"
#include "Maths.h"
#include "XMLSAXParser.h"
#include "XMLImageConfigHandler.h"


int snprintf_offsets(char s[], int len, char text[],
                     std::vector<AlignmentTransform *> transforms)
{
    int i;
    int msg_idx = 0;
    msg_idx = sprintf(s, "%-25s", text); 
    for (i = 0; i < int(transforms.size()); i++)
        msg_idx += snprintf(s+msg_idx, len-msg_idx, "%6.2f %6.2f  ", 
                           transforms[i]->XOffset,
                           transforms[i]->YOffset);
    return msg_idx;
}

void run_analysis(ImageList         ** image_list,
                  unsigned char      * base_list,
                  int                  firstcycle,
                  int                  ncycles,
                  TileInfo           & tileInfo,
                  RunInfo            & runInfo,
                  AnalysisParameters & params,
                  bool                 useRemappedIntensityFlag,
                  bool                 useInterpolatedPositionFlag,
                  const char           path_qc_output[],
                  const char           path_image_analysis[],
                  const char           sample[],
                  std::string          logfile,
                  const char           offset_filename[],
                  const std::string  & loadsuffix,
                  const std::string  & savesuffix,
                  IntensityData::CompressionType compression,
                  int                * is_deblocked,
                  std::vector<int>   & matrix_cycles)
{
    assert(ncycles > 0);
    assert(runInfo.NumChannels > 0);
    assert(runInfo.NumChannels <= 4);
    assert((params.GaussFactor >= 3.));
    assert((params.ClusterFwhm     > 1.) && (params.ClusterFwhm < 10.));
    assert((params.ClusterDistance > 1.) && (params.ClusterDistance < 10.));
    assert((params.PearsonThreshold > 0.) && (params.PearsonThreshold < 1.));

    /* set up logging object */
    Logger *logger = new Logger(logfile.c_str());
    logger->add_handler(new LogHandler());
    char msg[256];
    sprintf(msg, "%s%s", software, software_version);
    logger->set_source(msg); 
    logger->info(path_image_analysis);
    logger->info(logfile.c_str());

    int i, j, k;
    i = 0;
    while ((i < runInfo.NumChannels) && (image_list[i]->load()))
        i++; // failed to load image
    if (i == runInfo.NumChannels) {
        logger->error("Failed to load any first-cycle image; exiting...\n");
        delete logger;
        error("Failed to load any first-cycle image; exiting...\n");
        exit(1);
    }

    runInfo.NumTiles    = image_list[i]->size();
    runInfo.ImageWidth  = image_list[i]->at(0)->nx;
    runInfo.ImageHeight = image_list[i]->at(0)->ny;

    sprintf(msg, "Lane %i TileIndex %i Tile %s_%i_%04i " 
            "Cycles %i Channels %i Tiles %i",
            tileInfo.Lane+1, tileInfo.TileIndex+1, sample,
            tileInfo.Lane+1, tileInfo.TileIndex+1,
            ncycles, runInfo.NumChannels, runInfo.NumTiles);
    logger->info(msg);
    if (compression == IntensityData::compressionGzip)
        logger->info("Compressing output files with gzip.");
    if (compression == IntensityData::compressionBzip2)
        logger->info("Compressing output files with bzip2.");

    if (params.DetectionType == 1)
    {
        sprintf(msg, "Tile  %i %04i: nx ny minmatch fwhm dist %i %i %4.2f %5.1f %5.1f",
                tileInfo.Lane+1, tileInfo.TileIndex+1,
                runInfo.ImageWidth, runInfo.ImageHeight,
                params.PearsonThreshold, params.ClusterFwhm,
                params.ClusterDistance);
    }
    else
    {
        sprintf(msg, "Tile  %i %04i: nx ny thresh fwhm dist type %i %i %5.1f %5.1f %5.1f %i",
                tileInfo.Lane+1, tileInfo.TileIndex+1,
                runInfo.ImageWidth, runInfo.ImageHeight,
                params.GaussFactor, params.ClusterFwhm,
                params.ClusterDistance, params.FilterType);
    }

    logger->info(msg);

    // initialise objects for creation of filenames
    std::string path = std::string(path_qc_output) + std::string(sample);
    RunFolderFileName *qc_name = new RunFolderFileName(path.c_str());
    qc_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    // initialise objects for creation of filenames
    path = std::string(path_image_analysis) + std::string(sample);
    RunFolderFileName *intensity_name = new RunFolderFileName(path.c_str());
    intensity_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    TileAnalyser *analyser = new TileAnalyser(runInfo);
    analyser->SetParameters(params);
    Alignment *default_transforms = new Alignment [runInfo.NumChannels];
    if (loadAlignmentTransforms(offset_filename, default_transforms,
                                runInfo.NumChannels) < 0) {
        sprintf(msg, "Could not open offset file %s.", offset_filename);
        memset(default_transforms, 0, sizeof(Alignment)*runInfo.NumChannels);
    }
    else {
        sprintf(msg, "Successfully read offset file %s", offset_filename);
    }
    logger->info(msg);
    analyser->SetAlignment(default_transforms);
    delete [] default_transforms;

    int cycle;
    std::vector<AlignmentTransform * > transforms;
    std::vector<float *> intensities;
    std::vector<float *> noisevalues;
    st_fposition *referencepositions = 0;
    int nobj = 0;
    k = 0;

    // Read in existing data files. This is only reqired for extensions of 
    // previous analyses.
    if (loadsuffix.size() > 0) {
        int nobj2, ncol;
        std::string filename = qc_name->getIndexFilename(RunFolderFileName::POSITION_FILE, k);
        sprintf(msg, "Reading %s", filename.c_str());
        logger->info(msg);
        int status = read_positions(filename.c_str(),
                                    &referencepositions, &nobj);
        if (status) 
            error("Failed to read position information from %s.",
                  filename.c_str());

        filename = intensity_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE, k) 
                   + loadsuffix;
        IntensityData intfile(filename.c_str(), compression);
        status = intfile.read(intensities, &nobj2, &ncol);
        if (status == 0) {
            if (nobj != nobj2) {
                error("Number of objects in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
            if ((ncol != runInfo.NumChannels) && (ncol != 0)) {
                error("Number of channels in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
        } 
        filename = intensity_name->getIndexFilename(RunFolderFileName::NOISE_FILE, k) 
                   + loadsuffix;
        IntensityData nsefile(filename.c_str(), compression);
        status = nsefile.read(noisevalues, &nobj2, &ncol);
        if (status == 0) {
            if (nobj != nobj2) {
                error("Number of objects in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
            if ((ncol != runInfo.NumChannels) && (ncol != 0)) {
                error("Number of channels in file %s does not match.",
                      filename.c_str());
                exit(1);
            }
        } 
        firstcycle  = intensities.size();
        sprintf(msg, "Starting from cycle %i\n", firstcycle+1); 
    }

    // Now analyse new cycles.
    for (cycle = firstcycle; cycle < ncycles; cycle++) {
        i = cycle;
        // load image files 
        for (j = 0; j < runInfo.NumChannels; j++) {
            if (image_list[i*runInfo.NumChannels+j]->load()) {
                warning("Image load failed. Using empty image instead.");
                image_list[i*runInfo.NumChannels+j]->createEmptyImages(
                      runInfo.ImageWidth, runInfo.ImageHeight,
                      runInfo.NumTiles);
            }
            assert((int)image_list[i*runInfo.NumChannels+j]->size() 
                   == runInfo.NumTiles);
        }
        for (k = 0; k < runInfo.NumTiles; k++) {
            tileInfo.Cycle       = cycle;
            tileInfo.IsDeblocked = is_deblocked[cycle];
            analyser->SetTile(tileInfo);
            if (cycle > 0)
                analyser->SetReference((Position *)referencepositions, nobj);
            for (j = 0; j < runInfo.NumChannels; j++) {
                // single-image analysis
                Image *image = image_list[i*runInfo.NumChannels+j]->at(k);
                // image alignment
                image->useInterpolatedPositionFlag = useInterpolatedPositionFlag;
                analyser->GetAnalyser()->setUseRemappedIntensityFlag(useRemappedIntensityFlag);
                analyser->AnalyseImage(image, j,
                                       (char)base_list[j]);
                ImageData *data = image->data;
                sprintf(msg, "Image %i %04d %3i %i: %c #obj %5d offset %6.2f %6.2f %7.4f %7.4f exp %3.1f %3.1f",
                        tileInfo.Lane+1, tileInfo.TileIndex+k+1, cycle+1, j+1,
                        (char)(base_list[j]),
                        image->get_objects()->getSize(), 
                        data->transform->XOffset, data->transform->YOffset,
                        data->transform->XScale,  data->transform->YScale,
                        image->properties->ExposureTime,
                        data->estimated_fwhm);
                logger->info(msg);
            }

            transforms.clear();
            for (j = 0; j < runInfo.NumChannels; j++)
                transforms.push_back(image_list[i*runInfo.NumChannels+j]->at(k)
                                                       ->getData()->transform);
            char s[125];
            snprintf(s, sizeof(s),
                    "Tile  %i %04i %3i: Original offsets:     ",
                    tileInfo.Lane+1, tileInfo.TileIndex+1, cycle + 1);
            snprintf_offsets(msg, sizeof(msg), s, transforms);
            logger->info(msg);

            /*
             * In first cycle, generate the reference positions for all
             * later cycles and write them to a text file.
             */
            if (cycle == 0) {
                Position *newReference;
                analyser->BuildReference(&newReference, &nobj);
                referencepositions = new st_fposition [ nobj ];
                memcpy(referencepositions, newReference, nobj * sizeof(Position));
                analyser->SavePositions(tileInfo, path_qc_output,
                                        nobj, newReference);
                analyser->SaveIndexPositions(tileInfo, path_qc_output,
                                             nobj, newReference);
            }

            float *signal = new float [ runInfo.NumChannels*nobj ];
            float *noise  = new float [ runInfo.NumChannels*nobj ];
            Alignment *alignments  = new Alignment    [ runInfo.NumChannels ];
            TileQCValues *qcValues = new TileQCValues [ runInfo.NumChannels ];
            analyser->CompleteTile(signal, noise,
                                   (Position *)referencepositions, nobj, nobj,
                                   alignments, qcValues, 0);
            int valid_flag = qcValues[0].OffsetsValidated;
            delete [] alignments;
            delete [] qcValues;
            intensities.push_back(signal);
            noisevalues.push_back(noise);

            snprintf(s, sizeof(s),
                    "Tile  %i %04i %3i: Validated offsets (%i):",
                    tileInfo.Lane+1, tileInfo.TileIndex+1, cycle + 1,
                    valid_flag);
            snprintf_offsets(msg, sizeof(msg), s, transforms);
            logger->info(msg);

            /* write output */
            if (cycle == 0) {
                analyser->SaveClusterDetails(tileInfo, path_qc_output,
                                             nobj);
            }
            // save individual intensity/noise files per cycle
            // only if more than one tile given (stack file format)
            // or current cycle is specified in <matrix_cycles>
            int save_flag = (runInfo.NumTiles > 1);
            for (j = 0; j < (int)matrix_cycles.size(); j++)
                if (cycle == matrix_cycles[j])
                    save_flag = 1;
            if (save_flag) {
                IntensityData intfile(qc_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE, 
                                                             k, cycle).c_str());
                intfile.write(signal, nobj, runInfo.NumChannels);
                IntensityData nsefile(qc_name->getIndexFilename(RunFolderFileName::NOISE_FILE, 
                                                             k, cycle).c_str());
                nsefile.write(noise, nobj, runInfo.NumChannels);
            }
            analyser->SaveQCFile(tileInfo, path_qc_output);

            /* clear images to reduce memory footprint */            
            for (j = 0; j < runInfo.NumChannels; j++)
                image_list[i*runInfo.NumChannels+j]->at(k)->clear();
        }
    }
    logger->info("Writing output files...");
    if (runInfo.NumTiles == 1) {
        IntensityData nsefile(intensity_name->getIndexFilename(RunFolderFileName::NOISE_FILE,
                                                               0).c_str(),
                             compression);
        nsefile.write(noisevalues, nobj, runInfo.NumChannels,
                      intensity_name->getLane(0), intensity_name->getTile(0),
                      (Position *)referencepositions);
        std::string filename = intensity_name->getIndexFilename(RunFolderFileName::INTENSITY_FILE,
                                                                0) 
                               + savesuffix;
        IntensityData intfile(filename.c_str(), compression);
        intfile.write(intensities, nobj, runInfo.NumChannels,
                      intensity_name->getLane(0), intensity_name->getTile(0),
                      (Position *)referencepositions);
    }
    for (cycle = 0; cycle < ncycles; cycle++) {
        delete [] intensities[cycle];
        delete [] noisevalues[cycle];
    }
    delete [] referencepositions;
    // write out combined intensity and signal files
    analyser->GetAnalyser()->displayOffsets();

    logger->info("Run terminated successfully.");
    delete analyser;
    delete logger;
    delete qc_name;
    delete intensity_name;
}

int read_xml_parameters(const char filename[],
                        struct AnalysisParameters &params)
{
    FILE *fh = fopen(filename, "r");
    if (!fh) {
        error("Failed to open parameter file %s.", filename);
        return 0;
    }

    XMLImageConfigHandler *handler    = new XMLImageConfigHandler();
    XMLSAXParser          *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);
    handler->getParameters(params);
    delete handler;
    delete sax_parser;
    return 0;
}

int main(int argc, char *argv[]) 
{
    std::string offset_filename    = "default_offsets.txt";
    std::string parameter_filename = "";
    std::string tmp_prefix         = "tempseq";
    std::string tmp_prefix2        = "tempseq";
    std::string journal_filename   = "";
    char        tmp_filename[FILENAME_MAX] = "";
    char  sample[FILENAME_MAX]          = "s";
    IntensityData::CompressionType compression = IntensityData::compressionNone;
    int   ncycles     = 0;
    int   firstcycle  = 0;
    int   i, idummy;
    ImageList     **image_list;
    std::vector<std::string > filenames;
    unsigned char  *base_list;

    char *deblockedstring     = NULL;
    std::string logfile       = "";
    std::string loadsuffix    = "";
    std::string savesuffix    = "";
    std::vector<int> matrix_cycles;

    TileInfo tileInfo;
    tileInfo.Lane             = 0;
    tileInfo.TileIndex         = 0;

    RunInfo runInfo;
    runInfo.NumChannels       = 4;
    runInfo.NumTiles          = 1;
    runInfo.ImageWidth        = DEFAULT_PIXEL_DIM;
    runInfo.ImageHeight       = DEFAULT_PIXEL_DIM;

    AnalysisParameters params;
    AnalysisParameters cmdline_params(-1.0f);
    AnalysisParameters config_params(-1.0f);
    bool useInterpolatedPositionFlag = false;
    bool useRemappedIntensityFlag    = true;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-')
            switch (argv[i][1]) {
            case 'c':
                {
                    char comp[20];
                    sscanf(argv[i], "-compression=%s", comp);
                    if (!strcmp(comp, "gzip"))
                        compression = IntensityData::compressionGzip;
                    if (!strcmp(comp, "bzip2"))
                        compression = IntensityData::compressionBzip2;
                } break;
            case 'f': 
                switch (argv[i][2]) {
                case 'w': sscanf(argv[i], "-fwhm=%f", &(cmdline_params.ClusterFwhm)); 
                          break;
                case 'i': sscanf(argv[i], "-filter=%d", &(cmdline_params.FilterType)); 
                          break;
                case 'l': {
                              int flag = 0;
                              sscanf(argv[i], "-flag=%d", &flag);
                              switch (flag) {
                              case 1: useRemappedIntensityFlag = false; break;
                              case 2: useInterpolatedPositionFlag = true; break;
                              case 3: useRemappedIntensityFlag = false;
                                      useInterpolatedPositionFlag = true; break;
                              }
                          } break;
                } break;
            case 't': 
                switch (argv[i][2]) {
                case 'h': sscanf(argv[i], "-threshold=%f", &(cmdline_params.GaussFactor)); 
                          break;
                case 'i': 
                    if (sscanf(argv[i], "-tile=%i,%i", 
                               &(tileInfo.Lane), &(tileInfo.TileIndex)) != 2)
                        if (sscanf(argv[i], "-tile=%[^_]_%d_%d",
                                   sample, &(tileInfo.Lane), 
                                   &(tileInfo.TileIndex)) == 3) {
                            tileInfo.Lane --; tileInfo.TileIndex--;
                        }
                    break;
                case 'y': sscanf(argv[i], "-type=%d", &(params.DetectionType)); 
                    break;
                } break;
            case 'm': 
                switch (argv[i][2]) {
                case 'a':
                    switch (argv[i][3]) {
                    case 'x': sscanf(argv[i], "-maxshift=%f", 
                                     &params.MaxStageOffset); break;
                    case 't': sscanf(argv[i], "-matrix=%i", &idummy); 
                        matrix_cycles.push_back(idummy-1); break;
                    } break;
                case 'i': sscanf(argv[i], "-minmatch=%f", &(cmdline_params.PearsonThreshold));
                    break;
                }
                break;
            case 'o': if (sscanf(argv[i], "-offsets=%s", tmp_filename))
                          offset_filename = tmp_filename;
                      break;
            case 'd': 
                switch (argv[i][2]) {
                case 'i': sscanf(argv[i], "-dist=%f", &(cmdline_params.ClusterDistance)); break;
                case 'e': deblockedstring = new char [ strlen(argv[i]) + 1];
                          sscanf(argv[i], "-deblocked=%s", deblockedstring);
                          break;
                } break;
            case 'p': 
                if (sscanf(argv[i], "-prefix=%s", tmp_filename) == 1) {
                    tmp_prefix = tmp_filename;
                    if (tmp_prefix2 == "tempseq")
                        tmp_prefix2 = tmp_filename;
                }
                if (sscanf(argv[i], "-prefix2=%s", tmp_filename))
                    tmp_prefix2 = tmp_filename;

                if (sscanf(argv[i], "-params=%s", tmp_filename)) {
                     parameter_filename = tmp_filename;
                     read_xml_parameters(parameter_filename.c_str(), config_params);
                }
                break;
            case 's': 
                switch (argv[i][2]) {
                case 'a': 
                    switch(argv[i][3]) {
                    case 'm': sscanf(argv[i], "-sample=%s", sample); break;
                    case 'v':
                        if (sscanf(argv[i], "-savesuffix=%s", tmp_filename))
                            savesuffix = tmp_filename;
                        break;
                    } break;
                } break;
            case 'n':
                switch (argv[i][2]) {
                case 'x': sscanf(argv[i], "-nx=%i", &(runInfo.ImageWidth));
                    break;
                case 'y': sscanf(argv[i], "-ny=%i", &(runInfo.ImageHeight));
                    break;
                case 'c': sscanf(argv[i], "-nc=%i", &firstcycle); break;
                } break;
            case 'l': 
                switch (argv[i][2]) {
                case 'o':
                    switch (argv[i][3]) {
                    case 'g':
                        if (sscanf(argv[i], "-logfile=%s", tmp_filename))
                            logfile = tmp_filename;
                        break;
                    case 'a':
                        if (sscanf(argv[i], "-loadsuffix=%s", tmp_filename))
                            loadsuffix = tmp_filename;
                        break;
                    } break;
                } break;
            }   
        else {
            journal_filename = argv[i];
        }
    }

    cmdline_params.CopyIfPositive(config_params);
    params.CopyIfPositive(cmdline_params);

    if ((argc < 2) || (journal_filename == "")) {
        printf("Usage: run_image_analysis [options] <journal filename>\n"
              "Options:\n"
           "  -type=<n>                    set detection algorithm type 0,1\n"
           "  -filter=<n>                  set filter type 0,1,2,3\n"
           "  -fwhm=<x>                    set instrumental FWHM\n"
           "  -threshold=<x>               set 'Gauss factor'\n"
           "  -minmatch=<x>                set 'Pearson threshold'\n"
           "  -maxshift=<n>                set maximum shift for alignment (in pixels)\n"
           "  -dist=<x>                    set distance parameter for colocalisation\n"
           "  -params=<filename>           set the path to an XML parameters file\n"	
           "  -prefix=<filename>           set prefix for internal filenames\n"
           "  -logfile=<filename>          set the path to a log file\n"
           "  -sample=<text>               set sample id (must not contain underscores)\n"
           "  -tile=<lane>,<tile>          set lane/tile (0 indexed)\n"
           "  -tile=<sample>_<lane>_<tile> set sample/lane/tile (1 indexed)\n");
        exit(0);
    }

    // parse journal file (a description of which images to load)
    if (parse_journal(journal_filename.c_str(), &runInfo, &ncycles, filenames,
                      &base_list) < 0) {
        error("Cannot open journal file %s.", journal_filename.c_str());
        exit(1);
    }
    image_list = new ImageList * [ static_cast<int>(filenames.size()) ];
    for (i = 0; i < static_cast<int>(filenames.size()); i++) {
        image_list[i] = new ImageList;
        image_list[i]->addFile(filenames[i].c_str());
    }

    ncycles += firstcycle;

    // parse deblock string
    int *is_deblocked = new int [ncycles];
    for (i = 0; i < ncycles; i++)
        is_deblocked[i] = 0;
    if (deblockedstring) {
        for (i = 0; i < Maths::min(ncycles, (int)strlen(deblockedstring)); i++)
            if (toupper(deblockedstring[i]) == 'D')
                is_deblocked[i] = 1;
        delete [] deblockedstring;
    }

    // create name for log file
    if (logfile.length() == 0)
        logfile = std::string(tmp_prefix) + std::string(software)
                + std::string(".log");

    // run the image analysis
    run_analysis(image_list, base_list, firstcycle, ncycles,
                 tileInfo,  runInfo, params,
                 useRemappedIntensityFlag, useInterpolatedPositionFlag,
                 tmp_prefix.c_str(), tmp_prefix2.c_str(), sample, logfile,
                 offset_filename.c_str(), loadsuffix, savesuffix, compression,
                 is_deblocked, matrix_cycles);

    delete [] base_list;
    for (i = 0; i < runInfo.NumChannels * (ncycles-firstcycle); i++)
        delete image_list[i];
    delete [] image_list;
    delete [] is_deblocked;
    return 0;
}
