#!/usr/bin/env perl

# PROJECT: GERALD
# MODULE:  GERALD.pl
# AUTHOR:  A. J. Cox
#
# Copyright (c) 2005-2006 Solexa
#               2007      Illumina
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# GERALD.pl ... stands for
# Generation of Recursive Analyses Linked by Dependency

# Description: 
# Generate a makefile to auto-generate files of sequence analysis data
# The makefile contains details of dependencies between the various
# analyses

# Usage: GERALD.pl [--EXPT_DIR value] [--FORCE] configfile

use warnings;
use strict;
use File::Copy;
use File::Spec;
use File::Basename;
use XML::Simple;
use Data::Dumper;
use Getopt::Long qw(:config no_ignore_case);
use FindBin;
use lib "$FindBin::Bin/../";
use Gerald::Common;

my $numLanes           = 8;
my $analysisKey        = 'ANALYSIS';
my $useBasesKey        = 'USE_BASES';
my $indexBasesKey      = 'INDEX_BASES';
my $readLengthKey      = 'READ_LENGTH';
my $origReadLengthsKey = 'ORIG_READ_LENGTHS';
my $qcalSourceKey      = 'QCAL_SOURCE';
my $qtableRequiredKey  = 'QTABLE_REQUIRED';
my $qtablePathKey      = 'QTABLE_PATH';
my $srfArchiveRequired = 'SRF_ARCHIVE_REQUIRED';
my $srfQcalKey         = 'SRF_QCAL';
my $srf2ndCycleKey     = 'SRF_2ND_CYCLE';
my $exptDirKey         = 'EXPT_DIR';
my $exptDirRootKey     = 'EXPT_DIR_ROOT';
my $outDirKey          = 'OUT_DIR';
my $dataDirKey         = 'Data';
my $notDef             = "nOtDeFiNeD";

my $qtable_suffix      = '$(QTABLE_SUFFIX)';

#------------------------------------------------------------------------------

sub info($) {
    my $msg = shift;
    print STDERR $msg;
}


sub warning($) {
    my $msg = shift;
    print STDERR "Warning: ", $msg;
}


sub error($) {
    my $msg = shift;
    die "Error: ", $msg;
}


# auto-generate a unique folder name for a GERALD run
sub generateGeraldFolderName($$$) {
    my $exptdir    = shift;
    my $folderRoot = shift;
    my $logName    = shift;

    my ($d, $m, $y)=(localtime)[3,4,5];
    my $outRoot=sprintf("%02d-%02d-%04d_%s", 
			$d, $m+1, $y+1900, $logName);
    my $try=1;
    my $outName=$outRoot;
    while (-e File::Spec->catdir($exptdir, $folderRoot."_".$outName))
    {
	$try++;
	$outName="$outRoot.$try";
    } #while

    return File::Spec->catfile($exptdir, $folderRoot."_".$outName)
}


sub getLogName() {
    my $logName=$ENV{"LOGNAME"};
    $logName="unknown" unless defined($logName);
    return $logName;
}


# sort out EXPT_DIR, EXPT_DIR_ROOT and OUT_DIR
sub updateFolderNames($$$) {
    my $paramsRef  = shift;
    my $folderRoot = shift;
    my $logName    = shift;

    if ($paramsRef->{$exptDirKey} eq $notDef) {
	warning("Did not specify any value for ${exptDirKey}\n");
	warning("Assuming current working directory.\n");
	$paramsRef->{$exptDirKey} = '.';
    }

    # convert experiment directory to absolute path if necessary
    my $abspath = File::Spec->rel2abs($paramsRef->{$exptDirKey});

    # if path exists, separate out EXPT_DIR, EXPT_DIR_ROOT and OUT_DIR
    if (-d $abspath) {
	my @dirs = File::Spec->splitdir($abspath);
	my $idx;
	for ($idx = $#dirs; $idx > 0; $idx--) {
	    if ($dirs[$idx] eq $dataDirKey) {
		last;
	    }
	}
	if ($dirs[$#dirs] =~/${folderRoot}/) {
	    $paramsRef->{$exptDirKey} 
	         = File::Spec->catdir(@dirs[($idx-1) .. ($#dirs-1)]);
	    $paramsRef->{$outDirKey} 
	         = File::Spec->catdir($exptDirKey, $dirs[$#dirs]);
	}
	else {
	    $paramsRef->{$exptDirKey}
	         = File::Spec->catdir(@dirs[($idx-1) .. $#dirs]);
	}
	$paramsRef->{$exptDirRootKey} = "";
	if ($idx > 1) {
	    $paramsRef->{$exptDirRootKey} 
                 = File::Spec->catdir(@dirs[0 .. ($idx-2)]);
	}
    }

    # check if EXPT_DIR exists and is readable.
    # otherwise try to construct valid EXPT_DIR from EXPT_DIR_ROOT
    my $exptDirRootSearch=$paramsRef->{$exptDirRootKey}."/";
    $exptDirRootSearch=~s/\//\\\//g;

    if ((-d $paramsRef->{$exptDirKey})
	&& ($paramsRef->{$exptDirKey}=~/^$exptDirRootSearch/))
    { 
	error( "${exptDirKey} ".$paramsRef->{$exptDirKey}." present but not readable")
	    unless (-r $paramsRef->{$exptDirKey});
    } #if
    else
    {
	if (! File::Spec->file_name_is_absolute($paramsRef->{$exptDirKey})) {
	    $paramsRef->{$exptDirKey} 
	         = File::Spec->catdir($paramsRef->{$exptDirRootKey},
				      $paramsRef->{$exptDirKey});
	}
    }
    info("Will read sequence files from ".$paramsRef->{$exptDirKey}."\n");

    # update OUT_DIR if necessary
    if ($paramsRef->{$outDirKey} =~ /^$exptDirKey/)
    {
	if ($paramsRef->{$outDirKey} eq $exptDirKey)
	{
	    $paramsRef->{$outDirKey} 
	        = generateGeraldFolderName($paramsRef->{$exptDirKey},
					   $folderRoot, $logName);
	    info("Output folder ".$paramsRef->{$outDirKey}."\n");
	} #if
	else {
	    # replace string $exptDirKey with real path
	    my @s=File::Spec->splitdir($paramsRef->{$outDirKey});
	    $paramsRef->{$outDirKey} = File::Spec->catdir($paramsRef->{$exptDirKey},
						     $s[1]);

	    info("Will attempt to use directory ".$paramsRef->{$outDirKey}."\n");
	}
    } #if ($paramsRef->{"OUT_DIR"} =~ /^$exptDirKey/)
}


# read GERALD parameters from a key/value text config file
sub parseConfigTextFile($$$$) {
    my $content        = shift;
    my $paramsRef      = shift;
    my $multiParamsRef = shift;
    my $multiLaneRef   = shift;

    my $lastKey = "";

    foreach (split('\n', $content))
    {
	chomp;
	if ((/^\#/)||(/^\s*$/))
	{
	    info("$_\n") unless (/^\s*$/);
	    next;
	} #if
	my @s=split;
	$lastKey=shift(@s);
	if (defined($paramsRef->{$lastKey}))
	{
	    $paramsRef->{$lastKey}="@s";
	} # if
	else
	{
	    my @ss=split(':',$lastKey);

	    if ((scalar(@ss)!=2)
		||(!defined($paramsRef->{$ss[1]}))
		||($ss[0]!~/^[1-8]+$/))
	    {
		error( "Invalid key $lastKey specified in file\n");
	    } # if
	    $lastKey=$ss[1];
	    @ss=split(//,$ss[0]);
	    $multiLaneRef->{$lastKey}="" unless (defined($multiLaneRef->{$lastKey}));
	    foreach my $lane (@ss)
	    {
		error("Parameter $lastKey already defined for lane $lane\n")
		    if ($multiLaneRef->{$lastKey}=~/$lane/);
		$multiLaneRef->{$lastKey}.=$lane;
		$multiParamsRef->{$lastKey}{getLaneKey($lane)}="@s";
	    } # foreach
	} # else
	
    } #while
}


# read GERALD parameters from an XML config file
sub readConfigXMLFile($$$$$) {
    my $filename       = shift;
    my $xml            = shift;
    my $paramsRef      = shift;
    my $multiParamsRef = shift;
    my $multiLaneRef   = shift;

    my $exptHash=$xml->XMLin($filename);
 
    #print Dumper($test),"\n";

    %{$paramsRef}     =%{$exptHash->{ChipWideRunParameters}};
    %{$multiParamsRef}=%{$exptHash->{LaneSpecificRunParameters}};

    # next bit sets up multiParams so knows to expect per-lane definitions
    for my $param (keys %{$multiParamsRef})
    {
	for my $paramLane (keys %{$multiParamsRef->{$param}})
	{
	    $paramLane=~s/^s_//;
	    $multiLaneRef->{$param}.=$paramLane;
	} # for multiParam
    } # for param
} # sub readConfigXMLFile


# obsolete method to deduce read length from the Firecrest folder name
# This does not know about read pairs or missing cycles.
sub getReadLengthFromFolderName($) {
    my $exptdir = shift;

    my @readlengths = ();
    my $ss='^.*C([0-9]+)\-([0-9]+).*$';
    if ($exptdir =~ /$ss/) {
	push (@readlengths, $2 - $1 + 1);
    }
    return @readlengths;
}

# get lengths for all full reads. Use the parameter XML file if
# present, otherwise try to parse Firecrest folder name.
sub getReadLengths2($) {
    my $exptdir = shift;

    my @readlengths;
    my @dirs         = File::Spec->splitdir($exptdir);
    my $bustarddir   = $dirs[$#dirs];
    my $firecrestdir = File::Spec->catdir(@dirs[0 .. ($#dirs - 1)]); 
    my $paramsfile   = File::Spec->catfile($firecrestdir, ".params");
    if (! Gerald::Common::getReadLengths($paramsfile, $bustarddir, \@readlengths)) {
	warning ("Failed to read ${paramsfile}.\n");
	@readlengths = getReadLengthFromFolderName($exptdir);
    }
    info("Found ".scalar(@readlengths)." reads.\n");
    for (my $i = 0; $i < scalar(@readlengths); $i++) {
	info ("Read ".($i+1).": length ".$readlengths[$i].".\n");
    }
    return @readlengths;
}


# count the number of occurences of a character in a string
sub countChars($$) {
    my $s        = shift;
    my $chTarget = shift;
    my $count = 0;
    foreach (split ('', $s)) {
	if ($_ eq $chTarget) {
	    $count += 1;
	}
    }
    return $count;
}


# do the full expansion of the USE_BASES string, including
# wild cards and other short cuts, but excluding multiple reads.
sub expandUseBases($$$)
{
    my $useBases  = shift;
    my $length    = shift;
    my $maxLength = shift;

    if ($useBases eq 'all') {
	$useBases = 'Y*';
    }
    if ($useBases eq 'odd') {
	$useBases = 'YN' x ($maxLength/2);
	if (defined($length)) {
	    $useBases = 'YN' x ($length/2);
	}
    }
    if ($useBases =~ /[^IiYyNn\*\d]/) {
	error( "String ${useBases} contains invalid characters.\n");
    }
    $useBases = expandUseBasesString($useBases);
    my @chars = split('', $useBases);
    my ($numY, $numYast, $astYChar) = (0, 0, 'Y');
    my ($numN, $numNast, $astNChar) = (0, 0, 'n');
    my ($numI, $numIast, $astIChar) = (0, 0, 'I');
    my $lastChar = '';

    for (my $i = 0; $i < scalar(@chars); $i++) {
	my $ch = $chars[$i];
	# count number of Ys, Is, and Ns
	if ($ch =~ /[Yy]/) {
	    $numY += 1;
	}
	elsif ($ch =~ /[Ii]/) {
	    $numI += 1;
	}
	elsif ($ch =~ /[nN]/)  {
	    $numN += 1;
	}
	# count number of expansions (asterisks)
	elsif ($ch eq '*') {
	    if ($lastChar =~ /[Yy]/) {
		$numYast += 1;
		$astYChar = $lastChar;
	    }
	    elsif ($lastChar =~ /[Ii]/) {
		$numIast += 1;
		$astIChar = $lastChar;
	    }
	    else {
		$numNast += 1;
		$astNChar = $lastChar;
	    }
	}
	$lastChar = $ch;
    }
    if (($numYast > 1) || ($numNast+$numIast > 1)) {
	error( "Too many expansions specified in ${useBases}.\n");
    }
    if ((defined($length) && ($numY-$numYast > $length))
	|| ($numY-$numYast > $maxLength)
	|| ($numN-$numNast+$numI-$numIast > $maxLength - ($numY-$numYast))) {
	error( "Too many bases specified in ${useBases}\n");
    }

    # set read length for Y expansion
    my $newNumY = $length;
    if (!defined($newNumY)) { # deduce read length from total number of bases
	$newNumY = $numY;
	if ($numYast > 0) {
	    $newNumY = $maxLength-($numI-$numIast)-($numN-$numNast);
	}
    }

    # expand Y*
    if ($numYast > 0) {
	my $newChar = $astYChar x ($newNumY-($numY-$numYast));
	$useBases =~s/${astYChar}\*/${newChar}/;
    }

    # expand I*
    my $newNumI = $numI;
    if ($numIast > 0) {
	$newNumI = $maxLength-$newNumY-($numN-$numNast); 
	my $newChar = $astIChar x ($newNumI-($numI-$numIast));
	$useBases =~s/${astIChar}\*/${newChar}/;
    }

    # expand N*
    my $newNumN = $numN;
    if ($numNast > 0) {
	$newNumN = $maxLength - $newNumY - $newNumI;
	my $newChar = $astNChar x ($newNumN-($numN-$numNast));
	$useBases =~s/${astNChar}\*/${newChar}/;
    }

    if (length($useBases) > $maxLength) {
	error( "Too many bases specified after expansion to ${useBases}.\n"
	       ."Maximum length is ${maxLength}.\n");
    }
    $useBases .= 'n' x ($maxLength - length($useBases));
    return $useBases;
}

# full expansion of USE_BASES string, including multiple read expansion
sub expandFullUseBases($$$$$$) {
    foreach (@_) {
	if ($_ eq $notDef) {
	    $_ = undef;
	}
    }
    my $fullReadLengthsRef = shift;
    my $analysis           = shift;
    my $useBases           = shift;
    my $readLength         = shift;
    my $readLength1        = shift;
    my $readLength2        = shift;
    my @fullReadLengths    = @{$fullReadLengthsRef};

    my @parts = split(",", $useBases, -1);
    if (scalar(@parts) > scalar(@fullReadLengths)) {
	error("Too many reads specified in ${useBasesKey}=${useBases}.\n");
    }

    my $fullReadLength = 0;
    foreach (@fullReadLengths) {
	$fullReadLength += $_;
    }
    my $numUsedReads = 0;
    foreach (@parts) {
	if ($_ =~ /([Yy]|all|odd|even)/) {
	    $numUsedReads += 1;
	}
    }

    if ((lc $analysis eq 'eland_pair')
	||(lc $analysis eq 'sequence_pair')) {
	# if only one read specified, duplicate USE_BASES for both reads
	if ($numUsedReads eq 1) {
	    push (@parts, $parts[0]);
	    $numUsedReads += 1;
	}
	if ($numUsedReads != 2) {
	    error("Wrong number of reads specified for ${analysisKey} ${analysis}: ${useBasesKey}=$useBases.\n");
	}
	my $idxUsedReads = 0;
	for (my $i = 0;  $i < scalar(@parts); $i++) {
	    if ($parts[$i] =~ /([Yy]|all|odd|even)/) {
		if ($idxUsedReads == 0) {# first read
		    $parts[$i] = expandUseBases($parts[$i], $readLength1, 
						$fullReadLengths[$i]);
		    $parts[$i] =~ s/y/Y/g;
		}
		else { # second read
		    $parts[$i] = expandUseBases($parts[$i], $readLength2, 
						$fullReadLengths[$i]);
		    $parts[$i] =~ s/Y/y/g;
		}
		$idxUsedReads += 1;
	    }
	    else { # index or non-read
		$parts[$i] = expandUseBases($parts[$i], 0, 
					    $fullReadLengths[$i]);
	    }
	}
    }
    else { # lc $analysis ne 'eland_pair'
	if (($numUsedReads > 1) && defined($readLength)) {
	    error("Specification of READ_LENGTH is incompatible with multiple-read syntax for USE_BASES specified for ${analysisKey} ${analysis}: ${useBasesKey}=$useBases.\n");
	}
	# only one read specified
	if (scalar(@parts) == 1) {
	    $parts[0] = expandUseBases($parts[0], $readLength, $fullReadLength);
	    $parts[0] =~ s/y/Y/g;
	}
	# multiple reads specified
	else {
	    for (my $i = 0;  $i < scalar(@parts); $i++) {
		if ($parts[$i] =~ /([Yy]|all|odd|even)/) {
		    $parts[$i] = expandUseBases($parts[$i], $readLength, 
						$fullReadLengths[$i]);
		}
		else {
		    $parts[$i] = expandUseBases($parts[$i], 0, 
						$fullReadLengths[$i]);
		}
		$parts[$i] =~ s/y/Y/g;
		if (($i == 1) && (lc $analysis eq 'sequence')) {
		    $parts[$i] =~ s/Y/y/g;
		}
	    }
	}
    }
    my $result = join('',@parts);
    if (length($result) < $fullReadLength) {
	$result .= 'n' x ($fullReadLength - length($result));
    }
    return $result;
}


# full parsing of READ_LENGTH, USE_BASES and ANALYSYS.
# Expand wild cards and make sure all are mutually consistent.
sub parseUseBasesReadLengths($$$$) {
    my $paramsRef          = shift;
    my $multiParamsRef     = shift;
    my $multiLaneRef       = shift;
    my $fullReadLengthsRef = shift;
    my @fullReadLengths    = @{$fullReadLengthsRef};

    my $useBases;
    my %laneParams;
    my $key;
    my @keys = ($analysisKey, $useBasesKey, $readLengthKey, 
		$readLengthKey.'1', $readLengthKey.'2');

    
    # look at lane specific parameters. Because the auto-completion
    # and expansion rules are fairly complicated, need to recompute
    # for every lane that has any parameters different from the 
    # chip wide parameters
    for (my $lane = 1; $lane <= $numLanes; $lane++) {
	my $laneKey = getLaneKey($lane);
	# skip lane if parameters no different from chip-wide params
        # ANALYSIS parameter is always defined.
	my $skipFlag = 1;
	$laneParams{$analysisKey} = $multiParamsRef->{$analysisKey}{$laneKey};
	if (lc $laneParams{$analysisKey} ne lc $paramsRef->{$analysisKey}) {
	    $skipFlag = 0;
	}

	# check other parameters; fall back to chip-wide parameters
	# if necessary
	foreach $key (($useBasesKey,  $readLengthKey, 
		       $readLengthKey.'1', $readLengthKey.'2')) {
	    $laneParams{$key} = $paramsRef->{$key};
	    if (defined($multiLaneRef->{$key}) &&
		($multiLaneRef->{$key} =~ /$lane/)) {
		$laneParams{$key} = $multiParamsRef->{$key}{$laneKey};
		$skipFlag = 0;
	    }
	}
	# skip expansion if lane-specific parameters identical to
	# chip wide parameters
	if (($skipFlag == 1) || (lc $laneParams{$analysisKey} eq 'none')) {
	    $multiParamsRef->{$useBasesKey}{$laneKey} = '';
	    next;
	}
	$useBases = expandFullUseBases($fullReadLengthsRef,
				       $laneParams{$analysisKey}, 
				       $laneParams{$useBasesKey},
				       $laneParams{$readLengthKey}, 
				       $laneParams{$readLengthKey.'1'},
				       $laneParams{$readLengthKey.'2'});
	my @newKeys = ( $analysisKey, $useBasesKey );
	$multiParamsRef->{$useBasesKey}{$laneKey}       = $useBases;
	if ((lc $laneParams{$analysisKey} eq 'eland_pair')
	    ||(lc $laneParams{$analysisKey} eq 'sequence_pair')) {
	    $multiParamsRef->{$readLengthKey.'1'}{$laneKey} 
	        = countChars($useBases, 'Y');
	    $multiParamsRef->{$readLengthKey.'2'}{$laneKey} 
	        = countChars($useBases, 'y');
	    push @newKeys, $readLengthKey.'1';
	    push @newKeys, $readLengthKey.'2';
	}
	else {
	    $multiParamsRef->{$readLengthKey}{$laneKey}     
	        = countChars($useBases, 'Y');
	    push @newKeys, $readLengthKey;
	}

	if (countChars($useBases, 'I') > 0) {
	    $multiParamsRef->{$indexBasesKey}{$laneKey} = $useBases;
	    $multiParamsRef->{$indexBasesKey}{$laneKey} =~ s/[yY]/n/g;
	    $multiParamsRef->{$indexBasesKey}{$laneKey} =~ s/I/Y/g;
	    $multiParamsRef->{$useBasesKey}{$laneKey}   =~ s/I/n/g;
	    push @newKeys, $indexBasesKey;
	}
	
	foreach $key (@newKeys) {
	    if (!defined($multiLaneRef->{$key})) {
		$multiLaneRef->{$key} = $lane;
	    }
	    elsif ($multiLaneRef->{$key} !~ /$lane/) {
		$multiLaneRef->{$key} .= $lane;
	    }
	}
    }

    # parse ${params} chip wide parameters
    foreach $key (@keys) {
	$laneParams{$key} = $paramsRef->{$key};
    }
    $useBases = expandFullUseBases($fullReadLengthsRef,
				   $laneParams{$analysisKey}, 
				   $laneParams{$useBasesKey},
				   $laneParams{$readLengthKey},
				   $laneParams{$readLengthKey.'1'},
				   $laneParams{$readLengthKey.'2'});
    $paramsRef->{$useBasesKey} = $useBases;
    $paramsRef->{$readLengthKey."1"} = countChars($useBases, 'Y');
    $paramsRef->{$readLengthKey."2"} = countChars($useBases, 'y');
    $paramsRef->{$readLengthKey}     = countChars($useBases, 'Y');
    if (countChars($useBases, 'I') > 0) {
	$paramsRef->{$indexBasesKey} = $useBases;
	$paramsRef->{$indexBasesKey} =~ s/[yY]/n/g;
	$paramsRef->{$indexBasesKey} =~ s/I/Y/g;
	$paramsRef->{$useBasesKey}   =~ s/I/n/g;
    }
}


sub getLaneKey($) {
    my $lane = shift;
    return "s_${lane}";
}

#------------------------------------------------------------------------------

sub get_read_key($)
{
    my $read_num = shift;

    return (($read_num == 0) ? '' : "_${read_num}");
}

#------------------------------------------------------------------------------
# find a value for the given key in multiParams. If none is found,
# use value from params
sub getLaneKeyValue($$$$) {
    my $key            = shift;
    my $lane           = shift;
    my $paramsRef      = shift;
    my $multiParamsRef = shift;

    my $value = '';
    my $laneKey = getLaneKey($lane);
    if (defined ($multiParamsRef->{$key}) &&
	defined ($multiParamsRef->{$key}{$laneKey})) {
	$value = $multiParamsRef->{$key}{$laneKey};
    }
    elsif (defined($paramsRef->{$key})) {
	$value = $paramsRef->{$key};
    }
    
    return $value;
}

#------------------------------------------------------------------------------

sub set_lane_key_value($$$$)
{
    my $key = shift;
    my $lane_num = shift;
    my $value = shift;
    my $multi_params_ref = shift;

    if (!defined($multi_params_ref->{$key})) {
	my %undef_hash;
	%{$multi_params_ref->{$key}} = %undef_hash;
    }

    my $lane_key = getLaneKey($lane_num);
    $multi_params_ref->{$key}{$lane_key} = $value;
}

#------------------------------------------------------------------------------
# The model is of a base (read 0) key name that may have a non-zero read number
# appended to give read-specific keys for multi-read lanes.

sub set_lane_read_key_value($$$$$)
{
    my $key = shift;
    my $lane_num = shift;
    my $read_num = shift;
    my $value = shift;
    my $multi_params_ref = shift;

    my $full_key = $key . (($read_num == 0) ? '' : "$read_num");
    set_lane_key_value($full_key, $lane_num, $value, $multi_params_ref);
}

#------------------------------------------------------------------------------
# implement equivalent of Cwd->realpath(File::Spec->catdir()),
# but without traversing the directory tree to check existence of
# intermediate directories.
sub realpath($$) {
    my $path1 = shift;
    my $path2 = shift;
    my @parts1 = File::Spec->splitdir($path1);
    my @parts2 = File::Spec->splitdir($path2);
    my $count = 0;

    while (($count < scalar(@parts2)) 
	   && ($count < 4) # run-folder is 4 levels above Gerald folder
	   && ($parts2[$count] eq '..')) {
	$count += 1;
    }
    $path1 = File::Spec->catdir(@parts1[0 .. ($#parts1 - $count)]); 
    $path2 = File::Spec->catdir(@parts2[$count .. ($#parts2)]);
    return File::Spec->catdir($path1, $path2);
}

#------------------------------------------------------------------------------

sub get_lane_analysis($$)
{
    my $lane_num = shift;
    my $multi_params_ref = shift;

    my $lane_key = getLaneKey($lane_num);
    return lc $multi_params_ref->{$analysisKey}{$lane_key};
}

#------------------------------------------------------------------------------

sub will_not_analyse($)
{
    my $analysis = shift;

    return ($analysis eq 'none');
}

#------------------------------------------------------------------------------

sub will_not_align($)
{
    my $analysis = shift;

    return (will_not_analyse($analysis)
	    || ($analysis eq 'sequence')
	    || ($analysis eq 'sequence_pair'));
}

#------------------------------------------------------------------------------

sub will_not_produce_qtable($)
{
    my $analysis = shift;

    return (will_not_align($analysis)
	    || ($analysis eq 'monotemplate')
	    || ($analysis eq 'expression'));
}

#------------------------------------------------------------------------------

sub is_paired($)
{
    my $analysis = shift;

    return (($analysis eq 'eland_pair')
	    || ($analysis eq 'sequence_pair'));
}

#------------------------------------------------------------------------------

sub add_ele_to_uniq_sorted_list_str($$)
{
    my $list_str_ref = shift;
    my $val = shift;

    if (defined($$list_str_ref)) {
	my %map = split(/(:)/, $$list_str_ref . ':');
	$map{$val} = ':';
	$$list_str_ref = join(':', sort keys %map);
    } else {
	$$list_str_ref = "$val";
    }
}

#------------------------------------------------------------------------------
# check for presence of reference files
sub validateReferenceFiles($$$) {
    my $paramsRef       = shift;
    my $multiParamsRef = shift;
    my $multiLaneRef   = shift;

    for (my $lane = 1; $lane <= $numLanes; $lane++) {
	my $analysis = get_lane_analysis($lane, $multiParamsRef);
	next if (will_not_align($analysis));

	if (($analysis eq 'default') || ($analysis eq 'monotemplate')
	    || ($analysis eq 'expression')) {
	    my $genomeDir  = getLaneKeyValue('GENOME_DIR', $lane,
					     $paramsRef, $multiParamsRef);
	    my $genomeFile = getLaneKeyValue('GENOME_FILE', $lane,
					     $paramsRef, $multiParamsRef);
	    if (! File::Spec->file_name_is_absolute($genomeDir)) {
		$genomeDir = realpath($paramsRef->{$outDirKey},
				      $genomeDir);
	    }

	    $genomeFile = File::Spec->catdir($genomeDir, $genomeFile);
	    $genomeFile = File::Spec->canonpath($genomeFile);
	    error("Lane ${lane}: Reference file ${genomeFile} not found.\n")
		unless (-r $genomeFile);
	}

	if (($analysis eq 'eland') || ($analysis eq 'eland_pair')
	    || ($analysis eq 'eland_extended')) {
	    my $genomeDir  = getLaneKeyValue('ELAND_GENOME', $lane,
					     $paramsRef, $multiParamsRef);
	    if (! File::Spec->file_name_is_absolute($genomeDir)) {
		$genomeDir = realpath($paramsRef->{$outDirKey},
				      $genomeDir);
	    }
	    error("Lane ${lane}: Reference directory ${genomeDir} not found.\n")
		unless (-d $genomeDir);
	}
    }
}

#------------------------------------------------------------------------------
# This may be called once for a single-read lane or once for each read of a
# multi-read lane.

sub parse_one_lane_read_qcal_source($$$$$$$)
{
    my $params_ref = shift;
    my $multi_params_ref = shift;
    my $multi_lane_ref = shift;
    my $lane_qtables_used_ref = shift;
    my $lane_num = shift;
    my $read_num = shift;
    my $qtable_optional = shift;

    # For a single-read lane this is the expected source.
    # For a multi-read lane it is still a fallback if a read-specific source
    # is not found.
    my $qcal_source = getLaneKeyValue($qcalSourceKey, $lane_num,
				      $params_ref, $multi_params_ref);

    # Override with read-specific source if multi-read and available.
    if ($read_num != 0) {
	my $read_qcal_source
	    = getLaneKeyValue($qcalSourceKey . $read_num, $lane_num,
			      $params_ref, $multi_params_ref);

	if ($read_qcal_source) {
	    $qcal_source = $read_qcal_source;
	}
    }

    # Default to `auto' if qtable is required.
    # If qtable is optional (i.e. the mode does not use qcals as a prerequisite
    # for any other output files) and no source is specified (i.e. the user 
    # has not thereby requested qcal generation) there is nothing more to do.
    if (!$qcal_source) {
	if ($qtable_optional) {
	    return;
	} else {
	    $qcal_source = 'auto';
	}
    }

    my $qtable_path = $notDef;

    if ($qcal_source eq 'auto') {
	my $analysis = get_lane_analysis($lane_num, $multi_params_ref);

	if (will_not_produce_qtable($analysis)) {
	    error("Lane ${lane_num}: Analysis type $analysis does not "
		  . "include quality table generation, so `auto' is not "
		  . "a valid $qcalSourceKey for this lane.");
	}

	my $list_ref = \($lane_qtables_used_ref->{$lane_num});
	add_ele_to_uniq_sorted_list_str($list_ref, $read_num);
	set_lane_read_key_value($qtableRequiredKey, $lane_num, $read_num, 1,
				$multi_params_ref);
	$qtable_path 
	    = getLaneKey($lane_num) . get_read_key($read_num) . $qtable_suffix;
    } elsif ($qcal_source =~ /auto(\d+)/) {
	my $source_lane_num = $1;
	    
	if (($source_lane_num < 1) || ($source_lane_num > $numLanes)) {
	    error("Lane ${lane_num}: lane ($source_lane_num) specified "
		  . "as reference for quality calibration is invalid.");
	}

	my $source_lane_analysis = get_lane_analysis($source_lane_num,
						     $multi_params_ref);

	if (will_not_produce_qtable($source_lane_analysis)) {
	    error("Lane ${lane_num}: lane ($source_lane_num) cannot be "
		  . "used as reference for quality calibration because "
		  . "its analysis type ($source_lane_analysis) does not "
		  . "include quality table generation.");
	}

	# Destination read 0 : source lane single : use source read 0
	# Destination read n (> 0) : source lane single : use source read 0
	# Destination read 0 : source lane pair : use source read 1
	# Destination read n (> 0) : source lane pair : use source read n

	my $source_read_num = (is_paired($source_lane_analysis)
			       ? (($read_num > 0)
				  ? $read_num
				  : 1)
			       : 0);

	my $list_ref = \($lane_qtables_used_ref->{$source_lane_num});
        add_ele_to_uniq_sorted_list_str($list_ref, $source_read_num);

        set_lane_read_key_value($qtableRequiredKey, $source_lane_num,
				$source_read_num, 1, $multi_params_ref);
        $qtable_path = (getLaneKey($source_lane_num) 
			. get_read_key($source_read_num) . $qtable_suffix);
    } else {
	$qtable_path = File::Spec->canonpath($qcal_source);

        error("Lane ${lane_num}: "
	      . "External qtable `${qtable_path}' not found.\n")
	    unless (-r $qtable_path);
    }

    # QTABLE_PATH[12] are set at most once per lane, so no need to guard
    # against duplicates in the list of lanes.
    my $full_key = $qtablePathKey . (($read_num == 0) ? '' : "$read_num");
    $multi_lane_ref->{$full_key} .= $lane_num;

    set_lane_read_key_value($qtablePathKey, $lane_num, $read_num, $qtable_path,
                            $multi_params_ref);
}

#------------------------------------------------------------------------------
# Check SRF_ARCHIVE_REQUIRED values and expand into per-lane SRF_ARCHIVE_REQUIRED[12]

sub parse_srf_archive($$$$)
{
    my $params_ref = shift;
    my $multi_params_ref = shift;
    my $multi_lane_ref = shift;
    my $secondCycle = shift;

    my $srfArchiveHashref  = $multi_params_ref->{$srfArchiveRequired};
    my $srf2ndCycleHashref = $multi_params_ref->{$srf2ndCycleKey};
    for (my $lane_num = 1; $lane_num <= $numLanes; ++$lane_num) 
    {
	my $srfRequired = '';
        my $laneKey = "s_$lane_num";
        $srfRequired = $srfArchiveHashref ->{$laneKey} if defined $srfArchiveHashref ->{$laneKey};
        $srfRequired = lc($srfRequired);
        $srfRequired = '' unless $srfRequired eq 'yes';
        set_lane_key_value($srfArchiveRequired, $lane_num, $srfRequired, $multi_params_ref);
        my $srf2ndCycle = '';
        #$srf2ndCycle = $srf2ndCycleHashref->{$laneKey} if defined $srf2ndCycleHashref->{$laneKey};
        $srf2ndCycle = '-2 ' . $secondCycle  if $multi_params_ref->{$analysisKey}{$laneKey} =~ /_pair$/;
        set_lane_key_value($srf2ndCycleKey, $lane_num, $srf2ndCycle, $multi_params_ref);
    }
    $multi_lane_ref->{$srf2ndCycleKey} = '12345678';
}
#------------------------------------------------------------------------------
# Check QCAL_SOURCE values and expand into per-lane QTABLE_REQUIRED[12] and
# QTABLE_PATH[12] values.

sub parse_qcal_sources($$$)
{
    my $params_ref = shift;
    my $multi_params_ref = shift;
    my $multi_lane_ref = shift;

    # Initialise

    foreach my $read_str ('', '1', '2') {
	$params_ref->{$qtableRequiredKey . $read_str} = 0;
	$params_ref->{$qtablePathKey . $read_str} = '';
	$multi_lane_ref->{$qtableRequiredKey . $read_str} = '';
	$multi_lane_ref->{$qtablePathKey . $read_str} = '';
    }

    my %lane_qtables_used;

    for (my $lane_num = 1; $lane_num <= $numLanes; ++$lane_num) {
	set_lane_key_value($qtableRequiredKey, $lane_num, 0,
			   $multi_params_ref);
    }

    # Set 
    for (my $lane_num = 1; $lane_num <= $numLanes; ++$lane_num) {
	my $analysis = get_lane_analysis($lane_num, $multi_params_ref);
	next if (will_not_analyse($analysis));

	# May need review
	my $qtable_optional = will_not_produce_qtable($analysis);

	my @read_nums = (is_paired($analysis) ? (1, 2) : (0));

	foreach my $read_num (@read_nums) {
	    parse_one_lane_read_qcal_source($params_ref, $multi_params_ref,
					    $multi_lane_ref,
					    \%lane_qtables_used,
					    $lane_num, $read_num,
					    $qtable_optional);
	}
    }

    foreach my $qtable_lane_num (sort keys %lane_qtables_used) {
	# May be read 0 (i.e. not separate reads) or a list of reads.
	my @qtables_used = split(/:/, $lane_qtables_used{$qtable_lane_num});

	foreach my $qtable_used (@qtables_used) {
	    my $required_key = $qtableRequiredKey . (($qtable_used == 0)
						     ? '' : $qtable_used);
	    $multi_lane_ref->{$required_key} .= $qtable_lane_num;
	}
    }
}

#------------------------------------------------------------------------------

# print all parameters
sub printParams($$) {
    my $paramsRef      = shift;
    my $multiParamsRef = shift;

    foreach (sort keys %{$paramsRef}) {
	info(sprintf("%-28s %s\n", $_,  $paramsRef->{$_}));
    } # for

    for (my $lane = 1; $lane <= $numLanes; $lane++) {
	my $key;
	my $key2;
	foreach $key (sort keys %{$multiParamsRef}) {
	    foreach $key2 (keys %{$multiParamsRef->{$key}}) {
		if ($key2 eq getLaneKey($lane)) {
		    info(sprintf("%-28s %s\n", "${lane}:${key}", 
				 $multiParamsRef->{$key}{$key2}));
		}
	    }
	}
    }
}


sub writeTextFile($$) {
    my $filename = shift;
    my $content  = shift;
    open (OUT, ">$filename")
	or error("Unable to create temporary file $filename: $!");
    print OUT $content;
    close (OUT);
}

#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------

my $version = "0.1.9";
my $id;
{
    my $cvs_id='@(#) $Id: GERALD.pl,v 1.171 2008/05/19 17:36:14 mzerara Exp $';
    my $cvs_version='$Revision: 1.171 $';
    # Make CVS tag into a displayable string
    $id=$cvs_id;
    $id=~s/\$//g;
} #scope

my $makeCommand="make";
my $folderRoot ="GERALD";

# Try to get user name from environment
my $logName=getLogName();

my $tmpName="/tmp/tmp-GERALD-$>-$$.txt";
my $tmpName2="/tmp/tmp-GERALD-$>-$$.txt";

END
{
    if (-e $tmpName)
    {
        info("Removing temporary file $tmpName\n");
        unlink ($tmpName);
    } # if
} # END

my $laneTag='@LANE@';


my %params;
# $multiLane{"param"} specifies lanes for which param has a lane-specific
# definition 
my %multiLane;
# $multiParams{"param"}{"lane"} specifies definition of param for lane
my %multiParams;

# default parameters
my $thisDir=File::Spec->rel2abs(dirname($0));
my $timeNow=scalar(localtime);


# Create XML 
my $xml = new XML::Simple
    ( suppressempty=>'',
      XMLDecl=>1,
      Variables => { id=>$id, 
		     timeNow=>$timeNow,
		     thisDir=>$thisDir,
		     logName=>$logName } );


my $configHash=$xml->XMLin();

#print Dumper($test),"\n";

my $macroDefs=\%{$configHash->{MacroDefinitions}};
%params=%{$configHash->{DefaultRunParameters}};

# Read command line parameters
my %options;
my @optionKeys = keys %params;
foreach (@optionKeys) {
    $_ .= "=s";
}
foreach (("RUN", "FORCE", "make", "length=s")) {
    push (@optionKeys, $_);
}
GetOptions(\%options, @optionKeys);


# Read command line parameters
unless (@ARGV>0)
{
    print STDERR "Usage: $0 [config.txt|-] [--RUN] [--FORCE] [--EXPT_DIR dir] [--OUT_DIR dir]\n";
    print STDERR "Allowable config params and their defaults\n";
    for (sort keys %params)
    {
        info("$_ => $params{$_}\n");
    } # for
    die;
} # unless

error("Unable to write to /tmp directory - need write access!\n")
    unless ((-d '/tmp') && (-w '/tmp'));


# read config file
foreach my $configFile (@ARGV)
{
    if ($configFile !~ /.xml$/) {
	if ($configFile eq '-')
	{
	    open (IN, "<&=STDIN")
		or die "Problem aliasing STDIN: $!";
	    info("Will read config params from STDIN\n");
	} #if
	else
	{
	    open (IN, "${configFile}")
		or die "Problem opening config file ${configFile}: $!";
	    info("Will read config params from ${configFile}\n");
	}
 
	my $content = "";
	while (<IN>) {
	    $content .= $_;
	}
	close (IN);

	parseConfigTextFile($content, \%params, \%multiParams, \%multiLane);

	writeTextFile($tmpName, $content);

    } #elsif text config file
    else
    { # it's an XML file, hooray
        info("Will read config params from ${configFile}\n");
	readConfigXMLFile($configFile, $xml,
			  \%params, \%multiParams, \%multiLane);
    } # else

} #unless

# parse command-line options; they will overwrite the parameters
# obtained from the parameter file
foreach my $key (keys %options) {
    if (($key eq "RUN")||($key eq "FORCE"))
    {            
	$params{$key}=1;
	next;        
    } #if
    if (($key eq "make")) {
	$params{"FORCE"} = 1;
	next;
    }
    if (($key eq "length")) {
	next;
    }
    die ("Invalid key ${key} specified on command line")
	unless (defined($params{$key}));
        $params{$key}= $options{$key};
}

# strip whitespace 
foreach (sort keys %params)
{
    $params{$_}=~s/^\s+//;
    $params{$_}=~s/\s+$//;
} #foreach


# Ascertain naming convention of sequence files
# TILE_ROOT and TILE_REGEX now defined in config params
# Here we just check they are sensibly defined
unless ($params{"TILE_REGEX"}=~/^$params{"TILE_ROOT"}/)
{
    die "Parameter TILE_REGEX (value=$params{TILE_REGEX}) does not begin with parameter TILE_ROOT (value=$params{TILE_ROOT})";
} # unless


# sort out folder names and set variables EXPT_DIR, EXPT_DIR_ROOT, OUT_DIR
updateFolderNames(\%params, $folderRoot, $logName);


# if the length option is set, ignore non-existing directories -
# this is just a test run
if (!defined($options{"length"})) {
    error("Cannot find directory ".$params{$exptDirKey}."\n")
	unless (-r $params{$exptDirKey});
    info("Found directory ".$params{$exptDirKey}."\n");
}

# extract information about full read-lengths
my @fullReadLengths = getReadLengths2($params{"EXPT_DIR"});

# Set the original read lengths parameter before any overrides.
$params{$origReadLengthsKey} = join(':', @fullReadLengths);

# override read lengths if command-line option is specified
if (defined($options{"length"})) {
    info("Overriding lengths using option ${options{'length'}}.\n");
    @fullReadLengths = split(',', $options{'length'});
}

# Sort out values for ANALYSIS. These need to be in the multiParams
# so they are correctly expanded when the makefile is built from the
# template

#              $multiLane{$lastKey}.=$lane;
#              $multiParams{$lastKey}{"s_$lane"}="@s";

# After the conversion this will always be "12345678"
$multiLane{$analysisKey}="";

for (my $i=1; $i<=$numLanes; $i++)
{
    $multiLane{$analysisKey}.="$i";
    $multiParams{$analysisKey}{getLaneKey($i)}=$params{$analysisKey}
    unless (defined ($multiParams{$analysisKey}{getLaneKey($i)}));  
} # if


# ANALYSIS values should now be ready for expansion
# Sort out READ_LENGTH, USE_BASES
parseUseBasesReadLengths(\%params, \%multiParams, \%multiLane,
			 \@fullReadLengths);


# check that all reference files exist
validateReferenceFiles(\%params, \%multiParams, \%multiLane);

# Check QCAL_SOURCE values and expand into (internal) per-lane QTABLE_REQUIRED 
# and QTABLE_PATH values.
parse_qcal_sources(\%params, \%multiParams, \%multiLane);

####
# SRF ARCHIVE
####

#print keys %{$multiParams{$srfArchiveRequired}}, "\n";
#print values %{$multiParams{$srfArchiveRequired}}, "\n";
#die "debug";

parse_srf_archive(\%params, \%multiParams, \%multiLane, $fullReadLengths[0]+1);

#$params{$srf2ndCycleKey} = '';
#if ($params{$analysisKey} =~ /_pair$/)
#{
    #$params{$srf2ndCycleKey} = "-2 " . ($fullReadLengths[0]+1);
#}

#my $analysisHash = $multiParams{$analysisKey};
#foreach (keys %{$analysisHash})
#{
    #$multiParams{$srf2ndCycleKey}{$_} = '';
    #if ($analysisHash->{$_} =~ /_pair$/)
    #{
        #$multiParams{$srf2ndCycleKey}{$_} = "-2 " . ($fullReadLengths[0]+1);
    #}
#}

# Check everything is defined
foreach (sort keys %params)
{
    error("No definition given for parameter $_")
        unless ($params{$_} ne $notDef); 
} # for


# print all parameters
printParams(\%params, \%multiParams);

# exit unless FORCE option has been specified
if (!defined($params{"FORCE"})) {
    exit(0);
}
#########################################################################
# now create output folders and makefile
########################################################################
#print Dumper(\%params);
#print Dumper(\%multiParams);
#print Dumper(\%multiLane);
#die;

# now create output directory
# first untaint $params{"OUT_DIR"}
my $outdir;
if ($params{"OUT_DIR"} =~ /^([\w\-\/\.\,]+)$/) {
    $outdir = $1; # $1 is untainted
} else {
    die "folder ".$params{"OUT_DIR"}." has invalid characters.\n";
}

unless (-d $outdir)
{
    mkdir $outdir 
       || die "Failed to create directory ".$outdir.": $!";
    print STDERR "Created directory ".$outdir."\n";
} #unless
else
{
    die "OUT_DIR ".$outdir." present but not readable"
	unless (-r $outdir);
} #if

my $outFileName = File::Spec->catdir($outdir, "Makefile");
print "want to make $outFileName\n";
print "File $outFileName already exists - will overwrite"
    if ((!defined($params{"FORCE"}))&&(-e $outFileName));

open ( OUT, ">$outFileName")
    || die "Failed to open file $outFileName for writing: $!";

my $modVal;
my $multiVal;
my $laneName;
my $lastKey;
my $thisVal;

# Read in makefile and substitute params
# NB can only do one @...@ substitution per line
while($thisVal=<main::DATA>)
{
    if ($thisVal=~/(@\w+@)/)
    {
#        print "found key $1\n";
        $lastKey=$1;
        $lastKey=~s/^@//;
        $lastKey=~s/\@$//;
#        print "- $lastKey\n";
        die "Could not find value for $lastKey"
            unless (defined($params{$lastKey})); 

        # print out chip-wide definition
        if ($lastKey ne $analysisKey)
        {
            $modVal=$thisVal;
            $modVal=~s/\@$lastKey\@/$params{$lastKey}/g;
            print OUT $modVal;
        } # if

        # print any lane-specific definitions
        if (defined($multiLane{$lastKey}))
        {
            if ($lastKey ne $analysisKey)
            {
#                $modVal=$laneTag."%:$thisVal";

		# As the QTABLE_PATH variables have to be used in prerequisites
		# they must have unique lane-dependent names rather than 
		# having the same name but being defined per lane.
		# (QTABLE_REQUIRED variables have the same requirement.)
		if (($lastKey =~ /^${qtablePathKey}/)     || 
                    ($lastKey =~ /^${qtableRequiredKey}/) ||
                    ($lastKey =~ /^${srfQcalKey}/)        ||
                    ($lastKey =~ /^${srf2ndCycleKey}/)    ||
                    ($lastKey =~ /^${srfArchiveRequired}/)) {
		    $thisVal = join('_', $laneTag, $thisVal);
		} else {
		    $thisVal=$laneTag."%:$thisVal";
		}
            } # if

            foreach my $lane (split(//,$multiLane{$lastKey}))
            {
                die unless (defined ($multiParams{$lastKey}{"s_$lane"}));
                $laneName=$params{"TILE_ROOT"}."_".${lane};
                if ($lastKey eq $analysisKey)
                {
#                    $modVal=$analysis{$multiParams{$lastKey}{"s_$lane"}};
                    $modVal=$multiParams{$lastKey}{"s_$lane"};
		    $modVal='@'.$modVal.'@';
		    $macroDefs->{LANE}=$laneName;
		    while (1)
		    {
#			print "start:\n$line \n";
			if ($modVal=~/@(\w+)(=?)(\w*)@/)
			{
			    my $key=$1;
#			    print "GOT $key\n";
			    my $val;
			    if ($2 eq '=')
			    { # parameterized key
				$key=$1;
				$val=$3; 
				if (defined($macroDefs->{$key}))
				{
				    print "Changing value of macro $key from ",
				    $macroDefs->{$key}, " to $val\n";
				}
				else
				{
				    print "Setting macro $key to $val\n",
				}
				$macroDefs->{$key}=$val;
#				$modVal=~s/@$key=$val@//;
				$modVal=~s/$&//;
			    } # if parameterized key
			    else
			    { # normal key
				$val=$macroDefs->{$key};
				if (defined($val))
				{
				    $val=~s/\s$//;
				    $key='@'.$key.'@';
				    $modVal=~s/$key/$val/;
				} # if
				else
				{
				    die "Could not find a def for macro $1";
				}
			    } # else

			} # if
			else
			{
			    last;
			} # else
    
		    } # while

                } # if analysisKey ...
                else
                {
                    $modVal=$thisVal;
                    $modVal=~s/\@$lastKey\@/$multiParams{$lastKey}{"s_$lane"}/g;
                }
#                print "ZZ $laneName $thisVal $lastKey\n";
#                print $multiParams{$lastKey}{"s_$lane"}."\n";
#                $modVal=$params{"TILE_ROOT"}.${lane}."%:".$thisVal;
#                $modVal=~s/\@$lastKey\@/$multiParams{$lastKey}{"s_$lane"}/g;
                $modVal=~s/$laneTag/$laneName/g;
#                print "OUT $modVal\n";
                print OUT $modVal;
            } # for
        } # if
    } # if
    else
    {
        print OUT $thisVal;
    } # else
} #while

close (OUT);

print STDERR "Successfully created $outFileName\n";

if (-e $tmpName)
{
    print STDERR "Copying config file $tmpName to "
        .$outdir."/config.txt\n";
    copy ($tmpName, $outdir."/config.txt")
        or die "Failed to copy config file: $!";
} # if

print STDERR "Dumping config parameters into $tmpName to "
        .$outdir."/config.xml\n";

open (XML, ">$outdir/config.xml")
    or die "Failed to open file $outdir/config.xml";
print XML "<RunParameters>\n";
print XML XMLout(\%params,NoAttr=>1,
		 RootName=>'ChipWideRunParameters');
print XML XMLout(\%multiParams,NoAttr=>1,
		 RootName=>'LaneSpecificRunParameters');
print XML "</RunParameters>\n";
close (XML);


print STDERR "Running 'make self_test' on $outFileName\n";

$ENV{'PATH'} = "/usr/local/bin:/bin:/usr/bin"; # untaint PATH variable
if (system("cd ${outdir}; make -f $outFileName self_test")!=0)
{
    print STDERR "Problem running 'make self_test' on $outFileName\n\n"; 
    print STDERR "Investigate and fix errors, then run\n";
    print STDERR "$0 $params{OUT_DIR}/config.txt --FORCE\n";
    print STDERR "to build corrected Makefile in same folder\n\n";
    exit (-1);
} # if

print STDERR "Run 'make self_test' on $outFileName completed with no problems\n";



if (defined($params{"RUN"}))
{
    my $comString="cd ".$outdir."; $makeCommand all";
    print STDERR "Will run $comString\n";
    exec ($comString);
} #if
else
{
    print STDERR "Task complete, exiting\n";
} #else

# And now for the grand finale, the master makefile template
# Anything looking like @THIS@ gets substituted.

# NB In Emacs, may have to edit makefile template below as a Makefile then 
# cut-and-paste in the changes - doesn't do tabs correctly in perl mode
__END__
# File: Makefile 
# Description: file for auto-generation of sequence analysis
# Auto-generated by @SOFTWARE_VERSION@
# Auto-generated on @TIME_STAMP@

# Copyright (c) Solexa Limited 2005
# Author: A. J. Cox


# @(#) $Id: GERALD.pl,v 1.171 2008/05/19 17:36:14 mzerara Exp $

.SUFFIXES:

# General variable definitions

EXPT_DIR:=@EXPT_DIR@

cmdprefix?=@CMDPREFIX@

TOOLS_DIR:=@TOOLS_DIR@
BIN_DIR:=@TOOLS_DIR@/../bin
QCAL_TOOLS_DIR:=$(TOOLS_DIR)/../QCalibration
SRF_ARCHIVE_REQUIRED:=@SRF_ARCHIVE_REQUIRED@
QTABLE_REQUIRED:=@QTABLE_REQUIRED@
QTABLE_REQUIRED1:=@QTABLE_REQUIRED1@
QTABLE_REQUIRED2:=@QTABLE_REQUIRED2@
QTABLE_PATH=@QTABLE_PATH@
QTABLE_PATH1=@QTABLE_PATH1@
QTABLE_PATH2=@QTABLE_PATH2@
ELAND_DIR:=$(TOOLS_DIR)/../Eland/
GENOME_DIR:=@GENOME_DIR@
GENOME_FILE:=@GENOME_FILE@
GENOME=${GENOME_DIR}/${GENOME_FILE}
READ_LENGTH:=@READ_LENGTH@
READ_LENGTH1:=@READ_LENGTH1@
READ_LENGTH2:=@READ_LENGTH2@
ORIG_READ_LENGTHS:=@ORIG_READ_LENGTHS@
ELAND_SEED_LENGTH:=@ELAND_SEED_LENGTH@

CONTAM_DIR:=@CONTAM_DIR@
CONTAM_FILE:=@CONTAM_FILE@
SEQUENCE_FORMAT:=@SEQUENCE_FORMAT@
QUALITY_FORMAT:=@QUALITY_FORMAT@

USE_BASES_CRITERION:=@USE_BASES@
USE_BASES= -useBases ${USE_BASES_CRITERION}
PURE_BASES_VAL:=@PURE_BASES@
PURE_BASES= -pureBases @PURE_BASES@

INDEX_BASES_CRITERION:=@INDEX_BASES@
INDEX_BASES=$(addprefix -indexBases ,$(INDEX_BASES_CRITERION))

GROUP_LANES=@GROUP_LANES@

# QVAL_COMPR:=none
QVAL_COMPR:=gzip
# QVAL_COMPR:=bzip2

# If running under Cygwin, need to add a .exe suffix onto
# names of compiled executables

ifdef COMSPEC
EXE:=.exe
endif

# Make sure that errors get propagated when pipes are used in the make commands
#HAVE_BASH3:=$(shell if [ "$$BASH_VERSINFO" -a "$$BASH_VERSINFO" -gt 2 ]; then echo 1; fi)
HAVE_BASH3:=$(shell sh --version | grep "bash, version [345]\.")
ifneq ($(HAVE_BASH3),)
PIPECMD:=set -o pipefail;
QUIET_PIPECMD:=@set -o pipefail;
endif

UNPROTECTED_TARGET=$@
SAFEPIPETARGET=$@.tmp && mv $@.tmp $@
SAFEPIPETARGET2=$(subst .,.tmp.,$@) && mv $(subst .,.tmp.,$@) $@
SAFEPIPETARGET3=$(CURDIR)/$@.tmp && mv $(CURDIR)/$@.tmp $(CURDIR)/$@
SAFEPIPETARGET_SMT=$(CURDIR)/$@.tmp && mv $(CURDIR)/$@.tmp $(EXPT_DIR)/$@
AND=&&

# The quote marks below contain a tab character
SORT_BY_TAB:=sort -T . --field-separator="	"

TILE_FILE:=tiles.txt

# Parameters for called programs...

# ... phageAlign

PA:=${TOOLS_DIR}/phageAlign${EXE}

PA_PARAMS=$(GENOME) $(READ_LENGTH) $(READ_LENGTH) $(USE_BASES)
PA_PARAMS_MONO=-tag $(GENOME) $(READ_LENGTH) $(READ_LENGTH) $(USE_BASES)


# ... score.pl

SC:=${TOOLS_DIR}/score.pl
SC_PREREQS:=$(SC) $(TILE_FILE)
SC_PARAMS=
#SC_PARAMS=$(GENOME)

# ... buildCoverage.pl

BC:=${TOOLS_DIR}/buildCoverage.pl
# Second BC_PARAMS is score threshold below which to ignore alignments
# when building coverage. Leave it out to include them all
BC_PARAMS=$(GENOME) 200


# ... plotCoverage.pl

PC:=${TOOLS_DIR}/plotCoverage.pl

# ... quahog

QH:=${TOOLS_DIR}/quahog${EXE}
QH_PARAMS= ${USE_BASES} ${PURE_BASES}
# ... qualityFilter.pl

QF:=${TOOLS_DIR}/qualityFilter.pl
FILTER_CRITERION:=@QF_PARAMS@
QF_PARAMS= -stamp ${FILTER_CRITERION}

# ... create_tile_thumbnails.pl

CE:=${TOOLS_DIR}/create_tile_thumbnails.pl
CE_PARAMS=error
CE_PARAMS_SMT=all
CE_PARAMS_PSH=hist
CE_PARAMS_PEC=perfect
CE_PARAMS_PIN=info

# ... convertToFasta.pl

TF:=${TOOLS_DIR}/convertToFasta.pl
TF_PARAMS=${USE_BASES}

# ... convertFromELAND.pl

FI:=${TOOLS_DIR}/convertFromELAND.pl
FI_PREREQS:=$(FI) $(TILE_FILE)

# ... Signal_Means.pl
SM:=${TOOLS_DIR}/Signal_Means.pl
SM_PARAMS=

# ... plotIntensity_for_IVC.pl
PI:=${TOOLS_DIR}/plotIntensity_for_IVC.pl

# ... create_IVC_thumbnail.pl
CI:=${TOOLS_DIR}/create_IVC_thumbnail.pl

# ... Signal_Means_Tile.pl
SMT:=${TOOLS_DIR}/Signal_Means_Tile.pl
SMT_PARAMS=$(QF) ${FILTER_CRITERION}

# ... plotIntensity_tiles.pl
PIT:=${TOOLS_DIR}/plotIntensity_tiles.pl
PIT_PREREQS:=$(PIT) $(TILE_FILE) $(REQUIRES_SMT_DIR)
PIT_PRE_LANE_PARAMS:=${TILE_FILE}
# Suffixes are not defined until further down, so need lazy expansion.
PIT_POST_LANE_PARAMS=$(SMT_DIR) $(SMT_SUFFIX) $(SMT_PNG_SUFFIX)

# ... Signal_Means_Monotemplates.pl
SMM:=${TOOLS_DIR}/Signal_Means_Monotemplates.pl
SMM_PARAMS=$(QF) ${FILTER_CRITERION} ${USE_BASES_CRITERION} $(READ_LENGTH)

# ... runReport.pl
RR:=${TOOLS_DIR}/runReport.pl
RR_PARAMS=${CURDIR} @EMAIL_SERVER@ \
@EMAIL_DOMAIN@ \
@WEB_DIR_ROOT@ \
--strip @NUM_LEADING_DIRS_TO_STRIP@ \
@EMAIL_LIST@ 

# ... plotScoreHist.pl
PSH:=${TOOLS_DIR}/plotScoreHist.pl

# ... plot_error_graphs.pl
PEG:=${TOOLS_DIR}/plot_error_graphs.pl
PEG_PREREQS:=${PEG} ${TILE_FILE}
PEG_PRE_LANE_PARAMS:=${TILE_FILE}
# Suffixes are not defined until further down, so need lazy expansion.
PEG_POST_LANE_PARAMS=${RESCORE_SUFFIX} ${RESCORE_PNG_SUFFIX} ${PEC_PNG_SUFFIX}

# ... plotInformation.pl
PIN:=${TOOLS_DIR}/plotInformation.pl

# ... byTileThumbnailView.pl
BTT:=${TOOLS_DIR}/byTileThumbnailView.pl

# ... plotIntensity_monotemplates.pl
PIM:=${TOOLS_DIR}/plotIntensity_monotemplates.pl

# ... create_IVC_monotemplates_thumbnail.pl 
CMT:=${TOOLS_DIR}/create_IVC_monotemplates_thumbnail.pl 

# ... jerboa.pl
JER:=${TOOLS_DIR}/jerboa.pl
JER_PREREQS:=$(JER) $(TILE_FILE)

# ... reportQuality.pl
RQ:=${TOOLS_DIR}/reportQuality.pl
RQS:=${TOOLS_DIR}/reportQualitySAF.pl

# ... convertQuality
CQ:=${TOOLS_DIR}/convertQuality
CQ_PARAMS= --useBases ${USE_BASES_CRITERION} $(QUALITY_FORMAT) ${READ_LENGTH}

# ... saf2qref.pl
S2QRF:=${TOOLS_DIR}/saf2qref.pl
S2QRF_PREREQS:=$(S2QRF) $(TILE_FILE)
S2QRF_PARAMS=--use_bases ${USE_BASES_CRITERION} --orig_read_lengths ${ORIG_READ_LENGTHS}

# ... align2qref.pl
A2QRF:=${TOOLS_DIR}/align2qref.pl
A2QRF_PARAMS=--use_bases ${USE_BASES_CRITERION}

# ... extract_quality_predictors
EQP:=${QCAL_TOOLS_DIR}/extract_quality_predictors
EQP_PARAMS=--orig_read_lengths ${ORIG_READ_LENGTHS} --qval_compression ${QVAL_COMPR} --pure_bases ${PURE_BASES_VAL}

# ... QualityCalibration
QLC:=${QCAL_TOOLS_DIR}/QualityCalibration${EXE}
# ... QualityCalibration.xml (config file for QualityCalibration)
QLCX:=${QCAL_TOOLS_DIR}/QualityCalibration.xml
# QLC_PARAMS must be immediately followed by LANE flanked by at signs. 
# Including that in QLC_PARAMS would cause it to be prematurely expanded.
QLC_PARAMS=--cfg $(QLCX) --qval_compression ${QVAL_COMPR} --tile_list_file $(TILE_FILE) --tile_prefix
QLC_PREREQS:=$(QLC) $(QLCX) $(TILE_FILE)


# ... QualityApply
QLA:=${QCAL_TOOLS_DIR}/QualityApply${EXE}
QLA_PARAMS:=${QUALITY_FORMAT}
# QLA_QVAL_PARAMS must be immediately followed by LANE flanked by at signs. 
# Including that in QLA_QVAL_PARAMS would cause it to be prematurely expanded.
# Similarly QLA_QVAL_PARAMS must follow the lane-dependent qtable spec.
QLA_QVAL_PARAMS:=--qval_compression ${QVAL_COMPR} --tile_list_file $(TILE_FILE) --tile_prefix
QLA_PREREQS:=$(QLA) $(TILE_FILE)

# ... filterQualityValues.pl
FQV:=${TOOLS_DIR}/filterQualityValues.pl
FQV_PARAMS=${USE_BASES_CRITERION} ${QUALITY_FORMAT}

# ... buildSeq.pl
BS:=${TOOLS_DIR}/buildSeq.pl

# ... unsquash

UNSQ:=${TOOLS_DIR}/unsquash${EXE}
UNSQ_PARAMS=${ELAND_GENOME} ${READ_LENGTH}

# ... buildExtendedAlign.pl

BEA:=${TOOLS_DIR}/buildExtendedAlign.pl

# ... pickBestAlignment.pl

PBA:=${TOOLS_DIR}/pickBestAlignment.pl

# ... convertFromSAF.pl

CFS:=${TOOLS_DIR}/convertFromSAF.pl

# ... pickBestPair.pl

PBP:=${TOOLS_DIR}/pickBestPair.pl
PAIR_PARAMS:= @PAIR_PARAMS@

# ... squashGenome

SG:=${ELAND_DIR}/squashGenome${EXE}

# ... buildExport

BE:=${TOOLS_DIR}/buildExport.pl

# ... illumina2srf

SRF:=${BIN_DIR}/illumina2srf
SRF_PARAMS:= @SRF_RAW@
SRF_PARAMS+= @SRF_PROCESSED@ 
SRF_PARAMS+= -c @SRF_CHASTITY@ 
SRF_QCAL_PARAMS+= -qc @OUT_DIR@ 

SRF_OUT_DIR:=@OUT_DIR@
# ...

TOOLS:=$(PA) $(SC) $(BC) $(PC) $(RA) $(HM) $(XL) $(PCL) $(QH) \
$(QF) $(CE) $(TF) $(FI) $(SM) $(PI) $(CI) $(SMT) $(PIT) $(SMM) $(RR) $(PSH) \
$(PEG) $(PIN) $(BTT) $(PIM) $(CMT) $(JER) $(RQ) $(RQS) $(EQP) $(QLC) $(QLA) \
$(FQV) $(BS) $(UNSQ) $(BEA) $(PBA) $(CFS) $(SG) $(BE)



# Following are lists of directories, files and executables  
# that will be checked by make self_test
# These lists may be added to depending on what analysis is asked for
DIRS_TO_CHECK=$(EXPT_DIR) $(TOOLS_DIR) $(QCAL_TOOLS_DIR)
FILES_TO_CHECK=$(TOOLS) $(QLCX)
EXECS_TO_CHECK:=$(TOOLS)

#
# Find lists of files that are inputs to processing  
#

# Eventually this list will be obtained from an XML file

SIG_FILE_SUFFIX_DEPHASED:=@SIG_FILE_SUFFIX_DEPHASED@
SIG_FILE_SUFFIX_NOT_DEPHASED:=@SIG_FILE_SUFFIX_NOT_DEPHASED@

SEQ_FILE_SUFFIX:=@SEQ_FILE_SUFFIX@
SIG_FILE_SUFFIX:=$(SIG_FILE_SUFFIX_DEPHASED)
PRB_FILE_SUFFIX:=@PRB_FILE_SUFFIX@
QHG_FILE_SUFFIX:=@QHG_FILE_SUFFIX@
SRF_FILE_SUFFIX:=@SRF_FILE_SUFFIX@

TILE:=@TILE_REGEX@

# Get list of sequence files
ALL_SEQ_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(SEQ_FILE_SUFFIX))

# Get list of sig files, if blank try non-phase corrected suffix instead
ALL_SIG_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX))
ifeq ($(strip $(ALL_SIG_FILES)),)
WARNING_MESSAGES+=Info: no intensity files satisfying \
	$(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX_DEPHASED) found, trying \
	$(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX_NOT_DEPHASED) instead
SIG_FILE_SUFFIX:=$(SIG_FILE_SUFFIX_NOT_DEPHASED)
ALL_SIG_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX))
endif

ALL_PRB_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(PRB_FILE_SUFFIX))

ALL_QHG_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(QHG_FILE_SUFFIX))

QHG_DIR:=$(EXPT_DIR)
ifeq ($(strip $(ALL_QHG_FILES)),)
QHG_DIR:=.
endif

SMT_DIR:=$(EXPT_DIR)/SignalMeans
REQUIRES_SMT_DIR:=$(if $(wildcard $(SMT_DIR)), , $(SMT_DIR))

ALL_TILES:=$(patsubst $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX),%,$(ALL_SEQ_FILES))
ALL_SIG_TILES:=$(patsubst $(EXPT_DIR)/%$(SIG_FILE_SUFFIX),%,$(ALL_SIG_FILES))
ALL_PRB_TILES:=$(patsubst $(EXPT_DIR)/%$(PRB_FILE_SUFFIX),%,$(ALL_PRB_FILES))

# This will be blank if the two tiles lists are the same
DIFF_TILES_SIG:=$(filter-out $(ALL_TILES), $(ALL_SIG_TILES))\
$(filter-out $(ALL_SIG_TILES), $(ALL_TILES))

# This will be blank if the two tiles lists are the same ...
DIFF_TILES_PRB:=$(filter-out $(ALL_TILES), $(ALL_PRB_TILES))\
$(filter-out $(ALL_PRB_TILES), $(ALL_TILES))


BAD_TILES:=@BAD_TILES@
# Any tiles listed in BAD_TILES get completely ignored
TILES:=$(filter-out $(BAD_TILES), $(ALL_TILES) )

ALL_LANE_NUMS:=1 2 3 4 5 6 7 8
ALL_LANES:=$(patsubst %,@TILE_ROOT@_%,$(ALL_LANE_NUMS))
LANES:=$(foreach w, $(ALL_LANES), $(findstring $w, $(TILES)))
LANE_TARGETS:=$(patsubst %,%,$(LANES))

MAX_THUMBNAIL_OPTION:=--maxTiles=20

#
# All types of files that can be generated are listed here
#

SAF_SUFFIX:=_saf.txt
CALSAF_SUFFIX:=_calsaf.txt
EXPORT_SUFFIX:=_export.txt
ALIGN_SUFFIX:=_align.txt
QHG_SUFFIX:=_qhg.txt
QALIGN_SUFFIX:=_qalign.txt
REALIGN_SUFFIX:=_realign.txt
SCORE_SUFFIX:=_score.txt
RESCORE_SUFFIX:=_rescore.txt
SCORE_FILES_SUFFIX:=_score_files.txt
RESCORE_PNG_SUFFIX:=_rescore.png
ERROR_PNGS_SUFFIX:=_error_pngs.txt
COV_SUFFIX:=_cov.txt
COV_PNG_SUFFIX:=_cov.png
SM_SUFFIX:=_Signal_Means.txt
SMT_SUFFIX:=_all.txt
SMT_PNG_SUFFIX:=_all.png
SMT_PNGS_SUFFIX:=_all_pngs.txt
MONO_SUFFIX:=_monotemplate.txt
PSH_PNG_SUFFIX:=_hist.png
PEC_PNG_SUFFIX:=_errors.png
PIN_PNG_SUFFIX:=_info.png
CALIGN_SUFFIX:=_calign.txt
CDIFF_SUFFIX:=_cdiff.txt
CPREDIFF_SUFFIX:=_cprediff.txt
CREDIFF_SUFFIX:=_crediff.txt
PREALIGN_SUFFIX:=_prealign.txt
CREALIGN_SUFFIX:=_crealign.txt
QREP_SUFFIX:=_qreport.txt
QCALREP_SUFFIX:=_qcalreport.txt
QREF_SUFFIX:=_qref.txt
QREF_FILES_SUFFIX:=_qref_files.txt
QVAL_SUFFIX:=_qval.txt
QVAL_GZ_SUFFIX:=_qval.txt.gz
QVAL_FILE_PAIR_SUFFIX:=_qval_file_pair.txt
QTABLE_SUFFIX:=_qtable.txt
QCAL_SUFFIX:=_qcal.txt
UB_QCAL_SUFFIX:=_ub_qcal.txt
QRAW_SUFFIX:=_qraw.txt
SEQPRE_SUFFIX:=_seqpre.txt
SEQUENCE_SUFFIX:=_sequence.txt
FRAG_SUFFIX:=_frag.txt
TAGCOUNT_SUFFIX:=_tagcount.txt
FILT_SUFFIX:=_filt.txt
ANOMALY_SUFFIX:=_anomaly.txt
ANOMRAW_SUFFIX:=_anomraw.txt
REANOMRAW_SUFFIX:=_reanomraw.txt
RAWPAIR_SUFFIX:=_rawpair.xml
PAIR_SUFFIX:=_pair.xml
GENOMESIZE_SUFFIX:=_genomesize.xml

QVAL_ACTUAL_SUFFIX:=$(QVAL_SUFFIX)

ifeq ($(strip $(QVAL_COMPR)),gzip)
  QVAL_ACTUAL_SUFFIX:=$(QVAL_GZ_SUFFIX)
endif

ifeq ($(strip $(QVAL_COMPR)),bzip2)
  QVAL_ACTUAL_SUFFIX:=$(QVAL_BZ2_SUFFIX)
endif

OUTPUT_FILE_TYPES+=$(SAF_SUFFIX)
OUTPUT_FILE_TYPES+=$(CALSAF_SUFFIX)
OUTPUT_FILE_TYPES+=$(EXPORT_SUFFIX)
OUTPUT_FILE_TYPES+=$(ALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(QHG_SUFFIX)
OUTPUT_FILE_TYPES+=$(QALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(REALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(SCORE_SUFFIX)
OUTPUT_FILE_TYPES+=$(RESCORE_SUFFIX)
OUTPUT_FILE_TYPES+=$(SCORE_FILES_SUFFIX)
OUTPUT_FILE_TYPES+=$(RESCORE_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(COV_SUFFIX)
OUTPUT_FILE_TYPES+=$(COV_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(SM_SUFFIX)
OUTPUT_FILE_TYPES+=$(SMT_SUFFIX)
OUTPUT_FILE_TYPES+=$(SMT_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(MONO_SUFFIX)
OUTPUT_FILE_TYPES+=$(PSH_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(PIN_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(CALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(CDIFF_SUFFIX)
OUTPUT_FILE_TYPES+=$(CPREDIFF_SUFFIX)
OUTPUT_FILE_TYPES+=$(CREDIFF_SUFFIX)
OUTPUT_FILE_TYPES+=$(PREALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(CREALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(QREP_SUFFIX)
OUTPUT_FILE_TYPES+=$(QCALREP_SUFFIX)
OUTPUT_FILE_TYPES+=$(SEQPRE_SUFFIX)
OUTPUT_FILE_TYPES+=$(SEQUENCE_SUFFIX)
OUTPUT_FILE_TYPES+=$(FRAG_SUFFIX)
OUTPUT_FILE_TYPES+=$(TAGCOUNT_SUFFIX)
OUTPUT_FILE_TYPES+=$(FILT_SUFFIX)
OUTPUT_FILE_TYPES+=$(ANOMALY_SUFFIX)
OUTPUT_FILE_TYPES+=$(ANOMRAW_SUFFIX)
OUTPUT_FILE_TYPES+=$(REANOMRAW_SUFFIX)
OUTPUT_FILE_TYPES+=$(RAWPAIR_SUFFIX)
OUTPUT_FILE_TYPES+=$(PAIR_SUFFIX)
OUTPUT_FILE_TYPES+=$(GENOMESIZE_SUFFIX)

# Chip-wide HTML views
ERROR_THUMB:=Error.htm
SIGNAL_MEANS:=Signal_Means.txt
PLOT_INTENSITY:=plotIntensity_for_IVC
IVC_THUMB:=IVC.htm
SMT_THUMB:=All.htm
PSH_THUMB:=Hist.htm
PEC_THUMB:=Perfect.htm
PIN_THUMB:=Info.htm
BTT_THUMB:=Tile.htm
JER_THUMB:=Summary.htm
RUN_REPORT:=finished.txt

OUTPUT_FILE_TYPES+=$(ERROR_THUMB)
OUTPUT_FILE_TYPES+=$(SIGNAL_MEANS)
OUTPUT_FILE_TYPES+=$(PLOT_INTENSITY)
OUTPUT_FILE_TYPES+=$(IVC_THUMB)
OUTPUT_FILE_TYPES+=$(SMT_THUMB)
OUTPUT_FILE_TYPES+=$(PSH_THUMB)
OUTPUT_FILE_TYPES+=$(PEC_THUMB)
OUTPUT_FILE_TYPES+=$(PIN_THUMB)
OUTPUT_FILE_TYPES+=$(BTT_THUMB)
OUTPUT_FILE_TYPES+=$(JER_THUMB)
OUTPUT_FILE_TYPES+=$(RUN_REPORT)

# Definitions for ELAND
ELAND_RESULT:=_eland_result.txt
ELAND_RESULT_MULTI:=_eland_multi.txt
ELAND_RESULT_EXTENDED:=_eland_extended.txt
ELAND_QUERY:=_eland_query.txt
ELAND_SORTED:=_sorted.txt
#ELAND_DIR:=$(TOOLS_DIR)/../Eland/
#ELAND_EXEC:=eland_$(READ_LENGTH)
ELAND_EXEC:=eland
#ELAND_PATH:=$(ELAND_DIR)/$(ELAND_EXEC)
ELAND_GENOME:=@ELAND_GENOME@
ELAND_REPEAT:=@ELAND_REPEAT@
ELAND_MULTIPLE_INSTANCES:=@ELAND_MULTIPLE_INSTANCES@
ELAND_MAX_MATCHES:=@ELAND_MAX_MATCHES@

OUTPUT_FILE_TYPES+=$(ELAND_RESULT)
OUTPUT_FILE_TYPES+=$(ELAND_RESULT_MULTI)
OUTPUT_FILE_TYPES+=$(ELAND_RESULT_EXTENDED)
OUTPUT_FILE_TYPES+=$(ELAND_QUERY)
OUTPUT_FILE_TYPES+=$(ELAND_SORTED)

ifeq ($(ELAND_MULTIPLE_INSTANCES),2)
BREAK_DEPENDENCY_LANE_NUMS:=4
endif
ifeq ($(ELAND_MULTIPLE_INSTANCES),4)
BREAK_DEPENDENCY_LANE_NUMS:=2 4 6
endif
ifeq ($(ELAND_MULTIPLE_INSTANCES),8)
BREAK_DEPENDENCY_LANE_NUMS:=1 2 3 4 5 6 7
endif

BREAK_DEPENDENCY:=$(patsubst %,@TILE_ROOT@_%,$(BREAK_DEPENDENCY_LANE_NUMS))


MAX_ELAND_READ_LENGTH:=32
ALLOWED_ELAND_READ_LENGTHS:=15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32


vpath %$(QHG_SUFFIX) $(EXPT_DIR)

# Don't delete any intermediate files
#.SECONDARY:
.PRECIOUS: %$(SMT_SUFFIX) %$(CALIGN_SUFFIX) %$(CREALIGN_SUFFIX) %$(CREDIFF_SUFFIX) %$(CPREDIFF_SUFFIX) %$(RESCORE_SUFFIX) %$(SCORE_SUFFIX) %$(PREALIGN_SUFFIX) %$(REALIGN_SUFFIX) %$(QTABLE_SUFFIX)

all: self_test warn error $(RUN_REPORT) $(JER_THUMB) $(LANE_TARGETS)
# The following line can be used to work around a bug in make-3.81
# which means that the QVAL intermediates do not get removed.
	find . -maxdepth 1 \
            -name "*$(UB_QCAL_SUFFIX)" | xargs $(RM)
	echo Gerald analysis finished successfully. ; @POST_RUN_COMMAND@ 
warn:	
ifneq ($(strip $(WARNING_MESSAGES)),)
	$(warning Warning(s) reported: $(WARNING_MESSAGES))
endif


ifeq ($(strip $(ALL_SEQ_FILES)),)
error_no_seq:	
	$(error Error: no sequence files satisfying $(TILE)$(SEQ_FILE_SUFFIX) were found in EXPT_DIR=$(EXPT_DIR))
ERROR_TARGETS+=error_no_seq
endif

# .. die if lists of seq and sig files are not the same
ifneq ($(strip $(DIFF_TILES_SIG)),)
error_seq_sig_diff:	
	$(error Error: $(TILE)$(SEQ_FILE_SUFFIX) and $(TILE)$(SIG_FILE_SUFFIX)\
	 file lists in EXPT_DIR=$(EXPT_DIR) differ!\
	 Differences=$(DIFF_TILES_SIG))
ERROR_TARGETS+=error_seq_sig_diff
endif

# .. die if lists of seq and prb files are not the same
ifneq ($(strip $(DIFF_TILES_PRB)),)
error_seq_prb_diff:	
	$(error Error: $(TILE)$(SEQ_FILE_SUFFIX) and $(TILE)$(PRB_FILE_SUFFIX)\
	 file lists in EXPT_DIR=$(EXPT_DIR) differ!\
	 Differences=$(DIFF_TILES_PRB))
ERROR_TARGETS+=error_seq_prb_diff
endif

error: $(ERROR_TARGETS)	

$(TILE_FILE):
	echo "$(TILES)" > $(TILE_FILE)

SRF_QCAL:=@SRF_QCAL@
SRF_2ND_CYCLE:=@SRF_2ND_CYCLE@ 

### Analysis specification for each lane follows ###

@ANALYSIS@

### End of lane-specific analysis ###

# If grouping of results by lane is desired (currently for eland_tag only)
# add appropriate functionality
ifdef GROUP_LANES

s_%$(TAGCOUNT_SUFFIX): $(GROUP_PREREQ_LIST)
	echo making $@ from $^
	echo $(patsubst %,s_%$(TAGCOUNT_SUFFIX),$(foreach THIS_LANE,$(patsubst s_%,%,$(LANES)),$(findstring $(THIS_LANE),$*))) 
	cat $(patsubst %,s_%$(TAGCOUNT_SUFFIX),$(foreach THIS_LANE,$(patsubst s_%,%,$(LANES)),$(findstring $(THIS_LANE),$*))) | perl -lane '{ $$"="\t"; shift(@F); $$h{"@F"}+=$$_; } END { for (keys %h) { print "$$h{$$_}\t$$_"}}' | sort -T . -nr > $(SAFEPIPETARGET)

GROUP_TARGETS=$(foreach THIS_GROUP,$(GROUP_LANES),s_$(THIS_GROUP)$(TAGCOUNT_SUFFIX))

JER_PREREQS+=$(GROUP_TARGETS)
endif

# Always add lane count files to target list, whether or not GROUP_LANES is set
JER_PREREQS+=$(GROUP_PREREQ_LIST)



# If the above necessitates some eland analysis, then define
# all the targets necessary to run it
ifdef LAST_ELAND_RESULT
DIRS_TO_CHECK+=$(ELAND_DIR)
FILES_TO_CHECK+=$(ELAND_REPEAT)


# Build Fasta format query from sequence files
#$(ELAND_QUERY): $(patsubst %,$(EXPT_DIR)/%,$(ELAND_PREREQS))
#	cd $(EXPT_DIR); $(TF) $(TF_PARAMS) $(ELAND_PREREQS) > $(CURDIR)/$@

# Run ELAND to generate results
#$(ELAND_RESULT): $(ELAND_QUERY) $(ELAND_EXEC)
#	@echo running eland on  $^
#	$(ELAND_PATH) $(ELAND_QUERY) $(ELAND_GENOME) $(ELAND_RESULT)\
#	$(ELAND_REPEAT)

endif # ELAND_PREREQS

# If the above necessitates some monotemplate analysis, then define
# all the targets necessary to run it
ifdef MONO_PREREQS

# Definitions for monotemplate analysis
MONO_RESULT:=Monotemplate.txt

$(MONO_RESULT): $(MONO_PREREQS)
	cat $^ > $(SAFEPIPETARGET)

PLOT_INTENSITY_MONO:=plotIntensity_monotemplates

$(PLOT_INTENSITY_MONO): $(MONO_RESULT)
	$(cmdprefix) $(PIM) $(MONO_RESULT) $(CURDIR) && echo dummy > $(SAFEPIPETARGET)

MONO_THUMB:=Monotemplate.htm

$(MONO_THUMB): $(PLOT_INTENSITY_MONO)
	$(cmdprefix) $(CMT) $(CURDIR) $(MONO_LANES) > $(SAFEPIPETARGET)

JER_PREREQS+=$(MONO_THUMB)

OUTPUT_FILE_TYPES+=$(MONO_RESULT) $(PLOT_INTENSITY_MONO) $(MONO_THUMB)


endif # MONO_PREREQS

ifdef CONTAM_FILE

PA_PARAMS_CONTAM=-tag $(CONTAM_DIR)/$(CONTAM_FILE) $(READ_LENGTH) $(READ_LENGTH) $(USE_BASES)
 
contam: $(addsuffix $(CDIFF_SUFFIX), $(CONTAM_TILES))

# Add contaminant target (TBD - make this lane-by-lane)
all+=contam

# Alignment to contaminant
%$(CALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)
	$(cmdprefix) $(PIPECMD) $(PA) $(PA_PARAMS_CONTAM) $<\
	| perl -lane '{ print unless (/^\#/) }' > $(SAFEPIPETARGET)

# Alignment to data minus alignment to contaminant - negative for contaminants 
%$(CDIFF_SUFFIX): %$(CALIGN_SUFFIX) %$(ALIGN_SUFFIX)
	$(cmdprefix) $(PIPECMD) cat $*$(ALIGN_SUFFIX) | grep -v '^\#' | cut -d" " -f2 \
	| paste - $*$(CALIGN_SUFFIX)  \
	| perl -lane '{ print $$F[0]-$$F[2] }' > $(SAFEPIPETARGET)

# Alignment to contaminant using alignment scores from first pass
%$(CREALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)\
%$(SCORE_SUFFIX)
	$(cmdprefix) $(PIPECMD) $(PA) $(PA_PARAMS_CONTAM) -scoreFile $*$(SCORE_SUFFIX) $<\
	| perl -lane '{ print unless (/^\#/) }' > $(SAFEPIPETARGET)

# Alignment to data minus alignment to contaminant - realigned data
%$(CPREDIFF_SUFFIX): %$(CREALIGN_SUFFIX) %$(PREALIGN_SUFFIX)
	$(cmdprefix) $(PIPECMD) cat $*$(PREALIGN_SUFFIX) | grep -v '^\#' | cut -d" " -f2 \
	| paste - $*$(CREALIGN_SUFFIX)  \
	| perl -lane '{ print $$F[0]-$$F[2] }' > $(SAFEPIPETARGET)
# Alignment to data minus alignment to contaminant - realigned + filtered data
%$(CREDIFF_SUFFIX): %$(CPREDIFF_SUFFIX) $(QHG_DIR)/%$(QHG_SUFFIX)
	$(cmdprefix) $(PIPECMD) cat $< | $(QF) $(QF_PARAMS) $(QHG_DIR)/$*$(QHG_SUFFIX) > $(SAFEPIPETARGET)

endif



# Macro definition

GRAB_LANE_TILES=$(filter $(1)%, $(2))

#
# Generic rules to build files
#

%$(GENOMESIZE_SUFFIX):
	$(SG) $(ELAND_GENOME) > $(SAFEPIPETARGET)

# Make "first pass" alignments

# moved to GERALD.xml to define on per-lane basis, as make 3.78 won't override
# a global definition with a per-lane definition when needed
#%$(ALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)
#	$(PA) $(PA_PARAMS) $<\
#	> $@

#%$(QTABLE_SUFFIX): %$(QVAL_SUFFIX) $(QLC)
#	$(QLC) $< > $@

#%$(QCAL_SUFFIX): %$(QVAL_SUFFIX) %$(QTABLE_SUFFIX) $(QLC)
#	$(QLA) $(QLA_PARAMS) $*$(QTABLE_SUFFIX) $< | \
#	$(FQV) $(FQV_PARAMS) $(READ_LENGTH) > $@
 
%$(QRAW_SUFFIX): $(EXPT_DIR)/%$(PRB_FILE_SUFFIX) $(CQ)
	$(cmdprefix) $(CQ) $(CQ_PARAMS) $< > $(SAFEPIPETARGET)




# Make scored alignments from quality value files

%$(QALIGN_SUFFIX): $(EXPT_DIR)/%$(PRB_FILE_SUFFIX) $(PA)
	$(cmdprefix) $(PA) $(PA_PARAMS) $<\
	> $(SAFEPIPETARGET)

# Make QC parameter files
%$(QHG_SUFFIX): $(EXPT_DIR)/%$(SIG_FILE_SUFFIX) $(QH)
	$(cmdprefix) $(QH) $(QH_PARAMS) $<\
	> $(SAFEPIPETARGET)


# Realignment on per tile basis

# moved to GERALD.xml to define on per-lane basis, as make 3.78 won't override
# a global definition with a per-lane definition when needed
#%$(PREALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)\
#%$(SCORE_SUFFIX)
#	$(PA) $(PA_PARAMS) -scoreFile $*$(SCORE_SUFFIX) $< > $@

%$(REALIGN_SUFFIX): %$(PREALIGN_SUFFIX) $(QHG_DIR)/%$(QHG_SUFFIX)
	$(cmdprefix) $(PIPECMD) cat $< | $(QF) $(QF_PARAMS) $(QHG_DIR)/$*$(QHG_SUFFIX) > $(SAFEPIPETARGET)

# Make pngs from align, realign and prb files
%$(PSH_PNG_SUFFIX): %$(RESCORE_SUFFIX)
	$(cmdprefix) $(PSH) $@ $*$(ALIGN_SUFFIX) $*$(SCORE_SUFFIX) $*$(REALIGN_SUFFIX) $*$(RESCORE_SUFFIX) $(EXPT_DIR)/$*$(PRB_FILE_SUFFIX) $(USE_BASES_CRITERION)

# Make pngs of information content from align+realign+prb
%$(PIN_PNG_SUFFIX): %$(RESCORE_SUFFIX)
	$(PIN) $@ $*$(ALIGN_SUFFIX) $*$(REALIGN_SUFFIX) \
	$(EXPT_DIR)/$*$(PRB_FILE_SUFFIX) $(USE_BASES_CRITERION)



# Build lane coverage after second alignment
%_cov.txt: %_rescore
	$(cmdprefix) $(PIPECMD) cat $(call GRAB_LANE_TILES, $*, $(GOOD_REALIGN_FILES))\
	| $(BC) $(BC_PARAMS) > $(SAFEPIPETARGET)

all_cov.txt: $(LANES_BEST_ALIGN) $(LANES_COV)
	$(cmdprefix) $(PIPECMD) cat $(LANES_BEST_ALIGN) | $(BC) $(BC_PARAMS) > $(SAFEPIPETARGET)

$(SMT_DIR):
	mkdir $@

#%$(SMT_SUFFIX): $(SMT_DIR)/%$(SMT_SUFFIX) ;

$(SMT_DIR)/%$(SMT_SUFFIX): $(EXPT_DIR)/%$(SIG_FILE_SUFFIX) \
    $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(QHG_DIR)/%$(QHG_SUFFIX) $(REQUIRES_SMT_DIR) 
	$(SMT) $(SMT_PARAMS) $(wordlist 1, 3, $^) > $(SAFEPIPETARGET)


# Make error plots of all rescore files
$(ERROR_THUMB):  $(addsuffix $(RESCORE_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(cmdprefix) $(CE) $(CE_PARAMS) > $(addprefix Full,$@) && \
	$(CE) $(CE_PARAMS) $(MAX_THUMBNAIL_OPTION) --link='<a href="$(addprefix Full,$@)">Full output (Warning: may overload your browser!)</a>' > $(SAFEPIPETARGET)

# Prerequisites for SIGNAL_MEANS
$(SIGNAL_MEANS): $(SM_PREREQS)
	cat $^ > $(SAFEPIPETARGET)

# Run plotIntensity_for_IVC.pl
$(PLOT_INTENSITY): $(SIGNAL_MEANS) 
	$(PI) $(SIGNAL_MEANS) $(CURDIR) && echo dummy > $@

# Run create_IVC_thumbnail.pl
$(IVC_THUMB): $(PLOT_INTENSITY)
	$(CI) $(CURDIR) > $(SAFEPIPETARGET)


# Build thumbnail of Signal_Means_Tile.pl output
$(SMT_THUMB) : $(addsuffix $(SMT_PNG_SUFFIX), $(SMT_THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_SMT) > $(addprefix Full,$@) && \
	$(CE) $(CE_PARAMS_SMT)  $(MAX_THUMBNAIL_OPTION) --link='<a href="$(addprefix Full,$@)">Full output (Warning: may overload your browser!)</a>' > $(SAFEPIPETARGET)

# Build thumbnail of sequence quality histograms
$(PSH_THUMB) : $(addsuffix $(PSH_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_PSH) > $(addprefix Full,$@) && \
	$(CE) $(CE_PARAMS_PSH)  $(MAX_THUMBNAIL_OPTION) --link='<a href="$(addprefix Full,$@)">Full output (Warning: may overload your browser!)</a>' > $(SAFEPIPETARGET)

# Build thumbnail of sequence quality histograms
$(PEC_THUMB) : $(addsuffix $(PEC_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_PEC) > $(addprefix Full,$@) && \
	$(CE) $(CE_PARAMS_PEC)  $(MAX_THUMBNAIL_OPTION) --link='<a href="$(addprefix Full,$@)">Full output (Warning: may overload your browser!)</a>' > $(SAFEPIPETARGET)

# Build thumbnail of information content histograms
$(PIN_THUMB) : $(addsuffix $(PIN_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_PIN) > $(addprefix Full,$@) && \
	$(CE) $(CE_PARAMS_PIN)  $(MAX_THUMBNAIL_OPTION) --link='<a href="$(addprefix Full,$@)">Full output (Warning: may overload your browser!)</a>' > $(SAFEPIPETARGET)



s_1: $(s_1_TARGET_LIST)
	touch s_1_finished.txt
s_2: $(s_2_TARGET_LIST)
	touch s_2_finished.txt
s_3: $(s_3_TARGET_LIST)
	touch s_3_finished.txt
s_4: $(s_4_TARGET_LIST)
	touch s_4_finished.txt
s_5: $(s_5_TARGET_LIST)
	touch s_5_finished.txt
s_6: $(s_6_TARGET_LIST)
	touch s_6_finished.txt
s_7: $(s_7_TARGET_LIST)
	touch s_7_finished.txt
s_8: $(s_8_TARGET_LIST)
	touch s_8_finished.txt


# Build by_tile thumbnail view of whatever is there
JER_PREREQS+= $(IVC_THUMB) $(SMT_THUMB) $(LANE_TARGETS)
$(JER_THUMB): $(JER_PREREQS)
	$(JER) $(CURDIR) > $(SAFEPIPETARGET)

# Commented version will generate _hist.png, _info.png, Hist.htm, Info.htm.
# $(BTT_THUMB) : $(ERROR_THUMB) $(IVC_THUMB) $(SMT_THUMB) $(PSH_THUMB) $(PEC_THUMB) $(PIN_THUMB) $(JER_THUMB)
$(BTT_THUMB) : $(ERROR_THUMB) $(IVC_THUMB) $(SMT_THUMB) $(PEC_THUMB) $(JER_THUMB)
	$(BTT) $(CURDIR) > $(SAFEPIPETARGET)

$(RUN_REPORT): $(BTT_THUMB)
	$(cmdprefix) $(RR) $(RR_PARAMS) && echo dummy > $@



# target for debugging
s_%_self_test:
	$(cmdprefix) $(QUIET_PIPECMD) echo $(DIRS_TO_CHECK) | perl -lane \
	'{for (@F) {die "\nERROR: did not find dir $$_\n\n" unless (-d $$_)}}'
	$(QUIET_PIPECMD) echo $(FILES_TO_CHECK) | perl -lane \
	'{for (@F) {die "\nERROR: did not find file $$_\n\n" unless (-e $$_)}}'
	$(QUIET_PIPECMD) echo $(EXECS_TO_CHECK) | perl -lane \
	'{for (@F) {die "\nERROR: file $$_ not executable\n\n" unless(-x $$_)}}'
self_test: $(patsubst %,%_self_test,$(LANES))



qhg: $(ALL_QHG_FILES)


qalign: $(LANES:%=%_qalign)

rescore: $(LANES:%=%_rescore)

lra: $(ALL_LANES_RA) summary_lra.txt plot_calls 

tidy:
	find . -empty -exec rm {} \;

clean_qcal:
	find . -name "*$(QRAW_SUFFIX)" -or -name "*$(QTABLE_SUFFIX)" -or -name "*$(QCAL_SUFFIX)" -or -name  "*$(QREP_SUFFIX)" -or -name  "*$(QCALREP_SUFFIX)" | xargs $(RM) 

clean0:
	find . -maxdepth 1 -name "*.png" -or -name "*_coverage" | xargs $(RM)
	find . -maxdepth 1 -name "core" $(addprefix -or -name "*, $(addsuffix ", $(OUTPUT_FILE_TYPES)))  | xargs $(RM)
	find . -maxdepth 1 -name "*.tmp"  | xargs $(RM)
	find . -maxdepth 1 -name "*$(UB_QCAL_SUFFIX)" -or -name "$(TILE_FILE)" | xargs $(RM)
	find . -maxdepth 1 -name "*$(QVAL_ACTUAL_SUFFIX)" | xargs $(RM)
	-$(RM) core core.*

clean: clean_qcal clean0
	@echo Finished cleaning.

clean_intermediate:
	@echo Nothing to be cleaned.

recursive_msg:
	@echo No target \'recursive\' - using \'all\' instead.

recursive: recursive_msg all

compress:
	gzip *.txt

uncompress:
	gunzip *.txt.gz

# End of generated Makefile
