#!/usr/bin/env perl

# PROJECT: GERALD
# MODULE:  extractQualityParamsFromSAF.pl
# AUTHOR:  A. J. Cox
#
# Copyright (c) 2006 Solexa
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Takes a SAF format alignment file and extracts the parameters necessary
# for Phred-style recalibration
# Based on Klaus' extractQualityParams.py script in the QCalibration
# module. Placed in this module because it uses Common::parseUseBases

# Usage: extractQualityParamsFromSAF.pl safFile qrawFile useBasesString



use warnings;
use strict;
use File::Basename;

use FindBin;
use lib "$FindBin::Bin/../";
use Gerald::Common;


my $version;
my $id;
{
    my $cvs_id='@(#) $Id: extractQualityParamsFromSAF.pl,v 1.3 2007/10/04 09:50:35 km Exp $';
    my $cvs_version='$Revision: 1.3 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope


my $usage=
"Usage: paste qrawFile safFile | $0 [--useBases useBases]  [--read[1|2]] readLength exptDir sigSuffix";


my $useBases=undef;
my $readPairMode=0;
# $readPairMode==0 (default) denotes no pairing - treat 'Y' and 'y' the same
# $readPairMode==1 denotes pair1 - ignore 'y' bases
# $readPairMode==2 denotes pair2 - ignore 'Y' bases
my $isNumeric=0;

die $usage unless (@ARGV>=2);

if ($ARGV[0] eq '--useBases')
{
    shift (@ARGV);
    die $usage unless (@ARGV>=3);
    $useBases=shift(@ARGV); 
} # if

if ($ARGV[0]=~/^\-\-read([012])$/)
{
    $readPairMode=$1;
    shift (@ARGV);
} # if


if (($ARGV[0] eq '--numeric')||($ARGV[0] eq '--symbolic'))
{
    $isNumeric=1 if ($ARGV[0] eq '--numeric');
    shift (@ARGV);
}




die $usage unless (@ARGV==3);

#my $readLength=shift(@ARGV);
my ($readLength, $exptDir, $sigSuffix)=@ARGV;

my @baseIndex;

unless (defined($useBases))
{
    $useBases='all';
}

my $i;
my $j;
my $qv;

my $PURE_BASES=12;
my $fileName;

sub max
{
    my $max=$_[0]->[0];
    my $i=0;
    for (@{$_[0]})
    {
        $max=$_ if ($_>$max);
	last if (++$i==$PURE_BASES);
    } # for
    return $max;
} # sub max

sub changeToNextTile
{
    die unless (@_==4);
    my ($lane, $tile, $dir, $suffix)=@_;

#    close ($fh) if (defined($fh));
    
#    $fh=undef;
    close(LINE);
    $fileName=sprintf("%s/s_%1d_%04d%s",$dir,$lane,$tile,$suffix);
    print STDERR "About to open file $fileName\n";
    open (LINE, "$fileName") 
	|| die "Failed to open file $fileName for reading: $!";
} # sub changeToNextTile

my $lane =-1;
my $tile =-1;


my $callNone=".\t.\n";
my $callRight="A\tA\n";
my $callWrong="A\tC\n";

@baseIndex=parseUseBases($useBases, $readLength, $readPairMode);
die unless (@baseIndex==$readLength);

my @purity;
my @signals;
my $maxPurity;

#print scalar(@baseIndex), " $baseIndex[0] $baseIndex[-1]\n"; die;
my @s;
my @s1;
my ($qstring, $readID, $chrom, $pos, $strand, $type);
my $l;
my $expand;
my $call;

while (<STDIN>)
{
    
    ($qstring, $readID, $chrom, $pos, $strand, $type) = split;
    die unless (length($qstring)==$readLength);
    @s=split(':',$readID); # should be folder:lane:tile:x:y[ab]

    if (defined($type))
    {
	if ($type eq '-') 
	{
	    $type=undef;
	} # if
	else
	{
	    while ($type=~/(\d+)/)
	    {
		$expand = '.' x $1;
		$type=~s/$1/$expand/;
	    }
	    die "Could not expand type from $_" 
		unless (length($type)==$readLength);    
	} # else
#	print "Expanded type to $type\n";
    } # if


    if (($lane ne $s[-4])||($tile ne $s[-3]))
    {
        $lane=$s[-4];
	$tile=$s[-3];
        changeToNextTile($lane, $tile, $exptDir, $sigSuffix);
    } # if

    $l=<LINE>;
    @s1=split('\t',$l);
    die unless ($s1[0]==$lane);
    die unless ($s1[1]==$tile);
    for ($i=0;$i<scalar(@baseIndex);$i++)
    {
	@signals=split(' ',$s1[4+$baseIndex[$i]]);
	unless (@signals==4)
	{
	    print STDERR basename($0), ": Error parsing line: $_";
	    print STDERR basename($0),
	    ": can't find quality values for cycle ",
	    $baseIndex[$i]+1, " in file ${fileName} "; 
	    if (defined($s1[$baseIndex[$i]]))
	    {
		print STDERR
		    "- could not parse field $s1[$baseIndex[$i+4]]\n";
	    } # if
	    else
	    {
		print STDERR 
		    "- only ",scalar(@s1), 
		    " cycles of quality values present on line\n";
	    } # else			
	    exit (-1);
	} # unless
# Original from extractQualityParams.py
#            signals.sort()
#            sum1 = (signals[-1]+signals[-2])
#            if ((signals[-1] <= 0.) or (sum1 <= 0.)):
#                purity[j] = 1.
#            else:
#                purity[j] = 1. - signals[-1]/sum1
	@signals=sort {$a<=>$b} @signals;
	my $sum = ($signals[-1]+$signals[-2]);
	if (($signals[-1]<=0.) || ($sum <= 0.))
	{
	    $purity[$i]=1.0;
	} # if
	else
	{
	    $purity[$i]=1.0 - $signals[-1]/($sum);
	} # else

	
#	$qv=max(\@s3);
#	$qstring.=(($isNumeric)?" $qv":chr(int(64+$qv)));
    } # for i

    $maxPurity=max(\@purity);
    for ($i=0;$i<scalar(@baseIndex);$i++)
    {
	printf "%d\t%4.2f\t%4.2f\t%d\t",$i+1,
	$purity[$i],
	$maxPurity,
	-getQualFromChar(substr($qstring,$i,1));
	unless (defined($type))
	{
	    print $callNone;
	} # unless
	else
	{
	    $call=substr($type,$i,1);
	    if ($call eq '.')
	    {
		print $callRight;
	    } # if
	    elsif ($call=~/[ACGT]/)
	    {
		print $callWrong;
	    } # elsif
	    else
	    {
		print $callNone;
	    } # else
	} # else
    } # for
} # while (<STDIN>)
