#!/usr/bin/env perl

# File: qualityFilter.pl
# Copyright (c) 2005, 2006 Solexa
# Author: A. J. Cox
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Description: takes as input 
# i) a set of one-line-per-sequence data files - piped via STDIN
# ii) a quality criterion 
# iii) and a set of quality files

# Filters the data files by applying the filter criterion to the
# quality files (which can be sequences, intensities, alignments)
# and outputting only the data file lines that correspond to those lines
# that pass. These are sent to STDOUT

# Obviously the data files need to be piped to STDIN in the same order
# as the quality files are specified on the command line
# Specifying them with the same wildcard ought to ensure this, e.g.
# cat s_1_????_align.txt 
# | qualityFilter '(X_COORD>20)&&(X_COORD<1300)' s_1_????_qhg.txt
# > output.txt

# Quality file lines are split by field into array @F (as in Perl one liners)
# Quality criterion can refer to these directly or via aliases
# LANE for $F[0]
# TILE for $F[1]
# X_COORD for $F[2]
# Y_COORD for $F[3]
# PURITY for $F[4]
# CHASTITY for $F[5]
# SIMILARITY for $F[6]
# NEIGHBOUR for $F[7]
# NHOOD for $F[8]

# This quality file line format is as output by QUAHOG.cpp version 1.3 
use warnings;
use strict;

use FindBin;
use lib "$FindBin::Bin/../";
use Gerald::Common;

#------------------------------------------------------------------------------

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: qualityFilter.pl,v 1.8 2007/11/26 14:42:26 tc Exp $';
    my $cvs_version='$Revision: 1.8 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

conditionally_init_sigpipe_handler();

my %alias=
    ( "LANE" => "\$F[0]",
      "TILE" => "\$F[1]",
      "X_COORD" => "\$F[2]",
      "Y_COORD" => "\$F[3]",
      "PURITY" => "\$F[4]",
      "CHASTITY" => "\$F[5]",
      "SIMILARITY" => "\$F[6]",
      "NEIGHBOUR" => "\$F[7]",
      "NHOOD" => "\$F[8]"   );


my $thisFile;
my $qcLine;
my $dataLine;
my $printIt;
my $doStamp=undef;
my $appendStats=undef;
my $numPassed=0;
my $numRead=0;

sub printFilteredLinesOnly
{
	if ($printIt)
	{
	    $numPassed++;
	    print $dataLine; 
	}
} # printFilteredLinesOnly

sub printWhetherPassed
{
    if ($printIt)
    {
	$numPassed++;
	print "Y\n";
    } # if
    else
    {
	print "N\n";
    } # else
} # printWhetherPassed

my $printLine=\&printFilteredLinesOnly;

if (@ARGV>=1)
{
    if ($ARGV[0] eq "-stamp")
    {
	print "#RUN_TIME ". scalar(localtime)."\n";
	print "#SOFTWARE_VERSION $id\n";
	$doStamp=1;
	shift(@ARGV);
    }
    elsif ($ARGV[0] eq "-stats")
    {
	$appendStats = 1;
	shift(@ARGV);
    }
} #if

if (@ARGV>=1)
{
    if ($ARGV[0] eq '-passed')
    {
	$printLine=\&printWhetherPassed;
	shift(@ARGV);
    } # if
} # if


unless (@ARGV>=2)
{
    print STDERR "Usage: cat dataFiles.txt | $0 [-stamp | -stats] [-passed] ";
    print STDERR "'QC criterion' qcFiles.txt\n";
    print STDERR "Allowable QC parameters:\n";
    foreach my $thisAlias( sort keys %alias )
    {
        print "$thisAlias\n";
    } #foreach
    die;
} #unless

my $qcCrit=shift(@ARGV);



my @F;


print STDERR "Will use quality criterion $qcCrit\n";

print "#FILTER_CRITERION $qcCrit\n" if (defined($doStamp));

foreach my $thisAlias( keys(%alias))
{
    $qcCrit=~s/$thisAlias/$alias{$thisAlias}/g;
} #foreach

#$qcCrit=~s/LANE/\$F[0]/g;
#$qcCrit=~s/TILE/\$F[1]/g;
#$qcCrit=~s/X_COORD/\$F[2]/g;
#$qcCrit=~s/Y_COORD/\$F[3]/g;
#$qcCrit=~s/PURITY/\$F[4]/g;
#$qcCrit=~s/CHASTITY/\$F[5]/g;
#$qcCrit=~s/SIMILARITY/\$F[6]/g;
#$qcCrit=~s/NEIGHBOUR/\$F[7]/g;
#$qcCrit=~s/NHOOD/\$F[8]/g;

print STDERR "Quality criterion translated to $qcCrit\n";



while ($thisFile=shift(@ARGV))
{
    open (QC, $thisFile ) 
        || die "Error: problem opening QC file $thisFile: $!";
    print STDERR "opened $thisFile\n";
    while (defined($qcLine=<QC>))
    {
        if ($qcLine=~/^\#/)
        {
            print $qcLine if defined($doStamp);
            next;
        } #if
#        print STDERR "qc line $qcLine\n";
        do
        {
            $dataLine=<STDIN>;
            die "Error: more lines of QC info than data\n"
                unless (defined ($dataLine));
           print $dataLine if (defined($doStamp)&&($dataLine=~/^\#/));
        }
        while ($dataLine=~/^\#/);
 #       print STDERR "dt line $dataLine\n";
	$numRead++;

        chomp ($qcLine);
        @F=split('\t', $qcLine);
        @F=split(/\s/,$qcLine) if (@F==1);
        
        $printIt=eval($qcCrit);
        die "Error: problem evaluating $qcCrit on $qcLine: $@"
            unless (defined($printIt));
 #       print STDERR "$printIt\n";

	$printLine->();
#	if ($printIt)
#	{
#	    $numPassed++;
#	    print $dataLine; 
#	}
    } # while $qcLine
    close (QC);
} # while $thisFile

if (defined($appendStats))
{
    print "# PASSED $numPassed OF $numRead\n";
}

if ($numPassed!=0)
{
    print STDERR 
	"$numPassed sequences out of $numRead passed filter criteria\n";
}
else 
{
    warn "Warning: none of $numRead data sequences passed filter criteria '$qcCrit'";
}

conditionally_report_sigpipes();

#------------------------------------------------------------------------------

