#!/usr/bin/env perl

# PROJECT: GERALD
# MODULE:  reportQualitySAF.pl 
# AUTHOR:  Richard Shaw 
#
# Copyright (c) 2007 Solexa
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# A version of reportQuality.pl for use on _saf + _qraw files rather than
# _align + _qraw files.
#
# Filtering values are pasted in from a _filt.txt file (assumed available).
#
# As score.pl now extracts the corresponding statistics from export files
# for the calibrated results and the uncalibrated results are expected to 
# be discontinued soon, this script is intended to be temporary and basic.

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: reportQualitySAF.pl,v 1.6 2008/04/23 14:36:17 rshaw Exp $';
    my $cvs_version='$Rev$';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

use warnings;
use strict;
use File::Basename;

use FindBin;
use lib "$FindBin::Bin/../";
use Gerald::Common;

#------------------------------------------------------------------------------

# Filter type strings
my $raw_str = 'raw';
my $QF_str = 'QF';

#------------------------------------------------------------------------------

sub read_input($$$$$$$$$)
{
    my $qual_val_freqs_ref = shift;
    my $errs_by_qual_val_ref = shift;
    my $lane_num_ref = shift;
    my $tile_freqs_ref = shift;

    my $seqpre_filename = shift;
    my $saf_filename = shift;
    my $qraw_filename = shift;
    my $filt_filename = shift;
    my $one_of_multiple_reads = shift;

    my $pipe = 
	"paste $seqpre_filename $saf_filename $qraw_filename $filt_filename |";

    my $aligned_num_saf_fields = ($one_of_multiple_reads ? 9 : 6);

    my $read_seq_field_ind = 1; # seqpre
    my $id_field_ind = 2; # SAF
    my $strand_dir_field_ind = 5; # SAF
    my $mismatch_field_ind = 6; # SAF

    # 2 from seqpre + 1 each from qraw, filt
    my $aligned_num_fields = $aligned_num_saf_fields + 4;

    my $base_qual_field_ind = $aligned_num_saf_fields + 2; # qraw
    my $filtered_field_ind = $aligned_num_saf_fields + 3; # filt

    open(PIPE, $pipe) || die "Failed to open $pipe";

    my $line_num = 0;
    my $num_cycles = 0;
    my ($run_id, $lane_num, $tile_num) = ('', 0, 0);

    while (<PIPE>) {
        ++$line_num;

        chomp;
        my @fields = split('\t', $_);
        next if (@fields < $aligned_num_fields);

        my $id_str = $fields[$id_field_ind];
        # Could do checking, e.g. that lane is same for all reads.
        ($run_id, $lane_num, $tile_num) = split(/:/, $id_str);
        ++$tile_freqs_ref->{$tile_num};

        my @qv_data_types = ($raw_str);

        if ($fields[$filtered_field_ind] eq 'Y') {
            push(@qv_data_types, $QF_str);
        }

        my $mismatch_str = $fields[$mismatch_field_ind];
        next if ($mismatch_str eq '-');

	my $read_seq = $fields[$read_seq_field_ind];
	my $num_cycles = length($read_seq);

	my $ref_seq = apply_mismatch($read_seq, $num_cycles,
                                      $mismatch_str, $line_num);

        my $base_qual_seq = $fields[$base_qual_field_ind];

        if (length($base_qual_seq) != $num_cycles) {
            my $base_qual_seq_len = length($base_qual_seq);
            die("Length ($base_qual_seq_len) of base quality string "
                . "($base_qual_seq) at line $line_num "
                . "does not match alignment descriptor ($mismatch_str)");
        }

        for (my $cycle_ind = 0; $cycle_ind < $num_cycles; ++$cycle_ind) {
	    my $read_base = substr($read_seq, $cycle_ind, 1);
	    my $ref_base = substr($ref_seq, $cycle_ind, 1);

	    # Ignore any bases that are N in read or ref.
	    # Unlike for a score file, there is no requirement to count
	    # uncalled bases for a quality report.
	    next if (($read_base eq 'N') || ($ref_base eq 'N'));

            my $base_qual_char = substr($base_qual_seq, $cycle_ind, 1);
            my $base_qual_val = ord($base_qual_char) - 64;

            foreach my $data_type (@qv_data_types) {
                if (!defined($qual_val_freqs_ref->
                             {$data_type}{$base_qual_val})) {
                    $qual_val_freqs_ref->{$data_type}{$base_qual_val} = 0;
                }

                ++$qual_val_freqs_ref->{$data_type}{$base_qual_val};

                if (!defined($errs_by_qual_val_ref->
                             {$data_type}{$base_qual_val})) {
                    $errs_by_qual_val_ref->{$data_type}{$base_qual_val} = 0;
                }

		# Ns are excluded above, so this is strictly a miss-call test.
                if ($read_base ne $ref_base) {
                    ++$errs_by_qual_val_ref->{$data_type}{$base_qual_val};
                }
            }
        }
    }

    close(PIPE);

    $$lane_num_ref = $lane_num;
}

#------------------------------------------------------------------------------

sub write_quality_line($$$$$$)
{
    my $qual_val = shift;
    my $curr_err_count = shift;
    my $curr_freq = shift;
    my $cum_err_count = shift;
    my $cum_freq = shift;
    my $num_bases = shift;

    # % error expected if the quality values are 100% accurate
    my $expected_err = 100 / (1 + (10 ** ($qual_val / 10)));

    printf("%d\t%d\t%d\t%5.2f\t%5.2f\t|\t%d\t%d\t%5.2f\t%5.2f\t|\t%5.3f\n",
	   $qual_val,
	   $cum_err_count,
	   $cum_freq,
	   100 * $cum_err_count / $cum_freq,
	   100.0 * $cum_freq / $num_bases, 
	   $curr_err_count,
	   $curr_freq,
	   100 * $curr_err_count / $curr_freq,
	   100 * $curr_freq / $num_bases,
	   $expected_err);
}

#------------------------------------------------------------------------------

sub write_quality_report($$$$)
{
    my $qual_val_freqs_ref = shift;
    my $errs_by_qual_val_ref = shift;
    my $lane_num = shift;
    my $tile_freqs_ref = shift;

    my @tile_nums = sort { $a <=> $b } keys %{$tile_freqs_ref};
    my @tile_names = map { sprintf("s_%d_%04d", $lane_num, $_); } @tile_nums;

    my $tile_list = '';
    my $num_tiles = 0;

    foreach my $tile_name (@tile_names) {
        $tile_list .= $tile_name;
        $tile_list .= (((++$num_tiles % 8) == 0) ? "\n" : " ");
    }

    my @qv_data_types = ($raw_str, $QF_str);

    print "#RUN_TIME ". scalar(localtime)."\n";
    print "#SOFTWARE_VERSION $id\n";

    foreach my $data_type (@qv_data_types) {
	if ($data_type eq $raw_str) {
	    print " ** Quality value report **\n\n";
	    print "Tiles analysed:\n", $tile_list;
	    print "\n$num_tiles tiles in all\n\n"; 
	    print " ** Results for raw data **\n"; 
	} else {
	    print "\n\n\nResults for quality filtered data\n"; 
	}

	print "-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n";
	print "QV\tthis QV or higher\t\t|\tthis QV only\t\t\t|\texpected\n";
	print "\t# error\t# bases\t% error\t% total\t|"
	    . "\t# error\t# bases\t% error\t% total\t|\t% error\n";
	print "-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\t-\n";

	my $cum_err_count = 0;
	my $cum_freq = 0;
	my $num_bases = 0;

	for (keys %{$qual_val_freqs_ref->{$data_type}}) {
	    $num_bases += $qual_val_freqs_ref->{$data_type}{$_};
	}

	for (sort {$b <=> $a} keys %{$qual_val_freqs_ref->{$data_type}}) {
	    $errs_by_qual_val_ref->{$data_type}{$_} = 0
		unless defined($errs_by_qual_val_ref->{$data_type}{$_});

	    my $curr_err_count = $errs_by_qual_val_ref->{$data_type}{$_};
	    $cum_err_count += $curr_err_count;

	    my $curr_freq = $qual_val_freqs_ref->{$data_type}{$_};
	    $cum_freq += $curr_freq;

	    write_quality_line($_,
			       $curr_err_count, $curr_freq,
			       $cum_err_count, $cum_freq,
			       $num_bases);
	}
    }
}

#------------------------------------------------------------------------------

my $usage="Usage: $0 seqpreFile safFile qrawFile filtFile";

die $usage unless (@ARGV == 4);

my $seqpre_filename = $ARGV[0];
my $saf_filename = $ARGV[1];
my $qraw_filename = $ARGV[2];
my $filt_filename = $ARGV[3];

my $one_of_multiple_reads = 0;

unless (basename($saf_filename) =~ /s_([1-8])_.*\.txt$/) {
    die "Could not parse SAF file name $saf_filename";
}

my $expected_lane_num = $1;

if (basename($saf_filename) =~ /s_[1-8]_\d+_.*\.txt$/) {
    $one_of_multiple_reads = 1;
}

my %qual_val_freqs;
my %errs_by_qual_val;

my $lane_num = 0;
my %tile_freqs;

read_input(\%qual_val_freqs, \%errs_by_qual_val,
           \$lane_num, \%tile_freqs,
           $seqpre_filename, $saf_filename, $qraw_filename, $filt_filename,
           $one_of_multiple_reads);

if ($lane_num != $expected_lane_num) {
    print(STDERR "From SAF file name $saf_filename expected lane "
          . "$expected_lane_num but found lane $lane_num\n");
}

write_quality_report(\%qual_val_freqs, \%errs_by_qual_val,
                     $lane_num, \%tile_freqs);

#------------------------------------------------------------------------------

