#!/usr/bin/env perl
# File: runReport.pl
# Copyright (c) 2006 Solexa
# Author: A. J. Cox
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Upon completion of a run, send an email of links to "interested parties"

#Usage: runReport.pl [geraldDir|--test] 
#server[:port] domain webRoot [--strip fieldsToStrip] addressList
#
#geraldDir  : GERALD directory about to which to send summary email
#             or --test to send test email (in which case webRoot is ignored)
#server     : name or IP address of SMTP server 
#port       : port number for SMTP services (usually 25 - defaults to this)
#domain     : domain name (e.g. my.organisation.org)
#webRoot    : root of URL (e.g. http://myWebServer/myWebShare/)	
#--strip    : strip fields from file path of geraldDir before appending
#             to webRoot to create URL, e.g. 2 for /mnt/mountPoint (default 2)
#addressList: list of email addresses to send report to (space separated)

# Email is sent by talking to an SMTP server via the SMTP services port, 
# which is traditionally port 25.

# You can test if port 25 of yourserver will be a willing participant in this
# escapade by either:
# telnet yourserver 25 - you should get a friendly message
# nmap -p 25 yourserver (if you have nmap installed)

# Assuming this works run:
# runReport.pl --test yourserver:25 yourdomain anything youremail@yourdomain
# to get a test email

# The protocol used here works on the several different SMTP servers at Solexa.
# However protocols may differ slightly. Try telnetting to the SMTP server and
# typing the commands sent in the 'sendMessage' subroutine manually.
# Unfortunately we can't get involved in supporting/debugging different 
# SMTP protocols - talk to your systems people! 



use warnings;
use strict;
use File::Basename;
use File::Spec;
use IO::Socket;

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: runReport.pl,v 1.11 2007/02/27 10:47:36 km Exp $';
    my $cvs_version='$Revision: 1.11 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

unless ((@ARGV>=5)||((@ARGV>=3)&&($ARGV[0] eq '--test'))) 
{
    print STDERR << "END";
Usage: $0 [geraldDir|-test] 
server[:port] domain webRoot [--strip fieldsToStrip] addressList

geraldDir  : GERALD directory about to which to send summary email
             or --test to send test email (in which case webRoot is ignored)
server     : name or IP address of SMTP server 
port       : port number for SMTP services (usually 25 - defaults to this)
domain     : domain name (e.g. my.organisation.org)
webRoot    : root of URL (e.g. http://myWebServer/myWebShare/)	
--strip    : strip fields from file path of geraldDir before appending
             to webRoot to create URL, e.g. 2 for /mnt/mountPoint (default 2)
addressList: list of email addresses to send report to (space separated)
END

exit(0);
}

my @message;
my @transcript;

my @s;

my $curDir=shift(@ARGV);
my $serverName=shift(@ARGV);
my $domainName=shift(@ARGV);
my $serverPort=25; # usual port for SMTP service
my $maxServerWait=20; # wait this long for a connection, then die

my $socket;

@s=split(':',$serverName);
($serverName,$serverPort)=@s if (@s==2);

my $host=$ENV{HOST};
$host=$ENV{HOSTNAME} unless (defined($host));
$host="unknown_host" unless (defined($host));


sub sendLine
{
    my ($lineRef)=@_;
    # The "\r\n" is vital when talking to Windows SMTP servers
    # It may be the case that Linux/UNIX SMTP servers need "\n" only
    print $socket $lineRef. "\r\n";
    push @transcript, "> $lineRef\n"; 
} # sendLine

sub getResponse
{
    my $answer=<$socket>;
    push @transcript, "< $answer";
    #  All error codes are in range 500-599
    if ($answer=~/^5[0-9][0-9]/)
    {
	print STDERR "$_" foreach (@transcript);
	warn "Error in SMTP protocol, report not sent";
	exit (0);
   } # if
} # getResponse


sub sendMessage
{
    my $answer;
    unless ($socket=IO::Socket::INET->new
	    ( PeerAddr => $serverName,
	      PeerPort => $serverPort,
	      Proto => 'tcp',
	      Type  => SOCK_STREAM,
	      Timeout => $maxServerWait)) 
    { 
	warn "Could not connect to $serverName:$serverPort - $!"; 
	exit (0); 
    } # OR

    # Some servers demand a domain name here. Those that don't don't seem
    # to mind if one is there.
    sendLine("helo ${domainName}");

    getResponse();

    # Our servers demand that domainName is our domain name. Others might
    # be less fussy.
    sendLine ("mail from: <${host}\@${domainName}>");
    
    getResponse();

    # Our main email server only allows email to be sent to addresses within
    # our domain name
    foreach (@ARGV)
    {
	sendLine("rcpt to: <${_}>");
	getResponse();
        unshift @message, "To: Gerald Reports <${_}>";
    } # foreach

    sendLine('data');

    getResponse();

    foreach (@message)
    {
	sendLine($_);
    } # foreach

    sendLine ('.');

    getResponse();

    sendLine ('quit');

    getResponse();

    print "SMTP protocol completed, transcript follows:\n";
    foreach (@transcript)
    {
	print "$_";
    } # for
} # sub sendMessage

# start of main loop

push @message, "Report sent from $host via $serverName port $serverPort at ".scalar(localtime);

my $webRoot=shift(@ARGV);

if ($curDir eq '--test')
{
    unshift @message, "Subject: Test message";
    sendMessage();
    exit (0);
} # if


if ($webRoot !~ /\/$/)
{
    $webRoot.='/';
} # if

my $fieldsToStrip=2; # strip this many fields from file path

if ($ARGV[0] eq '--strip')
{
    shift (@ARGV);
    $fieldsToStrip=shift(@ARGV);
} # if

my $file;


my @filesToLink=qw/Error.htm IVC.htm All.htm Perfect.htm Hist.htm Prb.htm Info.htm Summary.htm Monotemplate.htm/;

my $fullDirName=File::Spec->rel2abs($curDir);


# Get rid of any Windows '\' foolishness
$fullDirName=~s/\\/\//g;

my $runFolderPath;
my $runFolderURL;
my $geraldFolderURL;
my $geraldFolderPath;

my @splitGeraldFolder=split('/',$fullDirName);
my @splitRunFolder=@splitGeraldFolder;

pop @splitRunFolder; # Bustard
pop @splitRunFolder; # Firecrest
pop @splitRunFolder; # Data
pop @splitRunFolder; # Run folder

$"='/';

$runFolderPath="@splitRunFolder";
$geraldFolderPath="@splitGeraldFolder";



# Need to do <= because the leading '/' gives an extra null field when split
for (my $i=0; $i<=$fieldsToStrip; $i++)
{
    shift @splitRunFolder;
    shift @splitGeraldFolder;
} # for


$runFolderURL="${webRoot}@splitRunFolder";
$geraldFolderURL="${webRoot}@splitGeraldFolder";

# TBD Ensure only one '/' character at end of webRoot

unshift @message, "Subject: GERALD run completed in @splitGeraldFolder";

my $filePath;

my @xls=glob ("$runFolderPath/*.xls");

if (@xls!=0)    
{
    push @message , "";
    push @message, "Run folder summary spreadsheets:";
    
    foreach $file (@xls)
    {
	$file=basename($file);
	$filePath="$runFolderURL/$file";
	push @message, "";
	push @message, "$filePath";
    } # foreach
} # if

push @message, "";
push @message, "Links to output files:";


foreach $file (@filesToLink)
{
    $filePath=File::Spec->catfile($curDir,$file);

    
    if (-e $filePath)
    {
        push @message , "";
        push @message , "${geraldFolderURL}/${file}";
        push @message , "";
    } # if


} # foreach

sendMessage();

# end of runReport.pl
