#!/usr/bin/env perl

# PROJECT: GERALD
# MODULE:  saf2qref.pl
# AUTHOR:  R. J. Shaw
#
# Copyright (c) 2008 Illumina
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Extracts reference bases from a SAF file by applying the alignment
# descriptor to the read base sequence.

use warnings;
use strict;

use FindBin;
use lib "$FindBin::Bin/../";
use Gerald::Common;


my $version;
my $id;
{
    my $cvs_id='$Id: saf2qref.pl,v 1.5 2008/04/22 11:05:17 rshaw Exp $';
    my $cvs_version='$Revision: 1.5 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

#------------------------------------------------------------------------------

sub get_read_use_bases_str($$$$)
{
    my $use_bases_str = shift;
    my $orig_read_lengths_str = shift;
    my $read_num = shift;
    my $read_use_bases_str_ref = shift;

    my @orig_read_lengths = split(/:/, $orig_read_lengths_str);
    my $start_cycle_ind = 0;

    for (my $read_ind = 0; $read_ind < ($read_num - 1); ++$read_ind) {
	$start_cycle_ind += $orig_read_lengths[$read_ind];
    }

    my $read_length = $orig_read_lengths[$read_num - 1];

    if (($start_cycle_ind + $read_length) > length($use_bases_str)) {
	my $end_cycle_ind = $start_cycle_ind + $read_length - 1;
	warn("Attempt to extract past end of USE_BASES (`$use_bases_str' "
	     . "$start_cycle_ind:$end_cycle_ind)\n");
	return 0;
    }

    $$read_use_bases_str_ref
	= substr($use_bases_str, $start_cycle_ind, $read_length);

    return 1;
}

#------------------------------------------------------------------------------

sub make_out_file_name($$$)
{
    my ($qref_prefix, $tile_num, $qref_suffix) = @_;

    my $out_file_name
	= $qref_prefix . sprintf("%04d", $tile_num) . $qref_suffix;

    return $out_file_name;
}

#------------------------------------------------------------------------------

# FIXME : This should be just a default, overridable on the command line.
my $tile_file_path = './tiles.txt';

my $read1_opt = '--read1';
my $read2_opt = '--read2';
my $use_bases_opt = '--use_bases';
my $orig_read_lengths_opt = '--orig_read_lengths';

my $usage
    = "Usage: paste seqpreFile safFile | "
    . "$0 [$read1_opt | $read2_opt] $use_bases_opt USE_BASES "
    . "$orig_read_lengths_opt ORIG_READ_LENGTHS "
    . "qref_prefix qref_suffix > qrefFile";

if (@ARGV == 1) {
    if (($ARGV[0] eq '-h') || ($ARGV[0] eq '--help')) {
	print $usage, "\n";
	exit(0);
    }
}

die $usage unless ((@ARGV == 6) || (@ARGV == 7));

my $read_num = 0;
my $arg_ind = 0;

if (@ARGV == 7) {
    my $arg_str = $ARGV[$arg_ind++];

    if ($arg_str eq $read1_opt) {
	$read_num = 1;
    } elsif ($arg_str eq $read2_opt) {
	$read_num = 2;
    } else {
	die $usage;
    }
}

die $usage unless ($ARGV[$arg_ind++] eq $use_bases_opt);
my $use_bases_str = $ARGV[$arg_ind++];
die $usage unless ($ARGV[$arg_ind++] eq $orig_read_lengths_opt);
my $orig_read_lengths_str = $ARGV[$arg_ind++];
my $qref_prefix = $ARGV[$arg_ind++];
my $qref_suffix = $ARGV[$arg_ind++];

if ($read_num != 0) {
    my $full_use_bases_str = $use_bases_str;

    if (!get_read_use_bases_str($full_use_bases_str, $orig_read_lengths_str,
				$read_num, \$use_bases_str)) {
	die ("Failed to extract subset for read $read_num from USE_BASES `"
	     . $use_bases_str . "' given original read lengths `"
	     . $orig_read_lengths_str . "'");
    }
}

my @used_cycle_inds = parseUseBases($use_bases_str, 0, $read_num);
my $orig_read_length = length($use_bases_str);
my $pad_ch = '.';


my $read_id_field_ind = 0;
my $read_seq_field_ind = 1;
my $mismatch_field_ind = 6;

my $lane_num_in_read_id_field_ind = 1;
my $tile_num_in_read_id_field_ind = 2;

my $line_num = 0;

my $not_aligned_base = '.';
my $read_length = 0;
my $not_aligned_read_ref_str = ($not_aligned_base . "\n") x $orig_read_length;

my $last_tile_num = 0; # guaranteed non-tile-number

my $aligned = 0;
my $ref_seq;
my $padded_seq;

my $lane_num = 0;
my %tile_freqs;

*OSTRM = *STDOUT;

while (<STDIN>) {
    ++$line_num;
    chomp;

    next if (/^\#/);

    my @fields = split(/\t/);
    my @read_id_fields = split(/:/, $fields[$read_id_field_ind]);

    # Could check that all lane numbers are the same but currently using
    # this just for writing empty files for missing tiles below.
    $lane_num = $read_id_fields[$lane_num_in_read_id_field_ind];

    my $tile_num = $read_id_fields[$tile_num_in_read_id_field_ind];
    ++$tile_freqs{$tile_num};

    if ($tile_num != $last_tile_num) {
	my $out_file_name
	    = make_out_file_name($qref_prefix, $tile_num, $qref_suffix);
	open(OSTRM, ">$out_file_name");
	$last_tile_num = $tile_num;
    }

    my $read_seq = $fields[$read_seq_field_ind];

    if ($read_length == 0) {
        # Extract read length from first data line.
        $read_length = length($read_seq);
    } elsif (length($read_seq) != $read_length) {
        my $curr_read_length = length($read_seq);
        die("Length of read ($curr_read_length) in line $line_num"
            . " differs from earlier read length ($read_length)");
    }

    if ($aligned = ($#fields >= $mismatch_field_ind)) {
        my $mismatch_str = $fields[$mismatch_field_ind];

        # `-' -> partner aligned but this read did not -> 
        if ($mismatch_str ne '-') {
            $ref_seq = apply_mismatch($read_seq, length($read_seq),
                                      $mismatch_str, $line_num);
	    # Pad reference with `.' representing cycles masked by USE_BASES.
	    $padded_seq = pad_seq_with_masked_cycles($ref_seq,
						     \@used_cycle_inds,
						     $orig_read_length,
						     $pad_ch);
        } else {
            $aligned = 0;
        }
    }

    print OSTRM ($aligned
                  ? (join("\n", split(//, $padded_seq)) . "\n")
                  : $not_aligned_read_ref_str);
}


# Now write an empty file for each of any tiles that were expected but not
# represented in the input. This prevents some downstream unhappiness.

my @missing_tile_nums;

if (!determine_missing_tile_nums($tile_file_path, \%tile_freqs, $lane_num,
				 \@missing_tile_nums)) {
    die("Failed to determine_missing_tile_nums");
}

foreach my $missing_tile_num (@missing_tile_nums) {
    my $out_file_name
	= make_out_file_name($qref_prefix, $missing_tile_num, $qref_suffix);
    open(OSTRM, ">$out_file_name");
}


close(OSTRM);

#------------------------------------------------------------------------------
