#!/usr/bin/env python
#
# Copyright (c) 2006 Solexa
# Authors: Mike Parkinson
#          Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import sys, os, math
sys.path.append(os.path.join(os.path.split(sys.argv[0])[0], ".."))
import common

"""Utility to estimate average phasing and prephasing rates as a
linear regression of the average correlation between the signal
in different cycles."""

def read_phasing_file(filename):
    """Read in text file containing phasing and prephasing as as
    a function of cycle for 4 different bases. File format:
    cycle<TAB>phasing prephasing<TAB>phasingA phasingC phasingG phasingT<TAB>prephasingA..."""
    cycles = []
    combined = []
    try:
        lines = open(filename).readlines()
    except:
        lines = []
        common.error("Could not read filename.")
    for line in lines:
        if line.lstrip()[0] == '#':
            continue
        parts = line.split("\t")
        cycles.append(int(parts[0]))
        summary        = map(float, parts[1].split())
        basephasing    = map(float, parts[2].split())
        baseprephasing = map(float, parts[3].split())
        record = summary+basephasing+baseprephasing
        if (len(combined) == 0):
            combined = [ [] for i in range(len(record)) ]
        if (len(record) != len(combined)):
            common.error("Wrong number of columns in file %s" % filename)
            sys.exit(0)
        for i in range(len(combined)):
            combined[i].append(record[i])
    return (cycles, combined)

def linregress(x, y):
    """Linear regression."""
    sum_x = 0.
    sum_y = 0.
    sum_xx = 0
    sum_yx = 0.
    for i in range(len(x)):
        sum_x  += x[i]
        sum_xx += x[i] * x[i]
        sum_y  += y[i]
        sum_yx += y[i] * x[i]
    denom = ((len(x) * sum_xx) - (sum_x * sum_x));
    if (math.fabs(denom) < 1.e-12):
        return [-1., -1.]
    gradient = ((len(x) * sum_yx - sum_x * sum_y) / denom);
    intercept = ((sum_y - gradient * sum_x) / len(x));
    return [gradient, intercept]

def fit_data(data, first, last):
    """Linear regression fit for cumulative phasing of a data set"""
    x = map(float, range(first+1, last+1))
    y = data[first+1:last+1]
    return linregress(x, y)


if __name__ == "__main__":
    phasing_files = sys.argv[1:]
    for file in phasing_files:
        lane, tile = map(int, os.path.split(file)[1].split("_")[1:3])
        cycles, combined = read_phasing_file(file)
        # do a linear fit for values between <first> and <last> cycle
        first = 1
        last  = 11
        mincycles = 6
        if len(cycles) < mincycles:
            common.warning("Not enough cycles for automated "
                           "phasing estimate for %s: "
                           "Need %i cycles, have %i. Skipping..." %
                           (file, mincycles, len(cycles)))
            continue
        if len(cycles) < last+1:
            last = len(cycles)-1
            common.warning("Not enough cycles for automated "
                           "phasing estimate for %s: "
                           "Reducing estimation length to  %i cycles.\n" %
                           (file, last+1))

        fits = []
        for i in range(len(combined)):
            if i in [0, 2, 3, 4, 5]: # phasing values
                fits.append(fit_data(combined[i], first, last))
            else: # prephasing values
                #print len(combined[i][first+1:last+1])
                fits.append(fit_data(combined[i], first-1, last-1))

        print "%i\t%i\t%.4f\t%.4f\t%.4f %.4f %.4f %.4f\t%.4f %.4f %.4f %.4f\t%.4f\t%.4f" %\
              (lane, tile,
               fits[0][0], fits[1][0], # grdient combined phasing/prephasing
               fits[2][0], fits[3][0], fits[4][0], fits[5][0],# phasing
               fits[6][0], fits[7][0], fits[8][0], fits[9][0],# prephasing
               fits[0][1], fits[1][1]) # intercepts combined
