#! /usr/bin/env python
#
# Copyright (c) Solexa 2007
# Author: Klaus Maisinger
#
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import sys, os, glob, re, types, time
import runfolder, XMLSimple
from CycleSelector import CycleSelector

"""Common routines between goat_pipeline and bustard."""

PairedEndConfigFile    = "PairedEndInfo.xml"
MaxSequencingAutoCycle = 1000 

class Params:
    def __init__(self):
        self.options = []

    def getDict(self):
        """Return variable settings as a dictionary."""
        dic = getDict(self)
        del dic["options"]
        return dic

    def parse(self, options):
        error("parse() function not implemented!")


class Read:
    def __init__(self, RunFolder, Index=-1, FirstCycle = -1,
                 LastCycle = -1):
        self.FirstCycle = FirstCycle
        self.LastCycle  = LastCycle
        self.Index      = Index
        self.RunFolder  = RunFolder

def readcmp(x, y):
    if x.FirstCycle > y.FirstCycle:
        return 1
    else:
        return -1
    return 0

class GeneralParams(Params):
    def __init__(self):
        self.IparFlag          = 0
        self.FirstRunOnlyFlag  = 0
        self.Compression       = "gzip"
        self.CompressionSuffix = ".gz"
        self.SelectedTiles     = []
        self.SelectedFolders   = []
        self.MakeFlag          = 0
        self.BasecallFlag      = 1
        self.MinCycle          = -1
        self.MaxCycle          = -1
        self.Deblocked         = 0
        self.SelectedCycles    = []
        self.AutoCycleFlag     = 0
        self.ImagingReads      = [ Read("", Index=1, FirstCycle=1) ]
        self.Reads             = [ Read("", Index=1, FirstCycle=1) ]
        self.Instrument        = ""
        self.InstrumentFolder  = os.getenv('INSTRUMENT_DIR')
        self.RunFolder         = ""
        self.DebugFlag         = 0
        self.NewReadCycleList  = []
        self.CycleSelector     = CycleSelector('auto')
        self.options           = [ "new-read-cycle=", "compression=",
                                   "make", "tiles=", "folders=",
                                   "nobasecall", "deblocked",
                                   "maxcycle=", "mincycle=",
                                   "cycles=", "debug",
                                   "first-run-only",
                                   "INSTRUMENT_DIR="]

    def getDict(self):
        """Return variable settings as a dictionary."""
        dic = getDict(self)
        del dic["options"]
        del dic["CycleSelector"]
        del dic["NewReadCycleList"]
        return dic

    def getPairedEndConfigFile(self, folder):
        """Concatenates the folder and the name of the paired end config file"""
        return os.path.join(folder, PairedEndConfigFile)

    def readPairedEndConfigFile(self, folder):
        """Return the length of the first read, as defined in the config file"""
        pairedEndConfigFile = os.path.join(folder, PairedEndConfigFile)
        data = XMLSimple.XMLin(pairedEndConfigFile, KeepRoot=1)
        return int(data['FirstRead']['Length'])
        
    
    def parseImageFolders(self, args):
        """Identify total number of reads and create an ImagingRead for each.
        Different reads can be recognised in two ways:
        1. They are in different run-folders
        2. From the configuration file \"PairedEndInfo.xml\""""
        allRunFolders      = map(lambda a: runfolder.get_run_folder(a), args)
        existingRunFolders = [folder for folder in allRunFolders if os.path.exists(folder)]
        ignoredRunFolders  = filter(lambda folder: folder not in existingRunFolders, allRunFolders)
        for folder in ignoredRunFolders:
            warning("Folder %s not found. Ignoring." % folder)
        if not existingRunFolders:
            error("You need to specify at least one Run-Folder for analysis!")
            raise "Error"
        # Process single folder paired ends. At this point, just duplicate the run folder
        singleFolderPairedEnds = False
        if 1 == len(existingRunFolders) and os.path.exists(self.getPairedEndConfigFile(existingRunFolders[0])):
            try:
                print "Found paired end config file in %s." % existingRunFolders[0]
                firstReadLength = self.readPairedEndConfigFile(existingRunFolders[0])
                print "First read has length %i." % firstReadLength
                existingRunFolders.append(existingRunFolders[0])
                singleFolderPairedEnds = True
            except:
                warning("Failed to parse Paired End Config File for %s. Ignoring." % existingRunFolders[0])
        # Create all the ImagingReads for all the runs (single or multiple folders)
        # Note: the LastCycle will have to be initialized later
        self.ImagingReads = map(lambda index: Read(existingRunFolders[index], Index = index, FirstCycle = 1),
                                range(len(existingRunFolders)))
        self.RunFolder    = existingRunFolders[0]
        self.Instrument   = get_instrument(self.RunFolder)
        if singleFolderPairedEnds:
            self.ImagingReads[0].LastCycle = firstReadLength
            self.ImagingReads[1].FirstCycle = self.ImagingReads[0].LastCycle + 1

        #folders = []
        #for a in args:
        #    folder = runfolder.get_run_folder(a)
        #    if not os.path.exists(folder):
        #        warning("Folder %s not found. Ignoring." % folder)
        #        continue
        #    if folder not in folders:
        #        folders.append(folder)
        #if (len(folders) > 0):
        #    self.ImagingReads = []
        #    self.addRead(folders[0], FirstCycle=1)
        #if len(folders) > 1: # more than one run-folder
        #    for i in range(1, len(folders)):
        #        folder = folders[i]
        #        self.addRead(folder, FirstCycle=1)
        #elif (len(folders) == 0):
        #    error("You need to specify at least one "
        #          "Run-Folder for analysis!")
        #    raise "Error"
        #elif len(folders) == 1: # one run-folder: look for config file
        #    folder = folders[0]
        #    pairedfile = os.path.join(folder, PairedEndConfigFile)
        #    if os.path.exists(pairedfile):
        #        print "Found paired end config file %s." % pairedfile
        #        LastCycle = 0
        #        try:
        #            data = XMLSimple.XMLin(pairedfile, KeepRoot=1)
        #            LastCycle = int(data['FirstRead']['Length'])
        #            print "First read has length %i." % LastCycle
        #        except:
        #            warning("Failed to parse %s. Ignoring." % pairedfile)
        #        #print data
        #        if LastCycle > 0:
        #            self.addRead(FirstCycle=LastCycle+1)
        #self.RunFolder = folders[0]
        #self.Instrument = get_instrument(self.RunFolder)

    def parseImageAnalysis(self, args):
        """Read run parameters from the .params file in a Firecrest folder."""
        if (len(args)) < 1:
            error("Need to specify image analysis folder!")
            raise "Error"
        analysis_folder = os.path.abspath(args[0])
        paramsfile = os.path.join(runfolder.get_data_folder(analysis_folder),
                                  ".params")
        if paramsfile:
            params = runfolder.get_params(paramsfile,
                                          ["ImageAnalysis", "data"])
            run = runfolder.get_run(params, analysis_folder)
            if not run:
                error("Could not find run in .params file")
                raise "Error"
            if run.has_key("RunParameters"):
                data = run["RunParameters"]
                for key in data.keys():
                    self.__dict__[key] = data[key]
                if type(self.ImagingReads) == types.DictType:
                    self.ImagingReads = [ self.ImagingReads ]
                for i in range(len(self.ImagingReads)):
                    read = Read("")
                    for key in self.ImagingReads[i].keys():
                        read.__dict__[key] = self.ImagingReads[i][key]
                    self.ImagingReads[i] = read

                if type(self.Reads) == types.DictType:
                    self.Reads = [ self.Reads ]
                for i in range(len(self.Reads)):
                    read = Read("")
                    for key in self.Reads[i].keys():
                        read.__dict__[key] = self.Reads[i][key]
                    self.Reads[i] = read

            self.Tiles        = runfolder.get_tiles_from_tile_selection(run)
            self.MaxCycle     = run["Cycles"]['Number']
            self.MakeFlag     = 0
            self.BasecallFlag = 1
            self.DebugFlag    = 0

    def getTileDictionary(self, folder):
        """Build the tile dictionary. Keys are the tile name (e.g. s_2_0034).
        Values are the list of image files (1 image for each base and each
        used cycle. This method uses the ImagingReads FirstCycle and LastCycle
        (from the .params) to build the UsedCycleFolders. The heuristic is that
        a new run folder restarts the numbering of the folders (at 1).
        The 'folder' should be one of the ImageReads folder."""
        import string
        tiledic = {}
        filteredTiles = self.Tiles
        if self.SelectedTiles:
            tileFilter = re.compile("(%s)" % "|".join(self.SelectedTiles))
            filteredTiles = filter(lambda x: tileFilter.match(x), filteredTiles)
        cycleList  = []
        for imagingRead in self.ImagingReads:
            if folder == imagingRead.RunFolder:
                cycleList += range(imagingRead.FirstCycle, imagingRead.LastCycle + 1)
        # This is for multi-folder runs. C1.1 is the first cycle of a folder
        # but not necessarily the first cycle of the run. Therefore, if the
        # folder is not the first one, all the cycles must be shifted back
        # to have the first cycle of the folder in C1.1.
        cycleShift = cycleList[0] - 1
        if self.SelectedCycles:
            cycleList = filter(lambda x:x in self.SelectedCycles, cycleList)
        usedCycleFolders = map(lambda cycle: 'C' + str(cycle - cycleShift) + '.1', cycleList)
        for tile in filteredTiles:
            (sample, laneId, tileNumber) = tile.split('_');
            tileNumber = str(string.atoi(tileNumber))
            prefix = os.path.join(runfolder.get_image_folder(folder), 'L00'+laneId)
            prefixList = map(lambda cycle: os.path.join(prefix, cycle), usedCycleFolders)
            tiledic[tile] = []
            # for base in ['a', 'c', 'g', 't']:
            #    tiledic[tile].extend(map(lambda folder: os.path.join(folder, '_'.join([sample, laneId, tileNumber, base]) + '.tif'), prefixList))
            for folder in prefixList:
                tiledic[tile].extend(map(lambda base: os.path.join(folder, '_'.join([sample, laneId, tileNumber, base]) + '.tif'),  ['a', 'c', 'g', 't']))
        return tiledic

    def printInstrument(self):
        print "Instrument:    %s" % self.Instrument
        
    def printReads(self):
        if (len(self.ImagingReads) > 1):
            for read in self.ImagingReads:
                print "Read %i:        Imaging cycles %3i-%3i" % \
                      (read.Index, read.FirstCycle, read.LastCycle)

    def addRead(self, RunFolder = "", FirstCycle = -1,
                LastCycle = -1):
        for read in self.ImagingReads:
            if FirstCycle > 1 and FirstCycle == read.FirstCycle:
                error("Cannot add new read at cycle %i: "
                      "Already have read starting at cycle %i!\n"
                      % (FirstCycle, read.FirstCycle))
                sys.exit(1)
        Index = len(self.ImagingReads)
        for i in range(len(self.ImagingReads)):
            if FirstCycle > 1:
                if self.ImagingReads[i] <= FirstCycle:
                    Index = i+1
        for i in range(Index, len(self.ImagingReads)):
            self.ImagingReads[i].Index += 1
        #if len(self.Reads) > 1:
        #    error("Cannot add new read: Already have more than one read!")
        #    sys.exit(1)
        if FirstCycle > -1 and len(self.ImagingReads) > 0:
            self.ImagingReads[Index-1].LastCycle = FirstCycle - 1
        self.ImagingReads.insert(Index, Read(RunFolder, Index = len(self.ImagingReads)+1,
                               FirstCycle=FirstCycle,
                               LastCycle=LastCycle))

    # OBSOLETE
    def completeImagingReads(self, LastCycle, RestartCycles=[]):
        """Fill in missing cycle information."""
        count = 0
        for read in self.ImagingReads:
            if read.FirstCycle == 1:
                count += 1
        if (count != 1 + len(RestartCycles)):
            error("GeneralParams::completeRead(): Inconsistent read info.")
            sys.exit(1)
        count = 0
        for read in self.ImagingReads:
            if read.FirstCycle == 1:
                if (count > 0):
                    read.FirstCycle = RestartCycles[count-1]
                count += 1
        
        # now sort reads and fill in last cycles
        self.ImagingReads.sort(readcmp)
        for i in range(len(self.ImagingReads)):
            self.ImagingReads[i].Index = i+1
        self.ImagingReads[-1].LastCycle = LastCycle
        for i in range(len(self.ImagingReads)-1):
            self.ImagingReads[i].LastCycle = self.ImagingReads[i+1].FirstCycle-1
            #print [self.ImagingReads[i].FirstCycle, self.ImagingReads[i].LastCycle]
        
        #print [self.ImagingReads[-1].FirstCycle, self.ImagingReads[-1].LastCycle]

    def isSingleFolder(self):
        """Checks across all the ImagingReads to determine if they are
        in the same folder or in different folders."""
        runFolderList = map (lambda imagingRead: imagingRead.RunFolder, self.ImagingReads)
        uniqueRunFolderList = {}.fromkeys(runFolderList).keys()
        if 1 == len(uniqueRunFolderList):
            return True
        if len(self.ImagingReads) != len(uniqueRunFolderList):
            common.error("The number of run folders does not match the number of ImagingReads")
        return False        
        
    def completeLogicalReads(self):
        """Map the SelectedReads, if any, onto the ImagingReads and create the corresponding Logical Reads.
        Note: the SelectedReads are assumed to be expanded, sorted, and without duplicated values."""

        self.Reads = []
        for imagingRead in self.ImagingReads:
            if self.SelectedCycles:
                if self.isSingleFolder():
                    # selected cycles are directly aligned against the ImagingReads and split
                    referenceRead = imagingRead
                else:
                    # selected cycles are truncated against the first ImagingRead then replicated
                    referenceRead = self.ImagingReads[0]
                localCycles = filter(lambda cycle:
                                     cycle >= referenceRead.FirstCycle and
                                     cycle <= referenceRead.LastCycle,
                                     self.SelectedCycles)
            else:
                localCycles = range(imagingRead.FirstCycle, imagingRead.LastCycle+1)
            if not localCycles:
                continue;
            logicalRead = Read(imagingRead.RunFolder, 1)
            logicalRead.FirstCycle = 1
            if self.Reads:
                logicalRead.Index      += self.Reads[-1].Index
                logicalRead.FirstCycle += self.Reads[-1].LastCycle
            logicalRead.LastCycle = logicalRead.FirstCycle + len(localCycles) - 1
            self.Reads.append(logicalRead)

    def completeReads(self, imagingProtocol, useLastCycle = True):
        self.ImagingReads = imagingProtocol.getImagingReadList()
        self.Reads        = self.CycleSelector.getReadList(self.ImagingReads, useLastCycle)

    def completeReads___Obsolete(self, LastCycle, RestartCycles=[]):
        """Fill in missing cycle information and ensure logical order
        of reads."""
        #self.completeImagingReads(LastCycle, RestartCycles)
        # DAP-173: check that the LastCycle is greater than the total of all reads
        # Since this happens only in SFPE, there should be no RestartCycle
        while (self.ImagingReads[-1].LastCycle < self.ImagingReads[-1].FirstCycle):
            print("Warning: Expecting at least %d cycles (only %d found): Removing one ImagingRead\n" %
                  (self.ImagingReads[-1].FirstCycle, self.ImagingReads[-1].LastCycle))
            self.ImagingReads.pop()
            if (RestartCycles):
                print("Warning: Expecting a single forder but got RestartCycles\n")
        if not self.ImagingReads:
            common.error("No imaging reads after analysis of the folders")
        if self.ImagingReads[-1].LastCycle > LastCycle:
            print("Warning: Expecting at least %d cycles (only %d found): reducing LastCycle for last ImagingRead\n" %
                  (self.ImagingReads[-1].LastCycle, LastCycle))
            self.ImagingReads[-1].LastCycle = LastCycle
        self.completeLogicalReads()
        
                
    def parse(self, options):
        for o, a in options:
            # the IparFlag will be set if the IPAR directory is present
            #if o == "--IPAR":
            #    self.IparFlag = 1
            if o == "--first-run-only":
                self.FirstRunOnlyFlag = 1
            if o == "--new-read-cycle":
                self.NewReadCycleList.append(int(a))
                self.NewReadCycleList.sort()
                #self.addRead(FirstCycle=int(a))
            if o == "--debug":
                self.DebugFlag = 1
            if o == "--compression":
                comp = a.lower()
                compression_options = { "none"  : "",
                                        "gzip"  : ".gz",
                                        "bzip2" : ".bz2" }
                if comp in compression_options.keys():
                    self.Compression = comp
                    self.CompressionSuffix = compression_options[comp]
            if o == "--make":
                self.MakeFlag = 1
            if o == "--tiles":
                self.SelectedTiles = a.split(",")
            if o == "--folders":
                self.SelectedFolders = a.split(",")
            if o == "--nobasecall":
                self.BasecallFlag = 0
            if o == "--maxcycle":
                self.MaxCycle = int(a)
            if o in [ "--mincycle", "-m"]:
                self.MinCycle = int(a)
            if o == "--deblocked":
                self.Deblocked = 1
            if o == "--cycles":
                a = a.lower()
                self.CycleSelector = CycleSelector(a)
                if (a.find("auto") >= 0):
                    self.AutoCycleFlag = 1
                if (a.find("auto") > 0):
                    a = a.replace('auto', "%i" % MaxSequencingAutoCycle)
                if not (a.find('auto') == 0):
                    self.SelectedCycles = expand_ranges(a)
            if o == "--INSTRUMENT_DIR":
                self.InstrumentFolder = os.path.abspath(a)


def getDict(obj):
    """Serialise an object to simple types (dictionaries),
    which can later be processed by XMLSimple."""
    result = None
    if type(obj) == types.ListType:
        result = obj[:]
        for i in range(len(result)):
            result[i] = getDict(result[i])
    elif type(obj) == types.DictType:
        result = obj.copy()
        for key in result.keys():
            result[key] = getDict(result[key])
    elif type(obj) == types.InstanceType:
        result = obj.__dict__.copy()
        for key in result.keys():
            result[key] = getDict(result[key])
    else:
        result = obj
    return result


def unique_list(a):
    """Return a reduced list whose elements are unique."""
    dic = {}
    for element in a:
        dic[element] = 1
    return dic.keys()


def expand_ranges(a):
    """Expand various short forms of lists of ordered numbers."""
    numbers = []
    if (type(a) == types.StringType):
        for arg in a.split(","):
            arg2 = map(int, arg.split("-"))
            if len(arg2) == 1:
                arg2.append(arg2[0])
            numbers += range(int(arg2[0]), int(arg2[1])+1)
    elif (type(a) == types.ListType):
        for arg in a:
            if type(arg) == types.IntType:
                numbers.append(arg)
            else:
                numbers.append(arg[0], arg[1]+1)
    numbers = unique_list(numbers)
    numbers.sort()
    return numbers

def collapse_ranges(numbers):
    """Given a list of numbers, collapse runs of consecutive numbers
    to a list denoting the first and the last number."""
    sortednumbers = unique_list(numbers)
    sortednumbers.sort()
    newnumbers = []
    rangelist = []
    first, last = sortednumbers[0], sortednumbers[-1]
    prev = first
    for item in sortednumbers[1:]:
        if item != prev+1:
            rangelist.append([ first, prev ])
            first = item
        prev = item
    rangelist.append([first, last])
    folderlist = []
    for pair in rangelist:
        if pair[0] == pair[1]:
            newnumbers.append(pair[0])
        else:
            newnumbers.append([pair[0], pair[1]])
    return newnumbers
        
def collapse_ranges_to_string():
    pass

def select_tiles(tiledic, selected_tiles =[]):
    """Select tiles from a dictionary whose keys refer to tiles. Only part
    of the key needs to match the selected tiles."""
    if selected_tiles:
        a = re.compile("(%s)" % "|".join(selected_tiles))
        for key in tiledic.keys():
            if not a.search(key):
                del tiledic[key]
    return tiledic

def count_intensity_files(analysis_folder):
    """Count the number of cycles per intesntiy file in the image analysis
    working directory."""
    searchfolder = os.path.join(analysis_folder, "*_int.txt")
    filenames = glob.glob(searchfolder)
    tiledic = {}
    if (len(filenames) == 0):
        warning("No intensity files found in %s." % analysis_folder)
        searchfolder = os.path.join(analysis_folder, "*_int.txt.*")
        if len(glob.glob(searchfolder)) > 0:
            warning("Found compressed files in %s. Cannot deduce "
                    "cycle number from compressed files." % analysis_folder)
    for filename in filenames:
        tile  = runfolder.FileName.tilestr(filename)
        line  = open(filename).readline()
        count = len(line.split("\t"))-4
        print filename, tile, line, count
        if count > 0:
            tiledic[tile] = count
    return tiledic

def group_dictionary(function, list):
    """Create a dictionary where each key corresponds to a function
    value and the corresponding values are all list elements with the
    same function value.""" 
    dict = {}
    for item in list:
        id = function(item)
        dict.setdefault(id, []).append(item)
    return dict
        

def group(function, list):
    """Create a list of lists each containing elements that yield the
    same result when the function is applied. This creates groups
    of elements that have the same function value."""
    # sort by key value and return array
    dic = group_dictionary(function, list)
    keys = dic.keys()
    keys.sort()
    return [ dic[key] for key in keys ]


def get_instrument(analysis_folder):
    """Return name of instrument from runfolder parameter file"""
    paramsfile = runfolder.get_run_params_filename(analysis_folder)
    instrument = ""
    if os.path.isfile(paramsfile):
        params = runfolder.get_params(paramsfile)
        try:
            instrument = params["Run"][0]["instrument"]
        except:
            print params
            warning("No instrument id found in parameter file %s."
                    % paramsfile)
    else:
        warning("Parameter file %s not found!\n" % paramsfile)
    return instrument

def get_matrix(matrixfilename, instrument_folder, analysis_folder):
    """Return path to matrix file"""
    matrixfile = matrixfilename or \
                 runfolder.find_instrument_matrix_file(analysis_folder,
                                                       instrument_folder,
                                                       get_instrument(analysis_folder))
    if not os.path.exists(matrixfile):
        warning("No matrix file %s found!" % matrixfile)
        return None
    else:
        return matrixfile

def logline(level, message):
    tm = time.gmtime(time.time())
    mytime = time.strftime("%H:%M:%S", tm)
    mydate = time.strftime("%d/%m/%Y", tm)
    return ("%s\t%s\t%s\t%s\t%s\t%s\t%s\n"
            % (level, mydate, mytime, sys.argv[0],
               os.getenv('LOGNAME') or os.getenv('USERNAME'),
               os.getenv('HOSTNAME'), message))

def log(filename, level, message):
    try:
        open(filename, "a").writelines([ logline(level, message)])
    except:
        sys.stderr.write("Warning: Failed to write to %s.\n" %
                         os.path.abspath(filename))
    

def error(message):
    """Central routine for handling error messages"""
    sys.stderr.write("Error: %s\n" % message)
    log("error.txt", 7, message)

def warning(message):
    """Central routine for handling warning messages"""
    sys.stderr.write("Warning: %s\n" % message)
    log("warning.txt", 6, message)
