#!/usr/bin/env python

# Copyright (c) 2005, 2006 Solexa
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import sys, time

"""Manipulate frequency matrices. The matrix is a dictionary that has
a set of elements stored in a single array in C order. Additional
fields keep track of bases and exposure times. The matrix can be
written to a text file in a defined format."""

def read_matrix(filename):
    """Read matrix from a file."""
    fd = open(filename)
    if fd:
        matrix = { "Element": [] }
        for line in fd.readlines():
            if line[0] == '#':
                continue
            if line[0] == ">":
                elements = line.split()
                matrix.setdefault("Base", []).append(elements[1].upper())
                if len(elements) > 2:
                    matrix.setdefault("Exposure",
                                      []).append(float(elements[2]))
            else:
                matrix["Element"] += map(float, line.split())
        fd.close()
        return matrix
    return []

def write_matrix(filename, matrix):
    """Write matrix to a file."""
    fd = open(filename, "w")
    if fd:
        fd.write("# Matrix generated %s\n" %
                 time.strftime("%d-%m-%y %H:%M:%S %Z",
                               time.localtime()))
        
        if matrix.has_key("Base"):
            if matrix.has_key("Exposure"):
                for i in range(len(matrix["Base"])):
                    fd.write("> %s %5.2f\n" % (matrix["Base"][i],
                                               matrix["Exposure"][i]))
            else:
                for i in range(len(matrix["Base"])):
                    fd.write("> %s\n" % (matrix["Base"][i]))
        for i in range(4):
            fd.write("%s\n" %
                     " ".join(map(lambda x: "%6.2f" % x,
                                  matrix["Element"][i*4:((i+1)*4)])))
        fd.close()

def adjust_matrix(matrix, exposure):
    """Adjust matrix for different exposure times by rescaling
    coefficients."""
    if ((not matrix.has_key("Exposure")) or (len(exposure) != 4)
        or (len(matrix["Element"]) != 16)):
        return
    for i in range(4):
        for j in range(4):
            matrix["Element"][i*4+j] *= exposure[i]/matrix["Exposure"][i]
        matrix["Exposure"][i] = exposure[i]
    return

def normalise_matrix(matrix):
    """Normalise matrix such that its trace equals 4. This is somewhat
    arbitrary."""
    elements = matrix["Element"]
    norm = 0.
    for i in range(4):
        norm += elements[i*4+i]
    norm /= 4.
    for i in range(16):
        elements[i] /= norm
    return


if __name__ == "__main__":
    matrix = read_matrix(sys.argv[1])
    print matrix
    adjust_matrix(matrix, [3., 2., 2., 3.])
    print matrix
    normalise_matrix(matrix)
    print matrix
    if len(sys.argv) > 2:
        write_matrix(sys.argv[2], matrix)
