#ifndef GZ_PACK
#define GZ_PACK

#include <string>
#include <stdexcept>

#include "IoUtils.hh"

namespace GzPack {

struct DataLayout
{
public:
    DataLayout(const unsigned int numberOfChannels, const unsigned int numberOfClusters);
    DataLayout(const DataLayout &dataLayout);
    const unsigned int channels;
    const unsigned int clusters;
};

class GzPackReader : private IoUtils::AutoGzFile
{
public:
    GzPackReader(const std::string fileName) throw (std::logic_error);
    ~GzPackReader() {delete [] localStorage_;}
    /**
     ** Return an array with all the intensity values for the next cycle.
     ** All channels of a cluster are stored consecutively.
     ** Note: the pointer is invalidated when the method is called again or
     ** when the GzPackReader instance get out of scope (the storage is 
     ** freed by the destructor)
     ** On error, throws a logic exception. 
     ** It is the responsibility of the user to keep track of the cycles.
     **/
    const float *getIntensityData() throw(std::logic_error);
    /**
     ** Store the intensity data for the next cycle in a user-provided storage.
     ** Note: it is the responsibility of the client to ensure that there is
     ** enough storage (numberOfClusters * numberOfChannels floats).
     **
     ** Example 1:
     ** ----------
     ** float *storage = new float[gpr.getNumberOfClusters() * gpr.getNumberOfChannels()];
     ** getIntensityData(storage);
     ** delete [] storage;
     **
     ** Example 2:
     ** ----------
     ** vector<float>storage(gpr.getNumberOfClusters() * gpr.getNumberOfChannels());
     ** getIntensityData(storage.begin());
     **/
    template <class Iterator>
    void getIntensityData(Iterator intensityStorage) throw(std::logic_error);
    inline const DataLayout &getDataLayout()  const {return dataLayout_;}
    inline unsigned int getNumberOfChannels() const {return dataLayout_.channels;}
    inline unsigned int getNumberOfClusters() const {return dataLayout_.clusters;}
    /**
     ** The eof method is NOT const because it has to skip comments. The
     ** alternative would be to make sure that all other methods skip the 
     ** comments after they are done with their processing, which would give
     ** serious bug opportunities.
     **/
    inline bool eof() {skipCommentLines(); return static_cast<const IoUtils::AutoGzFile *>(this)->eof();}
    inline bool fileError() const {return static_cast<const IoUtils::AutoGzFile *>(this)->fileError();}
    void skipCycle() throw (std::logic_error);
private:
    const DataLayout dataLayout_;
    /**
     ** There is a requirement to allow comment lines in the data file. This 
     ** implies that comment lines must be skipped is some specific situations,
     ** for instance before testing for EOF. Unfortunately, the zlib API does
     ** not support 'peek' or 'put back' functionalities. Therefore, there ia a
     ** need to cache at least one character. Since it is much more convenient 
     ** to work on a full line, the decision is to cache 1 full line, more
     ** accurately, to cache the last pointer returned by gets.
     **
     ** Note: there is also a requirement for the data layout line to be a 
     ** comment line. Therefore, it is important that the comments are NOT
     ** skipped until the data layout has been initialized.
     **/
    char *currentLine_;
    /**
     ** Local storage needed by "const float *getIntensityData()"
     **/
    float *localStorage_;
    /**
     ** private helper methods.
     **/
    DataLayout initDataLayout() throw (std::logic_error);
    void skipCommentLines();
};

template <class Iterator>
void GzPackReader::getIntensityData(Iterator intensityStorage) throw (std::logic_error)
{
    skipCommentLines();
    using namespace std;
    Iterator currentIntensity = intensityStorage;
    for (unsigned int cluster = 0; dataLayout_.clusters > cluster; ++cluster)
    {
        if (Z_NULL == currentLine_)
        {
            throw logic_error(string("GzPackReader: Failed to read the intensity data"));
        }
        const char *const currentLineEnd = currentLine_ + strlen(currentLine_);
        char *valueStart = currentLine_;
        char *valueEnd   = currentLine_;
        for (unsigned int channel = 0; dataLayout_.channels > channel; ++channel)
        {
            if (currentLineEnd <= valueStart)
            {
                throw logic_error(string("GzPackReader: Missing intensity value"));
            }
            errno = 0;
            //*currentIntensity = strtof(valueStart, &valueEnd);
            *currentIntensity = (float)strtod(valueStart, &valueEnd);
            if (0 != errno || valueEnd == valueStart)
            {
                throw logic_error(string("GzPackReader: Expected a float, got: ") + std::string(valueStart));
            }
            valueStart = valueEnd + strspn(valueEnd, " \t\n");
            ++currentIntensity;
        }
        if (currentLineEnd != valueStart)
        {
            throw logic_error(string("GzPackReader: Excess intensity value"));
        }
        currentLine_ = gets();
    }
    skipCommentLines();
}

} // namespace GzPack

#endif // #ifndef GZ_PACK

