/*
 * 1-d and 2-d Levenberg-Marquardt fitting routines with a Gaussian test case.
 * 1-d Brent's  method.
 *
 * Time-stamp: <20 Oct 2006 18:00 BST Optimise.cpp km>
 * Copyright (c) Solexa 2005, 2006
 * Authors: Klaus Maisinger, Lorenzo Milazzo
 * This source file is covered by the "Illumina Genome Analyzer
 * Software License Agreement" and bound by the terms therein.
 */
/*
 * Note: The function pointer mechanism used to map 1-d and 2-d optimisation
 * routines to the same interface is suspected to perform poorly on Windows.
 * This will most likely have to be changed.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include "LinAlg.h"
#include "Optimise.h"

#define SIGNUM(x) ((x) >= 0.f ? 1.f : -1.f);

/*
 * Bracketing procedure
 *
 * Given a function 'func', and given distinct initial points 'xa' and 'xb',
 * this function searches in the downhill direction (defined by the function
 * as evaluated at the initial points) and returns new points 
 * 'xa', 'xb', 'xc' that bracket a minimum of the function;
 * it also returns the function values
 * at the three points, 'fa', 'fb', and 'fc'.
 *
 * - Inputs:
 *    func                  function
 *    args                  arguments of the function 'func'
 *    xa, xb                initial points
 * - Outputs:
 *    xa, xb, xc            new points that bracket a minimum of the function 'func'
 *    fa, fb, fc   function values at the three points
 */
void bracket(float (*func) (float, void *), void *args, float *xa,
             float *xb, float *xc, float *fa, float *fb, float *fc)
{
    float xx, xx_lim;
    float fx;
    float temp;

    *fa = (*func) (*xa, args);
    *fb = (*func) (*xb, args);

    if (*fb > *fa) {
        temp = *xa; *xa = *xb; *xb = temp;
        temp = *fb; *fb = *fa; *fa = temp;
    }

    // initial guess for 'xc'
    *xc = (*xb) + 1.618034f * (*xb - *xa);

    *fc = (*func) (*xc, args);

    while (*fb > *fc) {

        // calculating the point 'xx' by using a parabolic extrapolation
        float fac1 = (*xb - *xa) * (*fb - *fc);
        float fac2 = (*xb - *xc) * (*fb - *fa);
        float diff = fac2 - fac1;
        if (fabs(diff) < 1.0e-20) diff = 1.0e-20 * SIGNUM(diff); 
        xx = (*xb) - ((*xb - *xc) * fac2 - (*xb - *xa) * fac1)  
	    / (2.f * diff);

        // maximum value allowed for 'xx'
        xx_lim = (*xb) + 100.f * (*xc - *xb);

        if ((*xb < xx) && (xx < *xc)) {
            fx = (*func) (xx, args);

            if (fx < *fc) {
                *xa = (*xb); *xb = xx;
                *fa = (*fb); *fb = fx;
                return;
            }
            else if (fx > *fb) {
                *xc = xx; *fc = fx;
                return;
            }
            xx = (*xc) + 1.618034f * (*xc - *xb);
            fx = (*func) (xx, args);
        }
        else if ((*xc < xx) && (xx < xx_lim)) {
            fx = (*func) (xx, args);
            if (fx < *fc) {
                *xb = *xc; *xc = xx;
                xx = *xc + 1.618034f * (*xc - *xb);
                *fb = *fc; *fc = fx;
                fx = (*func) (xx, args);
            }
        }
        else if ((xx <= xx_lim) && (xx_lim <= *xc)) {
            xx = xx_lim;
            fx = (*func) (xx, args);
        }
        else {
            xx = (*xc) + 1.618034f * (*xc - *xb);
            fx = (*func) (xx, args);
        }

        // eliminating old points
        *xa = *xb; *xb = *xc; *xc = xx;
        *fa = *fb; *fb = *fc; *fc = fx;
    }
}

/*
 * Brent's method in one dimension, using parabolic interpolation.
 * See NR for details.
 * ARGUMENTS:
 *    func      the function to be minimised
 *    args      any extra arguments to be passed to the function
 *    xbracket  an array with 3 elements such that the second elements
 *              lies between the other two and its function value is less
 *              than the other two function values
 *    xtol      fractional convergence tolerance
 *    xmin      the minimum
 *    fmin      function value at the minimum
 *    funcallcount  number of function evaluations used
 * RETURNS
 *    0         if converged
 *    1         otherwise
 */
int brent(float (*func)(float, void*), void *args, float *xbracket, float xtol,
          float *xmin, float *fmin, int *funcallcount)
{
    int iter;
    float xleft, xright;
    float xcurr, xbest, x2best, xprev2best;
    float fcurr, fbest, f2best, fprev2best;
    float xprevstep = 0., xstep = 0.;

    xleft  = (xbracket[0] < xbracket[2] ? xbracket[0] : xbracket[2]);
    xright = (xbracket[0] > xbracket[2] ? xbracket[0] : xbracket[2]);
    xbest = x2best = xprev2best = xbracket[1];
    fbest = f2best = fprev2best =(*func)(xbracket[1], args); 
    *funcallcount += 1;

    for (iter = 0; iter < 50; iter++) {
        float xmid      = (xright + xleft)/2.f;
        float xinterval = (xright - xleft);
        float abstol = xtol * fabs(xbest) + 1.e-10f;
        /* convergence check */
        if (fabs(xbest - xmid) + xinterval/2.f <= 2.f*abstol) {
            *xmin = xbest;
            *fmin = fbest;
            return 0;
        }
        int golden_section = 1;
        if (fabs(xprevstep) > abstol) { 
            /* try parabolic interpolation through points 
               xbest, x2best, xprev2best */
            float xfac1 = xbest - xprev2best, xfac2 = xbest - x2best;
            float ffac1 = fbest - fprev2best, ffac2 = fbest - f2best;
            float numerator = xfac1*xfac1*ffac2 - xfac2*xfac2*ffac1;
            float denominator = 2.f*(xfac1*ffac2 - xfac2*ffac1);
            if (denominator != 0.f) {
                float xnewstep = -numerator/denominator;
                float xnewcurr = xbest + xnewstep;
                if ((fabs(xnewstep) < 0.5f * fabs(xprevstep))
                    && (xnewcurr > xleft) && (xnewcurr < xright)) {
                    golden_section = 0;
                    xprevstep = xstep;
                    xstep = xnewstep;
                    if ((xnewcurr - xleft < 2.f*abstol) 
                        || (xright - xnewcurr < 2.f*abstol))
                      xstep = abstol * SIGNUM(xmid - xbest);
                }
            }
        }

        if (golden_section) {
            /* if parabolic interpolation failed, use golden section search */
            xprevstep = (xbest >= xmid ? xleft-xbest : xright-xbest);
            xstep = 0.3819660f * xprevstep;
        }

        /* make sure step is at least abstol */
        if (fabs(xstep) < abstol)
          xstep = abstol * SIGNUM(xstep);

        /* function evaluation */
        xcurr = xbest + xstep;
        fcurr = (*func)(xcurr, args); *funcallcount += 1;

        if (fcurr <= fbest) {       // if new minimum has been found
            if (xcurr >= xbest) 
                xleft  = xbest; 
            else 
                xright = xbest;
            xprev2best = x2best; x2best = xbest; xbest = xcurr;
             fprev2best = f2best; f2best = fbest; fbest = fcurr;
        }
        else {                      // no minimum found
            if (xcurr < xbest)
                xleft  = xcurr;
            else
                xright = xcurr;
            if ((fcurr <= f2best) || (x2best == xbest)) {
                xprev2best = x2best; x2best = xcurr;
                fprev2best = f2best; f2best = fcurr;
            } 
            else if ((fcurr <= fprev2best) 
                     || (xprev2best == xbest) || (xprev2best == x2best)) {
                xprev2best = xcurr;
                fprev2best = fcurr;
            }
       }
    }

    *xmin = xbest;
    *fmin = fbest;
    return 1;
}

/*
 * evaluate gradient and Hessian of the chi^2 function for those <np>
 * parameters <a> that are to be fitted.
 */
void LM_chi2_der(float x[], float x2[], float y[], float sigma[], int ndata, 
                 float a[], int ia[], int na, int np,
                 float *chi2, float *chi2grad, float *chi2curv,
                 float (*func)(float, float, float [], float [], int))
{
    // create index lookup for those parameters that are to be fitted
    int *index = new int [np];
    int i, j, k, l = 0;
    for (j = 0; j < na; j++)
        if (ia[j]) {
            index[l] = j;
            l++;
        }

    // initialise chi2, gradient and curvature
    *chi2 = 0.;
    for (j = 0; j < np; j++)
        chi2grad[j] = 0.;
    for (j = 0; j < np*np; j++)
        chi2curv[j] = 0.;

    // compute chi^2, gradient and curvature
    float *dyda = new float [na];
    for (i = 0; i < ndata; i++) {
        assert(sigma[i] != 0.);
        float weight = 1.f/(sigma[i]*sigma[i]);
        float yval = (*func)(x[i], x2[i], a, dyda, na);
        float dy = y[i] - yval;
        *chi2 += dy*dy*weight;
        for (j = 0; j < np; j++) {
            float dydaj = dyda[index[j]];
            chi2grad[j] += weight*dydaj*dy;
            for (k = 0; k < j+1; k++)
                chi2curv[j*np+k] += weight * dydaj*dyda[index[k]];
        }
    }
    // symmetrise curvature
    for (j = 0; j < np;j++)
        for (i = 0; i < j; i++)
            chi2curv[i*np+j] = chi2curv[j*np+i];
    delete [] dyda;

    delete [] index;
}

class LMMinimiser
{
  private:
    // external variables
    float *x, *x2, *y, *sigma;
    int ndata;
    float (*func)(float, float, float [], float [], int);
    // local variables
    float *anew, *chi2grad, *chi2gradnew, *chi2curvnew, *covariance;
    float lambda;
private:
    void free();
public:
    LMMinimiser() {};
    ~LMMinimiser() { free(); };
    void initialise(float x[], float x2[], float y[], float sigma[], 
                    int ndata, 
                    float a[], int ia[], int na, int np, 
                    float *chi2, float *chi2curv,
                    float (*func)(float, float, float [], float [], int));
    int iterate(float a[], int ia[], int na, int np, 
                float *chi2, float *chi2curv, float *lambda);
};


void LMMinimiser::initialise(float xv[], float xv2[], float yv[],
                             float sig[], int nd, 
                             float a[], int ia[], int na, int np,
                             float *chi2, float *chi2curv,
                             float (*fun)(float, float, float [], float [],
                                          int))
{
    x = xv; x2 = xv2; y = yv; sigma = sig; ndata = nd; func = fun;
    anew = new float [na];
    chi2grad = new float [na];
    chi2gradnew = new float [na];
    chi2curvnew = new float [np*np]; 
    covariance = new float [np*np]; 
    LM_chi2_der(x, x2, y, sigma, ndata, a, ia, na, np,
                chi2, chi2grad, chi2curv, func);
    lambda = 1.e-3f;
    int j;
    for (j = 0; j < na; j++)
        anew[j] = a[j];
}

void LMMinimiser::free()
{
    delete [] covariance;
    delete [] chi2curvnew;
    delete [] chi2gradnew;
    delete [] chi2grad;
    delete [] anew;
}

/*
 * 2d Levenberg-Marquardt minimisation method. Compare LM_1d for the meaning
 * of the parameters. <np> is the total number of parameters to be fitted
 * (<np> <= <na>). <chi2curv> is an <np>*<np> array
 * containing the current curvature. Invert this array to obtain 
 * the covariance matrix.
 */
int LMMinimiser::iterate(float a[], int ia[], int na, int np, 
                         float *chi2, float *chi2curv, float *lam)
{
    int i, l;

    /* initialise variables */
    float old_chi2 = *chi2;

    /* solve linear equations; this also produces the covariance matrix
     * as the inverse of the curvature */
    for (i = 0; i < np*np; i++)
        covariance[i] = chi2curv[i];
    for (i = 0; i < np; i++)
        covariance[i*np+i] *=  (1.f + lambda);
    for (i = 0; i < np; i++)
        chi2gradnew[i] = chi2grad[i];

    /*  both covariance and chi2gradnew get overwritten */
    if (gauss_jordan_inverse_solve(covariance, np, chi2gradnew, 1))
        return 1;

    /* update parameter values a */
    i = 0;
    for (l = 0; l < na; l++)
        if (ia[l]) {
            anew[l] = a[l] + chi2gradnew[i];
            i++;
        }

    /* evaluate chi^2 derivatives for new value of a */
    LM_chi2_der(x, x2, y, sigma, ndata, anew, ia, na, np,
                chi2, chi2gradnew, chi2curvnew, func);

    /* 
     * check for convergence of chi^2. if it converges, update parameters,
     * otherwise increase lambda
     */
    if (*chi2 < old_chi2) {
        lambda *= 0.1f;
        for (i = 0; i < np*np; i++)
            chi2curv[i] = chi2curvnew[i];
        for (i = 0; i < np; i++)
            chi2grad[i] = chi2gradnew[i];

        for (l = 0; l < na; l++)
            a[l] = anew[l];
    } else {
        lambda *= 10.f;
        *chi2 = old_chi2;
    }
    *lam = lambda;
    return 0;
}

/*********************************************************************/
/*
 * wrapper to map a 2-d prototype to a 1-d function
 */
static float (*myfunc)(float, float[], float [], int);

static float myfunc_wrapper(float x, float x2, 
                            float a[], float dyda[], int na)
{
    return (*myfunc)(x, a, dyda, na);
}

/*
 * provides a convenience wrapper for the Levenberg Marquardt routines.
 * <x>, <y> are <ndata>-vectors containing the data points, <sigma> is 
 * a vector containing the corresponding errors on <y>.
 * <a>  is the <na>-dim. parameter vector, and <ia>  is an array of flags
 * indicating whether the corresponding parameter is to be fitted or
 * kept constant. <func> is the function whose parameters are to be fitted
 * and <chisq> the chi^2-value of the fit. 
 */
int LM_1d(float x[], float y[], float sigma[], int ndata,
          float a[], int ia[], int na, float *chisq,
          float (*func)(float, float [], float [], int))
{
    myfunc = func;
    return LM_2d(x, x, y, sigma, ndata, a, ia, na, chisq, myfunc_wrapper);
}

/*
 * 2-dimensional equivalent of  LM_1d. <x2> is the second coordinate vector.
 * this version returns the curvature at the current position.
 */
int LM_2d(float x[], float x2[], float y[], float sigma[], int ndata,
          float a[], int ia[], int na, float *chisq, float *curv,
          float (*func)(float , float, float [], float [], int))
{
    float alambda = -1., oldchisq;
    int iter = 0, converged = 0;
    int failed = 0;
    int np = 0;
    int i;

    for (i = 0; i < na; i++)
      if (ia[i])
        np++;

    LMMinimiser *lm = new LMMinimiser();
    *chisq = -1.;
    lm->initialise(x, x2, y, sigma, ndata, a, ia, na, np, chisq, curv, 
                   func);
    do {
        oldchisq = *chisq;
        failed = lm->iterate(a, ia, na, np, chisq, curv, &alambda);

        /*
         * rules for convergence: 
         * stop if chisq decreases by a negligible amount for at least 
         * two times.
         * do not stop if chi^2 increases. this is based on comments in NR.
         * it seems to work reasonably well in practice.
         */
        if ((fabs((*chisq-oldchisq)/(oldchisq)) < 1.e-3)
            && (*chisq <= oldchisq))
            converged ++;
        else
            converged = 0;
        iter ++;
#if 0
        printf("chisq %f %f %f %i\n", oldchisq, *chisq, alambda, converged); 
#endif
#if 0
        if ((iter > 20) || (ndata <= na)) {
            for (i = 0; i < na; i++)
              printf("%g ", a[i]);
            printf("\n");
            printf("chisq %f %f %f %i\n",
                   oldchisq, *chisq, alambda, converged); 
        }
        if (iter > 60)
            exit(0);
#endif
    } while ((converged < 3) && (iter < 20) && (!failed));

    delete lm;
    return failed;
}

/*
 * 2-dimensional equivalent of  LM_1d. <x2> is the second coordinate vector.
 */
int LM_2d(float x[], float x2[], float y[], float sigma[], int ndata,
          float a[], int ia[], int na, float *chisq,
          float (*func)(float , float, float [], float [], int))
{
    float *curv;
    int np = 0;
    int i;

    for (i = 0; i < na; i++)
      if (ia[i])
        np++;
    curv = new float [np*np];
    int failed = LM_2d(x, x2, y, sigma, ndata, a, ia, na, chisq, curv, func);
    delete [] curv;
    return failed;
}

/***********************************************************************/

#define TINY 1.e-12f

/*
 * Gaussian in 1d: y = A e^{(x-b)^2/(2 s^2)} [ + c + d*x ]
 * <na> parameters:
 *   a[0] - amplitude A
 *   a[1] - offset b
 *   a[2] - standard deviation sigma (s)
 * optional parameters: 
 *   a[3] - constant offset c
 *   a[4] - linear coefficent d
 */
float LM_gauss_1d(float x, float a[], int na)
{
    float result = 0.0;
    int i;
    for (i = 0; i+2 < na; i+=3) {
        if (fabs(a[i+2]) < TINY) a[i+2] = TINY;
        float arg = (x-a[i+1])*(x-a[i+1]) / (2.f * a[i+2]*a[i+2]);
        result += a[i+0] * exp(-arg);
    }
    /* treat optional parameters */
    if (i < na)
        result += a[i];
    if (i+1 < na)
        result += a[i+1] * x;
    return result;
}

/*
 * y = A e^{(x-b)^2/(2 s^2)}, a[0] = A, a[1] = b, a[2] = s
 */
float LM_gauss_1d_der(float x, float a[], float dyda[], int na)
{
    float xs, s2, arg, ex;
    float result = 0.;
    int i;
    assert(na >= 3);

    for (i = 0; i+2 < na; i+=3) { 
        xs = x-a[i+1];
        if (fabs(a[i+2]) < TINY) a[i+2] = TINY;
        s2 = 2.f * a[i+2]*a[i+2];
        arg = xs*xs / s2;
        ex = exp(-arg);

        dyda[i+0] = ex; /* dy/dA */
        dyda[i+1] = a[i+0] * ex * 2.f * xs / s2;
        dyda[i+2] = a[i+0] * ex * arg * 2.f / (a[i+2]);

        result += a[i] * ex;
    }
    /* treat optional parameters */
    if (i < na) {
        result += a[i];
        dyda[i] = 1.;
    }
    if (i+1 < na) {
        result += a[i+1] * x;
        dyda[i+1] = x;
    }
    return result;
}


/*
 *  Gaussian in 2d
 */
float LM_gauss_2d(float x1, float x2, float a[], int na)
{
    float result = 0.0;
    int i;
    assert(na >= 5);
    for (i = 0; i+4 < na; i+=5) {
        if (fabs(a[i+3]) < TINY) a[i+3] = TINY;
        if (fabs(a[i+4]) < TINY) a[i+4] = TINY;
        float arg1 = (x1-a[i+1])*(x1-a[i+1]) / (2.f * a[i+3]*a[i+3]);
        float arg2 = (x2-a[i+2])*(x2-a[i+2]) / (2.f * a[i+4]*a[i+4]);
        result += a[i+0] * exp(-(arg1+arg2));
    }
    /* add constant offset if at least one more parameter */
    if (i < na)
      result += a[i];
    return result;
}

float LM_gauss_2d_der(float x1, float x2, float a[], float dyda[], int na)
{
    float xs1, xs2, s12, s22, arg1, arg2, ex;
    float result = 0.;
    int i;
    assert(na >= 5);

    for (i = 0; i+4 < na; i+=5) { 
        xs1 = x1-a[i+1];
        xs2 = x2-a[i+2];
        if (fabs(a[i+3]) < TINY) a[i+3] = TINY;
        if (fabs(a[i+4]) < TINY) a[i+4] = TINY;
        s12 = 2.f * a[i+3]*a[i+3];
        s22 = 2.f * a[i+4]*a[i+4];
        arg1 = xs1*xs1 / s12;
        arg2 = xs2*xs2 / s22;
        ex = exp(-(arg1+arg2));

        dyda[i+0] = ex; /* dy/dA */
        dyda[i+1] = a[i+0] * ex * 2.f * xs1 / s12;
        dyda[i+2] = a[i+0] * ex * 2.f * xs2 / s22;
        dyda[i+3] = a[i+0] * ex * arg1 * 2.f / (a[i+3]);
        dyda[i+4] = a[i+0] * ex * arg2 * 2.f / (a[i+4]);

        result += a[i] * ex;
    }
    if (i < na) {
        result += a[i];
        dyda[i] = 1.;
    }

    return result;
}

/*
 * Gaussian in 2d, spherical profile (sigma identical in all directions).
 * Map parameter vectors back to non-spherical case.
 */
float LM_gauss_spherical_2d(float x1, float x2, float a[], int na)
{
    int i;
    int na2 = na+na/4;
    float *a2 = new float [na2];
    for (i = 0; i+4 < na2; i+=5) {
        int offs = i/5;
        a2[i+0] = a[i+0-offs];
        a2[i+1] = a[i+1-offs]; a2[i+2] = a[i+2-offs];
        a2[i+3] = a[i+3-offs]; a2[i+4] = a[i+3-offs];
    }
    assert(i+4 >= na2);
    while (i < na2) {
        a2[i] = a[i-i/5];
        i++;
    }
    float result = LM_gauss_2d(x1, x2, a2, na2);
    delete [] a2;
    return result;
}

float LM_gauss_spherical_2d_der(float x1, float x2, float a[], float dyda[], int na)
{
    int i;
    int na2 = na+na/4;
    float *a2 = new float [na2];
    for (i = 0; i+4 < na2; i+=5) {
        int offs = i/5;
        a2[i+0] = a[i+0-offs];
        a2[i+1] = a[i+1-offs]; a2[i+2] = a[i+2-offs];
        a2[i+3] = a[i+3-offs]; a2[i+4] = a[i+3-offs];
    }
    assert(i+4 >= na2);
    while (i < na2) {
        a2[i] = a[i-i/5];
        i++;
    }
    float *dyda2 = new float [na2];
    float result = LM_gauss_2d_der(x1, x2, a2, dyda2, na2);
    for (i = 0; i+4 < na2; i+=5) {
        int offs = i/5;
        dyda[i+0-offs] =  dyda2[i+0];
        dyda[i+1-offs] =  dyda2[i+1]; dyda[i+2-offs] =  dyda2[i+2];
        dyda[i+3-offs] =  dyda2[i+3]; dyda[i+3-offs] += dyda2[i+4];
    }
    assert(i+4 >= na2);
    while (i < na2) {
        dyda[i-i/5] = dyda2[i]; i++;
    }
    delete [] a2;
    delete [] dyda2;
    return result;
}
