#ifndef CLUSTER_NEIGHBORHOOD_HH
#define CLUSTER_NEIGHBORHOOD_HH

#include <vector>
#include <utility>

#include "ClusterLocation.hh"

/**
 ** Namespace providing functionalities related to cluster neighborhoods.
 **
 ** Note: does not support unsigned coordinates (computes differences).
 ** However, the implementation will show poor performances if there is
 ** a large number of clusters with negative coordinates.
 **/
namespace Cluster { namespace Neighborhood {

/**
 ** Helper class encapsulating the different steps involved to filter an
 ** exact neighborhood from the original list of clusters.
 **/
template<class CoordinateT>
class Filter;

/**
 ** Find the neighborhood of each individual cluster, in a list of clusters.
 ** The neighborhood is sorted by increasing distance to the reference, and
 ** by increasing value of the cluster id.
 **
 ** Performance: the default implementation uses a regular grid to speedup 
 ** the filtering of the clusters. The dimensions (along x and y) of the cells
 ** in the grid are critical performance parameters: too small and the cost
 ** overhead of the management of the grid (many empty cells) will become 
 ** noticeable; too large, and each cell will contain a large number of non-
 ** neighbors.
 ** As a rule of thumb, when the distribution of clusters is reasonably 
 ** uniform across the tile, cells of the size of the neighborhood are close
 ** to the optimal.
 **
 ** Example:
 **     typedef ClusterLocation<int> Location;
 **     const vector<Location> locationList = initLocationList(); // client
 **     const unsigned int radius = 5;
 **     const Finder<int> finder(locationList, pair<int, int>(radius, radius));
 **     vector<Location> neighborhood;
 **     neighborhood.reserve(100);
 **     for (typename vector<Location>::const_iterator 
 **          location = locationList.begin(); locationList.end() > location; ++location)
 **     {
 **         finder.getNeighborhood(location, neighborhood);
 **         // use the neighborhood here
 **     }
 **/
template<class CoordinateT>
class Finder
{
public:
    typedef ClusterLocation<CoordinateT> Location;
    typedef std::vector<Location> LocationList;
    typedef typename LocationList::const_iterator LocationIterator;
    typedef std::pair<CoordinateT, CoordinateT> CoordinatePair;
    Finder(LocationList &tile, const CoordinatePair &binSize);
    // find the neighborhood in the given radius, sorted by distance to the reference, and store it into the provided vector
    void getNeighborhood(const LocationIterator center, const CoordinateT radius, LocationList &neighborhood) const;
private:
    const Filter<CoordinateT> filter_;
    const LocationList &tile_;
};

}} // namespace Cluster { namespace Neighborhood {

#include "ClusterNeighborhood.cpp"
#endif // #ifndef CLUSTER_NEIGHBORHOOD_HH

