// QUAHOG.cpp
// Copyright (c) 2005, 2006 Solexa
// Author: A. J. Cox
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.

#include <vector>
#include <errno.h>

#include "QhgOptions.hh"
#include "QhgParameters.hh"
#include "QhgFilter.hh"
#include "SignalMeansTile.hh"
#include "IntensityData.h"
#include "SequenceData.h"
#include "Message.h"
#include "IoUtils.hh"

#include "Legacy.hh"

static char* quahog_cpp_id="@(#) $Id: QUAHOG.cpp,v 1.21 2008/04/01 10:50:34 craczy Exp $";

/**
 ** Stands for _QUA_lity _H_euristic _O_utput _G_enerator
 ** Generates quality metrics from a cluster intensity file.
 ** Generate signal means for each cycle
 ** Metrics are mainly to do with detecting the possibility of cluster mixing.
 **/

using namespace std;

int quahog(const QhgOptions &qhgOptions);

int main(int argc, char * const argv[])
{
    QhgOptions qhgOptions;
    if (CommandLine::Options::RUN == qhgOptions.init(argc, argv))
    {
        quahog(qhgOptions);
    }
}

int computeSignalMeans(const QhgOptions &qhgOptions, 
                       const string &input, 
                       const vector<float *> &intensities, 
                       const std::vector<QhgParameters> &clusterPurity,
                       const int numClusters, 
                       const int numChannels,
                       const int lane);

int quahog(const QhgOptions &qhgOptions)
{
    typedef ClusterLocation<int> Location;
    for (vector<string>::const_iterator input = qhgOptions.inputList.begin(); qhgOptions.inputList.end() != input; ++input)
    {
        const string intensityFile  = qhgOptions.intensityPrefix  + (*input) + qhgOptions.intensitySuffix;
        // read the intensity data
        int numClusters = 0;
        int numChannels = 0;       
        PositionIndex *clusterPos = NULL;
        std::vector<float *> intensities;
        IntensityData  intensityData(intensityFile.c_str(), IntensityData::compressionNone);
        const int readInputErrorCode = intensityData.read(intensities, &numClusters, &numChannels, &clusterPos);
        if (readInputErrorCode != 0) 
        {
            error("Failed to read file %s with error code %i.", intensityFile.c_str(), readInputErrorCode);
            return 1;
        }
        info("Found %i clusters.", numClusters);
        // Convert the old cluster positions to a vector of the new template class
        vector<Location> clusterLocationList(numClusters);
        for (int cluster = 0; cluster < numClusters; ++cluster)
        {
           clusterLocationList[cluster] = Location(clusterPos[cluster].X, clusterPos[cluster].Y, cluster);
        }
        // Build the Qhg parameters and output to qhg file
        QhgParametersBuilder qhgParametersBuilder(qhgOptions.useBases, qhgOptions.pureBases, clusterLocationList, intensities, numClusters, numChannels);
        const std::vector<QhgParameters> clusterPurity = qhgParametersBuilder.getQhgParametersList();
        const string quahogFile = qhgOptions.quahogSuffix.length() > 0 ? qhgOptions.quahogPrefix + (*input) + qhgOptions.quahogSuffix : "";
        info("Will send output to file %s.", quahogFile.c_str());
        writeQhgFile(quahogFile.c_str(), clusterPos, clusterPurity, numClusters);
        if (qhgOptions.signalMeans)
        {
            computeSignalMeans(qhgOptions, *input, intensities, clusterPurity, numClusters, numChannels, numClusters? clusterPos[0].lane: 1);
        }
        // Cleanup the allocated data
        delete [] clusterPos;
        for (int i = 0; i < int(intensities.size()); i++)
            delete [] intensities[i];
    }
    return 0;
}

int computeSignalMeans(const QhgOptions &qhgOptions, 
                       const string &input, 
                       const vector<float *> &intensities, 
                       const std::vector<QhgParameters> &clusterPurity,
                       const int numClusters, 
                       const int numChannels,
                       const int lane)
{
    // Load the sequence data
    const string sequenceFile   = qhgOptions.sequencePrefix   + input + qhgOptions.sequenceSuffix;
    const int numCycles = intensities.size();
    vector<unsigned char>sequence_array(numClusters*numCycles);
    //const int readSequenceErrorCode = read_sequence_file(sequenceFile.c_str(), (unsigned char *)sequence_array.data(), numClusters, numCycles);
    // vector doesn't implement data() in older versions of g++
    unsigned char *sequence_data = sequence_array.empty() ? 0 : &(sequence_array.front());
    const int readSequenceErrorCode = read_sequence_file(sequenceFile.c_str(), sequence_data, numClusters, numCycles);
    if (!readSequenceErrorCode)
    {
        vector<SignalMeans> signalMeansList;
        int numberOfValidClusters;
        // compute the signal means values
        SignalMeans::numberOfChannels = numChannels;
        signalMeansTile(signalMeansList, numberOfValidClusters, intensities, sequence_array, clusterPurity, *qhgOptions.qhgFilter);
        // output the signal means values
        const string signalMeansFile = qhgOptions.signalMeansPrefix + input + qhgOptions.signalMeansSuffix;
        //FILE *fh = fopen(signalMeansFile.c_str(), "w");
        IoUtils::AutoFile fh(signalMeansFile, "w");
        if (fh != 0)
        {
            writeSignalMeansTile(fh, signalMeansList, numClusters, numberOfValidClusters, lane);
        }
        return 0;
    }
    else
    {
        error("Failed to read file %s with error code %i.", sequenceFile.c_str(), readSequenceErrorCode);
        return 1;
    }
}

// end of file QUAHOG.cpp
