#include <string>
#include <map>

#include "QhgFilter.hh"
#include "Legacy.hh"

// Template to build a QhgFilter from an Accessor and a Comparator.
// The Accessor extracts a specific value fron the QhgParameters. 
// The Comparator compares that value to the given threshold.
template <class Accessor, class Comparator>
class QhgFilterImpl : public QhgFilter
{
    inline bool operator() (const QhgParameters &qhgParameters) const
    {
        // LEGACY: the qhgParameter is truncated to two digits to reproduce the 
        // results from version 0.3 (values truncated to 2 digits in the txt file
        return Comparator::compare(Legacy::truncate<2>(Accessor::access(qhgParameters)), threshold_);
        //return Comparator::compare(Accessor::access(qhgParameters), threshold_);
    }
};

class ChastityAccessor {
public:
    static float access(const QhgParameters &qhgParameters) {return qhgParameters.chastity;}
};

class PurityAccessor {
public:
    static float access(const QhgParameters &qhgParameters) {return qhgParameters.purity;}
};

class SimilarityAccessor {
public:
    static float access(const QhgParameters &qhgParameters) {return qhgParameters.similarity;}
};

class LtComparator {
public:
    static bool compare(float lhs, float rhs) {return lhs < rhs;}
};

class LeComparator {
public:
    static bool compare(float lhs, float rhs) {return lhs <= rhs;}
};
    
class EqComparator {
public:
    static bool compare(float lhs, float rhs) {return lhs == rhs;}
};


class NeComparator {
public:
    static bool compare(float lhs, float rhs) {return lhs != rhs;}
};

class GeComparator {
public:
    static bool compare(float lhs, float rhs) {return lhs >= rhs;}
};

class GtComparator {
public:
    static bool compare(float lhs, float rhs) {return lhs > rhs;}
};

typedef std::map<const std::string, std::map<const std::string, QhgFilter *> > QhgFilterMap;

void initQhgFilterMap(QhgFilterMap &qhgFilterMap)
{
    // Chastity
    std::map<const std::string, QhgFilter *> chastityFilterMap;
    chastityFilterMap["lt"] = new QhgFilterImpl<ChastityAccessor, LtComparator>();
    chastityFilterMap["le"] = new QhgFilterImpl<ChastityAccessor, LeComparator>();
    chastityFilterMap["eq"] = new QhgFilterImpl<ChastityAccessor, EqComparator>();
    chastityFilterMap["ne"] = new QhgFilterImpl<ChastityAccessor, NeComparator>();
    chastityFilterMap["ge"] = new QhgFilterImpl<ChastityAccessor, GeComparator>();
    chastityFilterMap["gt"] = new QhgFilterImpl<ChastityAccessor, GtComparator>();
    qhgFilterMap["chastity"] = chastityFilterMap;
    // Purity
    std::map<const std::string, QhgFilter *> purityFilterMap;
    purityFilterMap["lt"] = new QhgFilterImpl<PurityAccessor, LtComparator>();
    purityFilterMap["le"] = new QhgFilterImpl<PurityAccessor, LeComparator>();
    purityFilterMap["eq"] = new QhgFilterImpl<PurityAccessor, EqComparator>();
    purityFilterMap["ne"] = new QhgFilterImpl<PurityAccessor, NeComparator>();
    purityFilterMap["ge"] = new QhgFilterImpl<PurityAccessor, GeComparator>();
    purityFilterMap["gt"] = new QhgFilterImpl<PurityAccessor, GtComparator>();
    qhgFilterMap["purity"] = purityFilterMap;
    // Similarity
    std::map<const std::string, QhgFilter *> similarityFilterMap;
    similarityFilterMap["lt"] = new QhgFilterImpl<SimilarityAccessor, LtComparator>();
    similarityFilterMap["le"] = new QhgFilterImpl<SimilarityAccessor, LeComparator>();
    similarityFilterMap["eq"] = new QhgFilterImpl<SimilarityAccessor, EqComparator>();
    similarityFilterMap["ne"] = new QhgFilterImpl<SimilarityAccessor, NeComparator>();
    similarityFilterMap["ge"] = new QhgFilterImpl<SimilarityAccessor, GeComparator>();
    similarityFilterMap["gt"] = new QhgFilterImpl<SimilarityAccessor, GtComparator>();
    qhgFilterMap["similarity"] = similarityFilterMap;
}

const QhgFilter &getQhgFilter(const std::string &accessor, const std::string &comparator, float threshold) throw (UnsupportedQhgFilter)
{
    static QhgFilterMap qhgFilterMap;
    if (qhgFilterMap.empty())
    {
        initQhgFilterMap(qhgFilterMap);
    }
    std::map<const std::string, QhgFilter*> accessorMap = qhgFilterMap[accessor];
    if (accessorMap.empty()) throw UnsupportedQhgFilter("Unsupported filter: " + accessor);
    QhgFilter *qhgFilter = accessorMap[comparator];
    if (0 == qhgFilter) throw UnsupportedQhgFilter("Unsupported filter: " + comparator + " (for accessor " + accessor + ")");
    qhgFilter->threshold_ = threshold;
    return *qhgFilter;
}


