#include<getopt.h>
#include <map>
#include <sstream>

#include "QhgOptions.hh"
#include "Message.h"

using namespace std;
using namespace CommandLine;

int QhgOptions::getOptions(int argc, char * const argv[])
        throw (UnknownOptionException, MissingOptionException, InvalidArgumentException)
{
    const int valIntensityPrefix   = 1;
    const int valIntensitySuffix   = 2;
    const int valSequencePrefix    = 3;
    const int valSequenceSuffix    = 4;
    const int valQuahogPrefix      = 5;
    const int valQuahogSuffix      = 6;
    const int valSignalMeansPrefix = 7;
    const int valSignalMeansSuffix = 8;
    const char *shortOptionList = "hf:r:t:c:u:p:s";
    const struct option longOptionList[] =
    {   {"help",                      no_argument,            0, 'h'}
    ,   {"filter",              required_argument,            0, 'f'}
    ,   {"relation",            required_argument,            0, 'r'}
    ,   {"threshold",           required_argument,            0, 't'}
    ,   {"compression",         required_argument,            0, 'c'}
    ,   {"use-bases",           required_argument,            0, 'u'}
    ,   {"pure-bases",          required_argument,            0, 'p'}
    ,   {"with-signal-means",         no_argument, &signalMeans,   1}
    ,   {"intensity-prefix",    required_argument,            0,  valIntensityPrefix}
    ,   {"intensity-suffix",    required_argument,            0,  valIntensitySuffix}
    ,   {"sequence-prefix",     required_argument,            0,  valSequencePrefix}
    ,   {"sequence-suffix",     required_argument,            0,  valSequenceSuffix}
    ,   {"quahog-prefix",       required_argument,            0,  valQuahogPrefix}
    ,   {"quahog-suffix",       required_argument,            0,  valQuahogSuffix}
    ,   {"signal-means-prefix", required_argument,            0,  valSignalMeansPrefix}
    ,   {"signal-means-suffix", required_argument,            0,  valSignalMeansSuffix}
    ,   {0,       0,                 0,          0}
    };
    int longOptionIndex;
    int option;
    // Note: "case 'h'" returns immediately
    while (-1 != (option = getopt_long(argc, argv, shortOptionList, longOptionList, &longOptionIndex)))
    {
        switch (option)
        {
            case 0:
                break;
            case valIntensityPrefix:
                intensityPrefix = optarg;
                break;
            case valIntensitySuffix:
                intensitySuffix = optarg;
                break;
            case valSequencePrefix:
                sequencePrefix = optarg;
                break;
            case valSequenceSuffix:
                sequenceSuffix = optarg;
                break;
            case valQuahogPrefix:
                quahogPrefix = optarg;
                break;
            case valQuahogSuffix:
                quahogSuffix = optarg;
                break;
            case valSignalMeansPrefix:
                signalMeansPrefix = optarg;
                break;
            case valSignalMeansSuffix:
                signalMeansSuffix = optarg;
                break;
            case 'f':
                filter = optarg;
                break;
            case 'r':
                relation = optarg;
                break;
            case 't':
                threshold = Options::convert<float>(optarg);
                break;
            case 'c':
                compressionName = optarg;
                break;
            case 'u':
                useBasesString = optarg;
                break;
            case 'p':
                pureBases = Options::convert<unsigned int>(optarg);
                break;
            case 'h':
                return usage(argc, argv);
                break;
            case '?':
            case ':':
            default:
                throw UnknownOptionException("");
                break;
        }
    };
    return RUN;
}

void QhgOptions::processOptions() throw (InvalidArgumentException)
{
    try
    {
        qhgFilter = &getQhgFilter(filter, relation, threshold);
    }
    catch(const UnsupportedQhgFilter &e)
    {
        throw InvalidArgumentException(e.what());
    }
    static map<string, Compression> compressionMap;
    compressionMap["none"] = IntensityData::compressionNone;
    compressionMap["gzip"] = IntensityData::compressionGzip;
    compressionMap["bzip2"] = IntensityData::compressionBzip2;
    const map<string, Compression>::const_iterator found = compressionMap.find(compressionName);
    if (compressionMap.end() == found)
    {
        string message = "compression must be ";
        for(map<string, Compression>::const_iterator i = compressionMap.begin(); i != compressionMap.end(); ++i)
            message += (compressionMap.begin() == i ? "" : ", ") + i->first;
        throw InvalidArgumentException(message);
    }
    compression = found->second;
    useBases.resize(0);
    const unsigned int maxReadLength = 200;
    if (0 == useBasesString.size() || 0 == useBasesString.compare("odd"))
    {
        useBases.resize(maxReadLength, false);
        // Note: from an user point of view, the 1st base is numbered 1
        for (vector<bool>::iterator base = useBases.begin(); base < useBases.end(); base += 2)
            *base = true;
    }
    else if (0 == useBasesString.compare("all"))
    {
        useBases.resize(maxReadLength, true);
    }
    else
    {
        if (string::npos != useBasesString.find_first_not_of("yYnN"))
            throw InvalidArgumentException(useBasesString);
        for (string::const_iterator c = useBasesString.begin(); c != useBasesString.end(); ++c)
        {
            useBases.push_back( 'y' == *c || 'Y' == *c );
        }
    }
}

void QhgOptions::getParameters(int argc,  char * const argv[]) throw (InvalidArgumentException)
{
    // get the input files stripped from their prefix and suffix
    while (argc > optind)
    {
        const string arg = argv[optind];
        if (arg.length() <= intensityPrefix.length() + intensitySuffix.length())
        {
            throw InvalidArgumentException(arg + "does not match input prefix " + intensityPrefix + " and suffix " + intensitySuffix);
        }
        if (arg.compare(0, intensityPrefix.length(), intensityPrefix))
        {
            throw InvalidArgumentException(arg + "does not match input prefix " + intensityPrefix);
        }
        if (arg.compare(arg.length() - intensitySuffix.length(), intensitySuffix.length(), intensitySuffix))
        {
            throw InvalidArgumentException(arg + "does not match input suffix " + intensitySuffix);
        }
        inputList.push_back(arg.substr(intensityPrefix.length(), arg.length() - intensityPrefix.length() - intensitySuffix.length()));
        ++optind;
    } 
}

int QhgOptions::usage(int argc, char * const argv[], const std::string &message)
{
    if (0 != message.size())
    {
        error(message.c_str());
    }
    const string program = argv[0];
    const string padding = string(program.length(), ' ');
    ostringstream os;
    os << argv[0] 
       << " [--help] [--with-signal-means]" << endl << padding
       << " [-f filter -r relation -t threshold]" << endl << padding
       << " [-u bases] [-p pure] [-c compression]" << endl << padding
       << " [--intensity-prefix prefix] [--intensity-suffix suffix]" << endl << padding
       << " [--sequence-prefix prefix] [--sequence-suffix suffix]" << endl << padding
       << " [--quahog-prefix prefix] [--quahog-suffix suffix]" << endl << padding
       << " [--signal-mean-prefix prefix] [--signal-mean-suffix suffix]" << endl << padding
       << " input-file1 input-file2..." << endl
       << endl
       << "-h: prints this help and exits" << endl
       << "-f filter: type of filter - chastity, purity or similarity. [chastity]" << endl
       << "-r relation: type of relation for the filter - lt, le, eq, ne, ge, or gt. [ge]" << endl
       << "-t threshold: value of the threshold to use for the filter. [0.6]" << endl
       << "-u bases: the bases to use - a sequence of 'Y' ans 'N', all (YYYYY...) or odd (YNYNYN...). [odd]" << endl
       << "-p pure: bases to purity filter on. [1]" << endl
       << "-c compression: none, gzip or bzip2. [none]" << endl
       << "--intensity-prefix prefix: prefix of the intensity files." << endl
       << "--intensity-suffix suffix: suffix of the intensity files. [_sig2.txt]" << endl
       << "--sequence-prefix prefix: prefix of the sequence files." << endl
       << "--sequence-suffix suffix: suffix of the sequence files. [_seq.txt]" << endl
       << "--quahog-prefix prefix: prefix of the quahog files." << endl
       << "--quahog-suffix suffix: suffix of the quahog files. [_qhg.txt]" << endl
       << "--signal-mean-prefix prefix: prefix of the signal-mean files." << endl
       << "--signal-mean-suffix suffix: suffix of the signal-mean files. [_all.txt]" << endl
       << "input-file: intensity data (sig2) of the clusters located on a tile" << endl
       << endl
       << "Generates quality metrics from a cluster intensity file. These metrics require " << endl
       << "the intensity data from the input file and the corresponding sequence data. They" << endl
       << "are split into two different output files: the quahog file (with chastity, purity," << endl
       << "and neighborhood of each cluster), and the signal-mean file, with the consolidated" << endl
       << "data for each individual cycle." << endl
       << "These metrics are mainly used to detecting the possibility of cluster mixing." <<endl
       << endl;
    info(os.str().c_str());
    return NO_RUN;
}

