#include <iostream>
#include <algorithm>
#include <map>
#include <cassert>
#include <math.h>

#include "QhgParameters.hh"
#include "ClusterNeighborhood.hh"

#include "Message.h"

using namespace std;

const int QhgParametersBuilder::neighborhoodRadius = 10;
const int QhgParametersBuilder::binSizeX = 5;
const int QhgParametersBuilder::binSizeY = 5;

/**************************************************
 * QhgParameters::QhgParameters
 **************************************************/
QhgParameters::QhgParameters
( const float       * ints,
  const int           numCycles,
  const int           numChannels,
  const uint          pureBases,
  const vector<bool>& basesToUse ) : purity(1000000), chastity(1000000),
                                  similarity(-1.), distance(9999*9999)
 {
    uint basesGot(0);
    neighbours.resize(0);
    vector<float> thisInten(numChannels);
    vector<bool>::const_iterator i(basesToUse.begin());

    for (int j = 0; j < numCycles; j++) {
        for (int k = 0; k < numChannels; k++)
            thisInten[k] = ints[j*numChannels+k];
        if (i==basesToUse.end()) {
            cerr << "Error: need " << pureBases
                    << " bases, but -useBases only gives locations for "
                    << basesGot << endl;
            exit(-1);
        }

        if ((*i)==true) {
            float thisPure, thisChaste;
            intensity.push_back(thisInten);
            sort(thisInten.begin(), thisInten.end() );
            if (thisInten[3]<=0.) {
                thisPure=0;
                thisChaste=0;
            } // ~if
            else {
                float sum = thisInten[0]+thisInten[1]+thisInten[2]+thisInten[3];
                if (sum > 0.)
                    thisPure = thisInten[3] / sum;
                else
                    thisPure = 0.;
                sum = thisInten[2]+thisInten[3];
                if (sum > 0.)
                    thisChaste = thisInten[3] / sum;
                else
                    thisChaste = 0.;
            } // ~else          

            if (thisPure<purity)
                purity=thisPure;
            if (thisChaste<chastity)
                chastity=thisChaste;

            if (++basesGot==pureBases)
                return;
        } // ~if
        i++;
    } // ~do
}

/**************************************************
 * getSimilarity: work out similarity between intensity vectors
 * of two clusters. Worked out as cos of angle between them.
 * This is 1 if they are in exactly same direction
 * This is 0 if they are at right angles (e.g perfect A vs. perfect G)
 * A negative value is a sign something weird is going on (e.g. with
 * matrix correction), since it can't happen for positive intensity values
 * Needs to be interpreted with care, as for single template experiment
 * we expect all 1st cycle intensities to be similar. For a random mix
 * of templates, any 2 templates have a 1 in 4 chance per base of being similar
 **************************************************/
float getSimilarity( const QhgParameters& c1, const QhgParameters& c2 )
{
  assert(c1.intensity.size()==c2.intensity.size());
  float similarity(999999);
  double dotProd(0.0), mag1(0.0), mag2(0.0);
  
  for (uint i(0);i<c1.intensity.size();i++)
  {
    dotProd=0.0;mag1=0.0;mag2=0.0;

    //  cout << c1.intensity[i][0] << " " << c1.intensity[i][1] << " " << c1.intensity[i][2] << " " << c1.intensity[i][3] << " - " << c2.intensity[i][0] << " " << c2.intensity[i][1] << " " << c2.intensity[i][2] << " " << c2.intensity[i][3] << endl;
  
    for (int j(0); j<4; j++)
    {
      mag1+=c1.intensity[i][j]*c1.intensity[i][j];
      mag2+=c2.intensity[i][j]*c2.intensity[i][j];
      dotProd+=c1.intensity[i][j]*c2.intensity[i][j];
    } // ~for
    mag1=sqrt(mag1);
    mag2=sqrt(mag2);
    mag1*=mag2;
    if (mag1!=0) dotProd/=mag1;
    if (dotProd<similarity) similarity=dotProd;
  }


  // if mag1==0 at least one of the vectors must have had zero
  // magnitude, so will return zero
  return similarity;
  
} 

QhgParametersBuilder::QhgParametersBuilder(const vector<bool> &basesToUse,
                                           const unsigned int pureBases,
                                           std::vector<ClusterLocation<int> > &tile,
                                           const std::vector<float * >   intensitiesList,
                                           const unsigned int numberOfClusters,
                                           const unsigned int numberOfChannels)
    : numberOfClusters_(numberOfClusters)
    , numberOfCycles_  (intensitiesList.size())
    , numberOfChannels_(numberOfChannels)
{
    typedef ClusterLocation<int> Location;
    typedef std::vector<Location> LocationList;
    typedef LocationList::const_iterator LocationIterator;
    initializeQhgParametersList(basesToUse, pureBases, intensitiesList);
    Cluster::Neighborhood::Finder<int> neighborhoodFinder(tile, pair<int, int>(binSizeX,binSizeY));
    LocationList neighborhood;
    neighborhood.reserve(1000);
    for(LocationIterator cluster = tile.begin(); cluster != tile.end(); ++cluster)
    {
        neighborhood.resize(0);
        neighborhoodFinder.getNeighborhood(cluster, neighborhoodRadius, neighborhood);
        assignNeighborhood(*cluster, neighborhood);
    }
}


void QhgParametersBuilder::initializeQhgParametersList(const vector<bool> &basesToUse,
                                     const unsigned int pureBases,
                                     const std::vector<float * >   intensitiesList)
{
    qhgParametersList_.resize(0);
    qhgParametersList_.reserve(numberOfClusters_);
    // storage for all the intensities for all cycles for a given cluster
    // using 2 a dimensional vector is 5% slower than using a flat one.
    //vector<vector<float> > clusterIntensitiesList(numberOfCycles_, vector<float>(numberOfChannels_));
    vector<float> clusterIntensitiesList(numberOfCycles_*numberOfChannels_);
    for (unsigned int cluster = 0; cluster < numberOfClusters_; cluster++)
    {
        for (unsigned int cycle = 0; cycle < numberOfCycles_; cycle ++)
        {
            for (unsigned int channel = 0; channel < numberOfChannels_; channel++)
            {
                clusterIntensitiesList[cycle*numberOfChannels_ + channel] = intensitiesList[cycle][cluster*numberOfChannels_ + channel];
            }
        }
        
        //qhgParametersList_.push_back(QhgParameters(clusterIntensitiesList.data(), numberOfCycles_, numberOfChannels_, pureBases, basesToUse));
        // data is not a standard member of the original STL vector
        float *intensitiesData = clusterIntensitiesList.empty() ? 0 : &(clusterIntensitiesList.front());
        qhgParametersList_.push_back(QhgParameters(intensitiesData, numberOfCycles_, numberOfChannels_, pureBases, basesToUse));
    }
}

void QhgParametersBuilder::assignNeighborhood(const ClusterLocation<int> &cluster, const vector<ClusterLocation<int> > &neighborhood)
{
    const unsigned int &clusterIndex = cluster.getId();
    if (!neighborhood.empty()) 
    {
        const int dx = cluster.getX() - neighborhood[0].getX();
        const int dy = cluster.getY() - neighborhood[0].getY();
        qhgParametersList_[clusterIndex].similarity = getSimilarity(qhgParametersList_[clusterIndex], qhgParametersList_[neighborhood[0].getId()]);
        qhgParametersList_[clusterIndex].distance   = dx*dx + dy*dy; // squared distance actually
        vector<int> neighbours(neighborhood.size());
        for(unsigned int i = 0; i < neighbours.size(); ++i) neighbours[i] = neighborhood[i].getId();
        qhgParametersList_[clusterIndex].neighbours = neighbours;
    }
    else 
    {
        // if no neighbours<maxDist, print out dummy similarity value
        // nearest neighbour found (may be an overestimate)
        // and dummy neighbourhood description
        //qhgParametersList_[i].neighbours.push_back(-1);
        qhgParametersList_[clusterIndex].neighbours = vector<int>(1,-1);
    } // ~else
}

int writeQhgFile(const char                    filename[],
                 const PositionIndex  * clusterPos,
                 const vector<QhgParameters> & clusterPurity,
                 const int                     numClusters)
{
    const char spacer = ':';
    // send output to stdout unless redirected
    FILE* pOut=stdout;
    if (strlen(filename)) {
        //info("Will send output to file %s.", filename);
        pOut=fopen(filename, "w");
        if (pOut==NULL) {
            error("Problem opening output file %s", filename);
            exit(-1);
        } // ~if
    } // ~if
    else
        info("Will send output to standard output.");
    for (int i(0); i < numClusters; i++) {
        fprintf(pOut, "%u\t%u\t%i\t%i" "",
                clusterPos[i].lane,
                clusterPos[i].tile,
                clusterPos[i].X,
                clusterPos[i].Y);
        fprintf(pOut, "\t%7.2f\t%7.2f\t%7.2f\t%7.2f\t",
                clusterPurity[i].purity,
                clusterPurity[i].chastity,
                clusterPurity[i].similarity,
                sqrt((double)clusterPurity[i].distance));
        for (uint j(0); j<clusterPurity[i].neighbours.size(); j++) {
            fprintf(pOut, "%i", clusterPurity[i].neighbours[j]);
            if (j < clusterPurity[i].neighbours.size() - 1)
                fprintf(pOut, "%c", spacer);
        }
        fprintf(pOut, "\n");
    } // ~for i
    if (strlen(filename)) {
        fclose(pOut);
    }
    return 0;
}

