/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <map>
#include <sstream>

#include <getopt.h>

#include "File_Buffer.h"
#include "Message.h"
#include "StringUtil.h"

#include "EQP_Options.h"

using namespace CommandLine;

/*****************************************************************************/

EQP_Options::EQP_Options()
    : CommandLine::Options(),
      my_num_pure_bases(my_default_num_pure_bases),
      my_multi_read_flag(false),
      my_sig2_compr_name("none"),
      my_sig2_compr(IntensityData::compressionNone),
      my_qval_compr_name("none"),
      my_qval_compr(File_Buffer::compressionNone),
      my_qual_fields_mask(0)
{
    const std::string in_read_cycle_str("in_read_cycle");
    const std::string unchastity_str("unchastity");
    const std::string max_early_unchastity_str("max_early_unchastity");
    const std::string raw_unquality_str("raw_unquality");

    my_str_field_map[in_read_cycle_str]
        = Quality_Data::FIELD_IN_READ_CYCLE_NUM;
    my_str_field_map[unchastity_str] = Quality_Data::FIELD_UNCHASTITY;
    my_str_field_map[max_early_unchastity_str]
        = Quality_Data::FIELD_MAX_EARLY_UNCHASTITY;
    my_str_field_map["max_local_unchastity"]
        = Quality_Data::FIELD_MAX_LOCAL_UNCHASTITY;
    my_str_field_map["signal_decay"]
        = Quality_Data::FIELD_SIGNAL_DECAY;
    my_str_field_map["homopol_len"] = Quality_Data::FIELD_HOMOPOL_LEN;
    my_str_field_map[raw_unquality_str] = Quality_Data::FIELD_RAW_UNQUALITY;

    my_default_str_list.push_back(in_read_cycle_str);
    my_default_str_list.push_back(unchastity_str);
    my_default_str_list.push_back(max_early_unchastity_str);
    my_default_str_list.push_back(raw_unquality_str);
}

/*****************************************************************************/

bool EQP_Options::get_read_cycle_range(unsigned int read_num,
                                       unsigned int& read_lo_cycle_num,
                                       unsigned int& read_hi_cycle_num)
{
    read_lo_cycle_num = 1;

    if (read_num == 0) {
        // Not considering reads separately, so use all cycles.

        read_hi_cycle_num = 0;

        for (Orig_Read_Lens_List_CIter
                 orig_read_len_citer(my_orig_read_lens_list.begin());
             orig_read_len_citer != my_orig_read_lens_list.end();
             ++orig_read_len_citer) {
            read_hi_cycle_num += *orig_read_len_citer;
        }

        return true;
    }

    if (read_num > my_orig_read_lens_list.size()) {
        return false;
    }

    Orig_Read_Lens_List_CIter
        orig_read_len_citer(my_orig_read_lens_list.begin());
    read_hi_cycle_num = *orig_read_len_citer++;

    for (unsigned int tmp_read_num(1); tmp_read_num < read_num;
         ++tmp_read_num) {
        read_lo_cycle_num = read_hi_cycle_num + 1;
        read_hi_cycle_num += *orig_read_len_citer++;
    }

    return true;
}

/*****************************************************************************/
// Simply uses the get_read_cycle_range method with read_num 0,
// i.e. no division into separate reads.

unsigned int EQP_Options::total_num_cycles()
{
    unsigned int read_num(0);
    unsigned int read_lo_cycle_num(0);
    unsigned int read_hi_cycle_num(0);

    get_read_cycle_range(read_num, read_lo_cycle_num, read_hi_cycle_num);
    return read_hi_cycle_num;
}

/*****************************************************************************/

int EQP_Options::getOptions(int argc, char* const argv[])
    throw (CommandLine::UnknownOptionException, 
           CommandLine::MissingOptionException, 
           CommandLine::InvalidArgumentException)
{
    const char* short_option_cstr = "hl:p:mz:Z:x:";

    const struct option long_option_arr[] =
        { {"help",                    no_argument, 0, 'h'},
          {"orig_read_lengths", required_argument, 0, 'l'},
          {"pure_bases",        required_argument, 0, 'p'},
          {"multi_read",              no_argument, 0, 'm'},
          {"sig2_compression",  required_argument, 0, 'z'},
          {"qval_compression",  required_argument, 0, 'Z'},
          {"predictors",        required_argument, 0, 'x'},
          {0, 0, 0, 0}
        };

    int long_option_index(0);
    int option(0);

    // Note: case 'h' returns immediately.
    while (-1 != (option = getopt_long(argc, argv,
                                       short_option_cstr, long_option_arr,
                                       &long_option_index))) {
        switch (option) {
        case 0:
            break;
        case 'h':
            return usage(argc, argv);
            break;
        case 'l':
            my_orig_read_lengths_str = optarg;
            break;
        case 'p':
            my_num_pure_bases = Options::convert<unsigned int>(optarg);
            break;
        case 'm':
            my_multi_read_flag = true;
            break;
        case 'z':
            my_sig2_compr_name = optarg;
            break;
        case 'Z':
            my_qval_compr_name = optarg;
            break;
        case 'x':
            my_qual_fields_str = optarg;
            break;

        case '?':
        case ':':
        default:
            throw UnknownOptionException("");
            break;
        }
    }

    return RUN;
}

/*****************************************************************************/
// FIXME : Either use File_Buffer::CompressionType in IntensityData
//         or at least move this functionality there.

bool resolve_compression(const std::string& compr_name,
                         IntensityData::CompressionType& compr_type,
                         std::string& msg_str)
{
    typedef std::map<std::string, IntensityData::CompressionType> 
        Str_Compr_Map;
    typedef Str_Compr_Map::const_iterator Str_Compr_Map_CIter;
    Str_Compr_Map str_compr_map;

    str_compr_map["none"] = IntensityData::compressionNone;
    str_compr_map["gzip"] = IntensityData::compressionGzip;
    str_compr_map["bzip2"] = IntensityData::compressionBzip2;

    Str_Compr_Map_CIter match_str_compr_citer(str_compr_map.find(compr_name));

    if (str_compr_map.end() == match_str_compr_citer) {
        msg_str += "compression must be";

        for (Str_Compr_Map_CIter str_compr_citer = str_compr_map.begin();
             str_compr_citer != str_compr_map.end(); ++str_compr_citer) {
            msg_str += (((str_compr_citer == str_compr_map.begin())
                         ? " " : " or ")
                        + str_compr_citer->first);
        }

        return false;
    }

    compr_type = match_str_compr_citer->second;
    return true;
}

/*****************************************************************************/

void EQP_Options::update_qual_fields_mask()
{
    for (Quality_Data::Field_List_CIter 
             field_citer(my_qual_field_list.begin());
         field_citer != my_qual_field_list.end(); ++field_citer) {
        my_qual_fields_mask += *field_citer;
    }
}

/*****************************************************************************/

bool EQP_Options::parse_qual_fields(std::string& msg_str)
{
    my_qual_field_list.clear();

    // Return default list if predictors not specified.
    if (my_qual_fields_str.empty()) {
        for (Str_List_CIter str_citer(my_default_str_list.begin());
             str_citer != my_default_str_list.end(); ++str_citer) {
            my_qual_field_list.push_back(my_str_field_map[*str_citer]);
        }

        my_qual_field_list.push_back(Quality_Data::FIELD_CALLED_BASE);
        update_qual_fields_mask();
        return true;
    }

    typedef std::vector<std::string> Str_Vec;
    typedef Str_Vec::const_iterator Str_Vec_CIter;

    Str_Vec qual_field_str_vec = tokenize_str(my_qual_fields_str, ":");

    unsigned int field_mask(0);

    for (Str_Vec_CIter str_citer = qual_field_str_vec.begin();
         str_citer != qual_field_str_vec.end(); ++str_citer) {
        Str_Field_Map_CIter match_citer(my_str_field_map.find(*str_citer));

        if (my_str_field_map.end() == match_citer) {
            msg_str += ("predictor list must be a colon-separated subset "
                        "of the following predictor names : ");

            for (Str_Field_Map_CIter str_field_citer(my_str_field_map.begin());
                 str_field_citer != my_str_field_map.end();
                 ++str_field_citer) {
                msg_str += (((my_str_field_map.begin() == str_field_citer)
                             ? " " : " or ")
                            + str_field_citer->first);
                return false;
            }
        }

        const Quality_Data::Field field(match_citer->second);

        if ((static_cast<unsigned int>(field) & field_mask) != 0) {
            msg_str += ("predictor `" + match_citer->first + 
                        "` is repeated in predictor list.");
            return false;
        }

        my_qual_field_list.push_back(field);
        field_mask |= static_cast<unsigned int>(field);
    }

    // The called base is always the last column.
    my_qual_field_list.push_back(Quality_Data::FIELD_CALLED_BASE);
    update_qual_fields_mask();
    return true;
}

/*****************************************************************************/

void EQP_Options::processOptions() 
    throw (CommandLine::InvalidArgumentException)
{
    // Parse colon-separated original read lengths list.

    typedef std::vector<std::string> Str_Vec;
    typedef Str_Vec::const_iterator Str_Vec_CIter;

    if (!my_orig_read_lengths_str.empty()) {
        unsigned int total_orig_read_len(0);
        Str_Vec read_len_str_vec = tokenize_str(my_orig_read_lengths_str, ":");
        unsigned int read_num(1);

        for (Str_Vec_CIter read_len_str_citer = read_len_str_vec.begin();
             read_len_str_citer != read_len_str_vec.end();
             ++read_len_str_citer, ++read_num) {
            const unsigned int orig_read_len
                = Options::convert<unsigned int>(*read_len_str_citer);
            total_orig_read_len += orig_read_len;
            my_orig_read_lens_list.push_back(orig_read_len);

            // If we are processing multiple reads separately, construct a 
            // list of them (assuming read numbers increment from 1).
            if (my_multi_read_flag) {
                my_read_num_vec.push_back(read_num);
            }
        }
    }


    // If we are not processing multiple reads separately, the list will
    // contain a single invalid read number (same convention as 
    // parseUseBases).
    if (!my_multi_read_flag) {
        my_read_num_vec.push_back(0);
    }

    if (my_multi_read_flag && (my_orig_read_lens_list.size() < 2)) {
        std::string msg_str("A list of per-read original read lengths (for "
                            "multiple reads) must be supplied when the "
                            "--multi_read option is specified.");
        throw InvalidArgumentException(msg_str);
    }


    // Parse compression types.

    std::string msg_str("sig2 ");

    if (!resolve_compression(my_sig2_compr_name, my_sig2_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }

    msg_str = "qval ";

    if (!resolve_compression(my_qval_compr_name, my_qval_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }


    // Parse predictor list.

    msg_str = "";

    if (!parse_qual_fields(msg_str)) {
        throw InvalidArgumentException(msg_str);
    }
}

/*****************************************************************************/

void EQP_Options::getParameters(int argc, char* const argv[])
        throw (CommandLine::InvalidArgumentException)
{
    bool sig2_file_found(false);
    bool prb_file_found(false);
    bool seq_file_found(false);

    while (argc > optind) {
        const std::string arg_str(argv[optind++]);

        if (!sig2_file_found) {
            my_sig2_path_str = arg_str;
            sig2_file_found = true;
        } else if (!prb_file_found) {
            my_prb_path_str = arg_str;
            prb_file_found = true;
        } else if (!seq_file_found) {
            my_seq_path_str = arg_str;
            seq_file_found = true;
        } else {
            my_qval_path_str_list.push_back(arg_str);
        }
    }

    if (!sig2_file_found) {
        throw MissingOptionException("Sig2 file not specified.");
    }

    if (!prb_file_found) {
        throw MissingOptionException("Prb file not specified.");
    }

    if (!seq_file_found) {
        throw MissingOptionException("Seq file not specified.");
    }

    if (my_multi_read_flag) {
        if (my_qval_path_str_list.size() == 1) {
            throw MissingOptionException("Only one output (qval) file "
                                         "supplied though --multi_read "
                                         "was specified.");
        }

        if (my_qval_path_str_list.size() != my_orig_read_lens_list.size()) {
            throw MissingOptionException("Number of original read lengths "
                                         "specified does not match number "
                                         "of output (qval) files.");
        }
    } else if (my_qval_path_str_list.size() > 1) {
        throw InvalidArgumentException("Multiple output (qval) files "
                                       "were specified without "
                                       "--multi_read option.");        
    } else if (my_qval_path_str_list.empty()) {
        throw MissingOptionException("No qval file specified.");
    }
}

/*****************************************************************************/

int EQP_Options::usage(int num_args, char* const arg_cstr_arr[],
                        const std::string& msg_str)
{
    if (0 != msg_str.size())
    {
        error(msg_str.c_str());
    }

    const std::string app_name(arg_cstr_arr[0]);
    std::ostringstream ostrm;

    ostrm << "Usage : " << app_name << " -h|--help" << std::endl
          << "   OR : " << app_name << std::endl
          << "[-l|--orig_read_lengths <ORIG_READ_LENGTHS> [-m|--multi_read]]"
          << std::endl
          << "[-p|--pure_bases <PURE_BASES>] " << std::endl
          << "[-m|--multi_read] " << std::endl
          << "[-z|--sig2_compression <gzip|bzip2>] " << std::endl
          << "[-Z|--qval_compression <gzip|bzip2>] " << std::endl
          << "[-x|--predictors <PREDICTOR_LIST>] " << std::endl
          << "<sig2_file> <prb_file> <seq_file> " << std::endl
          << "<qval_file | read1_qval_file read2_qval_file ...>" << std::endl
          << std::endl
          << "where :-" << std::endl
          << "ORIG_READ_LENGTHS is a colon-separated list " << std::endl
          << "PURE_BASES is the number of cycles over which " << std::endl
          << "  max_early_unchastity is calculated (default " 
          << my_default_num_pure_bases << ")" << std::endl
          << "PREDICTOR_LIST is a colon-separated list drawn from the "
          << std::endl
          << "  following valid predictor names :-" << std::endl;

    for (Str_Field_Map_CIter str_field_citer(my_str_field_map.begin());
         str_field_citer != my_str_field_map.end(); ++str_field_citer) {
        ostrm << "  " << str_field_citer->first << std::endl;
    }

    ostrm << std::endl
          << "The default list is :-" << std::endl;

    for (Str_List_CIter str_citer(my_default_str_list.begin());
         str_citer != my_default_str_list.end(); ++str_citer) {
        ostrm << ((str_citer == my_default_str_list.begin()) ? "  " : ":")
              << *str_citer;
    }

    ostrm << std::endl;
    info(ostrm.str().c_str());
    return NO_RUN;
}

/*****************************************************************************/
