/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <map>
#include <sstream>

#include <getopt.h>

#include "File_Buffer.h"
#include "Message.h"
#include "StringUtil.h"
#include "Tile_List.h"

#include "QApply_Options.h"

using namespace CommandLine;

/*****************************************************************************/

QApply_Options::QApply_Options()
    : CommandLine::Options(),
      my_read_num(0),
      my_num_cycles(0),
      my_qval_compr_name("none"),
      my_qval_compr(File_Buffer::compressionNone),
      my_numeric_output_flag(false),
      my_qval_suffix_str("_qval.txt")
{
    ;
}

/*****************************************************************************/

int QApply_Options::getOptions(int argc, char* const argv[])
    throw (CommandLine::UnknownOptionException, 
           CommandLine::MissingOptionException, 
           CommandLine::InvalidArgumentException)
{
    const char* short_option_cstr = "hZ:s:n:T:t:S:";

    const struct option long_option_arr[] =
        { {"help",                   no_argument,  0, 'h'},
          {"orig_read_lengths", required_argument, 0, 'l'},
          {"read",              required_argument, 0, 'r'},
          {"qval_compression",  required_argument, 0, 'Z'},
          {"symbolic",               no_argument,  0, 's'},
          {"numeric",                no_argument,  0, 'n'},
          {"tile_list_file",    required_argument, 0, 'T'},
          {"tile_prefix",       required_argument, 0, 't'},
          {"qval_suffix",       required_argument, 0, 'S'},
          {0, 0, 0, 0}
        };

    int long_option_index(0);
    int option(0);

    bool output_type_found(false);

    // Note: case 'h' returns immediately.
    while (-1 != (option = getopt_long(argc, argv,
                                       short_option_cstr, long_option_arr,
                                       &long_option_index))) {
        switch (option) {
        case 0:
            break;
        case 'h':
            return usage(argc, argv);
        case 'l':
            my_orig_read_lengths_str = optarg;
            break;

        case 'r':
            my_read_num = Options::convert<int>(optarg);
            break;

        case 'Z':
            my_qval_compr_name = optarg;
            break;

        case 's':
            // Output type must be specified at most once.
            if (output_type_found) {
                return usage(argc, argv);
            }

            output_type_found = true;
            my_numeric_output_flag = false;
            break;

        case 'n':
            // Output type must be specified at most once.
            if (output_type_found) {
                return usage(argc, argv);
            }

            output_type_found = true;
            my_numeric_output_flag = true;
            break;

        case 'T':
            my_tile_list_file_path_str = optarg;
            break;

        case 't':
            my_tile_prefix_str = optarg;
            break;

        case 'S':
            my_qval_suffix_str = optarg;
            break;

        case '?':
        case ':':
        default:
            throw UnknownOptionException("");
            break;
        }
    }

    return RUN;
}

/*****************************************************************************/

void QApply_Options::processOptions()
    throw (CommandLine::InvalidArgumentException)

{
    // Parse colon-separated original read lengths list.

    typedef std::vector<std::string> Str_Vec;
    typedef Str_Vec::const_iterator Str_Vec_CIter;

    Str_Vec read_len_str_vec = tokenize_str(my_orig_read_lengths_str, ":");

    if (my_read_num > read_len_str_vec.size()) {
        const std::string msg_str("Read number and original read lengths list "
                                  "are incompatible.");
        throw InvalidArgumentException(msg_str);
    }

    unsigned int read_num(1);

    for (Str_Vec_CIter read_len_str_citer = read_len_str_vec.begin();
         read_len_str_citer != read_len_str_vec.end();
         ++read_len_str_citer, ++read_num) {
        const unsigned int orig_read_len
                = Options::convert<unsigned int>(*read_len_str_citer);
        my_orig_read_lens_list.push_back(orig_read_len);

        // A read number of 0 means use all cycles (of all reads).
        if ((my_read_num == 0) || (read_num == my_read_num)) {
            my_num_cycles += orig_read_len;
        }
    }


    // Qval compression

    std::string msg_str("qval ");

    if (!resolve_compression(my_qval_compr_name, my_qval_compr, msg_str)) {
        throw InvalidArgumentException(msg_str);
    }

    // Tile file

    if (my_tile_list_file_path_str.empty() != my_tile_prefix_str.empty()) {
        msg_str = ("Either both or neither of --tile_list_file and "
                   "--tile_prefix must be specified.");
        throw InvalidArgumentException(msg_str);
    }
}

/*****************************************************************************/

void QApply_Options::getParameters(int argc, char* const argv[])
    throw (CommandLine::InvalidArgumentException)
{
    // qtable

    if (argc > optind) {
        my_qtable_path_str = argv[optind++];
    } else {
        throw InvalidArgumentException("No qtable specified");
    }

    // qval(s) : Support specification of qval files either by tile list file
    // OR by explicit listing but not both - on the assumption that an attempt
    // to do both is probably a user error.

    if (!my_tile_list_file_path_str.empty()) {
        Tile_List tile_list(my_tile_list_file_path_str);
        Tile_List::Tile_Name_List tile_name_list;

        if (tile_list.find_tiles(my_tile_prefix_str, tile_name_list,
                                 my_qval_suffix_str)) {
            my_qval_path_str_list
                = static_cast<Path_Str_List>(tile_name_list);
        } else {
            const std::string msg_str
                = ("Failed to read tile list from `"
                   + my_tile_list_file_path_str + "' for prefix `"
                   + my_tile_prefix_str + "'.");
            throw InvalidArgumentException(msg_str.c_str());
        }

        if (argc > optind) {
            const std::string msg_str
                ("Attempt to list one or more qval files explicitly after"
                 " specifying qval files by tile list file.");
            throw InvalidArgumentException(msg_str.c_str());
        }

        return;
    }

    // Explicit list

    const unsigned int qval_suffix_len(my_qval_suffix_str.length());

    while (argc > optind) {
        const std::string arg_str(argv[optind++]);

#ifdef CHECK_QVAL_SUFFIX_IN_EXPLICIT_LIST
        bool suffix_ok(true);

        if (arg_str.length() >= qval_suffix_len) {
            suffix_ok = (arg_str.compare(arg_str.length() - qval_suffix_len,
                                         qval_suffix_len,
                                         my_qval_suffix_str) == 0);
        } else {
            suffix_ok = false;
        }

        if (!suffix_ok) {
            throw InvalidArgumentException(arg_str 
                                           + " does not include suffix `"
                                           + my_qval_suffix_str + "'");
        }
#endif // ! CHECK_QVAL_SUFFIX_IN_EXPLICIT_LIST

        my_qval_path_str_list.push_back(arg_str);
    }

    if (my_qval_path_str_list.empty()) {
        throw InvalidArgumentException("At least one qval file must be "
                                       "specified using a tile list or "
                                       "by being explicitly listed.");
    }
}

/*****************************************************************************/

int QApply_Options::usage(int num_args, char* const arg_cstr_arr[],
                          const std::string& msg_str)
{
    if (0 != msg_str.size()) {
        error(msg_str.c_str());
    }

    const std::string app_name(arg_cstr_arr[0]);
    std::ostringstream ostrm;

    ostrm << "Usage : " << app_name << " -h|--help" << std::endl
          << "   OR : " << app_name << std::endl
          << "-l|--orig_read_lengths <ORIG_READ_LENGTHS>" << std::endl
          << "[-r|--read READ_NUM]" << std::endl
          << "[-Z|--qval_compression none|gzip|bzip2]" << std::endl
          << "[-n|--numeric | -s|--symbolic]" << std::endl
          << "<qtable-file>" << std::endl
          << "< <qvalue-file1> [<qvalue-file2> ...]" << std::endl
          << "  | -T|--tile_list_file <tiles.txt> " << std::endl
          << "    -t|--tile_prefix <lane_prefix e.g. s_3>" << std::endl
          << "   [-S|--qval_suffix _qval.txt] >" << std::endl
          << std::endl
          << "If the qval files are specified by means of a tile list file "
          << std::endl
          << "plus lane prefix filter, the qval_suffix is appended to the "
          << std::endl
          << "tile names to generate the qval filenames. A non-default "
          << std::endl
          << "suffix value may be required for paired read analysis, e.g. "
          << std::endl
          << "_1_qval.txt for read 1 and _2_qval.txt for read 2."
          << std::endl;

    info(ostrm.str().c_str());
    return NO_RUN;
}

/*****************************************************************************/
