/*
 * Copyright (c) Solexa 2006
 * Authors: Klaus Maisinger, A. J. Cox 
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#include <iostream>
#include <cstdio>
#include <string>

#include "DataSetFactory.h"
#include "Message.h"
#include "QApply_Options.h"
#include "QualityTable.h"

/*****************************************************************************/

// Identifier to embed in source and binary files
static char* quality_apply_id="@(#) $Id: QualityApply.cpp,v 1.13 2008/04/24 10:56:34 rshaw Exp $";

/*
 * Apply a quality calibration table as calculated by QualityCalibration.cpp
 * to a data set.
 */

const int bufferSize(128);

enum { ERR_OK=0, ERR_ARGS=1, ERR_DATA_FILE=2 };

/*****************************************************************************/

class QualityWriter
{
public:
    virtual void operator()(std::string& s, bool isFirst, int score)=0;
protected:
    char buffer[bufferSize];
}; // ~class QualityWriter

/*****************************************************************************/

// write numerical output as per Klaus' original code
class QualityWriterNumeric : public QualityWriter
{
public:
    virtual void operator()(std::string& s, bool isFirst, int score)
    {
        snprintf(buffer, sizeof(buffer), " %i", score);
        s += (buffer+(isFirst?1:0));
    }
}; // ~class QualityWriterNumeric

/*****************************************************************************/

// write compact ASCII-encoded output
class QualityWriterSymbolic : public QualityWriter
{
public:
    virtual void operator()(std::string& s, bool isFirst, int score)
    {
        snprintf(buffer, sizeof(buffer), "%c", 64+score);
        s+=buffer;
    }
}; // ~class QualityWriterSymbolic

/*****************************************************************************/

int writeQualityOutput(FILE *fp,
                       DataSet *set,
                       std::vector<int> *scores,
                       QualityWriter& writeQuality,
                       std::string& buf_str,
                       int& cycle_ind,
                       const int num_cycles,
                       bool is_first_data_set,
                       bool is_last_data_set)
{
    int i;
    std::vector<std::vector<float > > *values = set->getParameterValues();
    bool first_cycle_of_first_data_set(is_first_data_set);

    for (i = 0; i < scores->size(); i++) {
        const bool first_cycle_of_read(cycle_ind == 0);

        if (first_cycle_of_read && !first_cycle_of_first_data_set) {
            fprintf(fp, "%s\n", buf_str.c_str());
            buf_str = "";
        }

        writeQuality(buf_str,
                     (first_cycle_of_read || first_cycle_of_first_data_set),
                     scores->at(i));

        if (++cycle_ind == num_cycles) {
            cycle_ind = 0;
        }

        if (is_last_data_set && (i == scores->size() - 1)) {
            fprintf(fp, "%s\n", buf_str.c_str());
        }

        if (first_cycle_of_first_data_set) {
            first_cycle_of_first_data_set = false;
        }
    }
}

/*****************************************************************************/

int writeQualityOutput(const char filename[],
                       DataSet *set,
                       std::vector<int> *scores,
                       QualityWriter& writeQuality,
                       std::string& buf_str,
                       int& cycle_ind,
                       const int num_cycles,
                       bool is_first_data_set,
                       bool is_last_data_set)
{
    FILE *fp = fopen(filename, "r");
    if (!fp)
        return -1;
    int result = writeQualityOutput(fp, set, scores, writeQuality,
                                    buf_str, cycle_ind, num_cycles,
                                    is_first_data_set, is_last_data_set);
    fclose(fp);
    return result;
}

/*****************************************************************************/

int main(int num_args, char* arg_cstr_arr[])
{
    const std::string app_name(arg_cstr_arr[0]);
    int ret_val(ERR_OK);

    QApply_Options qapply_options;

    if (qapply_options.init(num_args, arg_cstr_arr)
        != CommandLine::Options::RUN) {
        return ERR_ARGS;
    }


    QualityTable *table = new QualityTable();
    QualityWriter* writer;

    if (qapply_options.my_numeric_output_flag) {
        writer = new QualityWriterNumeric();
    } else {
        writer = new QualityWriterSymbolic();
    }

    if (table->read(qapply_options.my_qtable_path_str.c_str()) < 0) {
        exit(ERR_DATA_FILE);
    }

#if 0
    table->print();
#endif

    int stored_num_params(0);
    bool is_first_data_set(true);
    std::string qval_path_str;
    bool at_eof(false);
    const int num_cycles(static_cast<int>(qapply_options.my_num_cycles));

    for (QApply_Options::Path_Str_List_CIter qval_path_str_citer
             = qapply_options.my_qval_path_str_list.begin();
         qval_path_str_citer != qapply_options.my_qval_path_str_list.end();
         ++qval_path_str_citer) {
        qval_path_str = *qval_path_str_citer;

        Data_Set_Factory data_set_factory(qval_path_str,
                                          qapply_options.my_qval_compr);
        DataSet* data_set_ptr(0);
        at_eof = false;

        std::string buf_str;
        int cycle_ind(0);

        while (data_set_factory.get_next_data_set(data_set_ptr, at_eof)) {
            const int detected_num_params(data_set_ptr->getNumberParameters());

            if (is_first_data_set) {
                info("Detected %i parameters.", detected_num_params);
                stored_num_params = detected_num_params;
            } else if (detected_num_params != stored_num_params) {
                error("Current number of parameters (%i) does not match "
                      "previously seen number of parameters (%i)",
                      detected_num_params, stored_num_params);
                ret_val = ERR_DATA_FILE;
                break;
            }

            std::vector<int> scores;

            table->applySolexa(data_set_ptr, scores);

            writeQualityOutput(stdout, data_set_ptr, &scores, *writer,
                               buf_str, cycle_ind, num_cycles,
                               is_first_data_set, at_eof);

            if (is_first_data_set) {
                is_first_data_set = false;
            }


            /* FIXME : rjs : reuse one DataSet to minimise reallocation? */
            delete data_set_ptr;
            data_set_ptr = 0;

            if (at_eof) {
                break;
            }
        }

        if (ret_val != ERR_OK) {
            break;
        }
    }

    if (!at_eof) {
        // get_next_data_set call failed but not because of EOF
        // -> failed to open file at all.
        error("Failure while attempting to read from `%s'",
              qval_path_str.c_str());
        ret_val = ERR_DATA_FILE;
    }

    delete table;
    delete writer;
    return ret_val;
} // ~main

/*****************************************************************************/
