/*
 * Copyright (c) Solexa 2006, Illumina 2008.
 * Author: Klaus Maisinger
 * Extended by: Richard Shaw
 * This source file is covered by the 
 * "Illumina Genome Analyzer Software License Agreement"
 * and bound by the terms therein.
 */

#include <cstdio>
#include <string>
#include <vector>

#include "QCal_Options.h"
#include "DataSetFactory.h"
#include "Message.h"
#include "QualityTable.h"
#include "XMLConfigHandler.h"
#include "XMLSAXParser.h"

/*****************************************************************************/

enum { ERR_OK=0, ERR_ARGS=1, ERR_CFG_FILE=2, ERR_XML_PARSE=3,
       ERR_DATA_FILE=4 };

/*****************************************************************************/

bool replace_suffix(const std::string& path_str,
                    const std::string& old_suffix_str,
                    const std::string& new_suffix_str,
                    std::string& new_path_str)
{
    const unsigned int old_suffix_len(old_suffix_str.length());

    if (path_str.length() < (old_suffix_len + 1)) {
        return false;
    }

    if (path_str.compare(path_str.length() - old_suffix_len,
                         old_suffix_len, old_suffix_str) != 0) {
        return false;
    }

    new_path_str = (path_str.substr(0, path_str.length() - old_suffix_len)
                    + new_suffix_str);
    return true;
}

/*****************************************************************************/

int main(int num_args, char* arg_cstr_arr[]) 
{
    const std::string app_name(arg_cstr_arr[0]);
    int ret_val(ERR_OK);

    QCal_Options qcal_options;
    if (qcal_options.init(num_args, arg_cstr_arr)
        != CommandLine::Options::RUN) {
        return ERR_ARGS;
    }


    // Open config file.

    FILE* cfg_file_hndl = fopen(qcal_options.my_cfg_path_str.c_str(), "r");

    if (!cfg_file_hndl) {
        error("Failed to open configuration file `%s'",
              qcal_options.my_cfg_path_str.c_str());
        return ERR_CFG_FILE;
    }

    // Parse config file.

    XMLConfigHandler cfg_handler;
    XMLSAXParser* sax_parser_ptr = new XMLSAXParser();
    sax_parser_ptr->setContentHandler(&cfg_handler);
    sax_parser_ptr->parse(cfg_file_hndl);
    fclose(cfg_file_hndl);

    Parameter_Config_List param_cfg_list;

    if (!cfg_handler.get_param_cfg_list(param_cfg_list)) {
        error("Failed to extract parameter configurations from `%s'",
              qcal_options.my_cfg_path_str.c_str());
        exit(ERR_XML_PARSE);
    }

    if (param_cfg_list.empty()) {
        error("No parameter configurations found in `%s'",
              qcal_options.my_cfg_path_str.c_str());
        exit(ERR_XML_PARSE);
    }


    // Populate bin_boundaries.

    const int no_params = param_cfg_list.size();
    unsigned int param_ind(0);

    std::vector<std::vector<float> > bin_boundaries;
    bin_boundaries.resize(no_params);

    for (Parameter_Config_List_CIter param_cfg_citer = param_cfg_list.begin();
         param_cfg_citer != param_cfg_list.end();
         ++param_cfg_citer, ++param_ind) {
        param_cfg_citer->get_bin_boundary_vec(bin_boundaries[param_ind]);
    }

#if 0
    // DEBUG : Dump bin boundaries
    typedef std::vector<float> Float_Vec;
    typedef Float_Vec::const_iterator Float_Vec_CIter;

    for (int param_ind(0); param_ind < no_params; ++param_ind) {
        for (Float_Vec_CIter float_citer(bin_boundaries[param_ind].begin());
             float_citer != bin_boundaries[param_ind].end(); ++float_citer) {
            std::cout << " " << *float_citer;
        }

        std::cout << std::endl << std::endl;
    }
#endif

    // add parameters to table
    QualityTable *table = new QualityTable();

    for (int i(0); i < bin_boundaries.size(); i++)
        table->addParameter(bin_boundaries[i]);

    // fill table with values

    const std::string qval_suffix_str(qcal_options.my_qval_suffix_str);
    const std::string qref_suffix_str(qcal_options.my_qref_suffix_str);
    File_Buffer::Compression_Type qval_file_compr(qcal_options.my_qval_compr);
    File_Buffer::Compression_Type qref_file_compr(qcal_options.my_qref_compr);

    DataSet* data_set_ptr(0);
    std::string curr_qref_file_path_str;

    QCal_Options::File_Name_List_CIter qval_name_citer
         = qcal_options.my_qval_file_name_list.begin();
    QCal_Options::File_Name_List_CIter qval_name_end_citer
        = qcal_options.my_qval_file_name_list.end();
    const unsigned int 
        num_qval_files(qcal_options.my_qval_file_name_list.size());

    for (unsigned int qval_file_ind(1);
         qval_name_citer != qval_name_end_citer;
         ++qval_name_citer, ++qval_file_ind) {
        const std::string curr_qval_file_path_str(*qval_name_citer);

        fprintf(stderr, "%u %u %s\n", num_qval_files, qval_file_ind,
                curr_qval_file_path_str.c_str());
        
        // qval suffix already checked in QCal_Options.
        replace_suffix(curr_qval_file_path_str, qval_suffix_str,
                       qref_suffix_str, curr_qref_file_path_str);

        Data_Set_Factory data_set_factory(curr_qval_file_path_str,
                                          qval_file_compr,
                                          curr_qref_file_path_str,
                                          qref_file_compr);
        bool at_eof(false);
        bool first_data_set(true);
        int detected_num_params(0);

        while (data_set_factory.get_next_data_set(data_set_ptr, at_eof)) {
            if (data_set_ptr->getNumberValues() == 0) {
                break;
            }

            detected_num_params = data_set_ptr->getNumberParameters();

            if (first_data_set) {
                info("Detected %i parameters.", detected_num_params);
                first_data_set = false;
            }

            if (detected_num_params != no_params) {
                error("Wrong number of parameters: %i", 
                      data_set_ptr->getNumberParameters());
                break;
            }

            // This addData adds data in the sense of updating statistics
            // using it rather than allocating additional space for it.
            table->addData(data_set_ptr);

            /* FIXME : rjs : reuse one DataSet to minimise reallocation? */
            delete data_set_ptr;
            data_set_ptr = 0;

            if (at_eof) {
                break;
            }
        }

        if (!at_eof) {
            // get_next_data_set call failed but not because of EOF
            // -> failed to open qval file at all or in reading qref.
            const std::string 
                msg_end_str(curr_qref_file_path_str.empty()
                            ? ""
                            : ((std::string(" and `")
                                + curr_qref_file_path_str + "'")));
            error("Failure while attempting to read from `%s'%s",
                  curr_qval_file_path_str.c_str(), msg_end_str.c_str());
            ret_val = ERR_DATA_FILE;
            break;
        }
    }

#if 0
    table->printDataTable();
    table->error_table.create(&(table->data_table));
    table->printErrorTable();
#endif
    table->create();
    table->print();
    delete table;

    return ret_val;
}

/*****************************************************************************/
