/*****************************************************************************/
//
// Copyright (c) Illumina 2008
// Author: R. J. Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
//
/*****************************************************************************/

#include <cassert>

#include "File_Buffer.h"
#include "Maths.h"
#include "StringUtil.h"

#include "QualityData.h"

/*****************************************************************************/

const std::string Quality_Data::hdr_str = "# PB PB P PB B T\n";

/*****************************************************************************/

Quality_Data::Quality_Data(const std::string& data_file_name,
                           const Field_List& field_list,
                           const File_Buffer::Compression_Type compression)
    : my_file_buffer(data_file_name, compression), my_field_list(field_list),
      my_have_written_flag(false)
{
    // field widths excluding tabs
    my_field_width_map[FIELD_IN_READ_CYCLE_NUM] = 3; // up to 999!
    my_field_width_map[FIELD_UNCHASTITY] = 10; // as per intensities
    my_field_width_map[FIELD_MAX_EARLY_UNCHASTITY] = 10; // as per intensities
    my_field_width_map[FIELD_MAX_LOCAL_UNCHASTITY] = 10; // as per intensities
    my_field_width_map[FIELD_SIGNAL_DECAY] = 10; // as per intensities
    my_field_width_map[FIELD_HOMOPOL_LEN] = 3; // unlikely but match cycle
    my_field_width_map[FIELD_RAW_UNQUALITY] = 5; // enough?
    my_field_width_map[FIELD_CALLED_BASE] = 1; // char
}

/*****************************************************************************/

Quality_Data::~Quality_Data()
{
    ;
}

/*****************************************************************************/

void write_quality_predictor(char*& buf_ch_ptr,
                             Quality_Predictors_Vec_CIter qual_preds_citer,
                             Quality_Data::Field field,
                             unsigned int expected_width)
{
    unsigned int width(0);

    // last char written is '\0' : limit includes this; count does not.

    switch (field) {
    case Quality_Data::FIELD_IN_READ_CYCLE_NUM:
        width = snprintf(buf_ch_ptr, expected_width + 1, "%i",
                         qual_preds_citer->in_read_cycle_num);
        break;
    case Quality_Data::FIELD_UNCHASTITY:
        width = snprintf(buf_ch_ptr, expected_width + 1, "%4.2f",
                         qual_preds_citer->unchastity);
        break;
    case Quality_Data::FIELD_MAX_EARLY_UNCHASTITY:
        width = snprintf(buf_ch_ptr, expected_width + 1, "%4.2f",
                         qual_preds_citer->max_early_unchastity);
        break;
    case Quality_Data::FIELD_MAX_LOCAL_UNCHASTITY:
        width = snprintf(buf_ch_ptr, expected_width + 1, "%4.2f",
                         qual_preds_citer->max_local_unchastity);
        break;
    case Quality_Data::FIELD_SIGNAL_DECAY:
        width = snprintf(buf_ch_ptr, expected_width + 1, "%4.2f",
                         qual_preds_citer->signal_decay);
        break;
    case Quality_Data::FIELD_HOMOPOL_LEN:
        width = snprintf(buf_ch_ptr, expected_width + 1, "%i",
                         qual_preds_citer->homopol_len);
        break;
    case Quality_Data::FIELD_RAW_UNQUALITY:
        width = snprintf(buf_ch_ptr, expected_width + 1, "%i",
                         qual_preds_citer->raw_unquality);        
        break;
    case Quality_Data::FIELD_CALLED_BASE:
        width = sprintf(buf_ch_ptr, "%c",
                        qual_preds_citer->called_base);
        break;
    }

    buf_ch_ptr += Maths::min(width, expected_width);
}

/*****************************************************************************/

int Quality_Data::write(const Quality_Predictors_Vec& quality_predictors_vec)
{
    int err_num(0);

    // Calculate the required size of buffer and allocate it.

    const unsigned int num_separs(my_field_list.size() - 1);
    const unsigned int line_end_width(1);

    const unsigned int num_lines(quality_predictors_vec.size());

    // Cache the expected field widths in an array to avoid a map lookup
    // for every single field of every single base.
    unsigned int* expected_width_arr_ptr
        = new unsigned int[my_field_list.size()];
    unsigned int* expected_width_ptr(expected_width_arr_ptr);

    unsigned int line_buf_size(num_separs + line_end_width);

    for (Field_List_CIter field_citer(my_field_list.begin());
             field_citer != my_field_list.end(); ++field_citer) {
        const unsigned int field_width(get_field_width(*field_citer));
        line_buf_size += field_width;
        *expected_width_ptr++ = field_width;
    }

    const unsigned int file_buf_size((my_have_written_flag
                                      ? 0 : Quality_Data::hdr_str.size())
                                     + line_buf_size * num_lines);

    char* file_buf_ptr = new char[file_buf_size + 1];
    char* buf_ch_ptr(file_buf_ptr);

    assert(file_buf_ptr != 0);
    if (file_buf_ptr == 0) {
        const int out_of_memory(51); // well above write_file_* ranges.
        return out_of_memory;
    }

    
    // Flatten the vector of Quality_Predictors structs into a buffer.

    if (!my_have_written_flag) {
        // Write hdr only at beginning of the file, i.e. during first write.
        buf_ch_ptr += sprintf(buf_ch_ptr, Quality_Data::hdr_str.c_str());
    }

    for (Quality_Predictors_Vec_CIter 
             qual_preds_citer(quality_predictors_vec.begin());
         qual_preds_citer != quality_predictors_vec.end();
         ++qual_preds_citer) {

        expected_width_ptr = expected_width_arr_ptr;

        for (Quality_Data::Field_List_CIter field_citer(my_field_list.begin());
             field_citer != my_field_list.end(); ++field_citer) {

            // Just checking.
            assert(qual_preds_citer->called_base != '\0');

            write_quality_predictor(buf_ch_ptr, qual_preds_citer,
                                    *field_citer,
                                    *expected_width_ptr++);
            
            // Assumes called base is always the last field.
            buf_ch_ptr += sprintf(buf_ch_ptr,
                                  ((*field_citer != FIELD_CALLED_BASE)
                                   ? "\t" : "\n"));
        }
    };

    delete[] expected_width_arr_ptr;

    const int num_useful_bytes(buf_ch_ptr - file_buf_ptr);


    // Write the buffer to the file (with compression if required).

    if (!my_file_buffer.write_buffer(file_buf_ptr, num_useful_bytes)) {
        // FIXME : central error codes?
        err_num = 1;
    }

    delete [] file_buf_ptr;
    my_have_written_flag = true;
    return err_num;
}

/*****************************************************************************/

int Quality_Data::read_chunk(Quality_Predictors_Vec& quality_predictors_vec)
{
    return read_quality_file_chunk(my_file_buffer, quality_predictors_vec);
}

/*****************************************************************************/

unsigned int Quality_Data::get_field_width(const Field field) const
{
    Field_Width_Map_CIter field_width_citer = my_field_width_map.find(field);

    assert(field_width_citer != my_field_width_map.end());

    if (field_width_citer == my_field_width_map.end()) {
        return 0;
    }

    return field_width_citer->second;
}

/*****************************************************************************/
// FIXME : assumes the traditional 4 predictors.
//         Currently Quality_Data::read_chunk is not used anyway.

int read_quality_file_chunk(File_Buffer& file_buffer,
                            Quality_Predictors_Vec& quality_predictors_vec)
{
    quality_predictors_vec.clear();

    int err_num(0);
    bool at_eof(false);

    File_Buffer::Line_Vec line_vec;

    if (!file_buffer.read_some_lines(line_vec, at_eof)) {
        err_num = 1;
        return err_num;
    }

    // Parse out the predictors.

    Quality_Predictors qual_preds;

    for (File_Buffer::Line_Vec_CIter line_citer(line_vec.begin());
         line_citer != line_vec.end(); ++line_citer) {
        const std::string line_str(*line_citer);

        if (line_str.empty()) {
            continue;
        }

        if (line_str[0] == '#') {
            // Skip line as comment (but only if `#' is very first character).
            continue;
        }

        if (sscanf(line_str.c_str(), "%d %f %f %d %c",
                   &(qual_preds.in_read_cycle_num),
                   &(qual_preds.unchastity),
                   &(qual_preds.max_early_unchastity),
                   &(qual_preds.raw_unquality),
                   &(qual_preds.called_base)) == 5) {
            quality_predictors_vec.push_back(qual_preds);
        } else {
            err_num = 42; // above the read_file range
            break;
        }
    }

    return err_num;
}

/*****************************************************************************/
