// -*- c++ -*-
/*****************************************************************************/
//
// Copyright (c) Illumina 2008
// Author: R. J. Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
//
/*****************************************************************************/

#ifndef QUALITY_DATA_H
#define QUALITY_DATA_H

/*****************************************************************************/

#include <string>
#include <vector>
#include <list>
#include <map>

#include "File_Buffer.h"

/*****************************************************************************/

// For the Phred-like algorithm, the predictor values have to increase
// in the direction of decreasing expected quality; hence `unchastity', etc.
// It is possible that called_base might in future be incorporated into the 
// prediction but currently it is included only for alignment scoring.

struct Quality_Predictors {
    int in_read_cycle_num;
    float unchastity;
    float max_early_unchastity;
    float max_local_unchastity;
    float signal_decay;
    unsigned int homopol_len;
    int raw_unquality;
    char called_base;
};

typedef std::vector<Quality_Predictors> Quality_Predictors_Vec;
typedef Quality_Predictors_Vec::iterator Quality_Predictors_Vec_Iter;
typedef Quality_Predictors_Vec::const_iterator Quality_Predictors_Vec_CIter;

/*****************************************************************************/

class Quality_Data {
 public:
    typedef enum {
        FIELD_IN_READ_CYCLE_NUM    =   1,
        FIELD_UNCHASTITY           =   2,
        FIELD_MAX_EARLY_UNCHASTITY =   4,
        FIELD_MAX_LOCAL_UNCHASTITY =   8,
        FIELD_SIGNAL_DECAY       =  16,
        FIELD_HOMOPOL_LEN          =  32,
        FIELD_RAW_UNQUALITY        =  64,
        FIELD_CALLED_BASE          = 128
    } Field;

    typedef std::list<Field> Field_List;
    typedef Field_List::const_iterator Field_List_CIter;

    Quality_Data(const std::string& data_file_name,
                 const Field_List& field_list,
                 const File_Buffer::Compression_Type compression 
                 = File_Buffer::compressionNone);
    ~Quality_Data();

    int write(const Quality_Predictors_Vec& quality_predictors_vec);
    int read_chunk(Quality_Predictors_Vec& quality_predictors_vec);

    unsigned int get_field_width(const Field field) const;

    static const std::string hdr_str;

    const std::string file_path_str()
    { return my_file_buffer.file_path_str(); }

 private:
    File_Buffer my_file_buffer;
    Field_List my_field_list;

    typedef std::map<Field, unsigned int> Field_Width_Map;
    typedef Field_Width_Map::const_iterator Field_Width_Map_CIter;

    Field_Width_Map my_field_width_map;

    bool my_have_written_flag;
};

/*****************************************************************************/

int read_quality_file_chunk(File_Buffer& file_buffer,
                            Quality_Predictors_Vec& quality_predictors_vec);

/*****************************************************************************/

#endif // ! QUALITY_DATA_H

/*****************************************************************************/
