/*****************************************************************************/
// Copyright (c) Illumina 2007
// Author: Richard Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <sstream>
#include <cstdlib>

#include "XMLConfigHandler.h"

/*****************************************************************************/
// helper fns
/*****************************************************************************/

bool parse_uint(const std::string val_str, unsigned int& val)
{
    char* end_ptr(0);
    const char* cstr_ptr(val_str.c_str());
    const long long_val(strtol(cstr_ptr, &end_ptr, 10));

    if (end_ptr == cstr_ptr) {
        // long not found
        return false;
    }

    if (long_val < 0) {
        return false; // looking for uint
    }

    val = static_cast<unsigned int>(long_val);
    return true;
}

/*****************************************************************************/

bool parse_float(const std::string val_str, float& val) {
    char* end_ptr(0);
    const char* cstr_ptr(val_str.c_str());
    val = static_cast<float>(strtod(cstr_ptr, &end_ptr));
            
    if (end_ptr == cstr_ptr) {
        // Float not found
        return false;
    }

    return true;
}

/*****************************************************************************/

bool parse_float_list(const std::string val_str, Float_Vec& float_vec)
{
    std::istringstream istrm(val_str);
    std::string str;
    float val;

    while (istrm >> str) {
        if (!parse_float(str, val)) {
            return false;
        }
        
        float_vec.push_back(val);
    }

    return true;
}

/*****************************************************************************/
// Parameter_Config
/*****************************************************************************/

Parameter_Config::Parameter_Config()
    : my_spec_type(SPEC_TYPE_UNKNOWN),
      my_first_bin_low_bound(0.0), my_first_bin_high_bound(0.0),
      my_last_bin_low_bound(0.0), my_last_bin_high_bound(0.0),
      my_num_bins(0)
{
    unsigned int spec(SPEC_TYPE_ANY);
    my_item_name_info_map["Name"] = Item_Info(TYPE_STRING, spec);

    spec = SPEC_TYPE_BOUNDARY_LIST;
    my_item_name_info_map["BinBoundaries"] = Item_Info(TYPE_FLOAT_LIST, spec);

    spec = SPEC_TYPE_FIRST_LAST_NUM_BINS;
    my_item_name_info_map["FirstBinLowBound"]  = Item_Info(TYPE_FLOAT, spec);
    my_item_name_info_map["FirstBinHighBound"] = Item_Info(TYPE_FLOAT, spec);
    my_item_name_info_map["LastBinLowBound"]   = Item_Info(TYPE_FLOAT, spec);
    my_item_name_info_map["LastBinHighBound"]  = Item_Info(TYPE_FLOAT, spec);
    my_item_name_info_map["NumBins"]           = Item_Info(TYPE_INT, spec);
}

/*****************************************************************************/

void Parameter_Config::reset()
{
    // Simply set all the items as undefined (no need to change values). 

    for (Item_Name_Info_Map_Iter item_name_info_iter 
             = my_item_name_info_map.begin();
         item_name_info_iter != my_item_name_info_map.end();
         ++item_name_info_iter) {
        item_name_info_iter->second.my_defined_flag = false;
    }

    my_name = "";
    my_spec_type = SPEC_TYPE_UNKNOWN;
    my_first_bin_low_bound = my_first_bin_high_bound = 0.0;
    my_last_bin_low_bound = my_last_bin_high_bound = 0.0;
    my_bin_boundary_vec.clear();
    my_num_bins = 0;
}

/*****************************************************************************/

bool Parameter_Config::set_item_val(const std::string& item_name,
                                    const std::string& val_str)
{
    Item_Name_Info_Map_Iter item_name_info_iter
        = my_item_name_info_map.find(item_name);

    if (item_name_info_iter == my_item_name_info_map.end()) {
        // Unrecognised item name
        std::cerr << "Unrecognised config item name `" << item_name 
                  << "'" << std::endl;
        return false;
    }

    if (item_name_info_iter->second.my_defined_flag) {
        // Item value already set
        std::cerr << "Value already set for config item `" << item_name
                  << "' when value '" << val_str << "' found" << std::endl;
        return false;
    }

    if (val_str.empty()) {
        // An empty value is treated as an error.
        std::cerr << "Empty value found for config item `" << item_name
                  << "'" << std::endl;
        return false;
    }

    const unsigned int 
        item_spec_mask(item_name_info_iter->second.my_spec_mask);

    if (item_spec_mask != SPEC_TYPE_ANY) {
        if (my_spec_type == SPEC_TYPE_UNKNOWN) {
            // Set my_spec_type to first specific item spec mask.
            my_spec_type = item_spec_mask;
        } else if (item_spec_mask != my_spec_type) {
            // Item is from a conflicting set to previous item(s).
            std::cerr << "Config item `" << item_name 
                      << "' conflicts with a previously found item."
                      << std::endl;
            return false;
        }
    }

    switch (item_name_info_iter->second.my_type) {
    case TYPE_STRING:
        // Must be Name
        my_name = val_str;
        break;

    case TYPE_INT:
        // Must be NumBins
        {
            unsigned int val(0);

            if (!parse_uint(val_str, val)) {
                std::cerr << "Failed to extract uint value from `" << val_str
                          << "' for config item `" << item_name
                          << "'" << std::endl;
                return false;
            }

            if (val < 1) {
                // Zero bins
                std::cerr << "Number of bins must be strictly positive, found "
                          << val << std::endl;
                return false;
            }

            my_num_bins = val;
        }
        break;

    case TYPE_FLOAT:
        {
            float val(0);

            if (!parse_float(val_str, val)) {
                std::cerr << "Failed to extract float value from `" << val_str
                          << "' for config item `" << item_name
                          << "'" << std::endl;
                return false;
            }

            if (item_name == "FirstBinLowBound") {
                my_first_bin_low_bound = val;
            } else if (item_name == "FirstBinHighBound") {
                my_first_bin_high_bound = val;
            } else if (item_name == "LastBinLowBound") {
                my_last_bin_low_bound = val;
            } else if (item_name == "LastBinHighBound") {
                my_last_bin_high_bound = val;
            }
        }
        break;

    case TYPE_FLOAT_LIST:
        // Must be BinBoundaries
        if (!parse_float_list(val_str, my_bin_boundary_vec)) {
            std::cerr << "Failed to extract float values from `" << val_str
                      << "' for config item `" << item_name
                      << "'" << std::endl;
            return false;
        }
        break;
    }

    item_name_info_iter->second.my_defined_flag = true;
    return true;
}

/*****************************************************************************/

bool Parameter_Config::is_ready() const
{
    if (my_spec_type == SPEC_TYPE_UNKNOWN) {
        std::cerr << "Very incomplete configuration for parameter `" 
                  << my_name << "'" << std::endl;
        return false;
    }

    bool complete_set_found(true);

    for (Item_Name_Info_Map_CIter item_name_info_citer 
             = my_item_name_info_map.begin();
         item_name_info_citer != my_item_name_info_map.end();
         ++item_name_info_citer) {
        const Item_Info item_info(item_name_info_citer->second);

        if (!(((item_info.my_spec_mask & my_spec_type) == 0)
              || item_info.my_defined_flag)) {
            std::cerr << "Missing item : " << item_name_info_citer->first
                      << std::endl;
            complete_set_found = false;
        }
    }

    if (!complete_set_found) {
        std::cerr << "Incomplete configuration for parameter `"
                  << my_name << "'" << std::endl;
        return false;
    }

    return true;
}

/*****************************************************************************/

bool Parameter_Config::is_valid() const
{
    switch (my_spec_type) {
    case SPEC_TYPE_UNKNOWN:
        std::cerr << "Missing configuration items for `"
                  << my_name << "'" << std::endl;
        return false;

    case SPEC_TYPE_BOUNDARY_LIST:
        if (my_bin_boundary_vec.size() < 2) {
            std::cerr << "Number of bins must be >= 1 : found " 
                      << my_bin_boundary_vec.size()
                      << " for `" << my_name << "'" << std::endl;
            return false;
        }

        {
            float last_val(my_bin_boundary_vec[0] - 0.1); // first - some delta

            for (Float_Vec_CIter float_citer = my_bin_boundary_vec.begin();
                 float_citer != my_bin_boundary_vec.end(); ++float_citer) {
                if (*float_citer <= last_val) {
                    std::cerr << "Bin boundaries must be supplied in "
                              << "strictly increasing order for `"
                              << my_name << "'" << std::endl;
                    return false;
                }

                last_val = *float_citer;
            }
        }
        break;

    case SPEC_TYPE_FIRST_LAST_NUM_BINS:
        if (my_num_bins < 1) {
            std::cerr << "Number of bins must be >= 1 : found " << my_num_bins
                      << " for `" << my_name << "'" << std::endl;
            return false;
        }

        if (my_first_bin_low_bound >= my_first_bin_high_bound) {
            std::cerr << "First bin low bound (" << my_first_bin_low_bound
                      << ") >= high bound (" << my_first_bin_high_bound
                      << ") for `" << my_name << "'" << std::endl;
            return false;
        }

        if (my_last_bin_low_bound >= my_last_bin_high_bound) {
            std::cerr << "Last bin low bound (" << my_last_bin_low_bound
                      << ") >= high bound (" << my_last_bin_high_bound
                      << ") for `" << my_name << "'" << std::endl;
            return false;
        }

        switch (my_num_bins) {
        case 1:
            // Is it valid to have just one bin?
            if ((my_first_bin_low_bound != my_last_bin_low_bound)
                || (my_first_bin_high_bound != my_last_bin_high_bound)) {
                std::cerr << "One bin inconsistently specified for `"
                          << my_name << "'" << std::endl;
                return false;
            }
            break;

        case 2:
            if (my_first_bin_high_bound != my_last_bin_low_bound) {
                std::cerr << "Only two bins but not contiguous for `" 
                          << my_name << "'" << std::endl;
                return false;
            }
            break;

        default:
            if (my_first_bin_high_bound >= my_last_bin_low_bound) {
                std::cerr << "Bin overlap for `" << my_name << "'"
                          << std::endl;
                return false;
            }
        }

        break;

    default:
        return false;
    }

    return true;
}

/*****************************************************************************/

bool Parameter_Config::get_bin_boundary_vec(Float_Vec& bin_boundary_vec) const
{
    if (!(is_ready() && is_valid())) {
        return false;
    }

    if (my_spec_type == SPEC_TYPE_BOUNDARY_LIST) {
        if (my_bin_boundary_vec.size() < 2) {
            return false;
        }

        bin_boundary_vec = my_bin_boundary_vec;
    } else {
        // SPEC_TYPE_FIRST_LAST_NUM_BINS
        if (my_num_bins < 1) {
            return false;
        }

        bin_boundary_vec.push_back(my_first_bin_low_bound);
        bin_boundary_vec.push_back(my_first_bin_high_bound);

        if (my_num_bins > 1) {
            if (my_num_bins > 2) {
                const float step_size((my_last_bin_low_bound
                                       - my_first_bin_high_bound)
                                      / (my_num_bins - 2));
                float bound(my_first_bin_high_bound);

                for (unsigned int bin_ind(1); bin_ind < (my_num_bins - 2);
                     ++bin_ind) {
                    bound += step_size;
                    bin_boundary_vec.push_back(bound);
                }

                bin_boundary_vec.push_back(my_last_bin_low_bound);
            }

            bin_boundary_vec.push_back(my_last_bin_high_bound);
        }
    }

    return true;
}

/*****************************************************************************/

std::ostream& operator<<(std::ostream& ostrm, 
                         const Parameter_Config& param_cfg)
{
    ostrm << param_cfg.my_name << " :";

    switch (param_cfg.my_spec_type) {
    case Parameter_Config::SPEC_TYPE_UNKNOWN:
    default:
        ostrm << " NOT YET SPECIFIED";
        break;

    case Parameter_Config::SPEC_TYPE_BOUNDARY_LIST:
        for (Float_Vec_CIter float_citer
                 = param_cfg.my_bin_boundary_vec.begin();
             float_citer != param_cfg.my_bin_boundary_vec.end();
             ++float_citer) {
            ostrm << " " << *float_citer;
        }
        break;

    case Parameter_Config::SPEC_TYPE_FIRST_LAST_NUM_BINS:
        ostrm << " " << param_cfg.my_num_bins << " bins" << std::endl
              << "first bin : " << param_cfg.my_first_bin_low_bound
              << " <> " << param_cfg.my_first_bin_high_bound << std::endl
              << "last bin : " << param_cfg.my_last_bin_low_bound
              << " <> " << param_cfg.my_last_bin_high_bound << std::endl;

        if (param_cfg.my_num_bins > 2) {
            ostrm << "(width of other bins : "
                  << ((param_cfg.my_last_bin_low_bound
                      - param_cfg.my_first_bin_high_bound)
                      / (param_cfg.my_num_bins - 2))
                  << ")";
        }
        break;
    }

    ostrm << std::endl;
    return ostrm; 
}

/*****************************************************************************/
// XMLConfigHandler
/*****************************************************************************/

XMLConfigHandler::XMLConfigHandler()
  : XMLContentHandler(), my_parse_error_flag(false), my_in_param_flag(false)
{
    ;
}

/*****************************************************************************/

void XMLConfigHandler::startDocument()
{
    ;
}

/*****************************************************************************/

void XMLConfigHandler::startElement(const char name[], const Attributes *attr)
{
    const std::string ele_name(name);

    if (ele_name == "Parameter") {
        my_in_param_flag = true;
    } else if (my_in_param_flag) {
        my_curr_ele_name = ele_name;
    }
}

/*****************************************************************************/

void XMLConfigHandler::endElement(const char name[])
{
    const std::string ele_name(name);

    if (ele_name == "Parameter") {
        if (!my_in_param_flag) {
            my_parse_error_flag = true;
            std::cerr << "Parameter tag closed but not opened" << std::endl;
            return;
        }

        if (my_curr_param_cfg.is_ready() && my_curr_param_cfg.is_valid()) {
            my_param_cfg_list.push_back(my_curr_param_cfg);
        } else {
            std::cerr << "Parameter tag closed but parameter configuration "
                      << "incomplete or invalid for `"
                      << my_curr_param_cfg.name() << "'" << std::endl;
            my_parse_error_flag = true;
        }

        my_curr_param_cfg.reset();
    } else if (my_in_param_flag) {
        my_curr_ele_name = "";
    }
}

/*****************************************************************************/

void XMLConfigHandler::endDocument()
{
    ;
}

/*****************************************************************************/

void XMLConfigHandler::characters(const char ch[], int start, int length)
{
    // XMLSaxParser calls characters just once per tag, passing the entire
    // value in one go.

    if (!my_curr_param_cfg.set_item_val(my_curr_ele_name, ch)) {
        std::cerr << "Invalid value `" << ch << "' for `"
                  << my_curr_ele_name << "'" << std::endl;
        my_parse_error_flag = true;
    }
}

/*****************************************************************************/

bool XMLConfigHandler::get_param_cfg_list(Parameter_Config_List& 
                                          param_cfg_list) const
{
    param_cfg_list = my_param_cfg_list;
    return !my_parse_error_flag;
}

/*****************************************************************************/
