#!/usr/bin/env python

# Copyright (c) Solexa  2006
#               Illumina 2007
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

"""The script extracts a set of quality predictors for a phred-style
quality calibration from the run-folder. The predictors used
are currently:
- cycle number
- 1. - chastity of current base
- 1. - chastity of first 12 bases in read
- -Qbustard
"""

import sys, os, string, getopt

skipColumns = 4

def log(filename, level, message):
    try:
        open(filename, "a").writelines([ logline(level, message)])
    except:
        sys.stderr.write("Warning: Failed to write to %s." %
                         os.path.abspath(filename))
    

def error(message):
    """Central routine for handling error messages"""
    sys.stderr.write("Error: %s\n" % message)
    log("error.txt", 7, message)


def readParameters(signalfile, prbfile, alignfile, useBases,
                   PURE_BASES):
    """Extract the quality predictors from the data files."""
    signaldata = open(signalfile).readlines()
    prbdata =    open(prbfile).readlines()
    signaldata = filter(lambda x: x[0] != '#', signaldata)
    prbdata =    filter(lambda x: x[0] != '#', prbdata)
    if (len(prbdata) != len(signaldata)):
        error("Wrong number of lines: Mismatch between signal and prb files.")
        sys.exit(1)
        

    # read alignment data if given
    aligndata = None
    if alignfile:
        aligndata =  open(alignfile).readlines()
        aligndata =  filter(lambda x: x[0] != '#', aligndata)
        if (len(prbdata) != len(aligndata)):
            error("Wrong number of lines: "
                  "Mismatch between prb and align files.")
            sys.exit(1)
        if (len(aligndata) > 0):
            readlength = len(aligndata[0].split()[0])

    nreads = len(prbdata)
    if nreads == 0:
        return
    length = len(prbdata[0].split())/4
    readlength = length
    if useBases == "all":
        useBases = 'Y' * length
    if useBases == "odd":
        useBases = 'YN' * min(length/2)
        if length%2:
            useBases += 'Y'
    if len(useBases) < length:
        useBases += "n" * (length - len(useBases))
    cycleNumbers = range(1, length+1) 


    if aligndata:
        alignCycle = 0
        useCycles = [ -1 ] * length
        for j in range(length):
            if useBases[j] in ['y', 'Y']:
                if alignCycle < readlength:
                    useCycles[j] = alignCycle
                alignCycle += 1
        #print useCycles, cycleNumbers
        

    print "# PB PB P PB B T"
    for i in range(nreads):
        signal = signaldata[i].split("\t")[skipColumns:]
        prb = prbdata[i].split("\t")

        purity  = [ 1.] * length
        quality = [ 5. ] * length
        seqsequence = [ '.' ] * length
        refsequence = [ '.' ] * length
        purity2 = [ ]
        
        for j in range(length):
            quality[j] = - max(map(int, prb[j].split()))
            signals = map(float, signal[j].split()) 
            signals.sort()
            sum1 = (signals[-1]+signals[-2])
            if ((signals[-1] <= 0.) or (sum1 <= 0.)):
                purity[j] = 1.
            else:
                purity[j] = 1. - signals[-1]/sum1
            idx = cycleNumbers[j]
            if  idx <= PURE_BASES:
                purity2.append(purity[j])
           
        if aligndata:
            alignment = aligndata[i].split()
            refseq = "." * readlength
            sequence = alignment[0]
            if (len(alignment) >= 6):
                refseq = alignment[5]
                if alignment[4] == 'R':
                    refseq = list(refseq)
                    refseq.reverse()
                    refseq = string.translate(string.join(refseq, ""),
                                              string.maketrans("ACGT", "TGCA"))
            for j in range(length):
                idx = useCycles[j]
                if idx != -1:
                    seqsequence[j] = sequence[idx]
                    refsequence[j] = refseq[idx]

        purity2 = max(purity2)
        for j in range(length):
            print ("%i\t%4.2f\t%4.2f\t%i\t%c\t%c" %
                   (cycleNumbers[j], purity[j], purity2,
                    quality[j], seqsequence[j], refsequence[j]))


def usage():
    sys.stderr.write("Usage: extractQualityParams.py <sigfile> <prbfile> [<alignfile>]\n"
                     "           [--USE_BASES=<s>] [--PURE_BASES=<n>]\n")

if __name__ == "__main__":
    try:
        (options, args) = getopt.gnu_getopt(sys.argv[1:], "",
                                            [ "USE_BASES=",
                                              "PURE_BASES="])
    except:
        usage()
        sys.exit(1)

    PURE_BASES = 12
    USE_BASES ="all"
    
    if len(args) < 2:
        usage()
        sys.exit(1)

    signalfile = args[0]
    prbfile    = args[1]
    alignfile  = None
    if (len(args) > 2):
        alignfile = args[2]
        if not os.path.exists(alignfile):
            error("File %s not found." % alignfile)
            sys.exit(1)

    for o, a in options:
        if o == "--PURE_BASES":
            PURE_BASES = int(a)
        if o == "--USE_BASES":
            USE_BASES = a
            
    #print signalfile, prbfile, USE_BASES, PURE_BASES
    if not os.path.exists(signalfile):
        error("File %s not found." % signalfile)
        sys.exit(1)
    if not os.path.exists(prbfile):
        error("File %s not found." % prbfile)
        sys.exit(1)
    
    readParameters(signalfile, prbfile, alignfile, USE_BASES, PURE_BASES)
    
