/*****************************************************************************/
//
// Copyright (c) Illumina 2008
// Author: R. J. Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
//
/*****************************************************************************/
//
// Extracts a set of quality predictors for a phred-style quality calibration 
// from the run-folder. The predictors used are currently:
// - cycle number
// - 1. - chastity of current base
// - 1. - chastity of first 12 bases in read
// - -Qbustard
//
/*****************************************************************************/

#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cassert>

#include "EQP_Options.h"
#include "IntensityData.h"
#include "Message.h"
#include "PrbData.h"
#include "SequenceData.h"
#include "QualityData.h"

/*****************************************************************************/

enum {ERR_OK = 0, ERR_USAGE = 1, ERR_OUT_OF_MEMORY = 2, ERR_FILE_READ = 3,
      ERR_QUALITY_PREDICTORS = 4, ERR_FILE_WRITE = 5};

typedef std::vector<unsigned int> Cycle_Ind_Vec;
typedef Cycle_Ind_Vec::const_iterator Cycle_Ind_Vec_CIter;

typedef std::vector<float*> Sig_Ptr_Vec;
typedef Sig_Ptr_Vec::const_iterator Sig_Ptr_Vec_CIter;

typedef std::vector<float> Sig_Vec;
typedef Sig_Vec::const_iterator Sig_Vec_CIter;

typedef std::vector<float*> Prob_Ptr_Vec;
typedef Prob_Ptr_Vec::const_iterator Prob_Ptr_Vec_CIter;

typedef std::vector<float> Unchastity_Vec;
typedef Unchastity_Vec::iterator Unchastity_Vec_Iter;
typedef Unchastity_Vec::const_iterator Unchastity_Vec_CIter;

typedef std::vector<unsigned int> Len_Vec;
typedef Len_Vec::iterator Len_Vec_Iter;
typedef Len_Vec::const_iterator Len_Vec_CIter;

const unsigned int default_num_pure_bases(12);

const int min_possible_base_qv(-40);

/*****************************************************************************/
// unchastity = 1.0 - chastity

void calc_unchastity(const Sig_Vec& sig_vec, float& unchastity)
{
    float gold_sig(sig_vec[0]);
    float silver_sig(sig_vec[1]);

    for (Sig_Vec_CIter sig_citer = sig_vec.begin() + 1;
         sig_citer != sig_vec.end(); ++sig_citer) {
        const float sig(*sig_citer);

        if (sig > gold_sig) {
            silver_sig = gold_sig;
            gold_sig = sig;
        } else if (sig > silver_sig) {
            silver_sig = sig;
        }
    }

    const float sum(gold_sig + silver_sig);

    unchastity = (((gold_sig > 0.0) && (sum > 0.0))
                  ? (1.0 - (gold_sig / sum))
                  : 1.0);
}

/*****************************************************************************/

void find_max_in_win(const Unchastity_Vec_CIter left_citer,
                     const Unchastity_Vec_CIter right_citer,
                     Unchastity_Vec_CIter& max_citer)
{
    max_citer = left_citer;

    if (right_citer == left_citer) {
        return;
    }

    Unchastity_Vec_CIter tmp_citer(left_citer);

    // Because of above check can inc tmp_citer without missing break. 
    for (++tmp_citer; /* break below */; ++tmp_citer) {
        if ((*tmp_citer) > (*max_citer)) {
            max_citer = tmp_citer;
        }

        if (tmp_citer == right_citer) {
            break;
        }
    }
}

/*****************************************************************************/

void calc_local_max(const Unchastity_Vec& unchastity_vec,
                    const unsigned int num_neighs,
                    Unchastity_Vec& max_local_unchastity_vec)
{
    assert(!unchastity_vec.empty());
    if (unchastity_vec.empty()) {
        return;
    }

    if (max_local_unchastity_vec.size() != unchastity_vec.size()) {
        max_local_unchastity_vec.resize(unchastity_vec.size());
    }

    Unchastity_Vec_CIter left_citer(unchastity_vec.begin());
    Unchastity_Vec_CIter mid_citer(unchastity_vec.begin());
    Unchastity_Vec_CIter right_citer(unchastity_vec.begin());
    Unchastity_Vec_CIter max_citer(left_citer);
    Unchastity_Vec_CIter tmp_citer(right_citer);

    Unchastity_Vec_Iter dst_iter(max_local_unchastity_vec.begin());
    
    unsigned int neigh_count(num_neighs);

    // Move right_citer to point to the end of the window
    // (or to the last ele of the vector if that comes sooner).
    while (neigh_count-- > 0) {
        if (++tmp_citer == unchastity_vec.end()) {
            break;
        }

        right_citer = tmp_citer;
    }

    // At the start of each loop set the current maximum.
    // Then shift window right one position until the middle of the window
    // passes the end of the vector.
    // At each shift only two possibilities can change the maximum :-
    // a) the maximum position exits from left, or
    // b) a new maximum enters from right.

    neigh_count = num_neighs; // for virtual left pos
    bool new_max_needed = true; // to force initialisation in first iteration.

    for (Unchastity_Vec_CIter mid_citer(unchastity_vec.begin()); 
         mid_citer != unchastity_vec.end(); ++mid_citer) {

        // Update the max pos if necessary.

        if (new_max_needed) {
            find_max_in_win(left_citer, right_citer, max_citer);
        } else if (*right_citer > *max_citer) {
            max_citer = right_citer;
        }

        // Store the current max.
        *dst_iter++ = *max_citer;

        // Set-up for the next iteration :-

        new_max_needed = (max_citer == left_citer);

        // The virtual left end of the window will be left of the vec start
        // for num_neighs iterations.

        if (neigh_count > 0) {
            --neigh_count;
        } else {
            ++left_citer;
        }

        // Move the right end of the window unless it would fall off the vec.

        tmp_citer = right_citer;

        if (++tmp_citer != unchastity_vec.end()) {
            right_citer = tmp_citer;
        }
    }
}

/*****************************************************************************/
// Could improve performance by removing loop if assume sig_vec size to be 4.

float vec_max(const Sig_Vec& sig_vec)
{
    if (sig_vec.empty()) {
        return 0.0;
    }

    Sig_Vec_CIter sig_citer(sig_vec.begin());
    float max_val(*sig_citer++);

    for ( ; sig_citer != sig_vec.end(); ++sig_citer) {
        if (*sig_citer > max_val) {
            max_val = *sig_citer;
        }
    }

    return max_val;
}

/*****************************************************************************/

void calc_homopol_lens(const std::string& base_str, Len_Vec& len_vec)
{
    const unsigned int base_str_len(base_str.size());

    if (len_vec.size() != base_str_len) {
        len_vec.resize(base_str_len);
    }

    Len_Vec_Iter len_iter = len_vec.begin();

    char last_ch('_');
    unsigned int run_len(0);

    for (unsigned int ch_ind(0); ch_ind < base_str_len; ++ch_ind) {
        if (base_str[ch_ind] == last_ch) {
            ++run_len;
        } else {
            for (unsigned int base_ind(0); base_ind < run_len; ++base_ind) {
                *len_iter++ = run_len;
            }

            last_ch = base_str[ch_ind];
            run_len = 1;
        }
    }

    // Last run.
    for (unsigned int base_ind(0); base_ind < run_len; ++base_ind) {
        *len_iter++ = run_len;
    }
}

/*****************************************************************************/

bool extract_quality_predictors(const Sig_Ptr_Vec& sig_ptr_vec,
                                const Prob_Ptr_Vec& prob_ptr_vec,
                                const unsigned char* seq_ch_arr_ptr,
                                int num_clusters, int num_channels,
                                long lo_cluster_ind,
                                long hi_cluster_ind,
                                unsigned int read_lo_cycle_num,
                                unsigned int read_hi_cycle_num,
                                unsigned int num_pure_bases,
                                unsigned int qual_fields_mask,
                                Quality_Predictors_Vec& quality_predictors_vec)
{
    const int num_cycles(sig_ptr_vec.size());
    quality_predictors_vec.clear();

    const int curr_num_clusters(hi_cluster_ind - lo_cluster_ind + 1);
    const int 
        num_cycles_used(read_hi_cycle_num - read_lo_cycle_num + 1);

    quality_predictors_vec.resize(curr_num_clusters * num_cycles_used);
    Quality_Predictors_Vec_Iter 
        quality_preds_iter = quality_predictors_vec.begin();

    Sig_Vec cluster_cycle_sig_vec(num_channels);

    // Max local unchastity set-up.
    const bool do_max_local_unchastity
        = ((qual_fields_mask & Quality_Data::FIELD_MAX_LOCAL_UNCHASTITY) != 0);
    const unsigned int num_neighs(2);
    Unchastity_Vec unchastity_vec;
    unchastity_vec.resize(num_cycles_used);
    Unchastity_Vec max_local_unchastity_vec;
    max_local_unchastity_vec.resize(num_cycles_used);
    Unchastity_Vec_CIter
        max_local_unchastity_citer(max_local_unchastity_vec.begin());

    // Signal decay set-up.
    const bool do_signal_decay
        = ((qual_fields_mask & Quality_Data::FIELD_SIGNAL_DECAY) != 0);
    float in_read_cycle1_max_sig(0.0);

    // Homopolymer run lengths set-up.
    const bool do_homopol_lens((qual_fields_mask
                                & Quality_Data::FIELD_HOMOPOL_LEN) != 0);
    std::string base_str(num_cycles_used, 'X');
    Len_Vec homopol_len_vec;
    Len_Vec_CIter homopol_len_citer(homopol_len_vec.begin());

    for (long cluster_ind(lo_cluster_ind);
         cluster_ind <= hi_cluster_ind; ++cluster_ind) {
        float max_early_unchastity(0);
        unsigned int cycle_num(1); // ignoring read boundaries (1-offset)
        unsigned int in_read_cycle_num(1); // within read (1-offset)

        // Keep a placeholder at the start of the cluster, so we can go
        // back and set the across-cycles max_early_unchastity 
        // (i.e. the max unchastity in the first pure_bases cycles).
        Quality_Predictors_Vec_Iter
            cluster_begin_quality_preds_iter = quality_preds_iter;

        Sig_Ptr_Vec_CIter sig_ptr_citer(sig_ptr_vec.begin());
        Prob_Ptr_Vec_CIter prob_ptr_citer(prob_ptr_vec.begin());

        char* seq_ch_ptr = (char*) (seq_ch_arr_ptr + cluster_ind);

        // Iterate over cycles.
        // [The sequence data has all 1st cycles together, then all 2nd, etc. 
        //  -> need to increase its ptr by num_clusters per cycle.]

        for ( ; sig_ptr_citer != sig_ptr_vec.end();
              ++sig_ptr_citer, ++prob_ptr_citer, seq_ch_ptr += num_clusters,
                  ++cycle_num) {

            // Skip cycles outside the read of interest (if multi-read).
            if (cycle_num < read_lo_cycle_num) {
                continue;
            } else if (cycle_num > read_hi_cycle_num) {
                break;
            }

            float* cycle_sig_ptr(*sig_ptr_citer);
            float* prb_sig_ptr(*prob_ptr_citer);
            float unchastity(0);
            int max_base_qv(min_possible_base_qv);

            for (int channel_ind(0); channel_ind < num_channels;
                 ++channel_ind) {
                cluster_cycle_sig_vec[channel_ind]
                    = cycle_sig_ptr[(cluster_ind * num_channels)
                                    + channel_ind];

                const int base_qv(static_cast<int>
                                  (prb_sig_ptr[(cluster_ind * num_channels)
                                               + channel_ind]));

                if (base_qv > max_base_qv) {
                    max_base_qv = base_qv;
                }
            }

            calc_unchastity(cluster_cycle_sig_vec, unchastity);

            quality_preds_iter->in_read_cycle_num = in_read_cycle_num; 
            quality_preds_iter->unchastity = unchastity;

            if ((in_read_cycle_num <= num_pure_bases)
                && ((unchastity > max_early_unchastity)
                    || (in_read_cycle_num == 1))) {
                max_early_unchastity = unchastity;
            }

            quality_preds_iter->raw_unquality = - max_base_qv;

            quality_preds_iter->called_base = *seq_ch_ptr;

            // Potentially used for local unchastity calculations below.
            if (do_max_local_unchastity) {
                unchastity_vec[in_read_cycle_num - 1] = unchastity;
            }

            // Signal decay
            if (do_signal_decay) {
                const float cycle_max_sig(vec_max(cluster_cycle_sig_vec));

                if (in_read_cycle_num == 1) {
                    in_read_cycle1_max_sig = cycle_max_sig;
                }

                const float signal_decay((in_read_cycle1_max_sig == 0.0)
                                         ? 0.0
                                         : (cycle_max_sig / 
                                            in_read_cycle1_max_sig));

                quality_preds_iter->signal_decay 
                    = 1.0 - ((signal_decay < 0.0)
                             ? 0.0
                             : ((signal_decay > 1.0) ? 1.0 : signal_decay));
            }

            // Potentially used for homopolymer run calculations below.
            if (do_homopol_lens) {
                base_str[in_read_cycle_num - 1] = *seq_ch_ptr;
            }

            // The quality pedictors iterator and within read cycle number are
            // updated only if main body of loop is executed (not skipped).
            ++quality_preds_iter;
            ++in_read_cycle_num;
        }

        if (do_max_local_unchastity) {
            calc_local_max(unchastity_vec, num_neighs,
                           max_local_unchastity_vec);
            max_local_unchastity_citer = max_local_unchastity_vec.begin();
        }

        if (do_homopol_lens) {
            calc_homopol_lens(base_str, homopol_len_vec);
            homopol_len_citer = homopol_len_vec.begin();
        }

        // Now we can go back and fill in the max_early_unchastity values.
        for (Quality_Predictors_Vec_Iter 
                 tmp_quality_preds_iter = cluster_begin_quality_preds_iter;
             tmp_quality_preds_iter != quality_preds_iter;
             ++tmp_quality_preds_iter) {
            tmp_quality_preds_iter->max_early_unchastity
                = max_early_unchastity;

            if (do_max_local_unchastity) {
                tmp_quality_preds_iter->max_local_unchastity
                    = *max_local_unchastity_citer++;
            }

            if (do_homopol_lens) {
                tmp_quality_preds_iter->homopol_len = *homopol_len_citer++;
            }
        }
    }

    return true;
}

/*****************************************************************************/

int process_block_of_clusters(const Sig_Ptr_Vec& sig_ptr_vec,
                              const Prob_Ptr_Vec& prob_ptr_vec,
                              const unsigned char* seq_ch_arr_ptr,
                              int num_clusters, int num_channels,
                              long lo_cluster_ind,
                              long hi_cluster_ind,
                              const std::string& read_msg_str,
                              unsigned int read_lo_cycle_num,
                              unsigned int read_hi_cycle_num,
                              const EQP_Options& eqp_options,
                              Quality_Data& quality_data)
{
    Quality_Predictors_Vec quality_predictors_vec;

    if (!extract_quality_predictors(sig_ptr_vec, prob_ptr_vec,
                                    seq_ch_arr_ptr, num_clusters,
                                    num_channels,
                                    lo_cluster_ind,
                                    hi_cluster_ind,
                                    read_lo_cycle_num,
                                    read_hi_cycle_num,
                                    eqp_options.my_num_pure_bases,
                                    eqp_options.my_qual_fields_mask,
                                    quality_predictors_vec)) {
        error("Failed to extract quality predictors%s",
              read_msg_str.c_str());
        return ERR_QUALITY_PREDICTORS;
    }

    const int status = quality_data.write(quality_predictors_vec);

    if (status != 0) {
        error("Failed to write quality predictors%s to `%s'",
              read_msg_str.c_str(), quality_data.file_path_str().c_str());
        return ERR_FILE_WRITE;
    }

    return ERR_OK;
}

/*****************************************************************************/

int process_clusters(const Sig_Ptr_Vec& sig_ptr_vec,
                     const Prob_Ptr_Vec& prob_ptr_vec,
                     const unsigned char* seq_ch_arr_ptr,
                     int num_clusters, int num_channels,
                     const std::string& read_msg_str,
                     unsigned int read_lo_cycle_num,
                     unsigned int read_hi_cycle_num,
                     const EQP_Options& eqp_options,
                     Quality_Data& quality_data)
{
    int status(ERR_OK);
    const long clusters_per_block(10000);

    long hi_cluster_ind(0);
    long lo_cluster_ind(0);
    int num_blocks
        = static_cast<int>((num_clusters - 1) / clusters_per_block + 1);

    for (int block_ind(0); block_ind < num_blocks; block_ind++) {
        lo_cluster_ind = block_ind * clusters_per_block;
        hi_cluster_ind = lo_cluster_ind + clusters_per_block - 1;

        if (hi_cluster_ind > num_clusters - 1) {
            hi_cluster_ind = num_clusters - 1;
        }

        status = process_block_of_clusters(sig_ptr_vec, prob_ptr_vec,
                                           seq_ch_arr_ptr,
                                           num_clusters, num_channels,
                                           lo_cluster_ind,
                                           hi_cluster_ind,
                                           read_msg_str,
                                           read_lo_cycle_num,
                                           read_hi_cycle_num,
                                           eqp_options,
                                           quality_data);

        if (status != ERR_OK) {
            break;
        }
    }

    return status;
}

/*****************************************************************************/

int main(int num_args, char* arg_cstr_arr[])
{
    const std::string app_name(arg_cstr_arr[0]);
    int ret_val(ERR_OK);

    EQP_Options eqp_options;

    if (eqp_options.init(num_args, arg_cstr_arr)
        != CommandLine::Options::RUN) {
        return ERR_USAGE;
    }

    const unsigned int expected_num_cycles(eqp_options.total_num_cycles());

    IntensityData intensity_data(eqp_options.my_sig2_path_str.c_str(),
                                 eqp_options.my_sig2_compr);
    PrbData prb_data(eqp_options.my_prb_path_str.c_str());
    SequenceData seq_data(eqp_options.my_seq_path_str.c_str());

    Sig_Ptr_Vec sig_ptr_vec;
    int num_clusters(0);
    int num_chans(0);
    PositionIndex* position_index_ptr(0);

    int status = intensity_data.read(sig_ptr_vec, &num_clusters, &num_chans,
                                     &position_index_ptr);

    if (status != 0) {
        error("Failed to read intensity data from `%s'",
              eqp_options.my_sig2_path_str.c_str());
        return ERR_FILE_READ;
    }

    const int num_cycles(sig_ptr_vec.size());

    // Check value found if an expected value was specified (i.e. is non-zero)
    // BUT only if some clusters were actually found -> non-empty sig2 file.
    if ((num_clusters != 0)
        && (expected_num_cycles != 0)
        && ((unsigned int) num_cycles != expected_num_cycles)) {
        error("Number of cycles (%d) of intensity data found in `%s' "
              "does not match total number of cycles specified (%u)",
              num_cycles, eqp_options.my_sig2_path_str.c_str(),
              expected_num_cycles);
        return ERR_FILE_READ;
    }

    Prob_Ptr_Vec prob_ptr_vec;
    unsigned int num_prb_clusters(0);
    unsigned int num_prb_chans(0);

    status = prb_data.read(prob_ptr_vec, num_prb_clusters, num_prb_chans);

    if (status != 0) {
        error("Failed to read probability data from `%s'",
              eqp_options.my_prb_path_str.c_str());
        return ERR_FILE_READ;
    }

    if (num_prb_clusters != num_clusters) {
        error("Mismatch in cluster numbers between `%s' (%u) and `%s' (%u)",
              eqp_options.my_sig2_path_str.c_str(), num_clusters,
              eqp_options.my_prb_path_str.c_str(), num_prb_clusters);
        return ERR_FILE_READ;
    }

    // Check value found if an expected value was specified (i.e. is non-zero)
    // BUT only if some clusters were actually found -> non-empty prb file.
    if ((num_clusters != 0)
        && (expected_num_cycles != 0)
        && (prob_ptr_vec.size() != expected_num_cycles)) {
        error("Number of cycles (%u) of probability data found in `%s' "
              "does not match total number of cycles specified (%u)",
              prob_ptr_vec.size(), eqp_options.my_prb_path_str.c_str(),
              expected_num_cycles);
        return ERR_FILE_READ;
    }

    unsigned char* seq_ch_arr_ptr
        = new unsigned char[num_clusters * num_cycles];

    if (seq_ch_arr_ptr == 0) {
        error("Out of memory : sequence data");
        return ERR_OUT_OF_MEMORY;
    }


    // N.B. : Silently pads.
    status = seq_data.read(seq_ch_arr_ptr, num_clusters, num_cycles);

    if (status != 0) {
        error("Failed to read sequence data from `%s'",
              eqp_options.my_seq_path_str.c_str());
    }


    // Extract the quality predictors (possibly per read) and write them to
    // one or more qval files (one per read if multi-read).

    EQP_Options::File_Path_Str_List_CIter
        qval_path_str_citer(eqp_options.my_qval_path_str_list.begin());

    for (EQP_Options::Read_Num_Vec_CIter read_num_citer 
             = eqp_options.my_read_num_vec.begin();
         read_num_citer != eqp_options.my_read_num_vec.end();
         ++read_num_citer, ++qval_path_str_citer) {

        const unsigned int read_num(*read_num_citer);
        unsigned int read_lo_cycle_num(0);
        unsigned int read_hi_cycle_num(0);

        Quality_Data quality_data(*qval_path_str_citer,
                                  eqp_options.my_qual_field_list,
                                  eqp_options.my_qval_compr);

        std::string read_msg_str;

        if (read_num != 0) {
            std::ostringstream ostrm;
            ostrm << " for read " << read_num;
            read_msg_str = ostrm.str();
        }

        if (!eqp_options.get_read_cycle_range(read_num, read_lo_cycle_num,
                                              read_hi_cycle_num)) {
            error("Failed to get cycle range%s", read_msg_str.c_str());
        }

        if ((status = process_clusters(sig_ptr_vec,
                                       prob_ptr_vec,
                                       seq_ch_arr_ptr,
                                       num_clusters,
                                       num_chans,
                                       read_msg_str,
                                       read_lo_cycle_num,
                                       read_hi_cycle_num,
                                       eqp_options,
                                       quality_data)) != ERR_OK) {
            return status;
        }
    }

    return ERR_OK;
}

/*****************************************************************************/

