/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <iostream>

#include "misc_types.h"

#include "DataSetFactory.h"

/*****************************************************************************/

enum {ERR_OK = 0, ERR_USAGE = 1, ERR_FILE_READ = 2};

typedef std::vector<BaseAlignment> Base_Pair_Vec;
typedef Base_Pair_Vec::const_iterator Base_Pair_Vec_CIter;

/*****************************************************************************/

int usage(const std::string& app_name, int ret_val)
{
    std::cerr << "Usage: " << app_name
              << " <qval path> [<qref path>]" << std::endl;
}

/*****************************************************************************/

File_Buffer::Compression_Type
deduce_compression_from_name(const std::string& file_path_str)
{
    File_Buffer::Compression_Type 
        compression_type(File_Buffer::compressionNone);

    const unsigned int file_path_str_len(file_path_str.size());

    if (file_path_str_len > 3) {
        if (file_path_str.substr(file_path_str_len - 3) == ".gz") {
            compression_type = File_Buffer::compressionGzip;
        }
    }

    if (file_path_str_len > 4) {
        if (file_path_str.substr(file_path_str_len - 4) == ".bz2") {
            compression_type = File_Buffer::compressionBzip2;
        }
    }

    return compression_type;
}

/*****************************************************************************/

void dump_data_set(DataSet* data_set_ptr)
{
    if (!data_set_ptr) {
        return;
    }

    Float_Vec_Vec* float_vec_vec_ptr = data_set_ptr->getParameterValues();
    Base_Pair_Vec* base_pair_vec_ptr = data_set_ptr->getBaseReads();


    // Need to transpose the values matrix.
    typedef std::vector<Float_Vec_CIter> Float_Vec_CIter_Vec;
    typedef Float_Vec_CIter_Vec::iterator Float_Vec_CIter_Vec_Iter;
    Float_Vec_CIter_Vec trans_vec;

    for (Float_Vec_Vec_CIter float_vec_citer = float_vec_vec_ptr->begin(); 
         float_vec_citer != float_vec_vec_ptr->end(); ++float_vec_citer) {
        trans_vec.push_back(float_vec_citer->begin());
    }

    for (Base_Pair_Vec_CIter base_pair_citer = base_pair_vec_ptr->begin();
         base_pair_citer != base_pair_vec_ptr->end(); ++base_pair_citer) {
        for (Float_Vec_CIter_Vec_Iter trans_vec_iter = trans_vec.begin();
             trans_vec_iter != trans_vec.end(); ++trans_vec_iter) {
            Float_Vec_CIter float_citer(*trans_vec_iter);
            std::cout << *float_citer << "\t";
            (*trans_vec_iter)++;
        }

        std::cout << base_pair_citer->base_read << "\t"
                  << base_pair_citer->base_aligned << std::endl;
    }
}

/*****************************************************************************/

int main(int num_args, const char* arg_cstr_arr[])
{
    // Parse args.

    const std::string app_name(arg_cstr_arr[0]);
    std::string qref_path_str;

    switch (num_args) {
    case 2:
        break;

    case 3:
        qref_path_str = arg_cstr_arr[2];
        break;

    default:
        return usage(app_name, ERR_USAGE);
    }

    const std::string qval_path_str(arg_cstr_arr[1]);


    // Set up the Data_Factory.

    File_Buffer::Compression_Type
        qval_compr(deduce_compression_from_name(qval_path_str));
    File_Buffer::Compression_Type
        qref_compr(deduce_compression_from_name(qref_path_str));
    
    Data_Set_Factory data_set_factory(qval_path_str, qval_compr,
                                      qref_path_str, qref_compr);

    DataSet* data_set_ptr(0);
    bool at_eof(false);
    bool first_data_set(true);
    int num_params(0);
    int detected_num_params(0);


    // Read data.

    while (data_set_factory.get_next_data_set(data_set_ptr, at_eof)) {
            detected_num_params = data_set_ptr->getNumberParameters();

            if (first_data_set) {
                std::cerr << "Detected " << detected_num_params 
                          << " parameters.\n";
                num_params = detected_num_params;
                first_data_set = false;
            }

            if (detected_num_params != num_params) {
                std::cerr << "Wrong number of parameters: "<< 
                    data_set_ptr->getNumberParameters() << std::endl;
                continue;
            }

            dump_data_set(data_set_ptr);

            delete data_set_ptr;
            data_set_ptr = 0;

            if (at_eof) {
                break;
            }
    }
}

/*****************************************************************************/
