/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <map>
#include <list>
#include <iostream>

#include <cstring>

#include "option_test_utils.h"
#include "StringUtil.h"

#include "EQP_Options.h"

/*****************************************************************************/

bool check_extra_args(Arg_List& extra_arg_list,
                      EQP_Options& eqp_options)
{
    bool all_match(true);
    unsigned int extra_arg_ind(0);

    EQP_Options::File_Path_Str_List_CIter qval_path_str_citer
        = eqp_options.my_qval_path_str_list.begin();

    for (Arg_List_CIter arg_citer = extra_arg_list.begin();
         arg_citer != extra_arg_list.end();
         ++arg_citer, ++qval_path_str_citer) {
        const std::string arg_str(*arg_citer);

        if (qval_path_str_citer == eqp_options.my_qval_path_str_list.end()) {
            std::cerr << "Ran out of parsed arguments before `"
                      << arg_str << "'" << std::endl;
            all_match = false;
            break;
        }

        const std::string found_arg_str(*qval_path_str_citer);

        if (found_arg_str != arg_str) {
            std::cerr << "Expected extra arg `" << arg_str
                      << "' but found `" << found_arg_str
                      << "'" << std::endl;
            all_match = false;
        }
    }

    return all_match;
}                     

/*****************************************************************************/

int main()
{
    EQP_Options eqp_options;

    const std::string no_arg_str("NOARG");
    const std::string expected_app_name("extract_quality_predictors");
    const std::string sig2_path_str("../s_3_0059_sig2.txt");
    const std::string prb_path_str("../s_3_0059_prb.txt");
    const std::string seq_path_str("../s_3_0059_seq.txt");
    const std::string read1_qval_path_str("s_3_0059_1_qval.txt");
    const std::string read2_qval_path_str("s_3_0059_2_qval.txt");

    Opt_Arg_Map opt_arg_map;
    opt_arg_map["orig_read_lengths"] = "36:36";
    opt_arg_map["pure_bases"] = "14";
    opt_arg_map["multi_read"] = no_arg_str;
    opt_arg_map["sig2_compression"] = "bzip2";
    opt_arg_map["qval_compression"] = "gzip";
    opt_arg_map["predictors"]
        = "in_read_cycle:max_local_unchastity:homopol_len:raw_unquality";

    Arg_List extra_arg_list;
    extra_arg_list.push_back(sig2_path_str);
    extra_arg_list.push_back(prb_path_str);
    extra_arg_list.push_back(seq_path_str);
    extra_arg_list.push_back(read1_qval_path_str);
    extra_arg_list.push_back(read2_qval_path_str);


    // Count NOARGs.

    int num_no_args(0);

    for (Opt_Arg_Map_Iter opt_arg_iter = opt_arg_map.begin();
         opt_arg_iter != opt_arg_map.end(); ++opt_arg_iter) {
        if (opt_arg_iter->second == no_arg_str) {
            ++num_no_args;
        }
    }

    int num_args = (1 + (opt_arg_map.size() * 2) - num_no_args
                    + extra_arg_list.size());


    // Populate arg_cstr_arr.
    // (Note that the cstrs have to be non-const to match the signature
    //  of getopt_long as used by CommandLine::Options.)

    char** arg_cstr_arr = new char*[num_args + 1];
    *(arg_cstr_arr + num_args) = 0;

    int arg_ind(0);
    arg_cstr_arr[arg_ind++] = copy_to_new_cstr(expected_app_name.c_str());

    for (Opt_Arg_Map_Iter opt_arg_iter = opt_arg_map.begin();
         opt_arg_iter != opt_arg_map.end(); ++opt_arg_iter) {
        arg_cstr_arr[arg_ind++] = copy_to_new_cstr(std::string("--") +
                                                   opt_arg_iter->first);

        if (opt_arg_iter->second != no_arg_str) {
            arg_cstr_arr[arg_ind++] = copy_to_new_cstr(opt_arg_iter->second);
        }
    }

    for (Arg_List_Iter arg_iter = extra_arg_list.begin();
         arg_iter != extra_arg_list.end(); ++arg_iter) {
        arg_cstr_arr[arg_ind++] = copy_to_new_cstr(*arg_iter);
    }

    // Dump arg_cstr_arr.
    std::cout << "Arguments tested (" << num_args << ") :-" << std::endl;

    for (arg_ind = 0; arg_ind < num_args; ++arg_ind) {
        std::cout << arg_cstr_arr[arg_ind] << std::endl;
    }

    eqp_options.init(num_args, arg_cstr_arr);


    // Tidy up arg_cstr_arr after use.
    for (arg_ind = 0; arg_ind < num_args; ++arg_ind) {
        delete[](arg_cstr_arr[arg_ind]);
    }


    // Check the values parsed.

    bool args_ok(true);

    if (eqp_options.my_orig_read_lens_list.size() != 2) {
        args_ok = false;
    }

    for (EQP_Options::Orig_Read_Lens_List_CIter read_len_citer
             = eqp_options.my_orig_read_lens_list.begin();
         read_len_citer != eqp_options.my_orig_read_lens_list.end();
         ++read_len_citer) {
        if (*read_len_citer != 36) {
            args_ok = false;
        }
    }

    if (eqp_options.my_num_pure_bases != 14) {
        args_ok = false;
    }

    if (!eqp_options.my_multi_read_flag) {
        args_ok = false;
    }

    if (eqp_options.my_sig2_compr != IntensityData::compressionBzip2) {
        args_ok = false;
    }

    if (eqp_options.my_qval_compr != File_Buffer::compressionGzip) {
        args_ok = false;
    }

    if (eqp_options.my_sig2_path_str != sig2_path_str) {
        args_ok = false;
    }

    if (eqp_options.my_prb_path_str != prb_path_str) {
        args_ok = false;
    }

    if (eqp_options.my_seq_path_str != seq_path_str) {
        args_ok = false;
    }

    // in_read_cycle:max_local_unchastity:homopol_len:raw_unquality
    bool in_read_cycle_found(false);
    bool max_local_unchastity_found(false);
    bool homopol_len_found(false);
    bool raw_unquality_found(false);
    bool called_base_found(false);

    for (Quality_Data::Field_List_CIter 
             field_citer(eqp_options.my_qual_field_list.begin());
         field_citer != eqp_options.my_qual_field_list.end();
         ++field_citer) {
        switch (*field_citer) {
        case Quality_Data::FIELD_IN_READ_CYCLE_NUM:
            in_read_cycle_found = true;
            break;
        case Quality_Data::FIELD_MAX_LOCAL_UNCHASTITY:
            max_local_unchastity_found = true;
            break;
        case Quality_Data::FIELD_HOMOPOL_LEN:
            homopol_len_found = true;
            break;
        case Quality_Data::FIELD_RAW_UNQUALITY:
            raw_unquality_found = true;
            break;
        case Quality_Data::FIELD_CALLED_BASE:
            // This one should be set automatically.
            called_base_found = true;
            break;
        default:
            std::cout << "Unexpected field set." << std::endl;
            args_ok = false;
            break;
        }
    }

    if (!(in_read_cycle_found && max_local_unchastity_found
          && homopol_len_found && raw_unquality_found)) {
        std::cout << "Expected field(s) not set." << std::endl;
        args_ok = false;
    }

    if (eqp_options.my_qual_fields_mask 
        != (Quality_Data::FIELD_IN_READ_CYCLE_NUM
            | Quality_Data::FIELD_MAX_LOCAL_UNCHASTITY
            | Quality_Data::FIELD_HOMOPOL_LEN
            | Quality_Data::FIELD_RAW_UNQUALITY
            | Quality_Data::FIELD_CALLED_BASE)) {
        std::cout << "Quality fields mask has wrong value ("
                  << eqp_options.my_qual_fields_mask << ")." << std::endl;
        args_ok = false;
    }


    EQP_Options::File_Path_Str_List_CIter qval_path_str_citer
        = eqp_options.my_qval_path_str_list.begin();
    unsigned int read_lo_cycle_num(0);
    unsigned int read_hi_cycle_num(0);

    for (EQP_Options::Read_Num_Vec_CIter read_num_citer
             = eqp_options.my_read_num_vec.begin();
         read_num_citer != eqp_options.my_read_num_vec.end();
         ++read_num_citer, ++qval_path_str_citer) {
        const unsigned int read_num(*read_num_citer);

        if (!eqp_options.get_read_cycle_range(read_num,
                                              read_lo_cycle_num,
                                              read_hi_cycle_num)) {
            args_ok = false;
        }

        std::cout << "read " << read_num
                  << " : read_lo_cycle_num " << read_lo_cycle_num
                  << " : read_hi_cycle_num " << read_hi_cycle_num
                  << std::endl;

        switch (read_num) {
        case 1:
            if (*qval_path_str_citer != read1_qval_path_str) {
                args_ok = false;
            }

            if ((read_lo_cycle_num != 1) || (read_hi_cycle_num != 36)) {
                args_ok = false;
            }

            break;
        case 2:
            if (*qval_path_str_citer != read2_qval_path_str) {
                args_ok = false;
            }

            if ((read_lo_cycle_num != 37) || (read_hi_cycle_num != 72)) {
                args_ok = false;
            }

            break;
        default:
            args_ok = false;
        }
    }

    if (args_ok) {
        std::cout << "Arguments were correctly parsed." << std::endl;
    } else {
        std::cout << "** Incorrect parsing of arguments. **" << std::endl;
    }

    return 0;
}

/*****************************************************************************/
