/*****************************************************************************/
// Copyright (c) Illumina 2008
// Author: Richard Shaw
// This source file is covered by the Illumina Genome Analyzer Software 
// License Agreement and bound by the terms therein.
/*****************************************************************************/

#include <string>
#include <map>
#include <list>
#include <iostream>

#include <cstring>

#include "option_test_utils.h"
#include "StringUtil.h"

#include "QCal_Options.h"

/*****************************************************************************/

bool check_extra_args(Arg_List& extra_arg_list,
                      QCal_Options& qcal_options)
{
    bool all_match(true);
    unsigned int extra_arg_ind(0);

    QCal_Options::File_Name_List_CIter qval_name_citer
        = qcal_options.my_qval_file_name_list.begin();

    for (Arg_List_CIter arg_citer = extra_arg_list.begin();
         arg_citer != extra_arg_list.end();
         ++arg_citer, ++qval_name_citer) {
        const std::string arg_str(*arg_citer);

        if (qval_name_citer == qcal_options.my_qval_file_name_list.end()) {
            std::cerr << "Ran out of parsed arguments before `"
                      << arg_str << "'" << std::endl;
            all_match = false;
            break;
        }

        const std::string found_arg_str(*qval_name_citer);

        if (found_arg_str != arg_str) {
            std::cerr << "Expected extra arg `" << arg_str
                      << "' but found `" << found_arg_str
                      << "'" << std::endl;
            all_match = false;
        }
    }

    return all_match;
}                     

/*****************************************************************************/

int main()
{
    QCal_Options qcal_options;

    const std::string expected_app_name("QualityCalibration");

    Opt_Arg_Map opt_arg_map;
    opt_arg_map["qval_suffix"] = "_qval.txt";
    opt_arg_map["qref_suffix"] = "_qref.txt";
    opt_arg_map["qval_compression"] = "gzip";
    opt_arg_map["qref_compression"] = "bzip2";
    opt_arg_map["cfg"] = "qc_cfg.xml";

    Arg_List extra_arg_list;
    extra_arg_list.push_back("s_3_0059_qval.txt");
    extra_arg_list.push_back("s_3_0060_qval.txt");
    extra_arg_list.push_back("s_3_0061_qval.txt");

    int num_args = 1 + (opt_arg_map.size() * 2) + extra_arg_list.size();


    // Populate arg_cstr_arr.
    // (Note that the cstrs have to be non-const to match the signature
    //  of getopt_long as used by CommandLine::Options.)

    char** arg_cstr_arr = new char*[num_args + 1];
    *(arg_cstr_arr + num_args) = 0;

    int arg_ind(0);
    arg_cstr_arr[arg_ind++] = copy_to_new_cstr(expected_app_name.c_str());

    for (Opt_Arg_Map_Iter opt_arg_iter = opt_arg_map.begin();
         opt_arg_iter != opt_arg_map.end(); ++opt_arg_iter) {
        arg_cstr_arr[arg_ind++] = copy_to_new_cstr(std::string("--") +
                                                   opt_arg_iter->first);
        arg_cstr_arr[arg_ind++] = copy_to_new_cstr(opt_arg_iter->second);
    }

    for (Arg_List_Iter arg_iter = extra_arg_list.begin();
         arg_iter != extra_arg_list.end(); ++arg_iter) {
        arg_cstr_arr[arg_ind++] = copy_to_new_cstr(*arg_iter);
    }

    // Dump arg_cstr_arr.
    std::cout << "Arguments tested (" << num_args << ") :-" << std::endl;

    for (arg_ind = 0; arg_ind < num_args; ++arg_ind) {
        std::cout << arg_cstr_arr[arg_ind] << std::endl;
    }

    qcal_options.init(num_args, arg_cstr_arr);


    // Tidy up arg_cstr_arr after use.
    for (arg_ind = 0; arg_ind < num_args; ++arg_ind) {
        delete[](arg_cstr_arr[arg_ind]);
    }


    // Check the values parsed.

    bool args_ok(true);

    args_ok = args_ok && check_opt_arg_match(opt_arg_map, "qval_suffix",
                                             qcal_options.my_qval_suffix_str);
    args_ok = args_ok && check_opt_arg_match(opt_arg_map, "qref_suffix",
                                             qcal_options.my_qref_suffix_str);
    args_ok = args_ok && check_opt_arg_match(opt_arg_map, "qval_compression",
                                             qcal_options.my_qval_compr_name);
    args_ok = args_ok && check_opt_arg_match(opt_arg_map, "qref_compression",
                                             qcal_options.my_qref_compr_name);
    args_ok = args_ok && check_opt_arg_match(opt_arg_map, "cfg",
                                             qcal_options.my_cfg_path_str);
                        
    args_ok = args_ok && check_extra_args(extra_arg_list, qcal_options);

    if (args_ok) {
        std::cout << "Arguments were correctly parsed." << std::endl;
    }

    return 0;
}

/*****************************************************************************/
