#!/bin/bash

# validate.sh
# Compile the Solexa analysis pipeline if necessary, then run it
# on the Solexa validation dataset
# Author: Anthony J. Cox
# Extended by: Richard J. Shaw

# Copyright (c) Solexa (UK) Limited 2007. All rights reserved.

#------------------------------------------------------------------------------

EXPTS=("071112_SLXA-EAS1_0089_FC20120_R1:071116_SLXA-EAS1_0090_FC20120_R2:2-37,39-74:config.txt" "071204_EAS188_0003_FC200VLAAXX:auto:config.txt")

GENOMES=("E_coli.fa" "phi.fa")

#------------------------------------------------------------------------------

log_file='' # global (but set below) so do not have to specify it repeatedly

RM="rm -f"
MKDIR="mkdir -p"
CHMOD="chmod"

#------------------------------------------------------------------------------

function init_log()
{
    if [ -f $log_file ]
    then
	mv -f $log_file $log_file.old
    fi
}

#------------------------------------------------------------------------------

function log_msg()
{
    echo `date` $1 >> $log_file
}

#------------------------------------------------------------------------------

function log_separator()
{
    echo "----------" >> $log_file
}

#------------------------------------------------------------------------------

function info()
{
    log_msg "$1"
    echo "$1"
}

#------------------------------------------------------------------------------

function warn()
{
    log_msg "WARNING: $1"
    echo "WARNING: $1"
}

#------------------------------------------------------------------------------

function die()
{
    log_msg "$1"
    echo "$1"
    echo "See $log_file for details."
    exit -1
}

#------------------------------------------------------------------------------

function copy_file_to_log()
{
    file_path=$1
    log_msg "$file_path :-"
    log_separator
    cat $file_path >> $log_file
    log_separator
}

#------------------------------------------------------------------------------

function run_cmd_logging_output()
{
    cmd=$1

    log_msg "Running '$cmd' :-"
    log_separator

    $cmd >> $log_file 2>&1
    cmd_res=$?

    log_separator
    log_msg "Result code was $cmd_res"
    log_separator

    return $cmd_res
}

#------------------------------------------------------------------------------

function run_cmd_logging_err()
{
    cmd="$1"

    log_msg "Running '$cmd' :-"
    log_separator

    eval "$cmd" " 2>> $log_file"
    cmd_res=$?

    log_separator
    log_msg "Result code was $cmd_res"
    log_separator

    return $cmd_res
}

#------------------------------------------------------------------------------

function make_tmp_filename()
{
    base_name=$1
    ext_str=$2

    timestamp=`date +%s`

    fn="/tmp/${base_name}_${timestamp}.${ext_str}"

    eval "$3=$fn"
}

#------------------------------------------------------------------------------

function push_dir()
{
    new_dir=$1

    if [ ! -d $new_dir ]
	then
	die "Folder $new_dir does not exist - cannot change to it"
    fi

    log_msg "Changing to folder $new_dir"
    pushd $new_dir > /dev/null 2>&1

    if [ $?  -ne 0 ]
	then
	die "Failed to change to folder $new_dir"
    fi

    return 0
}

#------------------------------------------------------------------------------

function pop_dir()
{
    popd > /dev/null 2>&1

    if [ $? -eq 0 ]
	then
	log_msg "Returned to folder `pwd`"
    else
	die "Failed to undo folder change : still in `pwd`"
    fi

    return 0
}

#------------------------------------------------------------------------------

function capture_sys_info()
{
    src_file=$1
    hdr_msg=$2

    if [ -f $src_file ]
    then
	log_msg "${hdr_msg}"
	copy_file_to_log "${src_file}"
    fi
}

#------------------------------------------------------------------------------

function check_binary_version()
{
    bin_name="$1"
    ver_opt="$2"
    min_ver="$3"

    min_major=`echo $min_ver | cut -d"." -f1`
    min_minor=`echo $min_ver | cut -d"." -f2`
    min_release=`echo $min_ver | cut -d"." -f3`

    if [ -z $min_release ]
	then
	min_release=0
    fi

    ver_info=''

    # gnuplot is a special case - earlier versions do not support a version
    # option and instead must be passed a version command to run.
    #
    # bzip2 is also a special case - there seems to be no way to persuade it
    # to output its version without doing some compression.
    # Then stderr has to be swapped with stdout, rather than tying them
    # so that the compressed output (originally to stdout) goes to /dev/null
    # but the version (originally to stderr) goes to stdout -> variable.

    if [ "x$bin_name" == "xgnuplot" ]
	then
	ver_info=`echo "show version" | gnuplot 2>&1 | fgrep -i Version`
    elif [ "x$bin_name" == "xbzip2" ]
	then
	ver_info=`echo junk | $bin_name $ver_opt 2> /dev/null 3>&1 1>&2 2>&3`
    else
	# Python also writes its version to stderr.
	ver_info=`$bin_name $ver_opt 2>&1`
    fi

    triple_part_ver=0
    found_major=''
    found_minor=''
    found_release=''

    if [ -z "$ver_info" ]
	then
	die "Failed to find version for $bin_name"
    fi

    ver=`echo $ver_info | grep -E -o -m 1 "[0-9]+\.[0-9]+\.[0-9]+"`

    if [ $? -eq 0 ]
	then
	triple_part_ver=1
	first_ver=`echo $ver | head -1 | cut -d" " -f1`
	found_major=`echo $first_ver | cut -d"." -f1`
	found_minor=`echo $first_ver | cut -d"." -f2`
	found_release=`echo $first_ver | cut -d"." -f3`
    else
	ver=`echo $ver_info | grep -E -o -m 1 "[0-9]+\.[0-9]+"`

	if [ $? -eq 0 ]
	    then
	    first_ver=`echo $ver | head -1 | cut -d" " -f1`
	    found_major=`echo $first_ver | cut -d"." -f1`
	    found_minor=`echo $first_ver | cut -d"." -f2`
	else
	    die "Failed to find version for $bin_name"
	fi
    fi

    if [ -z "$found_minor" ]
	then
	found_minor=0
    fi

    if [ -z "$found_release" ]
	then
	found_release=0
    fi

    disp_ver="$found_major.$found_minor"

    if [ $triple_part_ver -ne 0 ]
	then
	disp_ver="$disp_ver.$found_release"
    fi

    log_msg "Found $bin_name version $disp_ver"

    if [ $found_major -gt $min_major ]
	then
	return 0
    elif [ $found_major -lt $min_major ]
	then
	return 1
    fi

    if [ $found_minor -gt $min_minor ]
	then
	return 0
    elif [ $found_minor -lt $min_minor ]
	then
	return 1
    fi

    if [ $triple_part_ver -ne 0 ]
	then
	if [ $found_release -gt $min_release ]
	    then
	    return 0
	elif [ $found_release -lt $min_release ]
	    then
	    return 1
	fi
    fi

    return 0
}

#------------------------------------------------------------------------------

function check_lib()
{
    lib_name=$1
    hdr_name=$2
    lib_str=$3
    ver_expr_str=$4

    base_name="test_${lib_name}"
    src_ext_str="cpp"
    bin_ext_str="bin"
    tmp_src_fn=""
    tmp_bin_fn=""

    log_msg "Checking for required library $lib_name"
    log_msg "[header file $hdr_name, linking with -l$lib_str,"
    log_msg "  using $ver_expr_str in test code]"

    make_tmp_filename $base_name $src_ext_str tmp_src_fn
    make_tmp_filename $base_name $bin_ext_str tmp_bin_fn

    echo "#include <iostream>" > $tmp_src_fn
    echo "#include <${hdr_name}>" >> $tmp_src_fn
    echo "int main()" >> $tmp_src_fn
    echo "{" >> $tmp_src_fn
    echo "std::cout << ${ver_expr_str};" >> $tmp_src_fn
    echo "return 0;" >> $tmp_src_fn
    echo "}" >> $tmp_src_fn

    log_msg "Wrote source to ${tmp_src_fn}."

    cmd="g++ -o ${tmp_bin_fn} ${tmp_src_fn} -l${lib_str}"

    cmd_out=`$cmd 2>&1`
    cmd_res=$?

    $RM ${tmp_bin_fn}
    $RM ${tmp_src_fn}

    log_msg "Ran '$cmd' with result ${cmd_res}"
    log_msg "Output was: '$cmd_out'"

    if [ $cmd_res -eq 0 ]
	then
	log_msg "$lib_name : OK"
    else
	log_msg "$lib_name : CHECK FAILED"
    fi

    return $cmd_res
}

#------------------------------------------------------------------------------

function check_perl_lib()
{
    lib_name="$1"

    safe_lib_name=`echo $lib_name | tr ":" "_"`
    base_name="test_${safe_lib_name}"
    script_ext_str="pl"
    tmp_script_fn=""

    log_msg "Checking for required Perl library $lib_name"

    make_tmp_filename $base_name $script_ext_str tmp_script_fn

    echo "#!/usr/bin/env perl" > $tmp_script_fn
    echo "use ${lib_name};" >> $tmp_script_fn
    echo "print ${lib_name}->VERSION" >> $tmp_script_fn

    log_msg "Wrote test script to ${tmp_script_fn}."

    $CHMOD u+x $tmp_script_fn

    cmd=$tmp_script_fn

    cmd_out=`$cmd 2>&1`
    cmd_res=$?

    $RM ${tmp_script_fn}

    log_msg "Ran '$cmd' with result ${cmd_res}"
    log_msg "Output was: '$cmd_out'"

    if [ $cmd_res -eq 0 ]
	then
	log_msg "$lib_name : OK"
    else
	log_msg "$lib_name : CHECK FAILED"
    fi

    return $cmd_res
}

#------------------------------------------------------------------------------

function check_for_prerequisites()
{
    skip_fftw_check=$1

    info "Checking for prerequisites"

    ret_val=0

    # Executables

    for PREREQ in "perl -v 5.6" "python -V 2.2" "make -v 3.78" "gnuplot --version 3.7" "convert -version 5.4.7" "g++ --version 3.2.2" "bzip2 -V 1.0.2" "gzip -V 1.3.3"
      do
      set -- ${PREREQ} # parses PREREQ into fields
      log_msg "checking for prerequisite $1"

      if [ `which "$1"` ]
	  then
	  log_msg "prerequisite $1 found"
	  log_msg "checking version for prerequisite $1"
	  log_msg "minimum recommended version for $1 is $3"

	  check_binary_version $1 $2 $3

	  if [ $? != 0 ]
	      then
	      warn "prerequisite $1 is of lower version than recommended ($3)"
              ret_val=1
	  fi
      else
	  warn "prerequisite $1 not found"
	  ret_val=1
      fi
    done


    # Libraries

    PREREQS="zlib zlib.h z zlibVersion():bzlib bzlib.h bz2 BZ2_bzlibVersion()"

    # fftwf_version is a char array not a function so has no brackets.
    FFTW_PREREQ_ELE="fftw fftw3.h fftw3f fftwf_version"

    if [ $skip_fftw_check -eq 0 ]
	then
	PREREQS="$PREREQS:$FFTW_PREREQ_ELE"
    fi

    # Temporarily change IFS so for loop will split list on colons.
    OLD_IFS=$IFS
    IFS=":"

    for PREREQ in $PREREQS
      do
      IFS=${OLD_IFS}
      set -- ${PREREQ} # parses PREREQ into fields

      check_lib $1 $2 $3 $4

      if [ $? != 0 ]
	  then
	  warn "prerequisite library $1 not found"
	  warn "(Note that on platforms for which $1 is made available as"
	  warn " separate run-time and development packages, this may be due"
	  warn " to the development package not having been installed."
	  warn " Check that $2 can be found in the standard include paths.)"

	  if [ "X$1" == "Xfftw" ]
	      then
	      warn " "
	      warn "The --skip-fftw-check option may be used to bypass"
	      warn "checking for an FFTW installation, e.g. if some other"
	      warn "Pipeline-compatible FFT library is installed."
	  fi

	  ret_val=1
      fi
    done

    IFS=${OLD_IFS}
 

    # Perl libraries : XML::Simple

    for PREREQ in "XML::Simple"
      do
      check_perl_lib $PREREQ

      if [ $? != 0 ]
	  then
	  warn "prerequisite Perl library $PREREQ not found"
	  ret_val=1
      fi
    done

    log_msg "Finished checking for prerequisites"
    return $ret_val
}

#------------------------------------------------------------------------------

function obtain_sys_info()
{
    START_DIR=$1

    info "Obtaining system information"

# FIMXE : Is this file used any more?
    SYSTEM_FILE="${START_DIR}/system.txt"

    if [ -f $SYSTEM_FILE ]
	then
	rm -f $SYSTEM_FILE
    fi

    capture_sys_info /proc/cpuinfo "Obtaining CPU information"
    capture_sys_info /proc/meminfo "Obtaining RAM information"
    capture_sys_info /proc/version "Obtaining OS version information"
    log_msg "Finished obtaining system information"
}

#------------------------------------------------------------------------------

function check_pipeline_installation()
{
    PIPELINE_DIR=$1

    info "Checking pipeline installation"

    TOOLS_DIR=$PIPELINE_DIR/Gerald

    if [ -e "${TOOLS_DIR}" ] # Tested variable is quoted.
	then
	log_msg "${TOOLS_DIR} exists"
    else
	msg="ERROR - could not find ${TOOLS_DIR},"
	die "$msg ensure you have unpacked pipeline correctly"
    fi

    if [ -d "${TOOLS_DIR}" ] # Tested variable is quoted.
	then
	log_msg "${TOOLS_DIR} is a directory"
    else
	die "ERROR - ${TOOLS_DIR} exists but not a directory"
    fi

    push_dir ${PIPELINE_DIR}

    log_msg "Compiling pipeline"
    make_cmd="make install"
    run_cmd_logging_output "$make_cmd"

    if [ $? == 0 ]
	then
	log_msg "Install seems OK"
    else
	die "Install failed!"
    fi

    pop_dir

    log_msg "Finished checking pipeline installation"
}

#------------------------------------------------------------------------------

function build_database()
{
    genome_src_path=$1
    START_DIR=$2
    PIPELINE_DIR=$3

    info "Building ELAND database from $genome_src_path"

    ELAND_TEST_DIR="${START_DIR}/ELAND_test"

    push_dir ${PIPELINE_DIR}

    if [ ! -d $ELAND_TEST_DIR ]
	then
	log_msg "Creating ELAND folder"
	cmd="$MKDIR $ELAND_TEST_DIR"
	run_cmd_logging_output "$cmd"

	if [ $? -ne 0 ]
	    then
	    die "Failed to create ELAND folder '$ELAND_TEST_DIR'"
	fi
    fi


    if [ ${COMSPEC} ]
	then
	SQUASH_GENOME=squashGenome.exe
    else
	SQUASH_GENOME=squashGenome
    fi

    cmd="Eland/${SQUASH_GENOME} ${ELAND_TEST_DIR} $genome_src_path"
    run_cmd_logging_output "$cmd"

    if [ $? == 0 ]
	then
	log_msg "Finished building ELAND database from $genome_src_path."
    else
	die "Failed to build ELAND database from $genome_src_path!"
    fi

    pop_dir

    return 0
}

#------------------------------------------------------------------------------
# RUN_FOLDER_R2 should be specified only for paired end analysis.

function create_analysis_folders()
{
    # $1 is OLD_FIRECREST : set below
    # $2 is NEW_FIRECREST : set below
    cycles=$3
    config_file_path=$4
    PIPELINE_DIR=$5
    RUN_FOLDER_R1=$6
    RUN_FOLDER_R2=$7

    info "Generating analysis folders and Makefiles"
    
    cmd="${PIPELINE_DIR}/Goat/goat_pipeline.py --cycles=${cycles} --offsets=auto --GERALD=${config_file_path} --make ${RUN_FOLDER_R1} ${RUN_FOLDER_R2}"
    run_cmd_logging_output "$cmd"

    if [ $? == 0 ]
	then
	log_msg "Finished generating analysis folders and Makefiles"
    else
	die "Failed in generating analysis folders and Makefiles"
    fi

    # Check that a new Firecrest directory was created.

    push_dir ${RUN_FOLDER_R1}/Data
    OLD_FIRECREST=`ls -rdt C* | head -n 1`
    NEW_FIRECREST=`ls -rdt C* | tail -n 1`
    pop_dir

    if [ "x${OLD_FIRECREST}" == "x${NEW_FIRECREST}" ]
	then
	die "No new Firecrest directory was created!"
    fi

    eval "$1=$OLD_FIRECREST"
    eval "$2=$NEW_FIRECREST"

    log_msg "Newest Firecrest folder is ${NEW_FIRECREST}"
}

#------------------------------------------------------------------------------

function run_analysis()
{
    RUN_FOLDER_R1=$1
    NEW_FIRECREST=$2
    CONCURRENT_JOBS=$3

    push_dir ${RUN_FOLDER_R1}/Data/${NEW_FIRECREST}

    info "Running pipeline"

    cmd="make recursive -j ${CONCURRENT_JOBS}"
    run_cmd_logging_output "$cmd"

    if [ $? == 0 ]
	then
	log_msg "Pipeline run complete"
    else
	die "Pipeline run failed"
    fi

    pop_dir

    log_msg "Finished running pipeline on validation dataset"
}

#------------------------------------------------------------------------------

function list_folder()
{
    folder=$1
    list_file=$2
    excluded_prefix=$3

    cmd="ls ${folder} | sort"

    if [ -n $excluded_prefix ]
	then
	cmd="$cmd | egrep -v '^$excluded_prefix'"
    fi

    cmd="$cmd > ${list_file}"

    run_cmd_logging_err "$cmd"
}

#------------------------------------------------------------------------------

function compare_output_folder_versions()
{
    RUN_FOLDER_R1=$1
    folder_label=$2
    ref_folder=$3
    actual_folder=$4
    expected_subdir_prefix=$5

    if [ ! -d $ref_folder ]
	then
	die "Reference folder $ref_folder not found!"
    fi

    if [ ! -d $actual_folder ]
	then
	die "Actual folder $actual_folder not found!"
    fi

    info "Comparing contents of ${ref_folder} and ${actual_folder}"

    expected_list_file=${RUN_FOLDER_R1}/${folder_label}_contents_expected.txt
    actual_list_file=${RUN_FOLDER_R1}/${folder_label}_contents_actual.txt

    list_folder $ref_folder $expected_list_file $expected_subdir_prefix
    list_folder $actual_folder $actual_list_file $expected_subdir_prefix
    run_cmd_logging_output "diff ${expected_list_file} ${actual_list_file}"

    diff_result=$?

    if [ $diff_result -ne 0 ]
    then
	copy_file_to_log $expected_list_file
	copy_file_to_log $actual_list_file
    fi

    $RM $expected_list_file
    $RM $actual_list_file

    if [ $diff_result -ne 0 ]
    then
	die "Unexpected differences found between reference and latest $folder_label folders."
    fi
}

#------------------------------------------------------------------------------

function compare_output_with_reference()
{
    RUN_FOLDER_R1=$1
    OLD_FIRECREST=$2
    NEW_FIRECREST=$3

    info "Entering ${RUN_FOLDER_R1}/Data to begin comparisons."

    push_dir ${RUN_FOLDER_R1}/Data

    compare_output_folder_versions $RUN_FOLDER_R1 "Firecrest" \
        $OLD_FIRECREST $NEW_FIRECREST "Bustard"


    OLD_BUSTARD=`ls -rdt ${OLD_FIRECREST}/Bustard* | head -n 1`
    NEW_BUSTARD=`ls -rdt ${NEW_FIRECREST}/Bustard* | tail -n 1`

    compare_output_folder_versions $RUN_FOLDER_R1 "Bustard" \
        $OLD_BUSTARD $NEW_BUSTARD "GERALD"


    OLD_GERALD=`ls -rdt ${OLD_BUSTARD}/GERALD* | head -n 1`
    NEW_GERALD=`ls -rdt ${NEW_BUSTARD}/GERALD* | tail -n 1`

    compare_output_folder_versions $RUN_FOLDER_R1 "GERALD" \
         $OLD_GERALD $NEW_GERALD

    pop_dir

    log_msg "Finished comparing validation dataset with created dataset"
}

#------------------------------------------------------------------------------

function run_expt()
{
    expt=$1
    START_DIR=$2
    PIPELINE_DIR=$3
    CONCURRENT_JOBS=$4

    # Extract and check run folders.

    paired=0
    run_folder_r1_path=''
    run_folder_r2_path=''
    run_folder_r1_name=''
    cycles=''
    config_file_name=''

    junk=`echo $expt | grep -E -o -m 1 ".+:.+:.+:.+"`

    if [ $? -eq 0 ]
	then
	paired=1
	run_folder_r1_name=`echo $expt | cut -d":" -f1`
	run_folder_r2_name=`echo $expt | cut -d":" -f2`
	cycles=`echo $expt | cut -d":" -f3`
	config_file_name=`echo $expt | cut -d":" -f4`

	if [ -z $run_folder_r1_name ]
	    then
	    die "Expt description '$expt' suggests paired reads but R1 empty."
	fi

	if [ -z $run_folder_r2_name ]
	    then
	    die "Expt description '$expt' suggests paired reads but R2 empty."
	fi

	run_folder_r2_path=${START_DIR}/${run_folder_r2_name}

	if [ ! -d $run_folder_r2_path ]
	    then
	    die "Specified R2 folder '$run_folder_r2_path' is not a directory."
	fi
    else
	junk=`echo $expt | grep -E -o -m 1 ".+:.+:.+"`

	if [ $? -eq 0 ]
	    then
	    run_folder_r1_name=`echo $expt | cut -d":" -f1`
	    cycles=`echo $expt | cut -d":" -f2`
	    config_file_name=`echo $expt | cut -d":" -f3`
	else
	    die "Expt description '$expt' has invalid format."
	fi
    fi

    run_folder_r1_path=${START_DIR}/${run_folder_r1_name}

    if [ ! -d $run_folder_r1_path ]
	then
	die "Specified run folder '$run_folder_r1_path' is not a directory."
    fi

    config_file_path=$run_folder_r1_path/$config_file_name

    if [ ! -f $config_file_path ]
	then
	die "Specified config file '$config_file_path' not found."
    fi

    msg_folder_spec=''

    if [ $paired -ne 0 ]
	then
	msg_folder_spec="R1 ${run_folder_r1_name} R2 ${run_folder_r2_name}"
    else
	msg_folder_spec=${run_folder_r1_name}
    fi

    info "About to analyse: ${msg_folder_spec}"
    info "Will use cycle spec : ${cycles}"

    # Do the analysis 

    OLD_FIRECREST=''
    NEW_FIRECREST=''
    create_analysis_folders OLD_FIRECREST NEW_FIRECREST \
        $cycles $config_file_path $PIPELINE_DIR \
	$run_folder_r1_path $run_folder_r2_path

#exit -1;

    run_analysis $run_folder_r1_path $NEW_FIRECREST $CONCURRENT_JOBS

    compare_output_with_reference $run_folder_r1_path \
        $OLD_FIRECREST $NEW_FIRECREST

    info "Analysis completed OK for: ${msg_folder_spec}"
}

#------------------------------------------------------------------------------

function usage()
{
    script_name=$1
    ret_val=$2

    echo "Usage: ${script_name} pipeline_install_directory"
    echo "       e.g. : ${script_name} /usr/local/bin/SolexaPipeline-0.3"
    echo "   OR: ${script_name} --prereqs"
    echo "       to check for required utilities and libraries only"
    echo
    echo "   If some Pipeline-compatible FFT library other than FFTW is"
    echo "   installed then the --skip-fftw-check option may be specified"
    echo "   to bypass checking for an FFTW installation :-"
    echo "   ${script_name} --skip-fftw-check pipeline_install_directory"
    echo "   ${script_name} --skip-fftw-check --prereqs"
    echo
    echo "   To allow the validation script to run multiple jobs"
    echo "   simultaneously (assuming your machine supports it), append"
    echo "   the maximum number of jobs, e.g. to allow 4 concurrent jobs"
    echo "   ${script_name} pipeline_install_directory 4"
  

    if [ $ret_val -ne 0 ]
    then
      log_msg "Incorrect usage of $script_name"
    fi

    exit $ret_val;
}

#------------------------------------------------------------------------------

SCRIPT=$0
START_DIR=${PWD}
log_file=${START_DIR}/${SCRIPT}.out
init_log

# Basic argument checking
# (Do not forget CONCURRENT_JOBS.)

num_args=$#

if [ $num_args -lt 1 ] || [ $num_args -gt 3 ]
then
    usage $SCRIPT -1
fi

skip_fftw_check=0
PIPELINE_DIR=$1

if [ "x$1" == "x--skip-fftw-check" ]
then
    skip_fftw_check=1
    PIPELINE_DIR=$2
    info "Will skip check for FFTW as requested."
fi

if [ "x$PIPELINE_DIR" == "x--prereqs" ]
then
    check_for_prerequisites $skip_fftw_check

    if [ $? -eq 1 ]
    then
	die "Failed prerequisite checking."
    else
        info "Prerequisites OK"
	exit 0
    fi
elif [ "x$PIPELINE_DIR" == "x--help" ]
then
    usage $SCRIPT 0
fi

if [ -n "${PIPELINE_DIR}" ]
then
    log_msg "${SCRIPT} starting"
    log_msg "Pipeline dir specified as ${PIPELINE_DIR}"
else
    usage $SCRIPT -1
fi

CONCURRENT_JOBS=1

jobs_arg=$2

if [ $skip_fftw_check -eq 1 ]
then
   jobs_arg=$3
fi

if [ -n "${jobs_arg}" ]
then
    junk=`echo $jobs_arg | grep -E -o -m 1 "^[0-9]+$"`

    if [ $? -eq 0 ]
    then
	if [ $jobs_arg -ge 1 ]
	then
	    info "Found option '${jobs_arg}' so will run ${jobs_arg} jobs in parallel."
	    CONCURRENT_JOBS=$jobs_arg
	else
	    info "Number of concurrent jobs must be at least 1."
	    usage $SCRIPT -1
	fi
    else
	info "Non-numeric '${jobs_arg}' specified for number of concurrent jobs."
	usage $SCRIPT -1
    fi
fi

# Check for some prerequisites.
check_for_prerequisites $skip_fftw_check

if [ $? -eq 1 ]
then
    die "Failed prerequisite checking"
fi

# Try to capture some system information.
obtain_sys_info $START_DIR

# Check pipeline installation, compiling code if necessary.
check_pipeline_installation $PIPELINE_DIR

# Process any reference sequences that need to be compiled from ASCII.
GENOME_BASE_DIR=${START_DIR}/Genomes

for genome in "${GENOMES[@]}"
  do
  build_database $GENOME_BASE_DIR/$genome $START_DIR $PIPELINE_DIR
done

# Set up analysis, run analysis and check results for each specified expt.
for expt in "${EXPTS[@]}"
  do
  run_expt $expt $START_DIR $PIPELINE_DIR $CONCURRENT_JOBS
done

info "${SCRIPT} finished : Installation OK"

#------------------------------------------------------------------------------

